module NCSTAT_Utilities
!! #############################################################################
!!
!! Copyright 2021 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
! ***************************************************
! MODULE EXPORTING GENERAL AND COMPUTING UTILITIES  *
! FOR THE NCSTAT SOFTWARE.                          *
!                                                   *
! LATEST REVISION : 17/02/2021                      *
!                                                   *
! ***************************************************
!
!
! USED MODULES
! ============
!
    use netcdf,   only : nf90_max_name, nf90_int, nf90_open, nf90_nowrite, nf90_noerr,     &
                         nf90_strerror, nf90_max_var_dims, nf90_sizehint_default,          &
                         nf90_inq_varid, nf90_inquire_variable, nf90_inquire_dimension,    &
                         nf90_put_var
!
    use Statpack, only : i1b, i4b, stnd, lgl, false, one,                       & 
                         kint, kfix, kexp, find_field, string_to_val, merror
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public :: get_listarg, get_int_list, open_2_files, check_lon_domain,            &     
              get_variable_dimcoords,   get_variables_dimcoords,                    &
              get_variable_dims,        get_variables_dims,                         &         
              select_grid_points_2d,    select_grid_points_3d,                      &
              duplicate_orca_points_2d, duplicate_orca_points_3d,                   &
              write_file_3d,            write_file_4d,                              &
              prtunit, nchar, sp, dp, radius, def_missing_scal, def_missing,        &
              nlon_def2, nlat_def2, ndepth_def2, nlon_def4, nlat_def4, ndepth_def4, &
              days_between_dates, char_date_to_int_date              
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface get_listarg
        module procedure    get_listarg2_i, get_listarg3_i,   &
                            get_listarg2_r, get_listarg3_r
    end interface
!
    interface write_file_3d
        module procedure   write_file_3d_rv, write_file_3d_rm
    end interface
!
    interface write_file_4d
        module procedure   write_file_4d_rv, write_file_4d_rm
    end interface
!   
!   
! PARAMETERS 
! ==========
!
    integer(i4b), parameter ::                                                       &
!
!  prtunit IS THE PRINTING UNIT
!
                                   prtunit=6,                                        &
!
!  nchar IS THE MAXIMUM LENGTH OF A CHARACTER STRING ARGUMENT
!
                                   nchar=10*nf90_max_name,                           &
!
!  nlon_def2, nlat_def2 AND ndepth_def2 ARE THE DIMENSIONS OF THE ORCA R2 GRID
!  nlon_def4, nlat_def4 AND ndepth_def4 ARE THE DIMENSIONS OF THE ORCA R4 GRID
!
                                   nlon_def2=182, nlat_def2=149, ndepth_def2=31,     &
                                   nlon_def4=92,  nlat_def4=76,  ndepth_def4=31,     &
!
! SYMBOLIC NAMES FOR KIND TYPES OF SINGLE- AND DOUBLE-PRECISION REALS.
!
                                   sp = kind( 1.0   ),                               &
                                   dp = kind( 1.0d0 )
!   
! radius IS THE RADIUS OF THE EARTH IN METERS.
!
    real(stnd), parameter :: radius=6371229.0_stnd,      &
!
! def_missing_scal IS THE DEFAULT MISSING INDICATOR FOR THE SCALE FACTORS.
!
                             def_missing_scal=one,       &
!
!  def_missing IS THE DEFAULT MISSING INDICATOR
!
                             def_missing=1.e+20_stnd
!
!
! MODULE VARIABLES
! ================
!    
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! ===========================================================================================
!
        subroutine get_listarg2_i( namproc, argname, message1, message2, arg1, arg2 )
!
!
!   Purpose
!   _______
!
!       This subroutine is intended to extract two integer values ARG1 and ARG2 from
!       the argument string ARGNAME. The set of delimiters is specified by the character
!       parameter DELIMS.
!                                                                              
!
!   Arguments
!   _________
!
!       NAMPROC     (INPUT) character(len=*)
!                   Name of the calling procedure.
!
!       ARGNAME     (INPUT) character(len=*)
!                   The string to analyze.
!
!       MESSAGE1    (INPUT) character(len=*)
!                   First error message if the two values are not integers.
!
!       MESSAGE2    (INPUT) character(len=*)
!                   Second error message if the number of arguments is less than 2 or
!                   greater than 2 .
!
!       ARG1        (OUTPUT) integer(i4b)
!                   First integer argument.
!
!       ARG2        (OUTPUT) integer(i4b)
!                   Second integer argument.
!                                                                              
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*), intent(in)  :: namproc, argname, message1, message2
!
        integer(i4b), intent(out)  :: arg1, arg2
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b) :: k, j, j1, j2
!
        integer(i1b)  :: kcode
!
        character(len=14) :: fmt
!   
!
!   PARAMETERS
!   __________
!
        character(len=*), parameter :: delims=",;/"
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        k = 1
        do j = 1, 2
            call find_field( argname, j1, j2, delims=delims, isearch=k )
            if ( j1/=0 ) then
                call string_to_val( argname(j1:j2), kcode, fmt )
                if ( kcode/=kint ) then
                    call merror( namproc//message1 )
                else
                    read( argname(j1:j2), fmt ) k
                    if ( j==1 ) then
                        arg1 = k
                    else
                        arg2 = k
                    end if
                end if
                k = j2 + 1
            else
                call merror( namproc//message2 )
            end if
        end do                
!
        call find_field( argname, j1, j2, delims=delims, isearch=k )
        if ( j1/=0 ) call merror( namproc//message2 )
!
!
!   END OF SUBROUTINE get_listarg2_i
!   ________________________________
!
        end subroutine get_listarg2_i
!
! ===========================================================================================
!
        subroutine get_listarg3_i( namproc, argname, message1, message2, arg1, arg2, arg3 )
!
!
!   Purpose
!   _______
!
!       This subroutine is intended to extract three integers values ARG1, ARG2 and ARG3 from
!       the argument string ARGNAME. The set of delimiters is specified by the character
!       parameter DELIMS.
!                                                                              
!
!   Arguments
!   _________
!
!       NAMPROC     (INPUT) character(len=*)
!                   Name of the calling procedure.
!
!       ARGNAME     (INPUT) character(len=*)
!                   The string to analyze.
!
!       MESSAGE1    (INPUT) character(len=*)
!                   First error message if the three values are not integers.
!
!       MESSAGE2    (INPUT) character(len=*)
!                   Second error message if the number of arguments is less than 3 or
!                   greater than 3 .
!
!       ARG1        (OUTPUT) integer(i4b)
!                   First integer argument.
!
!       ARG2        (OUTPUT) integer(i4b)
!                   Second integer argument.
!                                                                              
!       ARG3        (OUTPUT) integer(i4b)
!                   Third integer argument.
!                                                                              
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*), intent(in)  :: namproc, argname, message1, message2
!
        integer(i4b), intent(out)  :: arg1, arg2, arg3
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b) :: k, j, j1, j2
!
        integer(i1b)  :: kcode
!
        character(len=14) :: fmt
!   
!
!   PARAMETERS
!   __________
!
        character(len=*), parameter :: delims=",;/"
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        k = 1
        do j = 1, 3
            call find_field( argname, j1, j2, delims=delims, isearch=k )
            if ( j1/=0 ) then
                call string_to_val( argname(j1:j2), kcode, fmt )
                if ( kcode/=kint ) then
                    call merror( namproc//message1 )
                else
                    read( argname(j1:j2), fmt ) k
                    if ( j==1 ) then
                        arg1 = k
                    else if ( j==2 ) then
                        arg2 = k
                    else
                        arg3 = k
                    end if
                end if
                k = j2 + 1
            else
                call merror( namproc//message2 )
            end if
        end do                
!
        call find_field( argname, j1, j2, delims=delims, isearch=k )
        if ( j1/=0 ) call merror( namproc//message2 )
!
!
!   END OF SUBROUTINE get_listarg3_i
!   ________________________________
!
        end subroutine get_listarg3_i
!
! ===========================================================================================
!
        subroutine get_listarg2_r( namproc, argname, message1, message2, arg1, arg2 )
!
!
!   Purpose
!   _______
!
!       This subroutine is intended to extract two real values ARG1 and ARG2 from
!       the argument string ARGNAME. The set of delimiters is specified by the character
!       parameter DELIMS.
!                                                                              
!
!   Arguments
!   _________
!
!       NAMPROC     (INPUT) character(len=*)
!                   Name of the calling procedure.
!
!       ARGNAME     (INPUT) character(len=*)
!                   The string to analyze.
!
!       MESSAGE1    (INPUT) character(len=*)
!                   First error message if the two values are not real numbers.
!
!       MESSAGE2    (INPUT) character(len=*)
!                   Second error message if the number of arguments is less than 2 or
!                   greater than 2 .
!
!       ARG1        (OUTPUT) real(stnd)
!                   First real argument.
!
!       ARG2        (OUTPUT) real(stnd)
!                   Second real argument.
!                                                                              
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*), intent(in)  :: namproc, argname, message1, message2
!
        real(stnd), intent(out)  :: arg1, arg2
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        real(stnd) :: tmp
!
        integer(i4b) :: k, j, j1, j2
!
        integer(i1b)  :: kcode
!
        character(len=14) :: fmt
!   
!
!   PARAMETERS
!   __________
!
        character(len=*), parameter :: delims=",;/"
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        k = 1
        do j = 1, 2
            call find_field( argname, j1, j2, delims=delims, isearch=k )
            if ( j1/=0 ) then
                call string_to_val( argname(j1:j2), kcode, fmt )
                if ( kcode/=kfix .and. kcode/=kexp ) then
                    call merror( namproc//message1 )
                else
                    read( argname(j1:j2), fmt ) tmp
                    if ( j==1 ) then
                        arg1 = tmp
                    else
                        arg2 = tmp
                    end if
                end if
                k = j2 + 1
            else
                call merror( namproc//message2 )
            end if
        end do                
!
        call find_field( argname, j1, j2, delims=delims, isearch=k )
        if ( j1/=0 ) call merror( namproc//message2 )
!
!
!   END OF SUBROUTINE get_listarg2_r
!   ________________________________
!
        end subroutine get_listarg2_r
!
! ===========================================================================================
!
        subroutine get_listarg3_r( namproc, argname, message1, message2, arg1, arg2, arg3 )
!
!
!   Purpose
!   _______
!
!       This subroutine is intended to extract three real values ARG1, ARG2 and ARG3 from
!       the argument string ARGNAME. The set of delimiters is specified by the character
!       parameter DELIMS.
!                                                                              
!
!   Arguments
!   _________
!
!       NAMPROC     (INPUT) character(len=*)
!                   Name of the calling procedure.
!
!       ARGNAME     (INPUT) character(len=*)
!                   The string to analyze.
!
!       MESSAGE1    (INPUT) character(len=*)
!                   First error message if the three values are not real numbers.
!
!       MESSAGE2    (INPUT) character(len=*)
!                   Second error message if the number of arguments is less than 3 or
!                   greater than 3 .
!
!       ARG1        (OUTPUT) real(stnd)
!                   First real argument.
!
!       ARG2        (OUTPUT) real(stnd)
!                   Second real argument.
!                                                                              
!       ARG3        (OUTPUT) real(stnd)
!                   Third real argument.
!                                                                              
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*), intent(in)  :: namproc, argname, message1, message2
!
        real(stnd), intent(out)  :: arg1, arg2, arg3
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        real(stnd) :: tmp
!
        integer(i4b) :: k, j, j1, j2
!
        integer(i1b)  :: kcode
!
        character(len=14) :: fmt
!   
!
!   PARAMETERS
!   __________
!
        character(len=*), parameter :: delims=",;/"
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        k = 1
        do j = 1, 3
            call find_field( argname, j1, j2, delims=delims, isearch=k )
            if ( j1/=0 ) then
                call string_to_val( argname(j1:j2), kcode, fmt )
                if ( kcode/=kfix .and. kcode/=kexp ) then
                    call merror( namproc//message1 )
                else
                    read( argname(j1:j2), fmt ) tmp
                    if ( j==1 ) then
                        arg1 = tmp
                    else if ( j==2 ) then
                        arg2 = tmp
                    else
                        arg3 = tmp
                    end if
                end if
                k = j2 + 1
            else
                call merror( namproc//message2 )
            end if
        end do                
!
        call find_field( argname, j1, j2, delims=delims, isearch=k )
        if ( j1/=0 ) call merror( namproc//message2 )
!
!
!   END OF SUBROUTINE get_listarg3_r
!   ________________________________
!
        end subroutine get_listarg3_r
!
! ===========================================================================================
!
        subroutine get_int_list(  namproc, string, start, nlist, list )
!
!
!   Purpose
!   _______
!                                                                              
!   Retrieve a list of integers from the string STRING(START:).
!   It is assumed that the substring STRING(1:START-1) gives the argument
!   associated with the string (eg, -l= ... ).
!
!   Arguments
!   _________
!
!
!       NAMPROC                (INPUT) character(len=*)
!                              Name of the calling procedure.
!
!       STRING                 (INPUT) character(len=*)
!                              The string to analyze.
!
!       START                  (INPUT) integer(i4b)
!                              The starting position for scanning the string.
!
!       NLIST                  (OUTPUT) integer(i4b)
!                              The numbers of integers in the list.
!
!       LIST                   (OUTPUT) integer(i4b), pointer, dimension(:)
!                              The list.
!                                                                              
!                   
! _________________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Statpack, only : string_count, allocate_error, arth, quick_sort 
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),                   intent(in)  :: namproc
        character(len=*),                   intent(in)  :: string
!
        integer(i4b),                       intent(in)  :: start
        integer(i4b),                       intent(out) :: nlist
        integer(i4b), pointer, dimension(:)             :: list
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)  ::  status, j, k, l, i1, i2, j1, j2, ncolon, nitem, nselection_list
        integer(i4b), dimension(4)                 :: ndelims
        integer(i4b), dimension(:),    allocatable :: selection_list
        integer(i4b), dimension(:,:),  allocatable :: selection_list2
!   
        integer(i1b) :: kcode
!   
        character(len=14) :: fmt
!   
!
!   PARAMETERS
!   __________
!
        character(len=*), parameter :: delims=",;/:"
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
!
!       COUNT THE NUMBER OF SEPARATORS.
!
        do j = 1, 4
            ndelims(j) = string_count( string(start:), delims(j:j) )
        end do
!
!       COUNT THE NUMBER OF SPECIFICATIONS OF THE FORM x:y.
!
        ncolon = ndelims(4)
        if ( ncolon/=0 ) then
            allocate( selection_list2(ncolon,2), stat=status )
            if ( status/=0 ) call merror( namproc//allocate_error )
        end if
!
!       COUNT THE WHOLE NUMBER OF ITEMS IN THE LIST.
!
        nitem = sum( ndelims(1:3) ) + 1
        nselection_list = nitem - ncolon
        if ( nselection_list/=0 ) then
            allocate( selection_list(nselection_list), stat=status )
            if ( status/=0 ) call merror( namproc//allocate_error )
        end if
!
!       READ THE ARGUMENT.
!
        k  = start
        i1 = 0
        i2 = 0
!
!       CYCLE THROUGH THE ITEMS IN THE LIST.
!
        do j = 1, nitem
            call find_field( string, j1, j2, delims=delims(1:3), isearch=k )
!
            if ( j1==0 ) exit
!
            call string_to_val( string(j1:j2), kcode, fmt )
            if ( kcode/=kint ) then
!
!               THE CURRENT ITEM IS OF THE FORM x:y, WHERE x AND y ARE INTEGERS.
!
                i1 = i1 + 1                                     
                k  = j1
                do l = 1, 2
                    call find_field( string, j1, j2, delims=delims(1:4), isearch=k )
                    call string_to_val( string(j1:j2), kcode, fmt )
!
                    if ( kcode/=kint )    &
    call merror( namproc//"Items are integer values and separators are , ; / for argument "//string(1:start-1) )
!
                    read( string(j1:j2), fmt ) selection_list2(i1,l)
                    k = j2 + 1
                end do
            else
!
!               THE CURRENT ITEM IS A SIMPLE INTEGER VALUE.
!
                i2 = i2 + 1                                     
                read( string(j1:j2), fmt ) selection_list(i2)
                k = j2 + 1
            end if
        end do
!
!       NOW CHECK THE ARGUMENT.
!
        if ( i1/=ncolon .or. i2/=nselection_list ) &
        call merror( namproc//"Wrong specification for argument "//string(1:start-1)  )
!
        if ( nselection_list/=0 ) then
            if ( any( selection_list(:)<0 ) )     &
            call merror( namproc//"Selected items are positive integers for argument "//string(1:start-1) )
        end if
!
        nlist = nitem
!
        if ( ncolon/=0 ) then
            if ( any( selection_list2(:,:)<0 )  ) &
      call merror( namproc//"Selected items are positive integers for argument "//string(1:start-1) )
!
            if ( any( selection_list2(:,1)>selection_list2(:,2) ) )     &
      call merror( namproc//"Items of the form x:y with x>y are not allowed for argument "//string(1:start-1) )
!
            nlist = nlist + sum( selection_list2(:,2)-selection_list2(:,1) )
        end if
!
!       ALLOCATE VARIABLE FOR THE TIME COORDINATES LIST.
!
        allocate( list(nlist), stat=status )
        if ( status/=0 ) call merror( namproc//allocate_error )
!
        list(1:nselection_list) = selection_list(1:nselection_list)
        k = nselection_list
        do j = 1, ncolon
            j1 = selection_list2(j,1)
            j2 = selection_list2(j,2) - j1 + 1
            i1 = k + 1
            i2 = k + j2
            list(i1:i2) = arth( j1, 1_i4b, j2 )
            k = i2
        end do
!
!       DEALLOCATE WORK ARRAYS
!
        if ( allocated(selection_list) )  deallocate( selection_list )
!
        if ( allocated(selection_list2) )  deallocate( selection_list2 )
!
!       CHECK THAT THE TIME INDICES ARE STRICTLY INCREASING.
!
        k = count( list(1:nlist-1) >= list(2:nlist) )
        if ( k/=0 ) then
!
!           SORT THE TIME INDICES IN ASCENDING ORDER.
!
            call quick_sort( list(:)  )
            k = count( list(1:nlist-1) >= list(2:nlist) )
            if ( k/=0 )    &
            call merror( namproc//"Duplicate items are not allowed for argument "//string(1:start-1) )
        end if
!
!
!   END OF SUBROUTINE get_int_list
!   ______________________________
!
        end subroutine get_int_list
!
! ===========================================================================================
!
        subroutine open_2_files( namproc, filname, filid )
!
!
!   Purpose
!   _______
!                                                                              
!   Open two netCDF files and return their netCDF IDs in argument ID.
!
!   Arguments
!   _________
!
!       NAMPROC     (INPUT) character(len=*)
!                   Name of the calling procedure.
!
!       FILNAME     (INPUT) character(len=*), dimension(2)
!                   Names of the two netCDF files to open in a vector of length 2.
!
!       FILID       (OUTPUT) integer(i4b), dimension(2)
!                   Returned netCDF IDs for the two netCDF files in a vector of length 2.
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),               intent(in)  :: namproc
        character(len=*), dimension(2), intent(in)  :: filname
!
        integer(i4b),     dimension(2), intent(out) :: filid
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)  :: status, k
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        k = nf90_sizehint_default
        status = nf90_open( trim(filname(1)), nf90_nowrite, filid(1), chunksize=k )
        if ( status/=nf90_noerr )                                        &
        call merror( namproc//trim( nf90_strerror( status ) ) )
!
        if ( filname(2)/=filname(1) ) then                             
            k = nf90_sizehint_default
            status = nf90_open( trim(filname(2)), nf90_nowrite, filid(2), chunksize=k )
            if ( status/=nf90_noerr )                                    &
            call merror( namproc//trim( nf90_strerror( status ) ) )
        else
!
!           DEFAULT ID FOR THE SECOND NETCDF FILE.
!
            filid(2) = filid(1)
!
        endif
!
!   END OF SUBROUTINE open_2_files
!   ______________________________
!
        end subroutine open_2_files
!
! ===========================================================================================
!
        subroutine get_variable_dimcoords(  namproc, varname, filid,                               &
                                            varid_out, dimids_out, dim_names_out, dimlens_out,     &
                                            coor_id_out, coor_natts_out, coor_type_out,            &
                                            use_coor_variable_out, natts )
!
!
!   Purpose
!   _______
!                                                                              
!   Get the ID and, IDs and lengths of the dimensions, of a NETCDF variable given its name and
!   the ID of the netCDF file containing the variable. In addition, the IDs, types and numbers
!   of attributes of associated coordinate variables are returned if they exist.
!
!   Arguments
!   _________
!
!
!       NAMPROC                (INPUT) character(len=*)
!                              Name of the calling procedure.
!
!       VARNAME                (INPUT) character(len=*)
!                              Name of the netCDF variable.
!
!       FILID                  (INPUT) integer(i4b)
!                              NetCDF ID of the netCDF file.
!
!       VARID_OUT              (OUTPUT) integer(i4b)
!                              netCDF ID for the netCDF variable.
!
!       DIMIDS_OUT             (OUTPUT) integer(i4b), dimension(nf90_max_var_dims)
!                              list of IDs of the dimensions for the netCDF variable.                 
!
!       DIM_NAMES_OUT          (OUTPUT) character(len=*), dimension(nf90_max_var_dims)
!                              list of names of the dimensions for the netCDF variables.            
!                                                                              
!       DIMLENS_OUT            (OUTPUT) integer(i4b), dimension(nf90_max_var_dims)
!                              list of lengths of the dimensions for the netCDF variables.
!
!       COOR_ID_OUT            (OUTPUT) integer(i4b), dimension(nf90_max_var_dims)
!                              netCDF IDs for the coordinate variables if they exist.
!
!       COOR_NATTS_OUT         (OUTPUT) integer(i4b), dimension(nf90_max_var_dims)
!                              Number of attributes for the coordinate variables.
!
!       COOR_TYPE_OUT          (OUTPUT) integer(i4b), dimension(nf90_max_var_dims)
!                              NC types of the coordinate variables.
!
!       USE_COOR_VARIABLE_OUT  (OUTPUT) logical(i4b), dimension(nf90_max_var_dims)
!                              Logical indicators for the coordinate variables :     
!                   
!                                  USE_COOR_VARIABLE_OUT(i) = TRUE if a coordinate variable of
!                                                             name DIM_NAMES_OUT(i) exists.
!                   
!                                  USE_COOR_VARIABLE_OUT(i) = FALSE otherwise.
!                                                                              
!       NATTS                  (OUTPUT, OPTIONAL) integer(i4b)
!                              Returned number of variable attributes assigned to the netCDF
!                              variable.              
!                   
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*), intent(in)   :: namproc
        character(len=*), intent(in)   :: varname
        character(len=*), dimension(nf90_max_var_dims), intent(out)  :: dim_names_out
!
        integer(i4b),                               intent(in)  :: filid
        integer(i4b),                               intent(out) :: varid_out
        integer(i4b), dimension(nf90_max_var_dims), intent(out) :: dimids_out, dimlens_out,      &
                                                                   coor_id_out, coor_natts_out,  & 
                                                                   coor_type_out
        integer(i4b), optional,                     intent(out) :: natts
!
        logical(lgl), dimension(nf90_max_var_dims), intent(out) :: use_coor_variable_out  
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)                                      :: status, dimnum, k, i 
        integer(i4b), dimension(nf90_max_var_dims)        :: dimids_tmp
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        varid_out        = 0
        dimids_out(:)    = 0
        dimlens_out(:)   = 0
        dim_names_out(:) = " "
!
        coor_id_out(:)           = 0
        coor_natts_out(:)        = 0
        coor_type_out(:)         = 0
        use_coor_variable_out(:) = false
!
        status = nf90_inq_varid( filid, trim(varname), varid_out )
        if ( status/=nf90_noerr )  call merror( namproc//trim( nf90_strerror( status ) ) )
!
        status = nf90_inquire_variable( filid, varid_out,                                  &
                                        ndims=dimnum, dimids=dimids_out(:), natts=natts )
        if ( status/=nf90_noerr ) &
        call merror( namproc//trim( nf90_strerror( status ) ) )
! 
        do i = 1, dimnum
!         
             status = nf90_inquire_dimension( filid, dimids_out(i),                         &
                                              name=dim_names_out(i), len=dimlens_out(i) )
             if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
!
!            GET THE COORDINATE VARIABLES IDS FROM THEIR NAME, IF ANY.
!
             status = nf90_inq_varid( filid, trim(dim_names_out(i)), coor_id_out(i) )
             if ( status==nf90_noerr ) then
!    
!                FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!                FOR THE COORDINATE VARIABLES.
!
                 status = nf90_inquire_variable( filid, coor_id_out(i),                            &
                                                 xtype=coor_type_out(i), ndims=k,                  &
                                                 dimids=dimids_tmp, natts=coor_natts_out(i)        )
                 if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
!
                 use_coor_variable_out(i) = k==1 .and. dimids_tmp(1)==dimids_out(i)
             else
                 use_coor_variable_out(i) = false
             end if
             if ( .not.use_coor_variable_out(i) ) coor_type_out(i) = nf90_int
!         
        end do
!
!   END OF SUBROUTINE get_variable_dimcoords
!   ________________________________________
!
        end subroutine get_variable_dimcoords
!
! ===========================================================================================
!
        subroutine get_variables_dimcoords(  namproc, varname, filid,                               &
                                             varid_out, dimids_out, dim_names_out, dimlens_out,     &
                                             coor_id_out, coor_natts_out, coor_type_out,            &
                                             use_coor_variable_out )
!
!
!   Purpose
!   _______
!                                                                              
!   Get the IDs and, IDs and lengths of the dimensions, of a list of NETCDF variables given their
!   names and the IDs of netCDF files containing the variables. In addition, the IDs, types and
!   numbers of attributes of associated coordinate variables are returned if they exist.
!
!   Arguments
!   _________
!
!
!       NAMPROC                (INPUT) character(len=*)
!                              Name of the calling procedure.
!
!       VARNAME                (INPUT) character(len=*), dimension(:)
!                              List of names of netCDF variables in a vector.
!
!       FILID                  (INPUT) integer(i4b), dimension(size(VARNAME))
!                              NetCDF IDs of the two netCDF files in a vector of length size(VARNAME).
!
!       VARID_OUT              (OUTPUT) integer(i4b), dimension(size(VARNAME))
!                              netCDF IDs for the netCDF variables in a vector of length size(VARNAME).
!
!       DIMIDS_OUT             (OUTPUT) integer(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              list of IDs of the dimensions for the netCDF variables.                 
!
!       DIM_NAMES_OUT          (OUTPUT) character(len=*), dimension(nf90_max_var_dims,size(VARNAME))
!                              list of names of the dimensions for the netCDF the variables.            
!                                                                              
!       DIMLENS_OUT            (OUTPUT) integer(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              list of lengths of the dimensions for the netCDF variables.
!
!       COOR_ID_OUT            (OUTPUT) integer(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              netCDF IDs for the coordinate variables if they exist.
!
!       COOR_NATTS_OUT         (OUTPUT) integer(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              Number of attributes for the coordinate variables.
!
!       COOR_TYPE_OUT          (OUTPUT) integer(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              NC types of the coordinate variables.
!
!       USE_COOR_VARIABLE_OUT  (OUTPUT) logical(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              Logical indicators for the coordinate variables :
!                                  USE_COOR_VARIABLE_OUT(i,j) = TRUE if a coordinate variable of
!                                                               name DIM_NAMES_OUT(i,j) exists.
!                   
!                                  USE_COOR_VARIABLE_OUT(i,j) = FALSE otherwise.
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),                                 intent(in)   :: namproc
        character(len=*), dimension(:),                   intent(in)   :: varname
        character(len=*), dimension(nf90_max_var_dims,size(varname)), intent(out)  :: dim_names_out
!
        integer(i4b), dimension(size(varname)),                   intent(in)  :: filid
        integer(i4b), dimension(size(varname)),                   intent(out) :: varid_out
        integer(i4b), dimension(nf90_max_var_dims,size(varname)), intent(out) :: dimids_out, dimlens_out,     &
                                                                                 coor_id_out, coor_natts_out, &
                                                                                 coor_type_out
!
        logical(lgl), dimension(nf90_max_var_dims,size(varname)), intent(out) :: use_coor_variable_out  
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)                                      :: status, nvars, l_dimnum, l_var, k, i 
        integer(i4b), dimension(nf90_max_var_dims)        :: dimids_tmp
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nvars = size( varname )
!
        varid_out(:)       = 0
        dimids_out(:,:)    = 0
        dimlens_out(:,:)   = 0
        dim_names_out(:,:) = " "
!
        coor_id_out(:,:)           = 0
        coor_natts_out(:,:)        = 0
        coor_type_out(:,:)         = 0
        use_coor_variable_out(:,:) = false
!
        do l_var = 1, nvars
!
            status = nf90_inq_varid( filid(l_var), trim(varname(l_var)), varid_out(l_var) )
            if ( status/=nf90_noerr )  call merror( namproc//trim( nf90_strerror( status ) ) )
!
            status = nf90_inquire_variable( filid(l_var), varid_out(l_var),               &
                                            ndims=l_dimnum, dimids=dimids_out(:,l_var) )
            if ( status/=nf90_noerr ) &
            call merror( namproc//trim( nf90_strerror( status ) ) )
! 
            do i = 1, l_dimnum
!         
                status = nf90_inquire_dimension( filid(l_var), dimids_out(i,l_var), &
                                                 name=dim_names_out(i,l_var), len=dimlens_out(i,l_var) )
                if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
!
!               GET THE COORDINATE VARIABLES IDS FROM THEIR NAME, IF ANY.
!
                status = nf90_inq_varid( filid(l_var), trim(dim_names_out(i,l_var)), coor_id_out(i,l_var) )
                if ( status==nf90_noerr ) then
!    
!                   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!                   FOR THE COORDINATE VARIABLES.
!
                    status = nf90_inquire_variable( filid(l_var), coor_id_out(i,l_var),             &
                                                    xtype=coor_type_out(i,l_var), ndims=k,          &
                                                    dimids=dimids_tmp, natts=coor_natts_out(i,l_var) )
                    if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
!
                    use_coor_variable_out(i,l_var) = k==1 .and. dimids_tmp(1)==dimids_out(i,l_var)
                else
                    use_coor_variable_out(i,l_var) = false
                end if
                if ( .not.use_coor_variable_out(i,l_var) ) coor_type_out(i,l_var) = nf90_int
!         
            end do
!
        end do
!
!   END OF SUBROUTINE get_variables_dimcoords
!   _________________________________________
!
        end subroutine get_variables_dimcoords
!
! ===========================================================================================
!
        subroutine get_variable_dims(  namproc, varname, filid,              &
                                       varid_out, dimids_out, dimlens_out, natts )
!
!
!   Purpose
!   _______
!                                                                              
!   Get the ID and, IDs and lengths of the dimensions of a NETCDF variable given its
!   name and the ID of the netCDF file containing the variable.
!
!   Arguments
!   _________
!
!
!       NAMPROC                (INPUT) character(len=*)
!                              Name of the calling procedure.
!
!       VARNAME                (INPUT) character(len=*)
!                              Names of the netCDF variable.
!
!       FILID                  (INPUT) integer(i4b)
!                              NetCDF ID of the netCDF file.
!
!       VARID_OUT              (OUTPUT) integer(i4b)
!                              netCDF ID for the netCDF variable.
!
!       DIMIDS_OUT             (OUTPUT) integer(i4b), dimension(nf90_max_var_dims)
!                              list of IDs of the dimensions for the netCDF variable.
!                                                                              
!       DIMLENS_OUT            (OUTPUT) integer(i4b), dimension(nf90_max_var_dims)
!                              list of lengths of the dimensions for the netCDF variable.              
!                                                                              
!       NATTS                  (OUTPUT, OPTIONAL) integer(i4b)
!                              Returned number of variable attributes assigned to the netCDF
!                              variable.              
!                                                                              
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),                           intent(in)   :: namproc
        character(len=*),                           intent(in)   :: varname
!
        integer(i4b),                               intent(in)  :: filid
        integer(i4b),                               intent(out) :: varid_out
        integer(i4b), dimension(nf90_max_var_dims), intent(out) :: dimids_out, dimlens_out
        integer(i4b), optional,                     intent(out) :: natts
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)  :: status, dimnum, i 
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        varid_out      = 0
        dimids_out(:)  = 0
        dimlens_out(:) = 0
!
        status = nf90_inq_varid( filid, trim(varname), varid_out )
        if ( status/=nf90_noerr )  call merror( namproc//trim( nf90_strerror( status ) ) )
!
        status = nf90_inquire_variable( filid, varid_out,                          &
                                        ndims=dimnum, dimids=dimids_out(:), natts=natts )
        if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
! 
        do i = 1, dimnum
!         
            status = nf90_inquire_dimension( filid, dimids_out(i), len=dimlens_out(i) )
            if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
!         
        end do
!
!   END OF SUBROUTINE get_variable_dims
!   ___________________________________
!
        end subroutine get_variable_dims
!
! ===========================================================================================
!
        subroutine get_variables_dims(  namproc, varname, filid,              &
                                        varid_out, dimids_out, dimlens_out )
!
!
!   Purpose
!   _______
!                                                                              
!   Get the IDs and, IDs and lengths of the dimensions of a list of  NETCDF variables given their
!   names and the IDs of the netCDF files containing the variables.
!
!   Arguments
!   _________
!
!
!       NAMPROC                (INPUT) character(len=*)
!                              Name of the calling procedure.
!
!       VARNAME                (INPUT) character(len=*), dimension(:)
!                              Names of the netCDF variables in a vector.
!
!       FILID                  (INPUT) integer(i4b), dimension(size(VARNAME))
!                              NetCDF IDs of the two netCDF files in a vector of length size(VARNAME).
!
!       VARID_OUT              (OUTPUT) integer(i4b), dimension(size(VARNAME))
!                              netCDF IDs for the netCDF variables in a vector of length size(VARNAME).
!
!       DIMIDS_OUT             (OUTPUT) integer(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              list of IDs of the dimensions for the netCDF variables.
!                                                                              
!       DIMLENS_OUT            (OUTPUT) integer(i4b), dimension(nf90_max_var_dims,size(VARNAME))
!                              list of lengths of the dimensions for the netCDF variables.              
!                                                                              
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),                                 intent(in)   :: namproc
        character(len=*), dimension(:),                   intent(in)   :: varname
!
        integer(i4b), dimension(size(varname)),                   intent(in)  :: filid
        integer(i4b), dimension(size(varname)),                   intent(out) :: varid_out
        integer(i4b), dimension(nf90_max_var_dims,size(varname)), intent(out) :: dimids_out, dimlens_out
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)  :: status, nvars, l_dimnum, l_var, i 
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nvars = size( varname )
!
        varid_out(:)       = 0
        dimids_out(:,:)    = 0
        dimlens_out(:,:)   = 0
!
        do l_var = 1, nvars
!
            status = nf90_inq_varid( filid(l_var), trim(varname(l_var)), varid_out(l_var) )
            if ( status/=nf90_noerr )  call merror( namproc//trim( nf90_strerror( status ) ) )
!
            status = nf90_inquire_variable( filid(l_var), varid_out(l_var),               &
                                             ndims=l_dimnum, dimids=dimids_out(:,l_var) )
            if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
! 
            do i = 1, l_dimnum
!         
                status = nf90_inquire_dimension( filid(l_var), dimids_out(i,l_var), len=dimlens_out(i,l_var) )
                if ( status/=nf90_noerr ) call merror( namproc//trim( nf90_strerror( status ) ) )
!         
            end do
!
        end do
!
!   END OF SUBROUTINE get_variables_dims
!   ____________________________________
!
        end subroutine get_variables_dims
!
! ===========================================================================================
!
        subroutine check_lon_domain( grid_type_in, nlon_in, nlon_def_in,    &
                                    lon1_inout, lon2_inout, lon3_out)
!
!
!   Purpose
!   _______
!                                                                              
!   Check the longitude definition of a geographical domain.
!
!   Arguments
!   _________
!
!
!       GRID_TYPE_IN        (INPUT) character
!                           Grid type associated with the field ('n','t','w','u','v','f').
!                                                                              
!       NLON_IN             (INPUT) integer(i4b)
!                           Number of longitude points of the field.                                                 
!                   
!       NLON_DEF_IN         (INPUT) integer(i4b)
!                           Number of longitude points of the ORCA grid ( R2 or R4 resolution is
!                           assumed). This argument is used if GRID_TYPE_IN/='n', eg if the field
!                           is on the ORCA grid.
!
!       LON1_INOUT          (INPUT/OUTPUT) integer(i4b)
!                           First longitude of the domain (may be negative).
!                   
!       LON2_INOUT          (INPUT/OUTPUT) integer(i4b)
!                           Last longitude of the domain.                                                 
!                                                                              
!       LON3_OUT            (OUTPUT) integer(i4b)
!                           On exit, the number of longitude points of the domain.                                                 
!                           If, LON3_OUT is negative or equal                                                   
!                           to zero, the longitude definition of the domain
!                           specified by LON1_INOUT and LON2_INOUT on entry is wrong.                                             
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character,  intent(in) :: grid_type_in
!   
        integer(i4b), intent(in)      :: nlon_in
        integer(i4b), intent(in)      :: nlon_def_in
!   
        integer(i4b), intent(inout)      :: lon1_inout
        integer(i4b), intent(inout)      :: lon2_inout
        integer(i4b), intent(out)        :: lon3_out
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        if ( grid_type_in/='n' ) then
!
            if ( lon1_inout<0 ) then
                if ( nlon_in==nlon_def_in ) then
                    lon1_inout = min( lon1_inout, -2 )
                    lon2_inout = max( lon2_inout, 2 )
                    lon3_out   = nlon_in + lon1_inout - lon2_inout + 1
                else
                    lon3_out = -1
                end if
            else
                if ( lon1_inout==1 .and. lon2_inout==nlon_def_in ) then
                    lon1_inout = 2
                    lon2_inout = lon2_inout - 1
                else
                    lon1_inout = max( lon1_inout, 1 )
                    lon2_inout = min( lon2_inout, nlon_in )
                end if
                lon3_out = lon2_inout - lon1_inout + 1
            end if
!
        else
!
            if ( lon1_inout<0 ) then
                lon2_inout = max( lon2_inout, 1 )
                lon3_out   = nlon_in + lon1_inout - lon2_inout + 1
            else
                lon1_inout = max( lon1_inout, 1 )
                lon2_inout = min( lon2_inout, nlon_in )
                lon3_out   = lon2_inout - lon1_inout + 1
            end if
!
        end if
!
!
!   END OF SUBROUTINE check_lon_domain
!   __________________________________
!
        end subroutine check_lon_domain
!
! ===========================================================================================
!
        subroutine select_grid_points_2d( grid_type_in,                &
                                          lon1_in, lon2_in,            &
                                          lat1_in, lat2_in,            &
                                          nlon_def_in, nlat_def_in,    &
                                          mask_inout,                  &
                                          j1_out,j2_out                )
!
!
!   Purpose
!   _______
!                                                                              
!   Select a 2D geographical domain and set the values of the logical mask, correspondingly.
!
!   Arguments
!   _________
!
!                                                                             
!       GRID_TYPE_IN        (INPUT) character
!                           Grid type associated with the field ('n','t','w','u','v','f').
!                                                                              
!       LON1_IN             (INPUT) integer(i4b)
!                           First longitude of the domain (may be negative).
!                   
!       LON2_IN             (INPUT) integer(i4b)
!                           Last longitude of the domain.                                                 
!                                                                              
!       LAT1_IN             (INPUT) integer(i4b)
!                           First latitude of the domain.                                                 
!                   
!       LAT2_IN             (INPUT) integer(i4b)
!                           Last latitude of the domain.                                                 
!                                                                              
!       NLON_DEF_IN         (INPUT) integer(i4b)
!                           Number of longitude points of the ORCA grid ( R2 or R4 resolution is
!                           assumed). This argument is used if GRID_TYPE_IN/='n', eg if the field
!                           is on the ORCA grid.
!
!       NLAT_DEF_IN         (INPUT) integer(i4b)
!                           Number of latitude points of the ORCA grid ( R2 or R4 resolution is
!                           assumed). This argument is used if GRID_TYPE_IN/='n', eg if the field
!                           is on the ORCA grid.
!                                                                              
!       J1_OUT              (OUTPUT) integer(i4b)
!                           First longitude index to read the corresponding NETCDF variable.
!                   
!       J2_OUT              (OUTPUT) integer(i4b)
!                           Second longitude index to read the corresponding NETCDF variable.
!                                                                              
!       MASK_INOUT          (INPUT/OUTPUT) logical(lgl)
!                           The 2D mask array.                                                
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character, intent(in)         :: grid_type_in
!
        integer(i4b), intent(in)      :: lon1_in
        integer(i4b), intent(in)      :: lon2_in
        integer(i4b), intent(in)      :: lat1_in
        integer(i4b), intent(in)      :: lat2_in
        integer(i4b), intent(in)      :: nlon_def_in   
        integer(i4b), intent(in)      :: nlat_def_in
!
        integer(i4b), intent(out)      :: j1_out
        integer(i4b), intent(out)      :: j2_out
!
        logical(lgl), dimension(:,:), intent(inout) :: mask_inout
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)   :: nlon, nlat
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nlon = size( mask_inout, 1)
        nlat = size( mask_inout, 2)
!
!       SELECT THE DOMAIN.
!
        if ( lon1_in>0 ) then
            mask_inout(1:lon1_in-1,:nlat) = false
            mask_inout(lon2_in+1:nlon,:nlat) = false
        else
            mask_inout(lon2_in+1:nlon+lon1_in,:nlat) = false
        end if
!
        mask_inout(:nlon,1:lat1_in-1) = false
        mask_inout(:nlon,lat2_in+1:nlat) = false
!
!       REMOVE DUPLICATE POINTS FROM THE ORCA GRID.
!
        if ( grid_type_in/='n' ) then
!
            if ( lon1_in<0 ) then
                mask_inout(1,:nlat) = false
                mask_inout(nlon,:nlat) = false
            end if
!
            if ( lat2_in>=nlat_def_in-1 .and. nlat==nlat_def_in ) then
!
                mask_inout(:nlon,nlat) = false
!
                select case ( grid_type_in )
                    case ( 't', 'w' )
                        if ( nlon==nlon_def_in ) then
                            mask_inout((nlon/2)+2:nlon,nlat-1) = false
                        else
!
!                           IF THE GEOGRAPHICAL DOMAIN IS NOT THE WHOLE GLOBE,
!                           WE CANT DETERMINE THE DUPLICATE POINTS. THE WHOLE
!                           LIGN OF LATITUDE IS REMOVED HERE, BUT NORMALLY
!                           WE NEED ONLY TO REMOVE THE SAME GRID POINTS THAN
!                           ABOVE.
!
                            mask_inout(:nlon,nlat-1) = false
                        end if
                    case ( 'u' )
                        if ( nlon==nlon_def_in ) then
                            mask_inout((nlon/2)+1:nlon,nlat-1) = false
                        else
!
!                           IF THE GEOGRAPHICAL DOMAIN IS NOT THE WHOLE GLOBE,
!                           WE CANT DETERMINE THE DUPLICATE POINTS. THE WHOLE
!                           LIGN OF LATITUDE IS REMOVED HERE, BUT NORMALLY
!                           WE NEED ONLY TO REMOVE THE SAME GRID POINTS THAN
!                           ABOVE.
!
                            mask_inout(:nlon,nlat-1) = false
                        end if
                    case ( 'v', 'f' )
                        mask_inout(:nlon,nlat-1) = false
                end select
            end if
!
        end if
!
        if ( lon1_in<0 ) then
           j1_out = 1
           j2_out = nlon
        else
           j1_out = lon1_in
           j2_out = lon2_in
        end if
!
!
!   END OF SUBROUTINE select_grid_points_2d
!   _______________________________________
!
        end subroutine select_grid_points_2d
!
! ===========================================================================================
!
        subroutine select_grid_points_3d( grid_type_in,                &
                                          lon1_in, lon2_in,            &
                                          lat1_in, lat2_in,            &
                                          depth1_in, depth2_in,        &
                                          nlon_def_in, nlat_def_in,    &
                                          mask_inout,                  &
                                          j1_out,j2_out                )
!
!
!   Purpose
!   _______
!                                                                              
!   Select a 3D geographical domain and set the values of the logical mask, correspondingly.
!
!   Arguments
!   _________
!
!                                                                             
!       GRID_TYPE_IN        (INPUT) character
!                           Grid type associated with the field ('n','t','w','u','v','f').
!                                                                              
!       LON1_IN             (INPUT) integer(i4b)
!                           First longitude of the domain (may be negative).
!                   
!       LON2_IN             (INPUT) integer(i4b)
!                           Last longitude of the domain.                                                 
!                                                                              
!       LAT1_IN             (INPUT) integer(i4b)
!                           First latitude of the domain.                                                 
!                   
!       LAT2_IN             (INPUT) integer(i4b)
!                           Last latitude of the domain.                                                 
!                                                                              
!       DEPTH1_IN           (INPUT) integer(i4b)
!                           First latitude of the domain.                                                 
!                   
!       DEPTH2_IN           (INPUT) integer(i4b)
!                           Last latitude of the domain.                                                 
!                                                                              
!       NLON_DEF_IN         (INPUT) integer(i4b)
!                           Number of longitude points of the ORCA grid ( R2 or R4 resolution is
!                           assumed). This argument is used if GRID_TYPE_IN/='n', eg if the field
!                           is on the ORCA grid.
!
!       NLAT_DEF_IN         (INPUT) integer(i4b)
!                           Number of latitude points of the ORCA grid ( R2 or R4 resolution is
!                           assumed). This argument is used if GRID_TYPE_IN/='n', eg if the field
!                           is on the ORCA grid.
!                                                                              
!       J1_OUT              (OUTPUT) integer(i4b)
!                           First longitude index to read the corresponding NETCDF variable.
!                   
!       J2_OUT              (OUTPUT) integer(i4b)
!                           Second longitude index to read the corresponding NETCDF variable.
!                                                                              
!       MASK_INOUT          (INPUT/OUTPUT) logical(lgl)
!                           The 3D mask array.                                                
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character, intent(in)         :: grid_type_in
!
        integer(i4b), intent(in)      :: lon1_in
        integer(i4b), intent(in)      :: lon2_in
        integer(i4b), intent(in)      :: lat1_in
        integer(i4b), intent(in)      :: lat2_in
        integer(i4b), intent(in)      :: depth1_in
        integer(i4b), intent(in)      :: depth2_in
        integer(i4b), intent(in)      :: nlon_def_in   
        integer(i4b), intent(in)      :: nlat_def_in
!
        integer(i4b), intent(out)      :: j1_out
        integer(i4b), intent(out)      :: j2_out
!
        logical(lgl), dimension(:,:,:), intent(inout) :: mask_inout
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)   :: nlon, nlat, ndepth
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nlon   = size( mask_inout, 1)
        nlat   = size( mask_inout, 2)
        ndepth = size( mask_inout, 3)
!
!       SELECT THE DOMAIN.
!
        if ( lon1_in>0 ) then
            mask_inout(1:lon1_in-1,:nlat,:ndepth) = false
            mask_inout(lon2_in+1:nlon,:nlat,:ndepth) = false
        else
            mask_inout(lon2_in+1:nlon+lon1_in,:nlat,:ndepth) = false
        end if
!
        mask_inout(:nlon,1:lat1_in-1,:ndepth) = false
        mask_inout(:nlon,lat2_in+1:nlat,:ndepth) = false
!
!       SELECT THE VERTICAL RESOLUTION.
!
        mask_inout(:nlon,:nlat,1:depth1_in-1) = false
        mask_inout(:nlon,:nlat,depth2_in+1:ndepth) = false
!
!       REMOVE DUPLICATE POINTS FROM THE ORCA GRID.
!
        if ( grid_type_in/='n' ) then
!
            if ( lon1_in<0 ) then
                mask_inout(1,:nlat,:ndepth) = false
                mask_inout(nlon,:nlat,:ndepth) = false
            end if
!
            if ( lat2_in>=nlat_def_in-1 .and. nlat==nlat_def_in ) then
!
                mask_inout(:nlon,nlat,:ndepth) = false
!
                select case ( grid_type_in )
                    case ( 't', 'w' )
                        if ( nlon==nlon_def_in ) then
                            mask_inout((nlon/2)+2:nlon,nlat-1,:ndepth) = false
                        else
!
!                           IF THE GEOGRAPHICAL DOMAIN IS NOT THE WHOLE GLOBE,
!                           WE CANT DETERMINE THE DUPLICATE POINTS. THE WHOLE
!                           LIGN OF LATITUDE IS REMOVED HERE, BUT NORMALLY
!                           WE NEED ONLY TO REMOVE THE SAME GRID POINTS THAN
!                           ABOVE.
!
                            mask_inout(:nlon,nlat-1,:ndepth) = false
                        end if
                    case ( 'u' )
                        if ( nlon==nlon_def_in ) then
                            mask_inout((nlon/2)+1:nlon,nlat-1,:ndepth) = false
                        else
!
!                           IF THE GEOGRAPHICAL DOMAIN IS NOT THE WHOLE GLOBE,
!                           WE CANT DETERMINE THE DUPLICATE POINTS. THE WHOLE
!                           LIGN OF LATITUDE IS REMOVED HERE, BUT NORMALLY
!                           WE NEED ONLY TO REMOVE THE SAME GRID POINTS THAN
!                           ABOVE.
!
                            mask_inout(:nlon,nlat-1,:ndepth) = false
                        end if
                    case ( 'v', 'f' )
                        mask_inout(:nlon,nlat-1,:ndepth) = false
                end select
            end if
!
        end if
!
        if ( lon1_in<0 ) then
           j1_out = 1
           j2_out = nlon
        else
           j1_out = lon1_in
           j2_out = lon2_in
        end if
!
!
!   END OF SUBROUTINE select_grid_points_3d
!   _______________________________________
!
        end subroutine select_grid_points_3d
!
! ===========================================================================================
!
        subroutine duplicate_orca_points_2d( grid_type_in, mask, temp )
!
!
!   Purpose
!   _______
!                                                                              
!   Add duplicate points to the 2D ORCA GRID (R2 or R4 resolution is assumed).
!
!   Arguments
!   _________
!                                                                              
!       GRID_TYPE_IN        (INPUT) character                                                      
!                           Grid type of the data field ('t','w','u','v','f').
!                                        
!       MASK                (INPUT) logical(lgl), dimension(:,:)
!                           Land-sea mask.
!                                                                                                 
!       TEMP                (INPUT/OUTPUT) real(stnd), dimension(size(mask,1),size(mask,2))
!                           The data field.
!                                                                                                 
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character, intent(in) :: grid_type_in
!
        logical(lgl), dimension(:,:), intent(in)    :: mask
!
        real(stnd), dimension(size(mask,1),size(mask,2)), intent(inout)   :: temp
!
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b) :: nlon, nlat
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
!       GET THE DIMENSIONS OF THE 2D GRID.
!
        nlon = size( mask,1)
        nlat = size( mask,2)
!
!       ADD DUPLICATE POINTS TO THE ORCA 2D GRID.
!
        select case ( grid_type_in )
!
            case ( 't', 'w' )
                temp(nlon,:nlat-2) = temp(2,:nlat-2)
                temp(nlon/2+2:nlon,nlat-1:nlat) = temp(nlon/2:2:-1,nlat-1:nlat-2:-1)
                temp(2:nlon/2,nlat) = temp(nlon:nlon/2+2:-1,nlat-2)
                temp(1,:nlat) = temp(nlon-1,:nlat)
            case ( 'u' )
                where ( mask(nlon-1:nlon/2+1:-1,nlat-2) )                               &
                temp(2:nlon/2,nlat) = -temp(nlon-1:nlon/2+1:-1,nlat-2)
                where ( mask(nlon/2:2:-1,nlat-1:nlat-2:-1) )                            &
                temp(nlon/2+1:nlon-1,nlat-1:nlat) = -temp(nlon/2:2:-1,nlat-1:nlat-2:-1)
!
                temp(1,:nlat) = temp(nlon-1,:nlat)
                temp(nlon,:nlat) = temp(2,:nlat)
           case ( 'v' )
                temp(nlon,:nlat-2) = temp(2,:nlat-2)
                where ( mask(nlon/2:2:-1,nlat-2:nlat-3:-1) )                           &
                temp(nlon/2+2:nlon,nlat-1:nlat) = -temp(nlon/2:2:-1,nlat-2:nlat-3:-1)
                where ( mask(nlon:nlon/2+2:-1,nlat-2:nlat-3:-1) )                      &
                temp(2:nlon/2,nlat-1:nlat) = -temp(nlon:nlon/2+2:-1,nlat-2:nlat-3:-1)
                temp(1,:nlat) = temp(nlon-1,:nlat)
           case ( 'f' )
                temp(1,:nlat-2) = temp(nlon-1,:nlat-2)
                temp(nlon/2+1:nlon,nlat-1:nlat) = temp(nlon/2:1:-1,nlat-2:nlat-3:-1)
                temp(nlon,:nlat-2) = temp(2,:nlat-2)
                temp(1:nlon/2,nlat-1:nlat) = temp(nlon:nlon/2+1:-1,nlat-2:nlat-3:-1)
        end select
!
!
!   END OF SUBROUTINE duplicate_orca_points_2d
!   __________________________________________
!
        end subroutine duplicate_orca_points_2d
!
! ===========================================================================================
!
        subroutine duplicate_orca_points_3d( grid_type_in, mask, temp )
!
!
!   Purpose
!   _______
!                                                                              
!   Add duplicate points to the 3D ORCA GRID (R2 or R4 resolution is assumed).
!
!   Arguments
!   _________
!                                                                              
!       GRID_TYPE_IN        (INPUT) character                                                      
!                           Grid type of the data field ('t','w','u','v','f').        
!                                        
!       MASK                (INPUT) logical(lgl), dimension(:,:,:)
!                           Land-sea mask.                                                                
!                                                                                                 
!       TEMP                (INPUT/OUTPUT) real(stnd), dimension(size(mask,1),size(mask,2),size(mask,3))
!                           The data field.
!                                                                                                 
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character, intent(in) :: grid_type_in
!
        logical(lgl), dimension(:,:,:), intent(in)    :: mask
!
        real(stnd), dimension(size(mask,1),size(mask,2),size(mask,3)), intent(inout)   :: temp
!
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b) :: nlon, nlat, ndepth
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
!       GET THE DIMENSIONS OF THE 3D GRID.
!
        nlon   = size( mask,1)
        nlat   = size( mask,2)
        ndepth = size( mask,3)
!
!       ADD DUPLICATE POINTS TO THE ORCA 3D GRID.
!
        select case ( grid_type_in )
!
             case ( 't', 'w' )
                 temp(nlon,:nlat-2,:ndepth) = temp(2,:nlat-2,:ndepth)
                 temp(nlon/2+2:nlon,nlat-1:nlat,:ndepth) = temp(nlon/2:2:-1,nlat-1:nlat-2:-1,:ndepth)
                 temp(2:nlon/2,nlat,:ndepth) = temp(nlon:nlon/2+2:-1,nlat-2,:ndepth)
                 temp(1,:nlat,:ndepth) = temp(nlon-1,:nlat,:ndepth)
             case ( 'u' )
                 where ( mask(nlon-1:nlon/2+1:-1,nlat-2,:ndepth) )                              &
                 temp(2:nlon/2,nlat,:ndepth) = -temp(nlon-1:nlon/2+1:-1,nlat-2,:ndepth)
                 where ( mask(nlon/2:2:-1,nlat-1:nlat-2:-1,:ndepth) )                           &
                 temp(nlon/2+1:nlon-1,nlat-1:nlat,:ndepth) = -temp(nlon/2:2:-1,nlat-1:nlat-2:-1,:ndepth)
!
                 temp(1,:nlat,:ndepth) = temp(nlon-1,:nlat,:ndepth)
                 temp(nlon,:nlat,:ndepth) = temp(2,:nlat,:ndepth)
             case ( 'v' )
                 temp(nlon,:nlat-2,:ndepth) = temp(2,:nlat-2,:ndepth)
                 where ( mask(nlon/2:2:-1,nlat-2:nlat-3:-1,:ndepth) )                          &
                 temp(nlon/2+2:nlon,nlat-1:nlat,:ndepth) = -temp(nlon/2:2:-1,nlat-2:nlat-3:-1,:ndepth)
                 where ( mask(nlon:nlon/2+2:-1,nlat-2:nlat-3:-1,:ndepth) )                     &
                 temp(2:nlon/2,nlat-1:nlat,:ndepth) = -temp(nlon:nlon/2+2:-1,nlat-2:nlat-3:-1,:ndepth)
                 temp(1,:nlat,:ndepth) = temp(nlon-1,:nlat,:ndepth)
             case ( 'f' )
                 temp(1,:nlat-2,:ndepth) = temp(nlon-1,:nlat-2,:ndepth)
                 temp(nlon/2+1:nlon,nlat-1:nlat,:ndepth) = temp(nlon/2:1:-1,nlat-2:nlat-3:-1,:ndepth)
                 temp(nlon,:nlat-2,:ndepth) = temp(2,:nlat-2,:ndepth)
                 temp(1:nlon/2,nlat-1:nlat,:ndepth) = temp(nlon:nlon/2+1:-1,nlat-2:nlat-3:-1,:ndepth)
        end select
!
!
!   END OF SUBROUTINE duplicate_orca_points_3d
!   __________________________________________
!
        end subroutine duplicate_orca_points_3d
!
! ===========================================================================================
!
        subroutine write_file_3d_rv( namproc, vec, mask, periodicity, missing, filid, varid ) 
!
!
!   Purpose
!   _______
!                                                                              
!   Unpack and write a packed vector to a 3D NETCDF variable in an existing NETCDF file.
!
!   Arguments
!   _________
!
!
!       NAMPROC             (INPUT) character(len=*)
!                           Name of the calling procedure.
!
!       VEC                 (INPUT) real(stnd), dimension(:)
!                           The packed vector.
!                                                                              
!       MASK                (INPUT) logical(lgl), dimension(:,:)
!                           The 2D mask array.                                                
!
!       PERIODICITY         (INPUT) integer(i4b)
!                           The periodicity of the packed vector.
!
!       MISSING             (INPUT) real(stnd)
!                           The missing value indicator.
!
!       FILID               (INPUT) integer(i4b)
!                           NetCDF ID of the netCDF file.
!
!       VARID               (INPUT) integer(i4b)
!                           netCDF ID of the 3D netCDF variable.
!                                                                              
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),          intent(in)  :: namproc
!
        real(stnd), dimension(:),  intent(in)  :: vec
        real(stnd),                intent(in)  :: missing
!
        integer(i4b),  intent(in)  :: periodicity, filid, varid
!
        logical(lgl), dimension(:,:), intent(in) :: mask
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)                :: i, i1, i2, nlon, nlat, nvar, status
        integer(i4b), dimension(3)  :: startwri
!
        real(stnd), dimension(size(mask,1),size(mask,2)) :: temp
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nlon = size( mask, 1 )
        nlat = size( mask, 2 )
!
        nvar = size( vec )/periodicity
!
!       WRITE THE PACKED VECTOR.
!
        startwri(1:2)  = 1
        i2 = 0
!
        do i = 1, periodicity
!
            i1          = i2 + 1
            i2          = i2 + nvar
            startwri(3) = i
!
            temp(:nlon,:nlat) = unpack( vec(i1:i2), mask(:nlon,:nlat), missing )
!
            status = nf90_put_var( filid, varid, temp(:nlon,:nlat), start=startwri )
            if ( status/=nf90_noerr ) call merror(  namproc//trim( nf90_strerror( status ) ) )
!
        end do
!
!   END OF SUBROUTINE write_file_3d_rv
!   ___________________________________
!
        end subroutine write_file_3d_rv
!
! ===========================================================================================
!
        subroutine write_file_3d_rm( namproc, mat, mask, periodicity, missing, filid, varid ) 
!
!
!   Purpose
!   _______
!                                                                              
!   Unpack and write a packed matrix to a 3D NETCDF variable in an existing NETCDF file.
!
!   Arguments
!   _________
!
!
!       NAMPROC             (INPUT) character(len=*)
!                           Name of the calling procedure.
!
!       MAT                 (INPUT) real(stnd), dimension(:,:)
!                           The packed matrix.
!                                                                              
!       MASK                (INPUT) logical(lgl), dimension(:,:)
!                           The 2D mask array.                                                
!
!       PERIODICITY         (INPUT) integer(i4b)
!                           The periodicity of the packed matrix.
!
!       MISSING             (INPUT) real(stnd)
!                           The missing value indicator.
!
!       FILID               (INPUT) integer(i4b)
!                           NetCDF ID of the netCDF file.
!
!       VARID               (INPUT) integer(i4b)
!                           netCDF ID of the 3D netCDF variable.
!                                                                              
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),          intent(in)  :: namproc
!
        real(stnd), dimension(:,:),  intent(in)  :: mat
        real(stnd),                  intent(in)  :: missing
!
        integer(i4b),  intent(in)  :: periodicity, filid, varid
!
        logical(lgl), dimension(:,:), intent(in) :: mask
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)                :: i, i1, i2, k, l, nlon, nlat, nvar, status, nperiodicity
        integer(i4b), dimension(3)  :: startwri
!
        real(stnd), dimension(size(mask,1),size(mask,2)) :: temp
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nlon = size( mask, 1 )
        nlat = size( mask, 2 )
!
        nperiodicity = size( mat, 1 )
        nvar         = size( mat, 2 )/periodicity
!
!       WRITE THE PACKED MATRIX.
!
        startwri(1:2)  = 1
!
        if ( periodicity==1 ) then
!
            do k = 1, nperiodicity
!
                startwri(3) = k
!
                temp(:nlon,:nlat) = unpack( mat(k,:nvar), mask(:nlon,:nlat), missing )
!
                status = nf90_put_var( filid, varid, temp(:nlon,:nlat), start=startwri )
                if ( status/=nf90_noerr ) call merror(  namproc//trim( nf90_strerror( status ) ) )
!
            end do
!
        else
!
            l = 0
!
            do k = 1, nperiodicity
!
                i2 = 0
!
                do i = 1, periodicity
!
                    i1          = i2 + 1
                    i2          = i2 + nvar
                    l           = l + 1
                    startwri(3) = l
!
                    temp(:nlon,:nlat) = unpack( mat(k,i1:i2), mask(:nlon,:nlat), missing )
!
                    status = nf90_put_var( filid, varid, temp(:nlon,:nlat), start=startwri )
                    if ( status/=nf90_noerr ) call merror(  namproc//trim( nf90_strerror( status ) ) )
!
                end do
!
            end do
!
        end if
!
!   END OF SUBROUTINE write_file_3d_rm
!   ___________________________________
!
        end subroutine write_file_3d_rm
!
! ===========================================================================================
!
        subroutine write_file_4d_rv( namproc, vec, mask, periodicity, missing, filid, varid ) 
!
!
!   Purpose
!   _______
!                                                                              
!   Unpack and write a packed vector to a 4D NETCDF variable in an existing NETCDF file. 
!
!   Arguments
!   _________
!
!
!       NAMPROC             (INPUT) character(len=*)
!                           Name of the calling procedure.
!
!       VEC                 (INPUT) real(stnd), dimension(:)
!                           The packed vector.
!                                                                              
!       MASK                (INPUT) logical(lgl), dimension(:,:,:)
!                           The 3D mask array.                                                
!
!       PERIODICITY         (INPUT) integer(i4b)
!                           The periodicity of the packed vector.
!
!       MISSING             (INPUT) real(stnd)
!                           The missing value indicator.
!
!       FILID               (INPUT) integer(i4b)
!                           NetCDF ID of the netCDF file.
!
!       VARID               (INPUT) integer(i4b)
!                           netCDF ID of the 4D netCDF variable.
!                                                                              
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),          intent(in)  :: namproc
!
        real(stnd), dimension(:),  intent(in)  :: vec
        real(stnd),                intent(in)  :: missing
!
        integer(i4b),  intent(in)  :: periodicity, filid, varid
!
        logical(lgl), dimension(:,:,:), intent(in) :: mask
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)                :: i, i1, i2, nlon, nlat, nvar, nlevel, status
        integer(i4b), dimension(4)  :: startwri
!
        real(stnd), dimension(size(mask,1),size(mask,2),size(mask,3)) :: temp
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nlon   = size( mask, 1 )
        nlat   = size( mask, 2 )
        nlevel = size( mask, 3 )
!
        nvar   = size( vec )/periodicity
!
!       WRITE THE PACKED VECTOR.
!
        startwri(1:3)  = 1
        i2 = 0
!
        do i = 1, periodicity
!
            i1          = i2 + 1
            i2          = i2 + nvar
            startwri(4) = i
!
            temp(:nlon,:nlat,:nlevel) = unpack( vec(i1:i2), mask(:nlon,:nlat,:nlevel), missing )
!
            status = nf90_put_var( filid, varid, temp(:nlon,:nlat,:nlevel), start=startwri )
            if ( status/=nf90_noerr ) call merror(  namproc//trim( nf90_strerror( status ) ) )
!
        end do
!
!   END OF SUBROUTINE write_file_4d_rv
!   ___________________________________
!
        end subroutine write_file_4d_rv
!
! ===========================================================================================
!
        subroutine write_file_4d_rm( namproc, mat, mask, periodicity, missing, filid, varid ) 
!
!
!   Purpose
!   _______
!                                                                              
!   Unpack and write a packed matrix to a 4D NETCDF variable in an existing NETCDF file.
!
!   Arguments
!   _________
!
!
!       NAMPROC             (INPUT) character(len=*)
!                           Name of the calling procedure.
!
!       MAT                 (INPUT) real(stnd), dimension(:,:)
!                           The packed matrix.
!                                                                              
!       MASK                (INPUT) logical(lgl), dimension(:,:,:)
!                           The 3D mask array.                                                
!
!       PERIODICITY         (INPUT) integer(i4b)
!                           The periodicity of the packed matrix.
!
!       MISSING             (INPUT) real(stnd)
!                           The missing value indicator.
!
!       FILID               (INPUT) integer(i4b)
!                           NetCDF ID of the netCDF file.
!
!       VARID               (INPUT) integer(i4b)
!                           netCDF ID of the 4D netCDF variable.
!                                                                              
!                   
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*),          intent(in)  :: namproc
!
        real(stnd), dimension(:,:),  intent(in)  :: mat
        real(stnd),                  intent(in)  :: missing
!
        integer(i4b),  intent(in)  :: periodicity, filid, varid
!
        logical(lgl), dimension(:,:,:), intent(in) :: mask
!   
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)                :: i, i1, i2, k, l, nlon, nlat, nlevel, nvar, status, nperiodicity
        integer(i4b), dimension(4)  :: startwri
!
        real(stnd), dimension(size(mask,1),size(mask,2),size(mask,3)) :: temp
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        nlon   = size( mask, 1 )
        nlat   = size( mask, 2 )
        nlevel = size( mask, 3 )
!
        nperiodicity = size( mat, 1 )
        nvar         = size( mat, 2 )/periodicity
!
!       WRITE THE PACKED MATRIX.
!
        startwri(1:3)  = 1
!
        if ( periodicity==1 ) then
!
            do k = 1, nperiodicity
!
                startwri(4) = k
!
                temp(:nlon,:nlat,:nlevel) = unpack( mat(k,:nvar), mask(:nlon,:nlat,:nlevel), missing )
!
                status = nf90_put_var( filid, varid, temp(:nlon,:nlat,:nlevel), start=startwri )
                if ( status/=nf90_noerr ) call merror(  namproc//trim( nf90_strerror( status ) ) )
!
            end do
!
        else
!
            l = 0
!
            do k = 1, nperiodicity
!
                i2 = 0
!
                do i = 1, periodicity
!
                    i1          = i2 + 1
                    i2          = i2 + nvar
                    l           = l + 1
                    startwri(4) = l
!
                    temp(:nlon,:nlat,:nlevel) = unpack( mat(k,i1:i2), mask(:nlon,:nlat,:nlevel), missing )
!
                    status = nf90_put_var( filid, varid, temp(:nlon,:nlat,:nlevel), start=startwri )
                    if ( status/=nf90_noerr ) call merror(  namproc//trim( nf90_strerror( status ) ) )
!
                end do
!
            end do
!
        end if
!
!   END OF SUBROUTINE write_file_4d_rm
!   ___________________________________
!
        end subroutine write_file_4d_rm
!
!
! ===========================================================================================
!
!   THE FOLLOWING SUBROUTINES WILL BE INCLUDED IN THE NEXT RELEASE OF STATPACK
!
! ===========================================================================================
!
        subroutine days_between_dates( start_date, end_date, duration )
!
!   Purpose
!   _______
!
!   Calculate day number between two dates 
!
!   Arguments
!   _________
!
!       START_DATE             (INPUT) integer(i4b)
!                              YYYYMMDD date of beginning
!
!       END_DATE               (INPUT) integer(i4b)
!                              YYYYMMDD date of ending
!
!       DURATION               (OUTPUT) integer(i4b)
!                              day numbers between start and end date
!                                                                                                 
! _________________________________________________________________________________________________
!                                                                                                 
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
         integer(i4b), intent(in) :: start_date
!
         integer(i4b), intent(in) :: end_date
!
         integer(i4b), intent(out) :: duration
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
         integer(i4b) :: year, month, day, end_year, end_month, end_day, t_step
!
         integer(i4b), dimension(12) :: cal_month = 30
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
!       INITIAL DATE.
!
        year = start_date / 10000
        month = start_date / 100  - year * 100
        day = start_date - month * 100 - year * 10000
!
!       FINAL DATE.
!
        end_year = end_date / 10000
        end_month = end_date / 100  - end_year * 100
        end_day = end_date - end_month * 100 - end_year * 10000
!
!       EVALUATE MONTH DURATION.
!
        cal_month( 1 ) = 31
        cal_month( 3 ) = 31
        cal_month( 5 ) = 31
        cal_month( 7 ) = 31
        cal_month( 8 ) = 31
        cal_month( 10 ) = 31
        cal_month( 12 ) = 31
        cal_month( 2 ) = 28
!
!       TIME STEP.
!
        t_step = 1
!
!       CHANGE FEBRUARY MONTH DURATION IF LEAP YEAR.
!
        if ( ( modulo (year, 4) .eq. 0 .and. modulo (year, 100) .ne. 0 ) .or. modulo (year, 400) .eq. 0 )   &
        cal_month ( 2 ) = 29
!
!       LOOP OVER DAYS.
!
        do while ( year .lt. end_year .or.  ( ( month * 1000 + day ) .lt. ( end_month * 1000 + end_day ) ) )
!
            day = day + 1
!
!           INCREASE MONTH NUMBER.
!
            if ( day .gt. cal_month ( month ) ) then
                day = 1
                month = month + 1
!
!               INCREASE YEAR NUMBER.
!
                if ( month .gt. 12 ) then
                    month = 1
                    year  = year + 1
!
!                   CHANGE FEBRUARY MONTH DURATION IF LEAP YEAR.
!
                    if ( ( modulo (year, 4) .eq. 0 .and.                  &
                         modulo (year, 100) .ne. 0 ) .or. modulo (year, 400) .eq. 0 ) then
                        cal_month ( 2 ) = 29
                    else
                        cal_month ( 2 ) = 28
                    end if
                end if
!
            end if
!
!           DAY COUNT.
!
            t_step = t_step + 1
!
        end do
!
        duration = t_step - 1
!
!   END OF SUBROUTINE days_between_dates
!   ____________________________________
!
        end subroutine days_between_dates
!
! =========================================================================================
!
        subroutine char_date_to_int_date( char_date , int_date )
!
!   Purpose
!   _______
!
!       This subroutine convert date from character format to integer YYYYMMDD
!
!
!   Arguments
!   _________
!
!       CHAR_DATE       (INPUT) character(len=*)
!                        Date YYYY-MM-DD or DD-MM-YYYY formatted
!
!       INT_DATE        (OUTPUT) integer(i4b)
!                        Date YYYYMMDD formatted
!
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        character(len=*), intent(in)  :: char_date
!
        integer(i4b), intent(out)  :: int_date
!
!
!   SPECIFICATIONS FOR VARIABLES
!   ____________________________
!
        integer(i4b)      :: day, month, year, i_index, i_max, i_index2
!
        character(len=4)  :: first_char, second_char, third_char
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        i_max = len_trim( char_date )
!
        i_index = scan( char_date, "-")
        first_char = char_date(1:i_index-1)
        i_index2 = scan( char_date(i_index+1:i_max), "-")
        i_index2 = i_index2 + i_index
        second_char = char_date(i_index+1:i_index2-1)
        i_index = scan( char_date(i_index2+1:i_max), " ")
        i_index = i_index + i_index2
        third_char = char_date(i_index2+1:i_index-1)
!
        if ( len_trim(first_char)==4 ) then
           read( first_char, '(I4)' ) year
           read( third_char, '(I2)' ) day
        else
           read( first_char, '(I2)' ) day
           read( third_char, '(I4)' ) year
        end if
!
        read( second_char, '(I2)' ) month
!
        int_date = year * 10000 + month * 100 + day
!
!   END OF SUBROUTINE char_date_to_int_date
!   _______________________________________
!
        end subroutine char_date_to_int_date
!
!
! ===========================================================================================
!
! ********************************
! END OF MODULE NCSTAT_Utilities *
! ********************************
!    
end module NCSTAT_Utilities
