program comp_cor_4d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_cor_4d:
!
! ===========
! comp_cor_4d
! ===========
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 13/09/2018
!
! Purpose
! =======
!
! Compute correlation and regression coefficients between an index time series and a
! fourdimensional variable extracted from a NetCDF dataset and perform statistical tests
! on these correlation and regression coefficients.
!
! The procedure first transforms the input fourdimensional NetCDF variable
! as a *ntime* by *nv* rectangular matrix of observed variables stored columnwise (e.g. the
! selected cells of the 3-D grid-mesh associated with the fourdimensional NetCDF variable)
! and then computes measures of association between each of these variables, say *X*\ , and
! the input index time series, say *Y*\ .
!
! If *X*\ (:) and *Y*\ (:) are vectors of *ntime* observations, the (Pearson) 
! correlation coefficient of the two variables *X* and  *Y* is defined by
!
! **COR** = **COV**\ ( *X*\ , *Y*\  )/\ [ **STD**\ (*X*\ ). **STD**\ (*X*\ ) ]
!
! where
!
!  * **COV**\ ( *X*\ , *Y*\  )
!    is the covariance between *X* and  *Y* (e.g.
!    dot_product( *X*\ (:) - **MEAN**\ (*X*\ ) , *Y*\ (:) - **MEAN**\ (*Y*\ ) )/\ *ntime* )
!  * **MEAN**\ (*X*\ ) and  **MEAN**\ (*Y*\ ) 
!    are the means of *X* and  *Y*\ , respectively, as computed by :ref:`comp_clim_4d`.
!  * **STD**\ (*X*\ ) and  **STD**\ (*Y*\ ) 
!    are the standard deviations of *X* and  *Y*\ , respectively, as computed by :ref:`comp_clim_4d`.
!
! From this definition, it follows that the correlation is a pure number invariant under
! changes of scale and origin of the variables *X* and  *Y*\ . Furthermore, it can be
! shown that **COR** cannot be less than ``-1`` or greater than ``1``.
!
! The Pearson correlation coefficient can be used to assess if the relationship between two
! variables is more or less linear (e.g. if the the two variables are "proportional" to each other).
! More precisely, abs( **COR**\  ) = ``1``\  if and only if there exists a linear relationship
! between *X*\  and *Y*\ , e.g., if
!
! *X*\ (``i``\ ) = ``a``\. *Y*\ (``i``\) + ``b``\    , for ``i``\ = ``1`` to *ntime*
! , with ``a``\ > ``0``\  if **COR**\  = ``1``\ and ``a``\ < ``0``\  if **COR**\  = ``-1``\ .
!
! If the correlation coefficient is high in absolute value, but not equal to ``1``\  or ``-1``\ ,
! it can also be "summarized" by a straight line (sloped upwards or downwards), which is called
! the regression line  between *X*\  and *Y*\ . This line is also a least squares line, because
! it can be determined such that the sum of the squared distances of all the data points from 
! the line is the lowest possible in the scatter plot of *X*\  and *Y*\ .
!
! The coefficients ``a``\  and ``b``\  of the regression line are called, respectively, the
! regression and intercept coefficients for predicting the dependent variable *X* by the
! independent variable *Y*\ . These coefficients are computed by comp_cor_4d (intercept
! coefficients are computed only if the argument *-intercept* is specified). Furthermore,
! the regression lines for predicting the index time series *Y*\  from each of the
! *X*\  variables from the 3-D grid-mesh associated with the fourdimensional NetCDF variable can
! also be estimated if the optional argument **-rg=**\ ``reg2`` is specified.
!
! As mentioned before, the correlation coefficient **COR** represents the linear relationship
! between two variables. If this correlation coefficient is squared, then the resulting value
! (the coefficient of determination) will represent the proportion of common variation (or
! shared variance) between the two variables (i.e., the "strength" of the relationship). More
! precisely, this value gives you the percentage of variance of the dependent variable *X* explained
! by the independent variable *Y*\ . In order to evaluate the correlation between variables, it is
! important to know this "strength" as well as the significance of the correlation which can be
! assess with the help of statistical tests.
!
! If *X* and  *Y* are independently distributed, their covariance, and hence their correlation,
! is zero, but the converse is not generally true.
! However, in the case of *X* and  *Y* follow a bivariate normal distribution, the nullity of
! the correlation coefficient implies the independence of the variables *X* and  *Y*\ .
!
! Furthermore, in the case of a random sample of *ntime* observation pairs from such
! bivariate normal population with a zero correlation coefficient, the distribution of
! the variate
!
! **T** = **COR**\ .sqrt( [*ntime*\  - ``2``\ ] / [``1``\  - **COR**\ .\ **COR**\ ] )
!
! has a Student-Fisher t distribution with *ntime*\  - ``2`` degrees of freedom
! (call it t[*ntime*\ -``2``\ ] in what follows)  [vonStorch_Zwiers]_.
!
! Given the sample correlation **COR**\ , we can thus test the hypothesis
! of no correlation in the bivariate normal parent population with the help of the Student-Fisher t
! distribution [vonStorch_Zwiers]_. Monte Carlo simulations suggest that this test remains valid if
! the couple *X* and *Y*  does not follow a bivariate normal distribution and the number of observations
! is big enough (e.g. *ntime* > ``30``\ ); but in this case it is not a test of the independence
! of the two variables *X* and  *Y*\ .
!
! If the correlation coefficient in the parent distribution is not assumed to be zero, the distribution
! of **COR**\  has a complicated form. In that case, for example for computing confidence intervals
! for the correlation coefficient or testing if the parent correlation is some number different of zero,
! it is better to use the monotonic transformation
!
! **Z** = (``1``\ /``2``\ ).log( [ ``1``\  + **COR**\  ]/[``1``\  - **COR**\  ] )
!
! , which is called the Fisher z transformation and is the inverse of the hyperbolic tangent function.
! The Fisher z transform produces an asymptotically normal variate with variance equal
! to ``1``\ /(*ntime* - ``3``\ ) [vonStorch_Zwiers]_.
!
! The critical probabilities associated with the correlation coefficients are estimated by
!
! **PROB** = **P**\ ( abs( t[ *ntime*\ -``2``\  ] )>abs(  **T**\  ) )
!
! if the argument **-a=**\ ``student`` is specified when calling the procedure (this is
! the default value for this argument).
!
! If your sample of *ntime* observation pairs cannot be assumed to be normal, **PROB**
! can also be estimated by permutation methods (e.g. by using the argument **-a=**\ ``permute``
! when calling comp_cor_4d). More precisely, in the case of *ntime*\  independent and 
! identically distributed observations from a bivariate population of unknown form, we may
! consider the permutation of *Y*\  coordinates to test for the hypothesis of no correlation
! in the parent population. Let S( *X*\ , *Y*\  ) be the set of points obtained by permuting the
! coordinates of *Y*\ (:) in all *ntime*\ ! possible ways. Then, there is no correlation between
! *X*\  and *Y*\  in each element of S( *X*\  , *Y*\  ) since any permutation uniformly makes sets
! of uncorrelated data. Hence, by randomly permuting the order of the elements of the *Y*\ (:) vector
! and recomputing the correlation coefficient between *X*\ (:) and this permuted vector many times (as
! determined by the **-nb=**\ ``number_of_shuffles`` argument), we can estimate the permutation 
! distribution of abs( **COR**\  ), conditionally on the *Y*\ (:) vector,
! and compute critical probabilities **PROB** useful for testing the hypothesis of no correlation
! in the parent population (e.g. by counting the number of times the permuted values of abs( **COR**\  )
! exceed the magnitude of the original correlation in the observed sample).
!
! If your sample of *ntime* observation pairs cannot be assumed to be a random bivariate sample and
! the observations are auto-correlated in time, bootstrap procedures both in the time or frequency
! domains are available for estimating the critical probabilities, **PROB**\ , associated with the
! correlations coefficients [Davison_Hinkley]_ .
!
! If the argument **-a=**\ ``bootstrap`` a blockwise bootstrap procedure in the time domain is used
! to estimate the critical probabilities, **PROB**\ . This is useful when the observations are serially
! correlated. In this algorithm, we consider the population of subsamples or overlapping blocks of length
! *bootstrap_block_length* formed from the time observations in the  *Y*\ (:) vector. These overlapping
! blocks form a finite set defined by
!
! **BLK**\ (``i``\ ) = Y( ``i``\  + ``1``\  : ``i``\  + *bootstrap_block_length*\  )
! for ``i``\  = ``0``\  to  *ntime*\  - *bootstrap_block_length*
!
! Blockwise bootstrap is then realized by resampling randomly the blocks **BLK**\ (``i``\ )
! and gluing them together to form a kind of surrogate time series of length *ntime*\ . Finally,
! the correlation coefficient between *X*\ (:) and this surrogate time series is computed.
! This procedure is iterated many times, as determined by the **-nb=**\ ``number_of_shuffles`` argument,
! to estimate the bootstrap distribution of abs( **COR**\  ) and compute critical
! probabilities **PROB** useful for testing the hypothesis of no correlation in the parent
! population (e.g. by counting the number of times the bootstrap values of abs( **COR**\  ) exceed
! the magnitude of the original correlation in the observed sample). The **-bp=**\ , **-bs=**\  and 
! **-bl=**\  arguments allow the user to determine the exact form of the blockwise bootstrap
! algorithm. The **-bp=**\  argument is particularly useful if your time series are cyclostationary
! since it forces all the blocks to start at specific observations which are the same day, month or
! season. The **-bl=**\  argument allows the user to choose the size of the blocks.
! See the remarks below for more details.
!
! If the argument **-a=**\  is set to ``theiler`` or ``scramble``, a frequency bootstrap procedure is used
! to generate independent surrogate time series with the same spectral characteristics as the original
! time series. The basic idea of these methods is to create simulated series by manipulating the Discrete
! Fourier Transform (DFT) of a given time series instead of resampling randomly blocks of this series in the
! time domain. There are many different ways for how to do these manipulations
! in the literature. Two of these methods are currently implemented in comp_cor_3d, the Theiler method
! [Theiler_etal]_ [Ebisuzaki]_ and the Davison and Hinkley method [Davison_Hinkley]_ [Braun_Kulperger]_ .
! Both methods assume that the input time series are stationary (e.g. they do not contain pure harmonic
! components such as a seasonal cycle or a well-defined trend).
!
! The Theiler method (used when the argument **-a=**\  is set to ``theiler`` when calling comp_cor_4d) consists
! of randomly shifting the phases in the DFT of the *Y*\ (:) vector and back-transforms it to obtain a bootstrap
! sample in the time domain. As an illustration, assuming that *ntime*\  is odd and the DFT of *Y*\ (:) is the
! complex vector *Z*\ (:), then define the complex vector *O*\ (:) of *ntime*\  elements as
!
! * *O*\ ( ``1``\  ) = ``1``
!
! * *O*\ ( ``k``\  ) = exp( ``i``\ .\ ``2``\ .\ ``pi``\ .\ **U**\ (``k``\ ) )
!   for ``k``\  = ``2``\  to [ *ntime*\  + ``1``\  ] / ``2``
!
! * *O*\ ( ``k``\  ) = - *O*\ ( *ntime*\  - ``k``\  )
!   for ``k``\  = [ *ntime*\  + ``1``\  ] / ``2``\  + ``1`` to *ntime*
!
! , where ``i``\  = sqrt( ``-1``\  ) and the **U**\ (``k``\ ) are a random sample drawn from an uniform distribution
! on [``0``\ , ``1``\ ].
! Then, the Theiler surrogate series is obtained by multiplying element by element the complex vectors *Z*\ (:) and
! *O*\ (:) and, finally, taking the inverse DFT of this new complex vector to obtain a bootstrap sample in
! the time domain. By construction, this simulated series is real-valued, independent of the vector *X*\ (:),
! but its sample mean and periodogram are identical to those of the vector *Y*\ (:).
!
! The Davison and Hinkley approach (used when the argument **-a=**\  is set to ``scramble`` when calling comp_cor_4d)
! amounts also to a randomization of the phases of the Fourier coefficients of the DFT of *Y*\ (:), but also includes
! an additional step that modifies the amplitudes of these Fourier coefficients. Let, again, the vector **U**\ (:) be
! a random sample of *ntime*\  observations drawn from an uniform distribution on [``0``\ , ``1``\ ], and, define,
! the complex vector *O*\ (:) as
!
! * *O*\ ( ``1``\  ) = ``0``
!
! * *O*\ ( ``k``\  ) = *Z*\ ( ``k``\  ) . exp( ``i``\ .\ ``2``\ .\ ``pi``\ .\ **U**\ ( ``k``\  ) )
!   for ``k``\  = ``2``\  to *ntime*
!
! and form the complex vector *A*\ (:) as
!
! * *A*\ ( ``1``\  ) = *Z*\ ( ``1``\  )
!
! * *A*\ ( ``k``\  ) = sqrt( ``0.5``\  ) . [ *O*\ ( ``k``\  ) + conjg( *O*\ ( *ntime*\  - ``k``\  ) ) ]
!   for ``k``\  = ``2``\  to *ntime*
!
! Then, the Davison and Hinkley surrogate series is obtained by taking the inverse DFT of the
! complex vector *A*\ (:). Again, by construction, this simulated series is real-valued, independent
! of the vector *X*\ (:), with a sample mean identical to that of the vector *Y*\ (:), but, now its
! periodogram will be different from that of the vector *Y*\ (:). Note, however, that the mean spectrum
! obtained by averaging the periodograms of many Davison and Hinkley surrogate series will tend on average
! to the periodogram of the vector *Y*\ (:).
!
! Finally, Davison and Hinkley [Davison_Hinkley]_ also discuss how to generate surrogate datasets with non-Gaussian
! distributions. Their approach can be used both in the context of the Theiler and, Davison and Hinkley methods,
! and is used to derived critical probabilities if the argument **-a=**\  is set to ``theiler2`` or ``scramble2``\ .
! These variations of the original methods are useful when the observations are serially correlated and have
! also an asymmetric marginal distribution.
!
! For more details on both the Theiler or Davison and Hinkley surrogate methods, consult
! the references cited below.
!
! By default, comp_cor_4d computes the sample correlation and regression coefficients, the
! associated critical probabilities for testing the nullity of the correlation coefficients
! and the z transforms of the correlation coefficients between the index time series and each point
! in the time series of the 3-D grid-mesh associated with the input fourdimensional NetCDF
! variable. The intercept coefficients of the regression lines between *X*\  and *Y*\  are also 
! computed if the optional argument *-intercept* is specified when calling comp_cor_4d.
!
! Moreover, these statistics may be computed by taking into account the periodicity of the input
! fourdimensional NetCDF variable if you suspect that the time series are cyclostationary (by using
! the **-p=**\ ``periodicity`` argument when calling the procedure). All the results are finally stored
! in an output NetCDF dataset, after repacking the statistics on the original 3-D grid of the
! input fourdimensional NetCDF variable. 
!
! Finally, if the NetCDF variable is tridimensional use :ref:`comp_cor_3d` instead of comp_cor_4d.
!
! This procedure is parallelized if OpenMP is used. Moreover, this procedure computes the
! correlation and regression coefficients with only one pass through the data and an
! out-of-core strategy which is highly efficient on huge datasets.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_cor_3d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -vi=index_netcdf_variable \
!         -fi=input_index_netcdf_file              (optional) \
!         -m=input_mesh_mask_netcdf_file           (optional) \
!         -g=grid_type                             (optional : n, t, u, v, w, f) \
!         -x=lon1,lon2                             (optional) \
!         -y=lat1,lat2                             (optional) \
!         -z=level1,level2                         (optional) \
!         -t=time1,time2                           (optional) \
!         -p=periodicity                           (optional) \
!         -a=type_of_analysis                      (optional : student, permute, bootstrap, \
!                                                              theiler, theiler2, \
!                                                              scramble, scramble2) \
!         -rg=type_of_regression                   (optional : reg1, reg2) \
!         -o=output_netcdf_file                    (optional) \
!         -ti=itime1,itime2                        (optional) \
!         -pi=iperiodicity,istep                   (optional) \
!         -ni=index_for_2d_index_netcdf_variable   (optional) \
!         -nb=number_of_shuffles                   (optional) \
!         -bp=bootstrap_periodicity                (optional) \
!         -bs=bootstrap_season                     (optional) \
!         -bl=bootstrap_block_length               (optional) \
!         -sm=smoothing_factor                     (optional) \
!         -mi=missing_value                        (optional) \
!         -regstd                                  (optional) \
!         -intercept                               (optional) \
!         -double                                  (optional) \
!         -bigfile                                 (optional) \
!         -hdf5                                    (optional) \
!         -tlimited                                (optional)
!
! By default
! ----------
!
!        *-fi=*
!         the same as the *-f=* argument
!        *-m=* 
!         an *input_mesh_mask_netcdf_file* is not used
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 3-D grid-mesh associated with
!         the input *netcdf_variable* is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-z=*
!         the whole vertical resolution associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-p=*
!         the *periodicity* is set to ``1``. 
!        *-ti=*
!         the whole time period associated with the *index_netcdf_variable*
!        *-pi=*
!         this parameter is not used
!        *-ni=*
!         if the *index_netcdf_variable* is bidimensional, the first time series is used
!        *-a=*
!         *type_of_analysis* is set to ``student``
!        *-rg=*
!         *type_of_regression* is set to ``reg1``
!        *-nb=*
!         *number_of_shuffles* is set to ``99``
!        *-bp=*
!         this parameter is set ``1``\ . 
!        *-bs=*
!         this parameter is not used
!        *-bl=*
!         the *bootstrap_block_length* is set ``1``\ . 
!        *-sm=*
!         no smoothing is applied to the *index_netcdf_variable*
!        *-mi=*
!         the *missing_value* is set to ``1.e+20`` in the *output_netcdf_file*
!        *-o=*
!         *output_netcdf_file* name is set to ``cor_``\ *netcdf_variable*\ ``.``\ *index_netcdf_variable*\ ``.nc``
!        *-regstd*
!         the regression coefficients are computed in units
!         of the input NetCDF variables. If **-regstd** is activated,
!         the regression coefficients are computed in units of the
!         *netcdf_variable* by standard-deviation of the *index_netcdf_variable*
!        *-intercept*
!         the intercept coefficients of the regressions are
!         not computed. If **-intercept** is activated, the
!         intercept coefficients of the regressions are computed
!         and stored in the *output_netcdf_file*
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! --------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable for which a
!      correlation analysis must be computed and the **-f=**\ *input_netcdf_file* argument
!      specifies that this NetCDF variable must be extracted from the NetCDF file,
!      *input_netcdf_file*.
!
!   #) The optional argument **-m=**\ *input_mesh_mask_netcdf_file* specifies the land-sea mask
!      to apply to *netcdf_variable* for transforming this fourdimensional NetCDF variable as a
!      rectangular matrix of observed variables before computing the correlation analysis. By default,
!      it is assumed that each cell in the 3-D grid-mesh associated with the input fourdimensional
!      NetCDF variable is a valid time series (e.g. missing values are not present).
!
!      The geographical shapes and vertical resolution of the *netcdf_variable* (in the *input_netcdf_file*)
!      and the mask (in the *input_mesh_mask_netcdf_file*) must agree if an *input_mesh_mask_netcdf_file* is used.
!
!      Refer to :ref:`comp_clim_4d` or :ref:`comp_mask_4d` for creating a valid
!      *input_mesh_mask_netcdf_file* NetCDF file for regular or gaussian grids before
!      using comp_cor_4d.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that
!      the input NetCDF variable is from an experiment with the ORCA model (R2, R4 or R05 resolutions).
!      This argument is also used to determined the name of the mesh_mask_variable if
!      an *input_mesh_mask_netcdf_file* is used.
!
!   #) If the **-x=**\ *lon1,lon2*, **-y=**\ *lat1,lat2*  and **-z=**\ *level1,level2* arguments
!      are missing the whole geographical domain and vertical resolution associated with the
!      *netcdf_variable* is used.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_4d` for transforming geographical coordinates as indices or
!      generating an appropriate mesh-mask before using comp_cor_4d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, data in the whole time period
!      associated with the *netcdf_variable* is taken into account.
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!      The selected time period (e.g.  *time2* - *time1* + ``1``) must be a whole multiple of
!      the *periodicity* if the **-p=** argument is specified.
!
!   #) The **-p=**\ *periodicity* argument gives the periodicity of the input data for the
!      *netcdf_variable*. For example, with monthly data **-p=**\ ``12`` should be specified,
!      with yearly data **-p=**\ ``1`` may be used, etc.
!
!      Note that the output NetCDF file will have *periodicity* time observations.
!   
!   #) The **-vi=**\ *index_netcdf_variable* specifies a time series for the correlation
!      analysis.
!      If the **-vi=**\ *index_netcdf_variable* is present, the **-fi=** argument
!      must also be present and this argument specifies the NetCDF dataset
!      which contains the *index_netcdf_variable*. However, if the NetCDF dataset which
!      contains the *index_netcdf_variable* is the same as the NetCDF dataset specified
!      by the **-f=** argument, it is not necessary to specify the **-fi=** argument.
!
!   #) The **-ni=** argument specifies the index (e.g. an integer) for selecting the time series if
!      the *index_netcdf_variable*  specified in the **-vi=** argument is a 2D NetCDF variable. By default,
!      the first time series is used, which is equivalent to set *index_for_2d_netcdf_variable* to ``1``\ .
!
!   #) If the **-ti=**\ *itime1,itime2* argument is missing, data in the whole time period
!      associated with the *index_netcdf_variable* is taken into account.
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!   #) The **-pi=** argument gives the periodicity and select the time step for the
!      *index_netcdf_variable*. For example, to compute correlations with the January
!      monthly time series extracted from the *index_netcdf_variable* which is assumed
!      to be sampled every month, **-pi=**\ ``12,1`` should be specified, with yearly data
!      **-pi=**\ ``1,1`` may be used, etc.
!
!   #) The selected time periods for the *netcdf_variable* and *index_netcdf_variable*
!      must agree. This means that the following equality must be verified
!
!      ``(time2 - time1 + 1)/periodicity = ceiling((itime2 - itime1 - istep + 2)/iperiodicity)``,
!
!      otherwise, an error message will be issued and the program will stop.
!
!   #) The **-a=** argument selects the method for computing critical probabilities associated with
!      the correlation coefficients: 
!
!      * If **-a=**\ ``student``, a classical Student-Fisher t test is used.
!
!      * If **-a=**\ ``permute``, a permutation test is used.
!
!      * If **-a=**\ ``bootstrap``,  a moving block bootstrap test is used.
!
!      * If **-a=**\ ``theiler``, a phase-scrambled bootstrap (Theiler method) test is used.
!
!      * If **-a=**\ ``theiler2``, a phase-scrambled bootstrap (Theiler method) test is used,
!        but when phase-scrambling the index time series exact empirical
!        margins are used instead of normal margins as in the ``theiler``  option.
!
!      * If **-a=**\ ``scramble``, a phase-scrambled bootstrap (Davison-Hinkley method) test is used.
!
!      * If **-a=**\ ``scramble2``, a phase-scrambled bootstrap (Davison-Hinkley method) test is used,
!        but when phase-scrambling the index time series exact empirical
!        margins are used instead of normal margins as in the ``scramble`` option.
!
!   #) The **-nb=**\ *number_of_shuffles* argument specifies the number of shuffles for the phase-scrambled,
!      bootstrap or permutation tests
!      if **-a=**\ ``permute``, ``bootstrap``, ``theiler``, ``theiler2``, ``scramble`` or ``scramble2``.
!
!   #) The **-bp=**\ *bootstrap_periodicity* argument specifies that the index, ``i``\ , of the first
!      observation of each selected block in the moving block bootstrap algorithm verifies the condition
!      ``i`` = ``1``\  +\  *bootstrap_periodicity*\ .\ ``j`` where ``j`` is a random positive integer.
!      *bootstrap_periodicity* must be greater than zero and less than the length of the time series.
!      By default, *bootstrap_periodicity* is set to ``1``.
!
!   #) The **-bs=**\ *bootstrap_season* argument specifies that the input time series is a repetition
!      of the same season for different years and bootstrap_season specifies the length of the season.
!      *bootstrap_season* must be greater than zero and the length of the time series must be a multiple
!      of *bootstrap_season*.
!      If the optional argument *bootstrap_periodicity* is used, *bootstrap_season* must also be greater or
!      equal to *bootstrap_periodicity*.
!      By default, *bootstrap_season* is set to the length of the time series.
!
!   #) The **-bl=**\ *bootstrap_block_length* argument specifies the size of the blocks in the moving
!      block bootstrap algorithm. *bootstrap_block_length* must be greater than zero and less than the
!      length of the time series.
!      If the optional argument *bootstrap_periodicity* is used, *bootstrap_block_length* must also be
!      greater or equal to *bootstrap_periodicity*. Moreover, if the optional argument *bootstrap_season*
!      is used, *bootstrap_block_length* must also be less than *bootstrap_season*.
!      By default, *bootstrap_block_length* is set to ``1`` or to *bootstrap_periodicity* if this optional
!      argument is used.
!
!   #) The **-rg=** argument selects the method for computing the regression 
!      coefficients:
!
!      * If **-rg=**\ ``reg1``, the coefficients of the regression equation for predicting the *netcdf_variable*
!        by the *index_netcdf_variable* are computed. This is the default.
!
!      * If **-rg=**\ ``reg2``, the coefficients of the regression equation for predicting the *index_netcdf_variable*
!        by the *netcdf_variable* are computed.
!
!   #) The **-intercept** argument specifies that the intercept coefficients of the regression equation
!      must be computed and stored in the output NetCDF file.
!      By default, the intercept coefficients are not computed.
!
!   #) The **-regstd** argument specifies that the regression coefficients of the regression equation
!      must be expressed in terms of units of the input NetCDF variable by standard-deviation 
!      of the *index_netcdf_variable*.
!      By default, the regression coefficients are expressed in units of the input NetCDF variables.
!
!   #) **-sm=**\ *smoothing_factor* means that the time series associated with the
!      *index_netcdf_variable* (e.g. the **-vi=** argument) must be smoothed with
!      a moving average of approximately ``2``\ .\ *smoothing_factor* + ``1`` terms
!      before computing the correlations with the *netcdf_variable* (e.g. the **-v=** argument).
!      *smoothing_factor* must be a strictly positive integer (>zero).
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the netcdf_variables in the *output_netcdf_file*.
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``\ .
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the data has no missing values.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on correlation and regression analysis in the climate literature, see
!
!      * "Statistical Analysis in Climate Research",
!        by von Storch, H., and Zwiers, F.W., Cambridge University press,
!        Cambridge, UK, Chapter 8, 484 pp., 2002. ISBN: `9780521012300`_
!
!      For more details on frequency or time series bootstrap procedures, see
!
!      * "Testing for nonlinearity in time series: the method of surrogate data."
!        by Theiler, J., Eubank, S., Longtin, A., Galdrikian, B., and Farmer, J.D.
!        Physica D, vol. 58, 77-94, 1992. doi: `10.1016/0167-2789(92)90102-s`_
!
!      * "A method to estimate the statistical significance of a
!        correlation when the data are serially correlated",
!        by Ebisuzaki, W.,
!        Journal of climate, vol. 10, 2147-2153, 1997. doi: `10.1175/1520-0442(1997)010<2147:AMTETS>2.0.CO;2`_
!
!      * "Properties of a fourier bootstrap method for time series",
!        by Braun, W.J., and Kulperger, R.J.,
!        Communications in Statistics - Theory and Methods, vol 26, 1329-1336, 1997. doi: `10.1080/03610929708831985`_
!
!      * "Bootstrap methods and their application",
!        by Davison, A.C., and Hinkley, D.V.,
!        Cambridge Univesity press, Cambridge, UK, 1997. doi: `10.1017/CBO9780511802843`_
!
! .. _`9780521012300` : http://www.cambridge.org/gb/academic/subjects/earth-and-environmental-science/climatology-and-climate-change/statistical-analysis-climate-research?format=PB
! .. _`10.1016/0167-2789(92)90102-s` : http://dx.doi.org/10.1016/0167-2789(92)90102-s
! .. _`10.1175/1520-0442(1997)010<2147:AMTETS>2.0.CO;2` : https://doi.org/10.1175/1520-0442(1997)010<2147:AMTETS>2.0.CO;2
! .. _`10.1080/03610929708831985` : https://doi.org/10.1080/03610929708831985
! .. _`10.1017/CBO9780511802843` :  https://doi.org/10.1017/CBO9780511802843
!
! Outputs
! -------
!
!   comp_cor_4d creates an output NetCDF file that contains the correlation and regression statistics and
!   critical probabilities associated with these coefficients, taking into account eventually the periodicity
!   of the data as determined by the **-p=**\ *periodicity* argument. The output NetCDF dataset contains the
!   following NetCDF variables (in the description below, nlev, nlat and nlon are the length of the vertical
!   and spatial dimensions of the input NetCDF variable) and *periodicity* time observations, if **-rg=**\ ``reg1``\  :
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _cor\ ``(periodicity,nlev,nlat,nlon)`` : the Pearson correlation coefficients between 
!      each point in the time series of the 3-D grid-mesh associated with the input NetCDF variable and the
!      *index_netcdf_variable* time series.
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _prob\ ``(periodicity,nlev,nlat,nlon)`` : the critical probabilities associated with
!      two-sided tests of the correlation coefficients (e.g. the absolute value of the correlation is tested).
!      These critical probabilities are computed under the null hypothesis that the corresponding correlation
!      coefficients in the parent population are zero.
!
!      The **-a=**\ *type_of_analysis* argument determines how these critical probabilities are
!      computed.
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _z\ ``(periodicity,nlev,nlat,nlon)`` : the Fisher z Transforms of the correlation
!      coefficients for each point in the time series of the 3-D grid-mesh associated with the input
!      NetCDF variable and the *index_netcdf_variable* time series.
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _reg\ ``(periodicity,nlev,nlat,nlon)`` : the regression coefficients for predicting 
!      each point in the time series of the 3-D grid-mesh associated with the input NetCDF variable by the
!      *index_netcdf_variable* time series.
!
!      By default, the regression coefficients are expressed in units of the input NetCDF variable
!      *netcdf_variable* by unit of the *index_netcdf_variable* time series.
!      However, if the **-regstd** argument is specified the regression coefficients are expressed in terms of
!      units of the input NetCDF variable *netcdf_variable* by standard-deviation of the *index_netcdf_variable*
!      time series.
!      Finally, if  **-rg**\ =\ ``reg2`` is specified the roles of the input NetCDF variables *netcdf_variable* and
!      *index_netcdf_variable* are interchanged and the fitted regression models are for predicting the
!      *index_netcdf_variable* by each time series of the 3-D grid-mesh associated with the input
!      NetCDF variable *netcdf_variable* .
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _int\ ``(periodicity,nlev,nlat,nlon)`` : the intercept coefficients in the regression
!      models for predicting each point in the time series of the 3-D grid-mesh associated with the input
!      NetCDF variable by the *index_netcdf_variable* time series.
!
!      This variable is stored only if the **-intercept** argument has been specified when calling
!      comp_cor_4d.
!      Finally, if  **-rg**\ =\ ``reg2`` is specified the roles of the input NetCDF variables *netcdf_variable* and
!      *index_netcdf_variable* are interchanged and the fitted regression models are for predicting the
!      *index_netcdf_variable* by each time series of the 3-D grid-mesh associated with the input NetCDF
!      variable *netcdf_variable* .
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _nobs\ ``(periodicity)`` : the number of observations used to compute
!      the correlation and regression coefficients.
!
!   All these statistics, excepted the *netcdf_variable*\ _\ *index_netcdf_variable*\ _nobs\  variable, are packed in fourdimensional
!   variables whose first, second and third dimensions are exactly the same as those associated with the input NetCDF variable
!   *netcdf_variable* even if you restrict the geographical domain with the **-x=**\ , **-y=** and **-z=** arguments. However,
!   outside the selected domain, the output NetCDF variables are filled with missing values.
!
!   If **-rg=**\ ``reg2``\  , the naming convention for the variables is reversed, the *index_netcdf_variable* will be listed first
!   and the *netcdf_variable* will appear after. For example, the name of the NetCDF variable storing the correlation coefficient will
!   be *index_netcdf_variable*\ _\ *netcdf_variable*\ _cor\  instead of *netcdf_variable*\ _\ *index_netcdf_variable*\ _cor\  if
!   **-rg=**\ ``reg2``\  .
!
! Examples
! --------
!
!   #)   For computing monthly lead correlations from a fourdimensional NetCDF variable
!        ``votemper`` in the NetCDF file ``ST7_1m_00101_20012_grid_T_votemper.nc``
!        and a December-January Nino34 SST index in the NetCDF file
!        ``ST7_sst_nino34_dj.nc`` and store the results in a NetCDF file named
!        ``cor_ST7_1m_votemper_nino34_dj_grid_T.nc``, use the following
!        commands (note that the critical probabilities associated with the correlations
!        are estimated with the help of the Theiler method using ``999`` surrogate
!        time series and cyclostationarity is assumed for the ``sosstsst`` variable
!        since  **-p**\ =\ ``12`` is specified ) :
!
!        ::
!
!          $ comp_cor_4d \
!            -f=ST7_1m_00101_20012_grid_T_votemper.nc \
!            -v=votemper \
!            -m=mesh_mask_ST7.nc \
!            -p=12 \
!            -fi=sst_nino34_dj.nc \
!            -vi=sosstsst \
!            -a=theiler \
!            -nb=999 \
!            -o=cor_ST7_1m_votemper_nino34_dj_grid_T.nc
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,    only : nf90_share
    use Statpack,  only : omp_limit, update_cor
    use omp_lib,   only : omp_get_num_procs, omp_get_max_threads, omp_get_thread_num
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,   &
                         nf90_int, nf90_sizehint_default, nf90_open,                               &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,      &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,    &
                         nf90_def_var, nf90_inq_varid, nf90_inq_attname, nf90_copy_att,            &
                         nf90_put_att, nf90_enddef, nf90_global, nf90_put_var, nf90_close
!
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, lamch, kfix, kexp,     & 
                         kint, tolower, case_change, string_to_val, val_to_string,                 &
                         allocate_error, merror, arth, comp_cor, permute_cor, phase_scramble_cor,  &
                         comp_smooth, bootstrap_cor
 !
   use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,                               &
                                get_listarg, get_variable_dims, get_variable_dimcoords,            &
                                write_file_4d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_cor_4d : ',                     &
!
                  cmethod(7)=(/"student  ","permute  ","scramble ","scramble2",    &
                               "theiler  ","theiler2 ", "bootstrap"/) ,            &
!
                  cregression(2)=(/"reg1","reg2"/)                 
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncidi, ncid_mask, ncid_cor, periodid,                  &
                     varid, maskid, varidi, corid, regid, intid, probid, zid, nobsid,           &
                     i, i1, i2, j, j1, j2, k, l, nlon, nlat, nlevel, nt, nidim, nti, nvar, nn,  &
                     lon1, lon2, lon3, lat1, lat2, lat3, level1, level2, level3,                &
                     time1, time2, time3, itime1, itime2, itime3,                               &
                     period, nperiod, iperiod, istep, niperiod, iindex, method, regression,     &
                     nrep, smooth_factor, output_type, output_mode, block_size, bseason,        &
                     bperiodicity
    integer(i4b), dimension(4)                 :: coor_id2, startlec, countlec
    integer(i4b), dimension(5)                 :: id_var
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:,:,:), allocatable  :: temp
    real(stnd), dimension(:,:),   allocatable  :: x, xstat
    real(stnd), dimension(:),     allocatable  :: y, xycor, xyreg, xyint, prob, z
    real(stnd), dimension(2)                   :: ystat
    real(stnd)                                 :: missing, tmp, tmp2, xyn
!
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                                :: first, use_tlimited, use_x, use_y, use_z,      &
                                                   use_std, use_int, use_mask, create_hdf5
    logical(lgl), dimension(nf90_max_var_dims)  :: use_coor_variable
    logical(lgl), dimension(:,:,:), allocatable :: mask
!
    character(len=nchar)                       :: name, name2, name3, history
    character(len=nf90_max_name)               :: fic_name, var_name, mask_name, fici_name, vari_name,   &
                                                  cor_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
    character                                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
    integer(i4b)  :: ncid2, varid2, ioffset
!
    real(stnd), dimension(:,:),   allocatable  :: xstat2
    real(stnd), dimension(:),     allocatable  :: xycor2
    real(stnd), dimension(2)                   :: ystat2
    REAL(stnd)                                 :: xyn2
!
    logical  :: test_par
#endif
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute correlation coefficients between a time series index'
        write (prtunit,*) ' and a fourdimensional variable extracted from a NetCDF dataset'
        write (prtunit,*) ' and perform statistical tests on these coefficients'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_cor_4d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -vi=index_netcdf_variable'
        write (prtunit,*) '        -fi=input_index_netcdf_file             (optional)'
        write (prtunit,*) '        -m=input_mesh_mask_netcdf_file          (optional)'
        write (prtunit,*) '        -g=grid_type                            (optional : n, t, u, v, w, f)'
        write (prtunit,*) '        -x=lon1,lon2                            (optional)'
        write (prtunit,*) '        -y=lat1,lat2                            (optional)'
        write (prtunit,*) '        -z=level1,level2                        (optional)'
        write (prtunit,*) '        -t=time1,time2                          (optional)'
        write (prtunit,*) '        -p=periodicity                          (optional)'
        write (prtunit,*) '        -a=type_of_analysis                     (optional : student, permute, bootstrap, scramble,'
        write (prtunit,*) '                                                            scramble2, theiler, theiler2, bootstrap)'
        write (prtunit,*) '        -rg=type_of_regression                  (optional : reg1, reg2)'
        write (prtunit,*) '        -o=output_netcdf_file                   (optional)'
        write (prtunit,*) '        -ti=itime1,itime2                       (optional)'
        write (prtunit,*) '        -pi=iperiodicity,istep                  (optional)'
        write (prtunit,*) '        -ni=index_for_2d_index_netcdf_variable  (optional)'
        write (prtunit,*) '        -nb=number_of_shuffles                  (optional)'
        write (prtunit,*) '        -bp=bootstrap_periodicity               (optional)'
        write (prtunit,*) '        -bs=bootstrap_season                    (optional)'
        write (prtunit,*) '        -bl=bootstrap_block_length              (optional)'
        write (prtunit,*) '        -sm=smoothing_factor                    (optional)'
        write (prtunit,*) '        -mi=missing_value                       (optional)'
        write (prtunit,*) '        -regstd                                 (optional)'
        write (prtunit,*) '        -double                                 (optional)'
        write (prtunit,*) '        -intercept                              (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '        -bigfile                                (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                                   (optional)'
#endif
        write (prtunit,*) '        -tlimited                               (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-fi= the same as the -f= argument'
        write (prtunit,*) '-m= an input_mesh_mask_netcdf_file is not used'
        write (prtunit,*) '-g= the grid type is set to n'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-z= the whole vertical resolution associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-p= the periodicity is set to 1'
        write (prtunit,*) '-ti= the whole time period  associated with the index_netcdf_variable'
        write (prtunit,*) '-pi=1,1 : by default, this parameter is not used'
        write (prtunit,*) '-ni=1   : by default, if the index_netcdf_variable is bidimensional,'
        write (prtunit,*) '          the first time series is used'
        write (prtunit,*) '-a=student'
        write (prtunit,*) '-rg=reg1'
        write (prtunit,*) '-nb=99'
        write (prtunit,*) '-bp=1'
        write (prtunit,*) '-bs=by default, this parameter is not used'
        write (prtunit,*) '-bl=1'
        write (prtunit,*) '-sm= by default, no smoothing is applied to the index_netcdf_variable'
        write (prtunit,*) '-mi= by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-o=cor_netcdf_variable.index_netcdf_variable.nc'
        write (prtunit,*) '-regstd   : by default, the regression coefficients are computed in units'
        write (prtunit,*) '            of the input netcdf variables'
        write (prtunit,*) '-intercept: by default, the intercept coefficients of the regressions are'
        write (prtunit,*) '            not computed'
        write (prtunit,*) '-double   : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile : by default, a netCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
#ifdef _OPENMP
        write (prtunit,*)
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>30 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name   = " "
    var_name   = " "
    fici_name  = " "
    vari_name  = " "
    mask_name  = " "
    cor_name   = " "
!
    history = "comp_cor_4d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
    method        = 1
    regression    = 1
    nrep          = 99
    period        = 1
    nperiod       = 0
    smooth_factor = 0
!
    missing = def_missing
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    level1    = 1
    level2    = huge( level2 )
    time1     = 1
    time2     = huge( time2 )
!
    itime1    = 1
    itime2    = huge( itime2 )
    iperiod   = 1
    istep     = 1
    niperiod  = 0
    iindex    = 1
!
    bperiodicity = 0
    bseason      = 0
    block_size   = 0
!
    use_mask     = false
    use_std      = false
    use_int      = false
    use_x        = false
    use_y        = false
    use_z        = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-z=')
!
!               GET THE LEVEL DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "level1 and level2 are integer values", &
                                  "Wrong level definition",               &
                                  level1, level2 )
!
                use_z = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
                use_mask = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                cor_name = name(4:)
!
            case ('-p=')
!
!               GET THE PERIODICITY.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The periodicity is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF TEST ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case ( cmethod(3) )
                        method = 3
                    case ( cmethod(4) )
                        method = 4
                    case ( cmethod(5) )
                        method = 5
                    case ( cmethod(6) )
                        method = 6
                    case ( cmethod(7) )
                        method = 7
                    case default
                        call merror( name_proc//"Unknown test analysis option" )
                end select
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-fi=')
!
!                       GET THE NAME OF THE NETCDF DATASET FOR THE INDEX VARIABLE.
!
                        fici_name = name(5:)
!
                    case ('-vi=')
!
!                       GET THE NAME OF THE NETCDF INDEX VARIABLE IN THE DATASET.
!
                        vari_name = name(5:)
!
                    case ('-ti=')
!
!                       GET THE PERIOD DEFINITION FOR THE INDEX VARIABLE.
!
                        call get_listarg( name_proc, name(5:),                            &
                                          "itime1 and itime2 are integer values",         &
                                          "Wrong period definition for index variable",   &
                                          itime1, itime2 )
!
                    case ('-pi=')
!
!                       GET THE PERIODICITY AND THE TIME STEP FOR THE INDEX.
!
                        call get_listarg( name_proc, name(5:),                                 &
                                          "iperiod and istep are integer values",              &
                                          "Wrong periodicity definition for index variable",   &
                                          iperiod, istep )
!
                        if ( iperiod<=0 .or. istep<=0 .or. istep>iperiod )   &
                call merror( name_proc//"The -pi= parameters are positive integers with istep<=iperiod" )
!
                    case ('-ni=')
!
!                       GET THE INDEX FOR A 2D INDEX VARIABLE.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) iindex
                        end if
                        if ( kcode/=kint .or. iindex<=0  )                           &
                        call merror( name_proc//"The index for a 2D index variable is a positive integer value" )
!
                    case ('-rg=')
!
!                       DETERMINE THE TYPE OF REGRESSION ANALYSIS.
!
                        call case_change( name(5:), tolower )
!
                        select case ( name(5:) )
                            case ( cregression(1) )
                                regression = 1
                            case ( cregression(2) )
                                regression = 2
                            case default
                                call merror( name_proc//"Unknown regression analysis option" )
                        end select
!
                    case ('-nb=')
!
!                       GET THE NUMBER OF SHUFFLES FOR THE PHASE-SCRAMBLING, BOOTSTRAP OR PERMUTATION TESTS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) nrep
                        end if
                        if ( kcode/=kint .or. nrep<=0  )                                      &
                        call merror( name_proc//"The number of shuffles is a positive integer value" )
!
                    case ('-sm=')
!
!                       DETERMINE THE SMOOTHING FACTOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) smooth_factor
                        end if
                        if ( smooth_factor<=0 )        &
         call merror( name_proc//"The -sm= argument is a strictly positive integer value " )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case ('-bp=')
!
!                       DETERMINE THE BOOTSTRAP PERIODICITY.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) bperiodicity
                        end if
                        if ( bperiodicity<=0 )        &
         call merror( name_proc//"The -bp= argument is a strictly positive integer value " )
!
                    case ('-bs=')
!
!                       DETERMINE THE BOOTSTRAP SEASON.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) bseason
                        end if
                        if ( bseason<=0 )        &
         call merror( name_proc//"The -bs= argument is a strictly positive integer value " )
!
                    case ('-bl=')
!
!                       DETERMINE THE BOOTSTRAP BLOCK LENGTH.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) block_size
                        end if
                        if ( block_size<=0 )        &
         call merror( name_proc//"The -bl= argument is a strictly positive integer value " )
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-double')
!
!                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
                            case ('-regstd')
!
!                               COMPUTE THE REGRESSION COEFFICIENTS IN TERMS OF STDS.
!
                                use_std = true
!
                            case ('-intercept')
!
!                               COMPUTE THE INTERCEPT COEFFICIENTS.
!
                                use_int = true
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( fici_name==" " ) fici_name = fic_name
!
    if ( vari_name==" " )                                                 &
    call merror( name_proc//"Input index NETCDF variable is missing" )
!
    if ( cor_name==" " ) cor_name="cor_"//trim(var_name)//"."//trim(vari_name)//".nc"
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr )                                                &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_mask ) then
        status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( fici_name/=fic_name ) then
        status = nf90_open( trim(fici_name), nf90_nowrite, ncidi )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    else
        ncidi = ncid
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                    &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names, &
                                  dimlens_out=dimlens, coor_id_out=coor_id,           &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type, &
                                  use_coor_variable_out=use_coor_variable             )
!
    if ( count(dimlens/=0)/=4 )           &
    call merror( name_proc//"Input NETCDF variable is not fourdimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nlevel = dimlens(3)
    nt     = dimlens(4)
!
    if ( use_mask ) then
!
        chr = merge( 't', grid_type, grid_type=='w' )
!
!       GET THE MASK ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
        call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
        call merror( name_proc//"Shape disagreement with the mask" )
!
        if ( .not.use_x ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
            if ( status==nf90_noerr ) lon1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
            if ( status==nf90_noerr ) lon2 = i
!
        end if
!
        if ( .not.use_y ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
            if ( status==nf90_noerr ) lat1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
            if ( status==nf90_noerr ) lat2 = i
!
        end if
!
        if ( .not.use_z ) then
!
            status = nf90_get_att( ncid_mask, maskid, "level1_First_level", i )
            if ( status==nf90_noerr ) level1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "level2_Last_level", i )
            if ( status==nf90_noerr ) level2 = i
!
        end if
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( lon1<0 ) then
        lon2 = max( lon2, 1 )
        lon3 = nlon + lon1 - lon2 + 1
    else
        lon1 = max( lon1, 1 )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
    end if
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    if ( lon1<0 ) then
       lon3 = nlon
       j1   = 1
       j2   = nlon
    else
       j1 = lon1
       j2 = lon2
    end if
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE LEVEL DEFINITION.
!
    level1 = max( level1, 1 )
    level2 = min( level2, nlevel )
    level3 = level2 - level1 + 1
    if ( level3<=0  ) call merror( name_proc//"Wrong vertical definition" )
!
!   CHECK THE TIME DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    if ( period>time3 )  call merror( name_proc//"-t=, -p= options mismatch" )
!
    nperiod = time3/period
    if ( time3/=nperiod*period )      &
    call merror( name_proc//"The selected time period is not a multiple of the periodicity" )
!
    if ( nperiod<=2 )                 &
    call merror( name_proc//"Not enough time observations" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat,nlevel), mask(nlon,nlat,nlevel), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    if ( use_mask ) then
!
!       GET THE MASK.
!
        status = nf90_get_var( ncid_mask, maskid, temp )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DETERMINE THE GRID POINTS USED IN THE CORRELATION ANALYSIS FROM THE MASK.
!
        mask(:nlon,:nlat,:nlevel) = temp(:nlon,:nlat,:nlevel)==one
!
    else
!
!       CONSTRUCT A DUMMY MASK.
!
        mask(:nlon,:nlat,:nlevel) = true
!
    end if
!
!   SELECT THE DOMAIN.
!
    if ( lon1>0 ) then
        mask(1:lon1-1,:nlat,:nlevel)    = false
        mask(lon2+1:nlon,:nlat,:nlevel) = false
    else
        mask(lon2+1:nlon+lon1,:nlat,:nlevel) = false
    end if
!
    mask(:nlon,1:lat1-1,:nlevel)    = false
    mask(:nlon,lat2+1:nlat,:nlevel) = false
!
!   SELECT THE VERTICAL RESOLUTION.
!
    mask(:nlon,:nlat,1:level1-1)      = false
    mask(:nlon,:nlat,level2+1:nlevel) = false
!
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   GET THE INDEX VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS FOR THE NETCDF INDEX VARIABLE.
!
    call get_variable_dims( name_proc, trim(vari_name), ncidi,    &
                            varidi, dimids, dimlens )
!
    nidim = count(dimlens/=0)
    if ( nidim/=1 .and. nidim/=2 )     &
    call merror( name_proc//"Input index NETCDF variable is not uni or bidimensional" )
!
    if ( nidim==2 ) then
        if ( iindex>dimlens(1) )     &
        call merror( name_proc//"Selected index for 2D index_variable is out of range " )
        nti = dimlens(2)
    else 
        nti = dimlens(1)
    end if
!
!   CHECK THE TIME DEFINITION FOR THE INDEX VARIABLE.
!
    itime1 = max( itime1, 1 )
    itime1 = itime1 + istep - 1
    itime2 = min( itime2, nti )
    itime3 = itime2 - itime1 + 1
    if ( itime3<=0  ) call merror( name_proc//"Wrong time definition for index NETCDF variable" )
!
!   CHECK THE PERIOD DEFINITION FOR THE INDEX VARIABLE.
!
    if ( iperiod>itime3 )  call merror( name_proc//"-ti=, -pi= options mismatch" )
!
    tmp      = real(itime3,stnd)/real(iperiod,stnd)
    niperiod = ceiling( tmp )
!
    if ( niperiod/=nperiod )  call merror( name_proc//"-t=, -p=, -ti=, -pi= options mismatch" )
!
!   ALLOCATE WORK ARRAYS.
!
    nn = period*nvar
    allocate( y(nperiod), xstat(nn,2), xycor(nn),       & 
              xyreg(nn), prob(nn), z(nn), stat=status   )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET INDEX VARIABLE.
!
    if ( nidim==1 ) then
        status = nf90_get_var( ncidi, varidi, y(:nperiod),                               &   
                               start=(/ itime1 /), count=(/ nperiod /), stride=(/ iperiod /) )
    else 
        status = nf90_get_var( ncidi, varidi, y(:nperiod),                               &   
                               start=(/ iindex, itime1 /), count=(/ 1, nperiod /), stride=(/ 1, iperiod /) )
    end if
!
    if ( status/=nf90_noerr )                                     &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE INDEX NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncidi, varidi, "missing_value", tmp )
    first = status==nf90_noerr
!
    if ( .not.first ) then
        status = nf90_get_att( ncidi, varidi, "_FillValue", tmp )
        first = status==nf90_noerr
    end if
!
    if ( first ) then
        if ( any( y(:nperiod)==tmp ) )      &
        call merror( name_proc//"Missing values are not allowed" )
    end if
!
!   SMOOTH THE INDEX TIME SERIES IF REQUIRED.
!
    if ( smooth_factor/=0 ) then
        call comp_smooth( y(:nperiod), smooth_factor )
    end if
!
    countlec(1)  = lon3
    countlec(2)  = lat3
    countlec(3)  = level3
    countlec(4)  = 1
!
#ifdef _PARALLEL_READ
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    test_par = i>1 .and. j>1 .and. time3*nvar>omp_limit .and. nperiod>=j
!
    if ( test_par ) then
!
        ioffset = time1 - 1
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
    end if
!
#endif
!
    if ( method==1 ) then
!
!       USE CLASSICAL STUDENT T TEST FOR COMPUTING CRITICAL PROBABILITIES.
!
#ifdef _PARALLEL_READ
!
        if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,l,k,i1,i2,status,startlec,ncid2,varid2,temp,x,xstat2,ystat2,xycor2,xyn2,first)      &
!$OMP         ,SHARED(y,mask,xstat,ystat,xycor,xyn,countlec)
!
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( temp(lon3,lat3,level3), x(nn,1), xstat2(nn,2), xycor2(nn), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!           OPEN THE NETCDF DATASET.
!
            status = nf90_open( trim(fic_name), nf90_share, ncid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE VARIABLE ID FROM ITS NAME.
!
            status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
            startlec(1)  = j1
            startlec(2)  = lat1
            startlec(3)  = level1
!
            first        = true
!
!$OMP DO SCHEDULE(STATIC)
            do i = 1, nperiod
!
                l = ioffset + (i-1)*period
                i2 = 0
!
                do k = 1, period
!
                    i1          = i2 + 1
                    i2          = i2 + nvar
                    l           = l + 1
                    startlec(4) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                    status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3,:level3), start=startlec, count=countlec )
                                           
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA.
!
                    x(i1:i2,1) = pack( temp(:lon3,:lat3,:level3), mask(j1:j2,lat1:lat2,level1:level2) )
!
                end do
!
                call comp_cor( x(:nn,1:1), y(i:i), first, false,             &
                               xstat2(:nn,:2), ystat2(:2), xycor2(:nn), xyn2 )
!
                first = false
!
            end do
!$OMP END DO
!
!           THE SUMS OF SQUARES AND CROSSPRODUCTS HAVE BEEN CALCULATED ON DIFFERENT CHUNKS OF DATA,
!           NOW THE RESULTS ARE MERGED TOGETHER.
!$OMP MASTER
            xstat(:nn,:2)  = xstat2(:nn,:2)
            ystat(:2)      = ystat2(:2)
            xycor(:nn)     = xycor2(:nn)
            xyn            = xyn2
!$OMP END MASTER
!
!$OMP BARRIER
!
!$OMP CRITICAL (sumstat)
            if ( omp_get_thread_num()/=0 ) then
!
                call update_cor( xstat, ystat, xycor, xyn, xstat2, ystat2, xycor2, xyn2 )
!
            end if
!$OMP END CRITICAL (sumstat)
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( temp, x, xstat2, xycor2 )
!
!$OMP CRITICAL (closenc)
!
!           CLOSE THE NETCDF DATASET.
!
            status = nf90_close( ncid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
!           REALLOCATE WORK ARRAY.
!
            allocate( x(nn,1), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
        else
!
#endif
!
!           ALLOCATE WORK ARRAY.
!
            allocate( x(nn,1), stat=status   )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           COMPUTE THE CORRELATIONS.
!
            l     = time1 - 1
            first = true
!
            startlec(1)  = j1
            startlec(2)  = lat1
            startlec(3)  = level1
!
            do i = 1, nperiod
!
                i2 = 0
!
                do k = 1, period
!
                    i1          = i2 + 1
                    i2          = i2 + nvar
                    l           = l + 1
                    startlec(4) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                    status = nf90_get_var( ncid, varid, temp(:lon3,:lat3,:level3),     &
                                           start=startlec, count=countlec      )     
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA.
!
                    x(i1:i2,1) = pack( temp(:lon3,:lat3,:level3), mask(j1:j2,lat1:lat2,level1:level2) )
!
                end do
!
                call comp_cor( x(:nn,1:1), y(i:i), first, false,           &
                               xstat(:nn,:2), ystat(:2), xycor(:nn), xyn  )
!
                first = false
!
            end do
!
#ifdef _PARALLEL_READ
!
        end if
!
#endif
!
!       NOW FINISH THE COMPUTATIONS OF THE CORRELATIONS.
!
        call comp_cor( x(:nn,1:0), y(1:0), false, true,            &
                       xstat(:nn,:2), ystat(:2), xycor(:nn), xyn,  &
                       prob=prob(:nn), z=z(:nn)                    )
!
    else
!
!       USE A PHASE-SCRAMBLED BOOTSTRAP, PERMUTATION OR MOVING BLOCK BOOTSTRAP TEST
!       FOR COMPUTING CRITICAL PROBABILITIES.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( x(nn,nperiod), stat=status   )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
#ifdef _PARALLEL_READ
!
        if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,l,k,i1,i2,status,startlec,ncid2,varid2,temp,xstat2,ystat2,xycor2,xyn2,first)      &
!$OMP         ,SHARED(y,x,mask,xstat,ystat,xycor,xyn)
!
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( temp(lon3,lat3,level3), xstat2(nn,2), xycor2(nn), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!           OPEN THE NETCDF DATASET.
!
            status = nf90_open( trim(fic_name), nf90_share, ncid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE VARIABLE ID FROM ITS NAME.
!
            status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
            startlec(1)  = j1
            startlec(2)  = lat1
            startlec(3)  = level1
!
            first        = true
!
!$OMP DO SCHEDULE(STATIC)
            do i = 1, nperiod
!
                l = ioffset + (i-1)*period
                i2 = 0
!
                do k = 1, period
!
                    i1          = i2 + 1
                    i2          = i2 + nvar
                    l           = l + 1
                    startlec(4) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                    status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3,:level3), start=startlec, count=countlec )
                                           
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA.
!
                    x(i1:i2,i) = pack( temp(:lon3,:lat3,:level3), mask(j1:j2,lat1:lat2,level1:level2) )
!
                end do
!
                call comp_cor( x(:nn,i:i), y(i:i), first, false,             &
                               xstat2(:nn,:2), ystat2(:2), xycor2(:nn), xyn2 )
!
                first = false
!
            end do
!$OMP END DO
!
!           THE SUMS OF SQUARES AND CROSSPRODUCTS HAVE BEEN CALCULATED ON DIFFERENT CHUNKS OF DATA,
!           NOW THE RESULTS ARE MERGED TOGETHER.
!
!$OMP MASTER
            xstat(:nn,:2)  = xstat2(:nn,:2)
            ystat(:2)      = ystat2(:2)
            xycor(:nn)     = xycor2(:nn)
            xyn            = xyn2
!$OMP END MASTER
!
!$OMP BARRIER
!
!$OMP CRITICAL (sumstat)
            if ( omp_get_thread_num()/=0 ) then
!
                call update_cor( xstat, ystat, xycor, xyn, xstat2, ystat2, xycor2, xyn2 )
!
            end if
!$OMP END CRITICAL (sumstat)
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( temp, xstat2, xycor2 )
!
!$OMP CRITICAL (closenc)
!
!           CLOSE THE NETCDF DATASET.
!
            status = nf90_close( ncid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
!           NOW FINISH THE COMPUTATIONS OF THE CORRELATIONS.
!
            call comp_cor( x(:nn,1:0), y(1:0), false, true,            &
                           xstat(:nn,:2), ystat(:2), xycor(:nn), xyn,  &
                           z=z(:nn)                                    )
!
        else
!
#endif
!
!           FIRST READ THE NETCDF VARIABLE.
!
            l            = time1 - 1
            startlec(1)  = j1
            startlec(2)  = lat1
            startlec(3)  = level1
!
            do i = 1, nperiod
!
                i2 = 0
!
                do k = 1, period
!
                    i1          = i2 + 1
                    i2          = i2 + nvar
                    l           = l + 1
                    startlec(4) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                    status = nf90_get_var( ncid, varid, temp(:lon3,:lat3,:level3),      &
                                           start=startlec, count=countlec )     
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA.
!
                    x(i1:i2,i) = pack( temp(:lon3,:lat3,:level3), mask(j1:j2,lat1:lat2,level1:level2) )
!
                end do
!
            end do
!
!           COMPUTE THE CORRELATIONS.
!
            call comp_cor( x(:nn,:nperiod), y(:nperiod), true, true,    &
                           xstat(:nn,:2), ystat(:2), xycor(:nn), xyn,   &
                           z=z(:nn)                                     )
!
#ifdef _PARALLEL_READ
!
        end if
!
#endif
!
        select case ( method )
!
           case ( 2 )
!
!              DO PERMUTATION TESTS FOR COMPUTING CRITICAL PROBABILITIES.
!
               call permute_cor( x(:nn,:nperiod), y(:nperiod),           &
                                 xstat(:nn,:2), ystat(:2), xycor(:nn),   &
                                 prob(:nn), nrep=nrep, initseed=true     )
!
           case ( 3:6 )
!
!              DO PHASE-SCRAMBLED BOOTSTRAP TESTS FOR COMPUTING CRITICAL PROBABILITIES.
!              USE NORMAL MARGINS IF method=3 OR 5 AND EMPIRICAL MARGINS IF method=4 OR 6.
!
               i     = merge( 2, 1, method==3 .or. method==4 )
               first = method==4 .or. method==6
!
               call phase_scramble_cor( x(:nn,:nperiod), y(:nperiod),                              &
                                        xstat(:nn,:2), ystat(:2), xycor(:nn),                      &
                                        prob(:nn), nrep=nrep, method=i, norm=first, initseed=true  )
!
           case ( 7 )
!
!              DO MOVING BLOCK BOOTSTRAP TESTS FOR COMPUTING CRITICAL PROBABILITIES.
!
!              SET DEFAULT VALUES FOR PARAMETERS OF THE MOVING BLOCK BOOTSTRAP ALGORITHM, IF REQUIRED.
!
               if (bperiodicity==0 ) then
                   bperiodicity=1
               end if
!
               if (bseason==0 ) then
                   bseason=nperiod
               end if
!
               if (block_size==0 ) then
                   block_size=bperiodicity
               end if
!
               call bootstrap_cor( x(:nn,:nperiod), y(:nperiod),             &
                                   xstat(:nn,:2), xycor(:nn),                &
                                   prob(:nn), nrep=nrep, initseed=true,      &
                                   periodicity=bperiodicity, season=bseason, &
                                   block_size=block_size                     )
       end select
!
    end if
!
!   DEALLOCATE WORK ARRAY IF REQUIRED.
!
    if ( allocated(temp) ) deallocate( temp )
!
!   COMPUTE REGRESSION COEFFICIENTS.
!
!   MULTIPLY THE VARIANCE OF THE TIME INDEX BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
    status = nf90_get_att( ncidi, varidi, "scale_factor", tmp2 )
!
    if ( status==nf90_noerr ) then
!
!       MULTIPLY THE VARIANCE OF THE TIME INDEX BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED. 
!
        ystat(2) = ystat(2)*tmp2**2
!
!       MULTIPLY THE MEAN OF THE TIME INDEX BY THE SCALE FACTOR IF REQUIRED.
!
        if ( use_int ) ystat(1) = ystat(1)*tmp2
!
    end if
!
!   COMPUTE STANDARD-DEVIATION OF THE TIME INDEX.
!
    ystat(2) = sqrt( ystat(2) )
!
    tmp = lamch('s')
!
    if ( ystat(2)>tmp ) then
!
!       MULTIPLY THE VARIANCE OF THE VARIABLE BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
        status = nf90_get_att( ncid, varid, "scale_factor", tmp2 )
!
        if ( status==nf90_noerr ) then
!
!           MULTIPLY THE VARIANCE OF THE VARIABLE BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
            xstat(:nn,2) = xstat(:nn,2)*tmp2**2
!
            if ( use_int ) then
!
!                MULTIPLY THE MEAN OF THE VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                 xstat(:nn,1) = xstat(:nn,1)*tmp2
            end if
!
        end if
!
!       COMPUTE STANDARD-DEVIATION OF THE VARIABLE.
!
        xstat(:nn,2) = sqrt( xstat(:nn,2) )
!
        if ( regression==1 ) then
!
!           COMPUTE REGRESSION COEFFICIENTS OF THE VARIABLE BY THE TIME INDEX.
!
            if ( use_std ) then
                xyreg(:nn) = xstat(:nn,2)
            else
                xyreg(:nn) = xstat(:nn,2)/ystat(2)
            end if
!
        else
!
!           COMPUTE REGRESSION COEFFICIENTS OF THE TIME INDEX BY THE VARIABLE.
!
            if ( use_std ) then
                xyreg(:nn) = ystat(2)
            else
                where( xstat(:nn,2)>tmp ) xyreg(:nn) = ystat(2)/xstat(:nn,2)
            end if
!
        end if
!
        where( xstat(:nn,2)>tmp )
            xyreg(:nn) = xycor(:nn)*xyreg(:nn)
        elsewhere
            xyreg(:nn)   = missing
            xycor(:nn)   = missing
            prob(:nn)    = missing
            z(:nn)       = missing
        end where
!
    else
!
        xyreg(:nn)   = missing
        xycor(:nn)   = missing
        prob(:nn)    = missing
        z(:nn)       = missing
!
    end if
!
!   COMPUTE INTERCEPT COEFFICIENTS IF REQUIRED.
!
    if ( use_int ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( xyint(nn), stat=status   )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid, varid, "add_offset", tmp )
        if ( status==nf90_noerr ) xstat(:nn,1) = xstat(:nn,1) + tmp
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE TIME INDEX, IF ANY.
!
        status = nf90_get_att( ncidi, varidi, "add_offset", tmp )
        if ( status==nf90_noerr ) ystat(1) = ystat(1) + tmp
!
        if ( regression==1 ) then
!
!           COMPUTE INTERCEPT COEFFICIENTS OF THE VARIABLE BY THE TIME INDEX.
!
            if ( use_std ) then
!
                where( xyreg(:nn)/=missing )
                    xyint(:nn) = xstat(:nn,1) -  xyreg(:nn)*( ystat(1)/ystat(2) )
                elsewhere
                    xyint(:nn) = missing
                end where
!
            else
!
                where( xyreg(:nn)/=missing )
                    xyint(:nn) =  xstat(:nn,1) - xyreg(:nn)*ystat(1)
                elsewhere
                    xyint(:nn) = missing
                end where
!
            end if
!
        else
!
!           COMPUTE INTERCEPT COEFFICIENTS OF THE TIME INDEX BY THE VARIABLE.
!
            if ( use_std ) then
!
                where( xyreg(:nn)/=missing )
                    xyint(:nn) = ystat(1) - ( xyreg(:nn)/xstat(:nn,2) )*xstat(:nn,1)
                elsewhere
                    xyint(:nn) = missing
                end where
!
            else
!
                where( xyreg(:nn)/=missing )
                    xyint(:nn) = ystat(1) - xyreg(:nn)*xstat(:nn,1)
                elsewhere
                    xyint(:nn) = missing
                end where
!
            end if
!
        end if
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE CORRELATION ANALYSIS.
!
    if ( create_hdf5 ) then
        status = nf90_create( cor_name, cmode=output_mode, ncid=ncid_cor )
    else
        k = nf90_sizehint_default
        status = nf90_create( cor_name, cmode=output_mode, ncid=ncid_cor, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_cor, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    dimlens(3)   = nlevel
    if ( use_tlimited ) then
        dimlens(4)   = period
    else
        dimlens(4)   = nf90_unlimited
    end if
!
    do i = 1, 4
!
        status = nf90_def_dim( ncid_cor, name=trim(dim_names(i)),                         &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 4
!
        status = nf90_def_var( ncid_cor, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    if ( regression==1 ) then
        name = trim(var_name)//"_"//trim(vari_name)
    else
        name = trim(vari_name)//"_"//trim(var_name)
    end if
!
    status = nf90_def_var( ncid_cor, trim(name)//"_cor", output_type, dimids=dimids(:4), varid=corid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, trim(name)//"_reg", output_type, dimids=dimids(:4), varid=regid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_int ) then
        status = nf90_def_var( ncid_cor, trim(name)//"_int", output_type, dimids=dimids(:4), varid=intid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_def_var( ncid_cor, trim(name)//"_prob", output_type, dimids=dimids(:4), varid=probid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, trim(name)//"_z", output_type, dimids=dimids(:4), varid=zid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, trim(name)//"_nobs", nf90_int, dimids=dimids(4), varid=nobsid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, "period", nf90_int, dimids=dimids(4), varid=periodid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 4
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_cor, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    status = nf90_put_att( ncid_cor, periodid, "long_name", "Period axis" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, periodid, "short_name", "Period" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( nidim==1 ) then
!
        if ( regression==1 ) then
            name = trim(var_name)//" with "//trim(vari_name)
        else
            name = trim(vari_name)//" with "//trim(var_name)
        end if
!
    else
!
        call val_to_string( iindex, fmt, i )
        if ( regression==1 ) then
            name = trim(var_name)//" with "//trim(vari_name)//"("//fmt(1:i)//")"
        else
            name = trim(vari_name)//"("//fmt(1:i)//")"//" with "//trim(var_name)
        end if
!
    end if
!
    status = nf90_put_att( ncid_cor, corid, "long_name",                   &
                           "Pearson Correlation of "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, corid, "short_name", "Correlation"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, regid, "long_name",                   &
                           "Regression of "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, regid, "short_name", "Regression"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_int ) then
!
        status = nf90_put_att( ncid_cor, intid, "long_name",                   &
                               "Intercept of "//trim(name)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, intid, "short_name", "Intercept"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_put_att( ncid_cor, probid, "long_name", "Prob > !COR! under h0: CORh0=0"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, probid, "short_name", "Probability"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, zid, "long_name",                   &
                           "Fisher Transform of correlation of "//trim(name)  )
!
    status = nf90_put_att( ncid_cor, zid, "short_name", "Z transform"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nobsid, "long_name", "Number of observations" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nobsid, "short_name", "Nobs" )
    if ( status/=nf90_noerr ) call merror( name_proc// trim( nf90_strerror( status ) ) )
!
    if ( use_int ) then
        i1 = 5
        id_var(1:5) = (/ corid, regid, probid, zid, intid /)
    else
        i1 = 4
        id_var(1:4) = (/ corid, regid, probid, zid /)
    end if
!
    if (  output_type==nf90_real8 ) then
!
        do i=1, i1
!
            status = nf90_put_att( ncid_cor, id_var(i), "missing_value", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_cor, id_var(i), "_FillValue", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        status = nf90_put_att( ncid_cor, probid, "valid_min", real( zero, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, probid, "valid_max", real( one, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_min", real( -one, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_max", real( one, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        do i=1, i1
!
            status = nf90_put_att( ncid_cor, id_var(i), "missing_value", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_cor, id_var(i), "_FillValue", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        status = nf90_put_att( ncid_cor, probid, "valid_min", real( zero, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, probid, "valid_max", real( one, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_min", real( -one, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_max", real( one, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    name2(1:nchar) = " "
    name3(1:nchar) = " "
!
    if ( regression==1 ) then
!
        i1 = nf90_get_att( ncid, varid, "units", name2 )
        i2 = nf90_get_att( ncidi, varidi, "units", name3 )
!
    else
!
        i1 = nf90_get_att( ncidi, varidi, "units", name2 )
        i2 = nf90_get_att( ncid, varid, "units", name3 )
!
    end if
!
    if ( i1==nf90_noerr ) then
!
        if ( use_int ) then
            status = nf90_put_att( ncid_cor, intid, "units", trim(name2)  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
        if ( use_std ) then
            status = nf90_put_att( ncid_cor, regid, "units", trim(name2)//" by STD" )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        else if ( i2==nf90_noerr ) then
            status = nf90_put_att( ncid_cor, regid, "units", trim(name2)//" by "//trim(name3) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_cor, nf90_global, "title", "Correlation analysis of "//trim(name) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "iperiod", iperiod )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "istep", istep )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "smoothing_factor", smooth_factor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "regression", cregression(regression) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>1 ) then
!
        status = nf90_put_att( ncid_cor, nf90_global, "number_of_shuffles", nrep )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        if ( method==7 ) then
!
            status = nf90_put_att( ncid_cor, nf90_global, "bootstrap_periodicity", bperiodicity )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_cor, nf90_global, "bootstrap_season", bseason )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_cor, nf90_global, "bootstrap_block_length", block_size )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end if
!
    status = nf90_put_att( ncid_cor, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_cor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW  NETCDF DATASET.
!
    dimlens(4) = period
!
!   ALLOCATE WORK ARRAY.
!
    i = max( nlon, nlat, nlevel, period )
    allocate( record(i), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    do j = 1, 4
!
        j2 = dimlens(j)
!
        if ( use_coor_variable(j) ) then
            status = nf90_get_var( ncid, coor_id(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_cor, coor_id2(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        else
            status = nf90_put_var( ncid_cor, coor_id2(j), arth(1_i4b,1_i4b,j2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( record )
!
!   WRITE THE PERIODICITY.
!
    status = nf90_put_var( ncid_cor, periodid, arth(1_i4b,1_i4b,period) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE NUMBERS OF OBSERVATIONS.
!
    status = nf90_put_var( ncid_cor, nobsid, (/ (nperiod, i=1, period) /) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE CORRELATION COEFFICIENTS.
!
    call write_file_4d( name_proc, xycor(:nn), mask(:nlon,:nlat,:nlevel), period, missing, ncid_cor, corid )
!
!   WRITE THE REGRESSION COEFFICIENTS.
!
    call write_file_4d( name_proc, xyreg(:nn), mask(:nlon,:nlat,:nlevel), period, missing, ncid_cor, regid )
!
!   WRITE THE INTERCEPT COEFFICIENTS IF REQUIRED.
!
    if ( use_int ) then
        call write_file_4d( name_proc, xyint(:nn), mask(:nlon,:nlat,:nlevel), period, missing, ncid_cor, intid )
    end if
!
!   WRITE THE PROBABILITIES.
!
    call write_file_4d( name_proc, prob(:nn), mask(:nlon,:nlat,:nlevel), period, missing, ncid_cor, probid )
!
!   WRITE THE Z TRANSFORMS.
!
    call write_file_4d( name_proc, z(:nn), mask(:nlon,:nlat,:nlevel), period, missing, ncid_cor, zid )
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( use_int ) then
        deallocate( mask, x, y, xycor, xyreg, xyint, prob, z, xstat )
    else
        deallocate( mask, x, y, xycor, xyreg, prob, z, xstat )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( ncid/=ncidi ) then
        status = nf90_close( ncidi )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( use_mask ) then
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_cor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_cor_4d
! ==========================
!
end program comp_cor_4d
