program comp_cor_miss_3d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_cor_miss_3d:
!
! ================
! comp_cor_miss_3d
! ================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 13/09/2018
!
! Purpose
! =======
!
! Compute correlation and regression coefficients between an index time series and a
! tridimensional variable extracted from a NetCDF dataset and perform statistical tests
! on these correlation coefficients. Missing values are allowed both in the input index
! time series and the input NetCDF tridimensional variable. However, if your data does not contain
! missing values excepted those associated with a constant land-sea mask use :ref:`comp_cor_3d`
! instead of comp_cor_miss_3d to estimate correlation  and regression coefficients
! from your dataset.
!
! As in :ref:`comp_cor_3d`, the procedure first transforms the input tridimensional NetCDF variable
! as a *ntime* by *nv* rectangular matrix of observed variables stored columnwise (e.g. the
! selected cells of the 2-D grid-mesh associated with the tridimensional NetCDF variable)
! and then computes measures of association between each of these variables, say *X*\ , and
! the input index time series, say *Y*\ . However since missing values are present, the
! number of observations used to compute the means and standard-deviations for each variable
! and the coefficients of correlation between each pair of variables *X*\  and *Y*\  may vary;
! This is an important difference with the statistics obtained from :ref:`comp_cor_3d`.
!
! By default, comp_cor_miss_3d computes the sample correlation and regression coefficients, the
! associated critical probabilities for testing the nullity of the correlation coefficients
! and the z transforms of the correlation coefficients between the index time series and each point
! in the time series of the 2-D grid-mesh associated with the input tridimensional NetCDF
! variable. The intercept coefficients of the regression line between *X*\  and *Y*\  are also
! computed if the optional argument *-intercept* is specified when calling comp_cor_miss_3d.
! Moreover, all these statistics may be computed by taking into account the periodicity of the input
! tridimensional NetCDF variable if you suspect that the time series are cyclostationary (by using
! the **-p=**\ ``periodicity`` argument when calling the procedure). All the results are finally stored
! in an output NetCDF dataset, after repacking the statistics on the original 2-D grid of the
! input tridimensional NetCDF variable.
!
! Refer to :ref:`comp_cor_3d`, for a basic definition of all these statistics, which is not
! repeated here. Refer to [vonStorch_Zwiers]_ for a general introduction on the correlation/regression
! coefficients and the z transform of the correlation coefficients and their use in climate analysis.
!
! Due to the presence of missing values, two different methods are, however, available
! to estimate the correlation and regression coefficients (and the Fisher z transform) in
! comp_cor_miss_3d.
!
! In the first method (used when the argument **-alg=**\  is set to ``miss1``\  ; this is
! the default), comp_cor_miss_3d estimates first the means and standard-deviations for the
! index time series and each point in the time series of the 2-D grid-mesh associated with the
! input NetCDF variable by using all the available observations for each time series.
! Since missing values are present, the number of observations used to compute 
! the means and standard-deviations may then vary from one point to another
! in the 2-D grid-mesh associated with the input NetCDF variable (as well as for
! the index time series). In a second step, comp_cor_miss_3d estimates the correlation
! and regression coefficients using the previously computed univariate statistics and all
! valid pairs of observations for each couple of variables. From this definition, it follows
! that the correlation coefficients computed from this method may be greater than ``1`` or
! less than ``-1`` in some cases when the number of missing values is very important. However,
! in such cases, the procedure adjusts the value of the correlation coefficients accordingly.
!
! In the second method (used when the argument **-alg=**\  is set to ``miss2``\  ), comp_cor_miss_3d
! computes both the univariate and bivariate statistics from all valid pairs of observations for each
! couple of variables separately. From this definition, it follows that the estimated correlation
! coefficient cannot be less than ``-1`` or greater than ``1``. However, the univariate statistics may
! be based on much fewer observations than in the first method (e.g. when **-alg=**\  ``miss1``\  ).
!
! Finally, note that only one method is available for computing critical probabilities associated
! with the correlation coefficients and test the significance of the correlations, since the permutation
! and bootstrap methods cannot easily be implemented if missing values are present in the time series.
! This is in contrast to the variety of approaches available in :ref:`comp_cor_3d`.
!
! This procedure is parallelized if OpenMP is used and the NCSTAT software has been built
! with the _PARALLEL_READ CPP macro. Moreover, this procedure computes the
! correlation and regression coefficients  with only one pass through the data and an
! out-of-core strategy which is highly efficient on huge datasets.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_cor_miss_3d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -vi=index_netcdf_variable \
!         -fi=input_index_netcdf_file              (optional) \
!         -m=input_mesh_mask_netcdf_file           (optional) \
!         -g=grid_type                             (optional : n, t, u, v, w, f) \
!         -x=lon1,lon2                             (optional) \
!         -y=lat1,lat2                             (optional) \
!         -t=time1,time2                           (optional) \
!         -p=periodicity                           (optional) \
!         -a=type_of_analysis                      (optional : student) \
!         -rg=type_of_regression                   (optional : reg1, reg2) \
!         -o=output_netcdf_file                    (optional) \
!         -ti=itime1,itime2                        (optional) \
!         -pi=iperiodicity,istep                   (optional) \
!         -ni=index_for_2d_index_netcdf_variable   (optional) \
!         -mi=missing_value                        (optional) \
!         -alg=algorithm                           (optional : miss1, miss2) \
!         -regstd                                  (optional) \
!         -intercept                               (optional) \
!         -double                                  (optional) \
!         -bigfile                                 (optional) \
!         -hdf5                                    (optional) \
!         -tlimited                                (optional)
!
! By default
! ----------
!
!        *-fi=*
!         the same as the *-f=* argument
!        *-m=* 
!         an *input_mesh_mask_netcdf_file* is not used
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input *netcdf_variable* is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-p=*
!         the *periodicity* is set to ``1``
!        *-ti=*
!         the whole time period associated with the *index_netcdf_variable*
!        *-pi=*
!         this parameter is not used
!        *-ni=*
!         if the *index_netcdf_variable* is bidimensional, the first time series is used
!        *-a=*
!         *type_of_analysis* is set to ``student``
!        *-rg=*
!         *type_of_regression* is set to ``reg1``
!        *-mi=*
!         the *missing_value* is set to ``1.e+20`` in the *output_netcdf_file*
!        *-alg=*
!         the method used to compute unvariate and bivariate statistics is ``miss1``
!        *-o=*
!         *output_netcdf_file* name is set to ``cor_``\ *netcdf_variable*\ ``.``\ *index_netcdf_variable*\ ``.nc``
!        *-regstd*
!         the regression coefficients are computed in units
!         of the input NetCDF variables. If **-regstd** is activated,
!         the regression coefficients are computed in units of the
!         *netcdf_variable* by standard-deviation of the *index_netcdf_variable*
!        *-intercept*
!         the intercept coefficients of the regressions are
!         not computed. If **-intercept** is activated, the
!         intercept coefficients of the regressions are computed
!         and stored in the *output_netcdf_file*
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! --------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable for which a
!      correlation analysis must be computed and the **-f=**\ *input_netcdf_file* argument
!      specifies that this NetCDF variable must be extracted from the NetCDF file,
!      *input_netcdf_file*.
!
!   #) The optional argument **-m=**\ *input_mesh_mask_netcdf_file* specifies the land-sea mask
!      to apply to *netcdf_variable* for transforming this tridimensional NetCDF variable as a
!      rectangular matrix of observed variables before computing the correlation analysis. By default,
!      it is assumed that each cell in the 2-D grid-mesh associated with the input tridimensional
!      NetCDF variable is a valid time series (e.g. a time series with some valid data for at least
!      some observations).
!
!      The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*) and the mask
!      (in the *input_mesh_mask_netcdf_file*) must agree if an *input_mesh_mask_netcdf_file* is used.
!
!      Refer to :ref:`comp_clim_miss_3d` or :ref:`comp_mask_3d` for creating a valid
!      *input_mesh_mask_netcdf_file* NetCDF file for regular or gaussian grids before
!      using comp_cor_miss_3d.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that
!      the NetCDF variable is from an experiment with the ORCA model (R2, R4 or R05 resolutions).
!      This argument is also used to determined the name of the mesh_mask_variable if
!      an *input_mesh_mask_netcdf_file* is used.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing the whole
!      geographical domain associated with the *netcdf_variable* is used.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices or
!      generating an appropriate mesh-mask before using comp_cor_miss_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, data in the whole time period
!      associated with the *netcdf_variable* is taken into account.
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!      The selected time period (e.g.  *time2* - *time1* + ``1``) must be a whole multiple of
!      the *periodicity* if the **-p=** argument is specified.
!
!   #) The **-p=**\ *periodicity* argument gives the periodicity of the input data for the
!      *netcdf_variable*. For example, with monthly data **-p=**\ ``12`` should be specified,
!      with yearly data **-p=**\ ``1`` may be used, etc.
!
!      Note that the output NetCDF file will have *periodicity* time observations.
!
!   #) The **-vi=**\ *index_netcdf_variable* specifies a time series for the correlation
!      analysis.
!      If the **-vi=**\ *index_netcdf_variable* is present, the **-fi=** argument
!      must also be present and this argument specifies the NetCDF dataset
!      which contains the *index_netcdf_variable*. However, if the NetCDF dataset which
!      contains the *index_netcdf_variable* is the same as the NetCDF dataset specified
!      by the **-f=** argument, it is not necessary to specify the **-fi=** argument.
!
!   #) The **-ni=** argument specifies the index (e.g. an integer) for selecting the time series if
!      the *index_netcdf_variable*  specified in the **-vi=** argument is a 2D NetCDF variable. By default,
!      the first time series is used, which is equivalent to set *index_for_2d_netcdf_variable* to ``1``\ .
!
!   #) If the **-ti=**\ *itime1,itime2* argument is missing, data in the whole time period
!      associated with the *index_netcdf_variable* is taken into account.
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!   #) The **-pi=** argument gives the periodicity and select the time step for the
!      *index_netcdf_variable*. For example, to compute correlations with the January
!      monthly time series extracted from the *index_netcdf_variable* which is assumed
!      to be sampled every month, **-pi=**\ ``12,1`` should be specified, with yearly data
!      **-pi=**\ ``1,1`` may be used, etc.
!
!   #) The selected time periods for the *netcdf_variable* and *index_netcdf_variable*
!      must agree. This means that the following equality must be verified
!
!      ``(time2 - time1 + 1)/periodicity = ceiling((itime2 - itime1 - istep + 2)/iperiodicity)``,
!
!      otherwise, an error message will be issued and the program will stop.
!
!   #) The **-alg=** argument selects the method for computing the correlation coefficients:
!
!      * If **-alg=**\ ``miss1``,  the means and standard-deviations of the *netcdf_variable* and
!        *index_netcdf_variable* are computed from all valid data. The correlation coefficients
!        are based on these univariate statistics and on all valid pairs of observations.
!
!      * If **-alg=**\ ``miss2``, the univariate and bivariate statistics are computed  
!        from all valid pairs of observations for each couple of variables separately.
!
!   #) The **-a=** argument selects the method for computing critical probabilities associated with
!      the correlation coefficients. 
!
!      If **-a=**\ ``student``, a classical Student-Fisher t test is used.
!
!      No other test options are included in this version of NCSTAT to test correlation coefficients
!      from data with missing values, but this optional parameter is still present for later use.
!
!   #) The **-rg=** argument selects the method for computing the regression coefficients:
!
!      * If **-rg=**\ ``reg1``, the coefficients of the regression equation for predicting the *netcdf_variable*
!        by the *index_netcdf_variable* are computed. This is the default.
!
!      * If **-rg=**\ ``reg2``, the coefficients of the regression equation for predicting the *index_netcdf_variable*
!        by the *netcdf_variable* are computed.
!
!   #) The **-intercept** argument specifies that the intercept coefficients of the regression equation
!      must be computed and stored in the output NetCDF file.
!      By default, the intercept coefficients are not computed.
!
!   #) The **-regstd** argument specifies that the regression coefficients of the regression equation
!      must be expressed in terms of units of the input NetCDF variable *netcdf_variable* by standard-deviation 
!      of the *index_netcdf_variable*.
!      By default, the regression coefficients are expressed in units of the input NetCDF variables.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the netcdf_variables in the *output_netcdf_file*.
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``\ .
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the specified *netcdf_variable* and *index_netcdf_variable* have a scalar
!      missing or _FillValue attributes and that missing values in the data are identified by
!      the values of these missing or _FillValue attributes.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on correlation and regression analysis in the climate literature, see
!
!      * "Statistical Analysis in Climate Research",
!        by von Storch, H., and Zwiers, F.W., Cambridge University press,
!        Cambridge, UK, Chapter 8, 484 pp., 2002. ISBN: `9780521012300`_
!
! .. _9780521012300: http://www.cambridge.org/gb/academic/subjects/earth-and-environmental-science/climatology-and-climate-change/statistical-analysis-climate-research?format=PB
!
! Outputs
! -------
!
!   comp_cor_miss_3d creates an output NetCDF file that contains the correlation and regression statistics and
!   critical probabilities associated with these coefficients, taking into account eventually the periodicity
!   of the data as determined by the **-p=**\ *periodicity* argument. The output NetCDF dataset contains the
!   following NetCDF variables (in the description below, nlat and nlon are the length of the spatial dimensions
!   of the input NetCDF variable) and *periodicity* time observations, if **-rg=**\ ``reg1``\  :
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _cor\ ``(periodicity,nlat,nlon)`` : the Pearson correlation coefficients between 
!      each point in the time series of the 2-D grid-mesh associated with the input NetCDF variable and the
!      *index_netcdf_variable* time series.
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _prob\ ``(periodicity,nlat,nlon)`` : the critical probabilities associated with
!      two-sided tests of the correlation coefficients (e.g. the absolute value of the correlation is tested).
!      These critical probabilities are computed under the null hypothesis that the corresponding correlation
!      coefficients in the parent population are zero.
!
!      The **-a=**\ *type_of_analysis* argument determines how these critical probabilities are
!      computed.
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _z\ ``(periodicity,nlat,nlon)`` : the Fisher z Transforms of the correlation
!      coefficients for each point in the time series of the 2-D grid-mesh associated with the input
!      NetCDF variable and the *index_netcdf_variable* time series.
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _reg\ ``(periodicity,nlat,nlon)`` : the regression coefficients for predicting 
!      each point in the time series of the 2-D grid-mesh associated with the input NetCDF variable by the
!      *index_netcdf_variable* time series.
!
!      By default, the regression coefficients are expressed in units of the input NetCDF variable
!      *netcdf_variable* by unit of the *index_netcdf_variable* time series.
!      However, if the **-regstd** argument is specified the regression coefficients are expressed in terms of
!      units of the input NetCDF variable *netcdf_variable* by standard-deviation of the *index_netcdf_variable*
!      time series.
!      Finally, if  **-rg**\ =\ ``reg2`` is specified the roles of the input NetCDF variables *netcdf_variable* and
!      *index_netcdf_variable* are interchanged and the fitted regression models are for predicting the
!      *index_netcdf_variable* by each time series of the 2-D grid-mesh associated with the input
!      NetCDF variable *netcdf_variable* .
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _int\ ``(periodicity,nlat,nlon)`` : the intercept coefficients
!      in the regression models for predicting each point in the time series of the 2-D grid-mesh associated with the input
!      NetCDF variable by the *index_netcdf_variable* time series.
!
!      This variable is stored only if the **-intercept** argument has been specified when calling
!      comp_cor_miss_3d.
!      Finally, if  **-rg**\ =\ ``reg2`` is specified the roles of the input NetCDF variables *netcdf_variable* and
!      *index_netcdf_variable* are interchanged and the fitted regression models are for predicting the
!      *index_netcdf_variable* by each time series of the 2-D grid-mesh associated with the input NetCDF
!      variable *netcdf_variable* .
!
!   #) *netcdf_variable*\ _\ *index_netcdf_variable*\ _nobs\ ``(periodicity,nlat,nlon)`` : the number of observations used to compute
!      the correlation and regression coefficient for each point in the time series of the 2-D grid-mesh
!      associated with the input NetCDF variable.
!
!   All these statistics are packed in tridimensional variables whose first and second dimensions are exactly the same
!   as those associated with the input NetCDF variable *netcdf_variable* even if you restrict the geographical domain with
!   the **-x=** and **-y=** arguments. However, outside the selected domain, the output NetCDF variables are filled with
!   missing values.
!
!   If **-rg=**\ ``reg2``\  , the naming convention for the variables is reversed, the *index_netcdf_variable* will be listed first
!   and the *netcdf_variable* will appear after. For example, the name of the NetCDF variable storing the correlation coefficient will
!   be *index_netcdf_variable*\ _\ *netcdf_variable*\ _cor\  instead of *netcdf_variable*\ _\ *index_netcdf_variable*\ _cor\  if
!   **-rg=**\ ``reg2``\  .
!
! Examples
! --------
!
!   #)   For computing monthly lead correlations from a tridimensional NetCDF variable
!        ``sst`` in the NetCDF file ``HadISST2_sst.nc``
!        and a December-January Nino34 SST index in the NetCDF file
!        ``HadISST2_sst_nino34_dj.nc`` and store the results in a NetCDF file named
!        ``cor_HadISST2_1m_sst_nino34_dj.nc``, use the following
!        commands (note that cyclostationarity is assumed for the ``sst`` variable
!        since  **-p**\ =\ ``12`` is specified ) :
!
!        ::
!
!          $ comp_cor_miss_3d \
!            -f=HadISST2_sst.nc \
!            -v=sst \
!            -m=mesh_mask_HadISST2.nc \
!            -p=12 \
!            -fi=HadISST2_sst_nino34_dj.nc \
!            -vi=sst \
!            -o=cor_HadISST2_1m_sst_nino34_dj.nc
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,    only : nf90_share
    USE Statpack,  only : omp_limit, update_cor_miss2
    use omp_lib,   only : omp_get_num_procs, omp_get_max_threads, omp_get_thread_num
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,   &
                         nf90_int, nf90_sizehint_default, nf90_open,                               &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,      &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,    &
                         nf90_def_var, nf90_inq_varid, nf90_inq_attname, nf90_copy_att,            &
                         nf90_put_att, nf90_enddef, nf90_global, nf90_put_var, nf90_close
!
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, kfix, kexp,      & 
                         kint, tolower, case_change, string_to_val, val_to_string,           &
                         allocate_error, merror, arth, comp_cor_miss, comp_cor_miss2
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,                        &
                                 get_listarg, get_variable_dims, get_variable_dimcoords,     &
                                 write_file_3d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
!
    character(len=*), parameter :: name_proc='comp_cor_miss_3d : ',                           &
!
                            cmethod(7)=(/"student  ","permute  ","scramble ","scramble2",     &
                                         "theiler  ","theiler2 ", "bootstrap"/) ,             &
!
                            calgorithm(2)=(/"miss1","miss2"/) ,                               &
!
                            cregression(2)=(/"reg1","reg2"/)                         
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncidi, ncid_mask, ncid_cor, periodid,              &
                     varid, maskid, varidi, corid, regid, intid, probid, zid, nobsid,       &
                     i, i1, i2, j, j1, j2, k, l, nlon, nlat, nt, nidim, nti, nvar, nn,      &
                     lon1, lon2, lon3, lat1, lat2, lat3,                                    &
                     time1, time2, time3, itime1, itime2, itime3,                           &
                     period, nperiod, iperiod, istep, niperiod, iindex, method, alg,        &
                     regression, output_type, output_mode
    integer(i4b), dimension(3)                 :: coor_id2, startlec, countlec
    integer(i4b), dimension(5)                 :: id_var
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:,:),   allocatable  :: temp, x, xstat, xycor
    real(stnd), dimension(:),     allocatable  :: y, prob, z, xyreg, xyint
    real(stnd), dimension(4)                   :: ystat
    real(stnd)                                 :: missing, missing2, missingi, tmp, tmp2
!
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: first, use_tlimited, use_missingv, use_missingi,       &
                                                  use_x, use_y, use_std, use_int, use_mask, create_hdf5
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
    logical(lgl), dimension(:,:), allocatable  :: mask
!
    character(len=nchar)                       :: name, name2, name3, history
    character(len=nf90_max_name)               :: fic_name, var_name, mask_name, fici_name, vari_name,   &
                                                  cor_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
    character                                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
    integer(i4b)  :: ncid2, varid2, ioffset
!
    real(stnd), dimension(:,:),   allocatable  :: xstat2, xycor2
!
    logical  :: test_par
#endif
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute correlation and regression coefficients between a time'
        write (prtunit,*) ' series index and a tridimensional variable extracted from a'
        write (prtunit,*) ' NetCDF dataset and perform statistical tests on these coefficients.'
        write (prtunit,*) ' Missing values are allowed in both the time series index and the'
        write (prtunit,*) ' tridimensional variable'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_cor_miss_3d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -vi=index_netcdf_variable'
        write (prtunit,*) '        -fi=input_index_netcdf_file             (optional)'
        write (prtunit,*) '        -m=input_mesh_mask_netcdf_file          (optional)'
        write (prtunit,*) '        -g=grid_type                            (optional : n, t, u, v, w, f)'
        write (prtunit,*) '        -x=lon1,lon2                            (optional)'
        write (prtunit,*) '        -y=lat1,lat2                            (optional)'
        write (prtunit,*) '        -t=time1,time2                          (optional)'
        write (prtunit,*) '        -p=periodicity                          (optional)'
        write (prtunit,*) '        -a=type_of_analysis                     (optional : student)'
        write (prtunit,*) '        -rg=type_of_regression                  (optional : reg1, reg2)'
        write (prtunit,*) '        -o=output_netcdf_file                   (optional)'
        write (prtunit,*) '        -ti=itime1,itime2                       (optional)'
        write (prtunit,*) '        -pi=iperiodicity,istep                  (optional)'
        write (prtunit,*) '        -ni=index_for_2d_index_netcdf_variable  (optional)'
        write (prtunit,*) '        -mi=missing_value                       (optional)'
        write (prtunit,*) '        -alg=algorithm                          (optional : miss1, miss2)'
        write (prtunit,*) '        -regstd                                 (optional)'
        write (prtunit,*) '        -intercept                              (optional)'
        write (prtunit,*) '        -double                                 (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '        -bigfile                                (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                                   (optional)'
#endif
        write (prtunit,*) '        -tlimited                               (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-fi= the same as the -f= argument'
        write (prtunit,*) '-m= an input_mesh_mask_netcdf_file is not used'
        write (prtunit,*) '-g= the grid type is set to n'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-p= the periodicity is set to 1'
        write (prtunit,*) '-ti= the whole time period  associated with the index_netcdf_variable'
        write (prtunit,*) '-pi=1,1 : by default, this parameter is not used'
        write (prtunit,*) '-ni=1   : by default, if the index_netcdf_variable is bidimensional,'
        write (prtunit,*) '          the first time series is used'
        write (prtunit,*) '-a=student'
        write (prtunit,*) '-rg=reg1'
        write (prtunit,*) '-mi= by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-alg=miss2'
        write (prtunit,*) '-o=cor_netcdf_variable.index_netcdf_variable.nc'
        write (prtunit,*) '-regstd   : by default, the regression coefficients are computed in units'
        write (prtunit,*) '            of the input NetCDF variables'
        write (prtunit,*) '-intercept: by default, the intercept coefficients of the regressions are'
        write (prtunit,*) '            not computed'
        write (prtunit,*) '-double   : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : by default, a netCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
#ifdef _PARALLEL_READ
        write (prtunit,*)
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>24 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name      = " "
    var_name      = " "
    fici_name     = " "
    vari_name     = " "
    mask_name     = " "
    cor_name      = " "
!
    history = "comp_cor_miss_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
    regression    = 1
    method        = 1
    alg           = 2
    period        = 1
    nperiod       = 0
!
    missing2      = def_missing
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    itime1    = 1
    itime2    = huge( itime2 )
    iperiod   = 1
    istep     = 1
    niperiod  = 0
    iindex    = 1
!
    use_mask     = false
    use_std      = false
    use_int      = false
    use_x        = false
    use_y        = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
                use_mask = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                cor_name = name(4:)
!
            case ('-p=')
!
!               GET THE PERIODICITY.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The periodicity is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF TEST ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2), cmethod(3), cmethod(4), cmethod(5), cmethod(6), cmethod(7) )
                        call merror( name_proc//name(4:)//" test analysis option is not allowed" )
                    case default
                        call merror( name_proc//"Unknown test analysis option" )
                end select
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-fi=')
!
!                       GET THE NAME OF THE NETCDF DATASET FOR THE INDEX VARIABLE.
!
                        fici_name = name(5:)
!
                    case ('-vi=')
!
!                       GET THE NAME OF THE NETCDF INDEX VARIABLE IN THE DATASET.
!
                        vari_name = name(5:)
!
                    case ('-ti=')
!
!                       GET THE PERIOD DEFINITION FOR THE INDEX VARIABLE.
!
                        call get_listarg( name_proc, name(5:),                           &
                                          "itime1 and itime2 are integer values",        &
                                          "Wrong period definition for index variable",  &
                                          itime1, itime2 )
!
                    case ('-pi=')
!
!                       GET THE PERIODICITY AND THE TIME STEP FOR THE INDEX.
!
                        call get_listarg( name_proc, name(5:),                                &
                                          "iperiod and istep are integer values",             &
                                          "Wrong periodicity definition for index variable",  &
                                          iperiod, istep )
!
                        if ( iperiod<=0 .or. istep<=0 .or. istep>iperiod )   &
                call merror( name_proc//"The -pi= parameters are positive integers with istep<=iperiod" )
!
                    case ('-ni=')
!
!                       GET THE INDEX FOR A 2D INDEX VARIABLE.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) iindex
                        end if
                        if ( kcode/=kint .or. iindex<=0  )                           &
                        call merror( name_proc//"The index for a 2D index variable is a positive integer value" )
!
                    case ('-rg=')
!
!                       DETERMINE THE TYPE OF REGRESSION ANALYSIS.
!
                        call case_change( name(5:), tolower )
!
                        select case ( name(5:) )
                            case ( cregression(1) )
                                regression = 1
                            case ( cregression(2) )
                                regression = 2
                            case default
                                call merror( name_proc//"Unknown regression analysis option" )
                        end select
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing2
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5 ) )
!
                            case ('-alg=')
!
!                               DETERMINE THE ALGORITHM FOR THE COMPUTATION.
!
                                select case ( name(6:) )
                                    case ( calgorithm(1) )
                                        alg = 1
                                    case ( calgorithm(2) )
                                        alg = 2
                                    case default
                                        call merror( name_proc//"Unknown algorithm option" )
                                end select
!
                            case default
!
                                select case ( name(1:) )
!
                                    case ('-double')
!
!                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                        output_type = nf90_real8
!
                                    case ('-regstd')
!
!                                       COMPUTE THE REGRESSION COEFFICIENTS IN TERMS OF STDS.
!
                                        use_std = true
!
                                    case ('-intercept')
!
!                                       COMPUTE THE INTERCEPT COEFFICIENTS.
!
                                        use_int = true
!
#ifdef _USE_NETCDF36
                                    case ('-bigfile')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
                                        create_hdf5 = true
#endif
!
                                    case ('-tlimited')
!
!                                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                                        use_tlimited = true
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( fici_name==" " ) fici_name = fic_name
!
    if ( vari_name==" " )                                                 &
    call merror( name_proc//"Input index NETCDF variable is missing" )
!
    if ( cor_name==" " ) cor_name="cor_"//trim(var_name)//"."//trim(vari_name)//".nc"
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_mask ) then
        status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
        if ( status/=nf90_noerr )                                                &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( fici_name/=fic_name ) then
        status = nf90_open( trim(fici_name), nf90_nowrite, ncidi )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    else
        ncidi = ncid
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                    &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names, &
                                  dimlens_out=dimlens, coor_id_out=coor_id,           &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type, &
                                  use_coor_variable_out=use_coor_variable             )
!
    if ( count(dimlens/=0)/=3 )           &
    call merror( name_proc//"Input NETCDF variable is not tridimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nt     = dimlens(3)
!
    if ( use_mask ) then
!
        chr = merge( 't', grid_type, grid_type=='w' )
!
!       GET THE MASK ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
        call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
        call merror( name_proc//"Shape disagreement with the mask" )
!
        if ( .not.use_x ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
            if ( status==nf90_noerr ) lon1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
            if ( status==nf90_noerr ) lon2 = i
!
        end if
!
        if ( .not.use_y ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
            if ( status==nf90_noerr ) lat1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
            if ( status==nf90_noerr ) lat2 = i
!
        end if
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( lon1<0 ) then
        lon2 = max( lon2, 1 )
        lon3 = nlon + lon1 - lon2 + 1
    else
        lon1 = max( lon1, 1 )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
    end if
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    if ( lon1<0 ) then
       lon3 = nlon
       j1   = 1
       j2   = nlon
    else
       j1 = lon1
       j2 = lon2
    end if
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE TIME DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    if ( period>time3 )  call merror( name_proc//"-t=, -p= options mismatch" )
!
    nperiod = time3/period
    if ( time3/=nperiod*period )      &
    call merror( name_proc//"The selected time period is not a multiple of the periodicity" )
!
    if ( nperiod<=2 )                 &
    call merror( name_proc//"Not enough time observations" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    if ( use_mask ) then
!
!       GET THE MASK.
!
        status = nf90_get_var( ncid_mask, maskid, temp )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DETERMINE THE GRID POINTS USED IN THE CORRELATION ANALYSIS FROM THE MASK.
!
        mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
    else
!
!       CONSTRUCT A DUMMY MASK.
!
        mask(:nlon,:nlat) = true
!
    end if
!
!   SELECT THE DOMAIN.
!
    if ( lon1>0 ) then
        mask(1:lon1-1,:nlat)    = false
        mask(lon2+1:nlon,:nlat) = false
    else
        mask(lon2+1:nlon+lon1,:nlat) = false
    end if
!
    mask(:nlon,1:lat1-1)    = false
    mask(:nlon,lat2+1:nlat) = false
!
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   GET THE INDEX VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS FOR THE NETCDF INDEX VARIABLE.
!
    call get_variable_dims( name_proc, trim(vari_name), ncidi,    &
                            varidi, dimids, dimlens )
!
    nidim = count(dimlens/=0)
    if ( nidim/=1 .and. nidim/=2 )     &
    call merror( name_proc//"Input index NETCDF variable is not uni or bidimensional" )
!
    if ( nidim==2 ) then
        if ( iindex>dimlens(1) )     &
        call merror( name_proc//"Selected index for 2D index_variable is out of range " )
        nti = dimlens(2)
    else 
        nti = dimlens(1)
    end if
!
!   CHECK THE TIME DEFINITION FOR THE INDEX VARIABLE.
!
    itime1 = max( itime1, 1 )
    itime1 = itime1 + istep - 1
    itime2 = min( itime2, nti )
    itime3 = itime2 - itime1 + 1
    if ( itime3<=0  ) call merror( name_proc//"Wrong time definition for index NETCDF variable" )
!
!   CHECK THE PERIOD DEFINITION FOR THE INDEX VARIABLE.
!
    if ( iperiod>itime3 )  call merror( name_proc//"-ti=, -pi= options mismatch" )
!
    tmp      = real(itime3,stnd)/real(iperiod,stnd)
    niperiod = ceiling( tmp )
!
    if ( niperiod/=nperiod )  call merror( name_proc//"-t=, -p=, -ti=, -pi= options mismatch" )
!
!   ALLOCATE WORK ARRAYS.
!
    nn = period*nvar
    i = merge( 4, 2, alg==1 )
!
    allocate( y(nperiod), xstat(nn,4),                    & 
              xycor(nn,i), xyreg(nn), prob(nn), z(nn), stat=status   )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET INDEX VARIABLE.
!
    if ( nidim==1 ) then
        status = nf90_get_var( ncidi, varidi, y(:nperiod),                               &   
                               start=(/ itime1 /), count=(/ nperiod /), stride=(/ iperiod /) )
    else 
        status = nf90_get_var( ncidi, varidi, y(:nperiod),                               &   
                               start=(/ iindex, itime1 /), count=(/ 1, nperiod /), stride=(/ 1, iperiod /) )
    end if
!
    if ( status/=nf90_noerr )                                     &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE MISSING INDICATOR VALUE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "missing_value", missing )
    use_missingv = status==nf90_noerr
    if ( .not. use_missingv ) then
        status = nf90_get_att( ncid, varid, "_FillValue", missing )
        use_missingv = status==nf90_noerr
    end if
!
!   FIND THE MISSING INDICATOR VALUE ASSOCIATED WITH THE INDEX NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncidi, varidi, "missing_value", missingi )
    use_missingi = status==nf90_noerr
    if ( .not. use_missingi ) then
        status = nf90_get_att( ncidi, varidi, "_FillValue", missingi )
        use_missingi = status==nf90_noerr
    end if
!
    if ( .not. use_missingv ) then
!
        if ( use_missingi ) then
            missing = missingi
        else
            if ( status/=nf90_noerr ) &
    call merror( name_proc//"No missing or _FillValue attributes for the specified variable and time index" )
        end if
!
    else 
!
        if ( use_missingi ) then
!
            if ( missing/=missingi ) then
                if ( any( y(:nperiod)==missing ) )    &
    call merror( name_proc//"Missing attribute for the specified variable is an observed value in the time index" )
                where( y(:nperiod)==missingi ) y(:nperiod) = missing
            end if
!
        else
!
            if ( any( y(:nperiod)==missing ) )    &
    call merror( name_proc//"Missing attribute for the specified variable is an observed value in the time index" )
!
        end if
!
   end if
!
!   COMPUTE THE CORRELATIONS.
!
    countlec(1) = lon3
    countlec(2) = lat3
    countlec(3) = 1
!
    if ( alg==1 ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( x(nn,1), stat=status   )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        l           = time1 - 1
        startlec(1) = j1
        startlec(2) = lat1
        first       = true
!
        do i = 1, nperiod
!
            i2 = 0
!
            do k = 1, period
!
                i1          = i2 + 1
                i2          = i2 + nvar
                l           = l + 1
                startlec(3) = l
!
!               FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                status = nf90_get_var( ncid, varid, temp(:lon3,:lat3),                         &
                                       start=startlec, count=countlec )     
                if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               PACK THE DATA.
!
                x(i1:i2,1) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
            end do
!
!           COMPUTE THE CORRELATIONS WITH UNIVARIATE STATISTICS FROM ALL AVAILABLE DATA.
!
            call comp_cor_miss( x(:nn,1:1), y(i:i), first, false,                 &
                                xstat(:nn,:4), ystat(:4), xycor(:nn,:4), missing  )
!
            first = false
!
        end do
!
        call comp_cor_miss( x(:nn,1:0), y(1:0), false, true,                   &
                            xstat(:nn,:4), ystat(:4), xycor(:nn,:4), missing,  &
                            prob=prob(:nn), z=z(:nn)                           )
!
!       COMPUTE REGRESSION COEFFICIENTS.
!
        status = nf90_get_att( ncid, varid, "scale_factor", tmp2 )
!
        if ( status==nf90_noerr ) then
!
!           MULTIPLY THE VARIANCE OF THE VARIABLE BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
            where( xstat(:nn,2)/=missing ) xstat(:nn,2) = xstat(:nn,2)*tmp2**2
!
            if ( use_int ) then
!
!                MULTIPLY THE MEAN OF THE VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                 where( xstat(:nn,1)/=missing ) xstat(:nn,1) = xstat(:nn,1)*tmp2
            end if
!
        end if
!
!       COMPUTE STANDARD-DEVIATION OF THE VARIABLE.
!
        where( xstat(:nn,2)/=missing ) xstat(:nn,2) = sqrt( xstat(:nn,2) )
!
        status = nf90_get_att( ncidi, varidi, "scale_factor", tmp2 )
!
        if ( status==nf90_noerr ) then
!
!           MULTIPLY THE VARIANCE OF THE TIME INDEX BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
            if ( ystat(2)/=missing ) ystat(2) = ystat(2)*tmp2**2
!
!           MULTIPLY THE MEAN OF THE TIME INDEX BY THE SCALE FACTOR IF REQUIRED.
!
            if ( use_int .and. ystat(1)/=missing ) ystat(1) = ystat(1)*tmp2
!
        end if
!
!       COMPUTE STANDARD-DEVIATION OF THE TIME INDEX.
!
        if ( ystat(2)/=missing ) ystat(2) = sqrt( ystat(2) )
!
        xyreg(:nn) = missing
!
        if ( regression==1 ) then
!
!           COMPUTE REGRESSION COEFFICIENTS OF THE VARIABLE BY THE TIME INDEX.
!
            if ( use_std ) then
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*xstat(:nn,2)
            else
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*( xstat(:nn,2)/ystat(2) )
            end if
!
        else
!
!           COMPUTE REGRESSION COEFFICIENTS OF THE TIME INDEX BY THE VARIABLE.
!
            if ( use_std ) then
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*ystat(2)
            else
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*( ystat(2)/xstat(:nn,2) )
            end if
!
        end if
!
!       COMPUTE INTERCEPT COEFFICIENTS IF REQUIRED.
!
        if ( use_int ) then
!
!           ALLOCATE WORK ARRAY.
!
            allocate( xyint(nn), stat=status   )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid, varid, "add_offset", tmp )
            if ( status==nf90_noerr ) then
                where( xstat(:nn,1)/=missing ) xstat(:nn,1) = xstat(:nn,1) + tmp
            end if
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE TIME INDEX, IF ANY.
!
            status = nf90_get_att( ncidi, varidi, "add_offset", tmp )
            if ( status==nf90_noerr .and. ystat(1)/=missing ) ystat(1) = ystat(1) + tmp
!
            if ( regression==1 ) then
!
!               COMPUTE INTERCEPT COEFFICIENTS OF THE VARIABLE BY THE TIME INDEX.
!
                if ( use_std ) then
                    where( xyreg(:nn)/=missing )
                        xyint(:nn) = xstat(:nn,1) -  xyreg(:nn)*( ystat(1)/ystat(2) )
                    elsewhere
                        xyint(:nn) = missing2
                    end where
                else
                    where( xyreg(:nn)/=missing )
                         xyint(:nn) =  xstat(:nn,1) - xyreg(:nn)*ystat(1)
                    elsewhere
                         xyint(:nn) = missing2
                    end where
                end if
!
            else
!
!               COMPUTE INTERCEPT COEFFICIENTS OF THE TIME INDEX BY THE VARIABLE.
!
                if ( use_std ) then
                    where( xyreg(:nn)/=missing )
                        xyint(:nn) = ystat(1) - ( xyreg(:nn)/xstat(:nn,2) )*xstat(:nn,1)
                    elsewhere
                        xyint(:nn) = missing2
                    end where
                else
                    where( xyreg(:nn)/=missing )
                        xyint(:nn) = ystat(1) - xyreg(:nn)*xstat(:nn,1)
                    elsewhere
                        xyint(:nn) = missing2
                    end where
                end if
!
            end if
!
        end if
!
    else
!
#ifdef _PARALLEL_READ
!
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        test_par = i>1 .and. j>1 .and. time3*nvar>omp_limit .and. nperiod>=j
!
        if ( test_par ) then
!
            ioffset = time1 - 1
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( temp )
!
!$OMP PARALLEL PRIVATE(i,l,k,i1,i2,status,startlec,ncid2,varid2,temp,x,xstat2,xycor2,first)      &
!$OMP         ,SHARED(y,mask,xstat,xycor,countlec,missing)
!
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( temp(lon3,lat3), x(nn,1), xstat2(nn,4), xycor2(nn,2), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!           OPEN THE NETCDF DATASET.
!
            status = nf90_open( trim(fic_name), nf90_share, ncid2  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE VARIABLE ID FROM ITS NAME.
!
            status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
            startlec(1)  = j1
            startlec(2)  = lat1
!
            first        = true
!
!$OMP DO SCHEDULE(STATIC)
            do i = 1, nperiod
!
                l = ioffset + (i-1)*period
                i2 = 0
!
                do k = 1, period
!
                    i1 = i2 + 1
                    i2 = i2 + nvar
                    l  = l + 1
                    startlec(3) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                    status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3), start=startlec, count=countlec )
                                           
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA.
!
                    x(i1:i2,1) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
                end do
!
!               COMPUTE CORRELATIONS AND UNIVARIATE STATISTICS FROM ALL VALID PAIRS OF DATA.
!
                call comp_cor_miss2( x(:nn,1:1), y(i:i), first, false,                                         &
                                     xstat2(:nn,1:2), xstat2(:nn,3:4), xycor2(:nn,1), xycor2(:nn,2), missing   )
!
                first = false
!
            end do
!$OMP END DO
!
!           THE SUMS OF SQUARES AND CROSSPRODUCTS HAVE BEEN CALCULATED ON DIFFERENT CHUNKS OF DATA,
!           NOW THE RESULTS ARE MERGED TOGETHER.
!
!$OMP MASTER
            xstat(:nn,:4)  = xstat2(:nn,:4)
            xycor(:nn,:2)  = xycor2(:nn,:2)
!$OMP END MASTER
!
!$OMP BARRIER
!
!$OMP CRITICAL (sumstat)
!
            if ( omp_get_thread_num()/=0 ) then
!
                call update_cor_miss2( xstat(:nn,1:2),  xstat(:nn,3:4),  xycor(:nn,1),  xycor(:nn,2),   &
                                       xstat2(:nn,1:2), xstat2(:nn,3:4), xycor2(:nn,1), xycor2(:nn,2)   )
!
            end if
!
!$OMP END CRITICAL (sumstat)
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( temp, x, xstat2, xycor2 )
!
!$OMP CRITICAL (closenc)
!
!           CLOSE THE NETCDF DATASET.
!
            status = nf90_close( ncid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
!           REALLOCATE WORK ARRAYS.
!
            allocate( x(nn,1), temp(nlon,nlat), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
        else
!
#endif
!
!           ALLOCATE WORK ARRAY.
!
            allocate( x(nn,1), stat=status   )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           COMPUTE THE CORRELATIONS.
!
            l     = time1 - 1
            first = true
!
            startlec(1)  = j1
            startlec(2)  = lat1
!
            do i = 1, nperiod
!
                i2 = 0
!
                do k = 1, period
!
                    i1          = i2 + 1
                    i2          = i2 + nvar
                    l           = l + 1
                    startlec(3) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                    status = nf90_get_var( ncid, varid, temp(:lon3,:lat3),                         &
                                           start=startlec, count=countlec )     
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA .
!
                    x(i1:i2,1) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
                end do
!
!               COMPUTE CORRELATIONS AND UNIVARIATE STATISTICS FROM ALL VALID PAIRS OF DATA.
!
                call comp_cor_miss2( x(:nn,1:1), y(i:i), first, false,                                     &
                                     xstat(:nn,1:2), xstat(:nn,3:4), xycor(:nn,1), xycor(:nn,2), missing   )
!
                first = false
!
            end do
!
#ifdef _PARALLEL_READ
!
        end if
!
#endif
!
        call comp_cor_miss2( x(:nn,1:0), y(1:0), false, true,                                      &
                             xstat(:nn,1:2), xstat(:nn,3:4), xycor(:nn,1), xycor(:nn,2), missing,  &
                             prob=prob(:nn), z=z(:nn)                                              )
!
!       COMPUTE REGRESSION COEFFICIENTS.
!
        status = nf90_get_att( ncid, varid, "scale_factor", tmp2 )
!
        if ( status==nf90_noerr ) then
!
!           MULTIPLY THE VARIANCE OF THE VARIABLE BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
            where( xstat(:nn,2)/=missing ) xstat(:nn,2) = xstat(:nn,2)*tmp2**2
!
            if ( use_int ) then
!
!                MULTIPLY THE MEAN OF THE VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                 where( xstat(:nn,1)/=missing ) xstat(:nn,1) = xstat(:nn,1)*tmp2
            end if
!
        end if
!
!       COMPUTE STANDARD-DEVIATION OF THE VARIABLE.
!
        where( xstat(:nn,2)/=missing ) xstat(:nn,2) = sqrt( xstat(:nn,2) )
!
!       MULTIPLY THE VARIANCE OF THE TIME INDEX BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
        status = nf90_get_att( ncidi, varidi, "scale_factor", tmp2 )
!
        if ( status==nf90_noerr ) then
!
!           MULTIPLY THE VARIANCE OF THE VARIABLE BY THE SQUARE OF THE SCALE FACTOR IF REQUIRED.
!
            where( xstat(:nn,4)/=missing ) xstat(:nn,4) = xstat(:nn,4)*tmp2**2
!
            if ( use_int ) then
!
!                MULTIPLY THE MEAN OF THE VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                 where( xstat(:nn,3)/=missing ) xstat(:nn,3) = xstat(:nn,3)*tmp2
            end if
!
        end if
!
!       COMPUTE STANDARD-DEVIATION OF THE TIME INDEX.
!
        where( xstat(:nn,4)/=missing ) xstat(:nn,4) = sqrt( xstat(:nn,4) )
!
        xyreg(:nn) = missing
!
        if ( regression==1 ) then
!
!           COMPUTE REGRESSION COEFFICIENTS OF THE VARIABLE BY THE TIME INDEX.
!
            if ( use_std ) then
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*xstat(:nn,2)
            else
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*( xstat(:nn,2)/xstat(:nn,4) )
            end if
        else
!
!           COMPUTE REGRESSION COEFFICIENTS OF THE TIME INDEX BY THE VARIABLE.
!
            if ( use_std ) then
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*xstat(:nn,4)
            else
                where( xycor(:nn,1)/=missing ) xyreg(:nn) = xycor(:nn,1)*( xstat(:nn,4)/xstat(:nn,2) )
            end if
        end if
!
!       COMPUTE INTERCEPT COEFFICIENTS IF REQUIRED.
!
        if ( use_int ) then
!
!           ALLOCATE WORK ARRAY.
!
            allocate( xyint(nn), stat=status   )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid, varid, "add_offset", tmp )
            if ( status==nf90_noerr ) then
                where( xstat(:nn,1)/=missing ) xstat(:nn,1) = xstat(:nn,1) + tmp
            end if
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE TIME INDEX, IF ANY.
!
            status = nf90_get_att( ncidi, varidi, "add_offset", tmp )
            if ( status==nf90_noerr ) then
                where( xstat(:nn,3)/=missing ) xstat(:nn,3) = xstat(:nn,3) + tmp
            end if
!
            if ( regression==1 ) then
!
!               COMPUTE INTERCEPT COEFFICIENTS OF THE VARIABLE BY THE TIME INDEX.
!
                if ( use_std ) then
                    where( xyreg(:nn)/=missing )
                        xyint(:nn) = xstat(:nn,1) -  ( xyreg(:nn)/xstat(:nn,4) )*xstat(:nn,3)
                    elsewhere
                        xyint(:nn) = missing2
                    end where
                else
                    where( xyreg(:nn)/=missing )
                         xyint(:nn) =  xstat(:nn,1) - xyreg(:nn)*xstat(:nn,3)
                    elsewhere
                         xyint(:nn) = missing2
                    end where
                end if
!
            else
!
!               COMPUTE INTERCEPT COEFFICIENTS OF THE TIME INDEX BY THE VARIABLE.
!
                if ( use_std ) then
                    where( xyreg(:nn)/=missing )
                        xyint(:nn) = xstat(:nn,3) - ( xyreg(:nn)/xstat(:nn,2) )*xstat(:nn,1)
                    elsewhere
                        xyint(:nn) = missing2
                    end where
                else
                    where( xyreg(:nn)/=missing )
                        xyint(:nn) = xstat(:nn,3) - xyreg(:nn)*xstat(:nn,1)
                    elsewhere
                        xyint(:nn) = missing2
                    end where
                end if
!
            end if
!
        end if
!
    end if
!
!   UPDATE THE MISSING VALUE INDICATOR FOR THE STATISTICS.
!
    where( xycor(:nn,1)==missing )
        xycor(:nn,1) = missing2
        xyreg(:nn)   = missing2
        prob(:nn)    = missing2
        z(:nn)       = missing2
    end where
!
!   CREATE THE NEW NETCDF DATASET FOR THE CORRELATION ANALYSIS.
!
    if ( create_hdf5 ) then
        status = nf90_create( cor_name, cmode=output_mode, ncid=ncid_cor )
    else
        k = nf90_sizehint_default
        status = nf90_create( cor_name, cmode=output_mode, ncid=ncid_cor, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_cor, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    if ( use_tlimited ) then
        dimlens(3)   = period
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    do i = 1, 3
!
        status = nf90_def_dim( ncid_cor, name=trim(dim_names(i)),                         &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 3
        status = nf90_def_var( ncid_cor, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end do
!
    if ( regression==1 ) then
        name = trim(var_name)//"_"//trim(vari_name)
    else
        name = trim(vari_name)//"_"//trim(var_name)
    end if
!
    status = nf90_def_var( ncid_cor, trim(name)//"_cor", output_type, dimids=dimids(:3), varid=corid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, trim(name)//"_reg", output_type, dimids=dimids(:3), varid=regid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_int ) then
        status = nf90_def_var( ncid_cor, trim(name)//"_int", output_type, dimids=dimids(:3), varid=intid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_def_var( ncid_cor, trim(name)//"_prob", output_type, dimids=dimids(:3), varid=probid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, trim(name)//"_z", output_type, dimids=dimids(:3), varid=zid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, trim(name)//"_nobs", nf90_int, dimids=dimids(:3), varid=nobsid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_cor, "period", nf90_int, dimids=dimids(3), varid=periodid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_cor, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    status = nf90_put_att( ncid_cor, periodid, "long_name", "Period axis" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, periodid, "short_name", "Period" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( nidim==1 ) then
        if ( regression==1 ) then
            name = trim(var_name)//" with "//trim(vari_name)
        else
            name = trim(vari_name)//" with "//trim(var_name)
        end if
    else
        call val_to_string( iindex, fmt, i )
        if ( regression==1 ) then
            name = trim(var_name)//" with "//trim(vari_name)//"("//fmt(1:i)//")"
        else
            name = trim(vari_name)//"("//fmt(1:i)//")"//" with "//trim(var_name)
        end if
    end if
!
    status = nf90_put_att( ncid_cor, corid, "long_name",                   &
                           "Pearson Correlation of "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, corid, "short_name", "Correlation"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, regid, "long_name",                   &
                           "Regression of "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, regid, "short_name", "Regression"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_int ) then
        status = nf90_put_att( ncid_cor, intid, "long_name",                   &
                               "Intercept of "//trim(name)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, intid, "short_name", "Intercept"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_put_att( ncid_cor, probid, "long_name", "Prob > !COR! under h0: CORh0=0"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, probid, "short_name", "Probability"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, zid, "long_name",                   &
                           "Fisher Transform of correlation of "//trim(name)  )
!
    status = nf90_put_att( ncid_cor, zid, "short_name", "Z transform"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nobsid, "long_name", "Number of observations" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nobsid, "short_name", "Nobs" )
    if ( status/=nf90_noerr ) call merror( name_proc// trim( nf90_strerror( status ) ) )
!
    if ( use_int ) then
        i1 = 5
        id_var(1:5) = (/ corid, regid, probid, zid, intid /)
    else
        i1 = 4
        id_var(1:4) = (/ corid, regid, probid, zid /)
    end if
!
    if (  output_type==nf90_real8 ) then
!
        do i=1, i1
            status = nf90_put_att( ncid_cor, id_var(i), "missing_value", real( missing2, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_cor, id_var(i), "_FillValue", real( missing2, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end do
!
        status = nf90_put_att( ncid_cor, probid, "valid_min", real( zero, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, probid, "valid_max", real( one, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_min", real( -one, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_max", real( one, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        do i=1, i1
            status = nf90_put_att( ncid_cor, id_var(i), "missing_value", real( missing2, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_cor, id_var(i), "_FillValue", real( missing2, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end do
!
        status = nf90_put_att( ncid_cor, probid, "valid_min", real( zero, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, probid, "valid_max", real( one, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_min", real( -one, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_cor, corid, "valid_max", real( one, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    name2(1:nchar) = " "
    name3(1:nchar) = " "
!
    if ( regression==1 ) then
        i1 = nf90_get_att( ncid, varid, "units", name2 )
        i2 = nf90_get_att( ncidi, varidi, "units", name3 )
    else
        i1 = nf90_get_att( ncidi, varidi, "units", name2 )
        i2 = nf90_get_att( ncid, varid, "units", name3 )
    end if
!
    if ( i1==nf90_noerr ) then
!
        if ( use_int ) then
            status = nf90_put_att( ncid_cor, intid, "units", trim(name2)  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
        if ( use_std ) then
            status = nf90_put_att( ncid_cor, regid, "units", trim(name2)//" by STD" )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        else if ( i2==nf90_noerr ) then
            status = nf90_put_att( ncid_cor, regid, "units", trim(name2)//" by "//trim(name3) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_cor, nf90_global, "title", "Correlation analysis of "//trim(name) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "iperiod", iperiod )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "istep", istep )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "algorithm", calgorithm(alg) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "regression", cregression(regression) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_cor, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_cor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW  NETCDF DATASET.
!
    dimlens(3) = period
!
!   ALLOCATE WORK ARRAY.
!
    i = max( nlon, nlat, period )
    allocate( record(i), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    do j = 1, 3 
!
        j2 = dimlens(j)
!
        if ( use_coor_variable(j) ) then
            status = nf90_get_var( ncid, coor_id(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_cor, coor_id2(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        else
            status = nf90_put_var( ncid_cor, coor_id2(j), arth(1_i4b,1_i4b,j2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( record )
!
!   WRITE THE PERIODICITY.
!
    status = nf90_put_var( ncid_cor, periodid, arth(1_i4b,1_i4b,period) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE CORRELATION COEFFICIENTS.
!
    call write_file_3d( name_proc, xycor(:nn,1), mask(:nlon,:nlat), period, missing2, ncid_cor, corid )
!
!   WRITE THE REGRESSION COEFFICIENTS.
!
    call write_file_3d( name_proc, xyreg(:nn), mask(:nlon,:nlat), period, missing2, ncid_cor, regid )
!
!   WRITE THE INTERCEPT COEFFICIENTS IF REQUIRED.
!
    if ( use_int ) then
        call write_file_3d( name_proc, xyint(:nn), mask(:nlon,:nlat), period, missing2, ncid_cor, intid )
    end if
!
!   WRITE THE NUMBERS OF OBSERVATIONS.
!
    call write_file_3d( name_proc, xycor(:nn,2), mask(:nlon,:nlat), period, zero, ncid_cor, nobsid )
!
!   WRITE THE PROBABILITIES.
!
    call write_file_3d( name_proc, prob(:nn), mask(:nlon,:nlat), period, missing2, ncid_cor, probid )
!
!   WRITE THE Z TRANSFORMS.
!
    call write_file_3d( name_proc, z(:nn), mask(:nlon,:nlat), period, missing2, ncid_cor, zid )
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( use_int ) then
        deallocate( temp, mask, x, y, xycor, xyreg, xyint, prob, z, xstat )
    else
        deallocate( temp, mask, x, y, xycor, xyreg, prob, z, xstat )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( ncid/=ncidi ) then
        status = nf90_close( ncidi )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( use_mask ) then
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_cor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_cor_miss_3d
! ===============================
!
end program comp_cor_miss_3d
