program comp_eof_miss_3d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_eof_miss_3d:
!
! ================
! comp_eof_miss_3d
! ================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 13/09/2018
!
! Purpose
! =======
!
! Compute an Empirical Orthogonal Function (EOF) analysis , also
! known as Principal Component Analysis (PCA) from
! a tridimensional variable with missing values extracted from a NetCDF dataset.
!
! The procedure first transforms the input tridimensional NetCDF variable
! as a *ntime* by *nv* rectangular matrix, **X**\ , of observed variables (e.g. the
! selected cells of the 2-D grid-mesh associated with the tridimensional NetCDF variable)
! and then performs an EOF analysis of this rectangular matrix. Since missing values
! are present, there are irregularly distributed gaps in the matrix **X** and the
! procedure used here is different than the one used in :ref:`comp_eof_3d`\ .
!
! A first basic assumption of the estimation procedures used in comp_eof_miss_3d is that
! each row and column of the matrix, **X**\ , of the observed variables have at least one
! non-missing entry.
!
! Two alternative strategies are available in comp_eof_miss_3d for estimating the first *k* eigenvectors
! and principal component time series of an EOF model with gappy data :
!
! #) If the argument **-alg=** is set to ``var``\ , comp_eof_miss_3d first computes an
!    estimate of the matrix product (transpose(**X**\ ).\ **X**\ )/\ *ntime* (as long as every
!    pair of observed variables has at least one nonmissing entry). The elements of this matrix
!    estimate are calculated from all valid observations for every pair of observed variables.
!    If fewer than one valid observation is present for some pair of variables, the procedure prints
!    an error message and stops. In a second step, the first *k* eigenvectors, *B* (*B* is a *k* by *nv*
!    matrix with eigenvectors stored rowwise) and eigenvalues of this matrix estimate
!    are computed by inverse iteration. These eigenvectors are preliminary estimates of the
!    first *k* eigenvectors of the matrix (transpose(**X**\ ).\ **X**\ )/\ *ntime*.
!    Note that because the matrix product estimate (derived from the data matrix **X**
!    with missing values) is not necessarily positive definite some of these eigenvalues
!    may be negative. In a third step, comp_eof_miss_3d obtains preliminary least square estimates,
!    *A* (*A* is a *ntime* by *k* matrix), of the first *k* principal component scores by regressing the
!    observations onto the eigenvectors, *B*\ , using only the valid entries in each observation.
!    Note that these preliminary estimates of the first *k* principal component scores are
!    not necessarily jointly uncorrelated if missing values are present.
!
! #) If the argument **-alg=** is set to ``obs``\ , comp_eof_miss_3d first computes an
!    estimate of the matrix product (**X**\ .transpose(**X**\ ))/\  *nv* (as long as every
!    pair of observations has at least one nonmissing entry). The elements of this matrix
!    estimate are calculated from all valid observed variables for every pair of observations.
!    If fewer than one valid variable is present for some pair of observations, the procedure prints
!    an error message and stops. In a second step, the first *k* eigenvectors, *A* (*A* is a *ntime* by *k*
!    matrix with eigenvectors stored columnwise) and eigenvalues of this matrix estimate
!    are computed by inverse iteration. These eigenvectors are preliminary estimates of the
!    first *k* (standardized) principal component scores.
!    Note that because the matrix product estimate derived from the data matrix **X**
!    with missing values is not necessarily positive definite some of the eigenvalues
!    may be negative. In a third step, comp_eof_miss_3d obtains preliminary least square estimates,
!    *B* (*B* is a *k* by *nv* matrix), of the first *k* eigenvectors of the EOF model by regressing the
!    variables onto the preliminary estimates of the principal component scores, *B*\ , using only the
!    valid entries in each variable.
!    Note that these preliminary estimates of the first *k* eigenvectors are not necessarily
!    orthogonal if missing values are present.
!
! After these alternative preliminary steps, it is an easy task to obtain suitable orthonormalization of the
! *A* and *B* matrices of the estimated *k*\ -component model similar to the traditional ones in the
! restricted *k* EOF model by computing the Singular Value Decomposition (SVD) of the *A*\ *B* matrix
! product. Note that there is no need to compute the *A*\ *B* product in this final computation since
! the SVD of this matrix product can be easily deduced from the two smallest SVD of *A* and  *B*\ , respectively.
! Let
!
!           *A*\ *B*  = **U**\ **S**\ **V**
!
! where
!
!  * **U**
!    is a *ntime* by *k* matrix with orthonormal columns (the left singular vectors stored columnwise)
!  * **S**
!    is a square *k* by *k* matrix with nonnegative elements on its principal diagonal and zeros
!    elsewhere (the diagonal elements of **S** are the singular values of *A*\ *B*\ )
!  * **V**
!    is a *k* by *nv* matrix with orthonormal rows (the right singular vectors stored rowwise)
!
! Note that this SVD has no more than *k* terms with a singular value distinct from zero since *A*\ *B*
! is of rank inferior or equal to *k*. With these notations, the first *k* unstandardized "principal components"
! of the *k*\ -EOF model are just **A**\ =\ **U**\ **S** and the first *k* "eigenvectors" of this model may be
! defined as **B**\ =\ **V** (up to a constant scaling factor).
!
! With these final estimates of the EOF model with gappy data, observed that the *k* eigenvectors stored rowwise in **B**
! are orthogonal and that the principal component time series stored columnwise in **A** are jointly uncorrelated
! as for the traditional EOF model without missing values in the data.
!
! More sophisticated methods are available in the literature for estimating PCA/EOF models with missing values [Terraya]_
! [Terrayb]_ [Schneider]_ and some of them will be implemented in future NCSTAT releases. 
!
! An output NetCDF dataset containing singular values, eigenvectors **B** and standardized principal component time
! series **U** is created. The eigenvectors are repacked as a tridimensional variable in the output NetCDF dataset.
!
! Optionally, comp_eof_miss_3d may also compute estimates of the missing values in the data matrix, **X**\ ,
! of observed variables with the help of the estimated *k*\ -EOF model **A**\ **B** and stored the resulting matrix
! in an output NetCDF file. This filled matrix is also repacked as a tridimensional variable in a second output NetCDF
! dataset which may be used for further analyses in NCSTAT procedures.
!
! This procedure is parallelized if OpenMP is used.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_eof_miss_3d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -g=grid_type                      (optional : n, t, u, v, w, f) \
!         -r=resolution                     (optional : r2, r4) \
!         -b=nlon_orca, nlat_orca           (optional) \
!         -x=lon1,lon2                      (optional) \
!         -y=lat1,lat2                      (optional) \
!         -t=time1,time2                    (optional) \
!         -a=type_of_analysis               (optional : scp, cov, cor) \
!         -c=input_climatology_netcdf_file  (optional) \
!         -d=type_of_distance               (optional : dist2, ident) \
!         -alg=algorithm                    (optional : var, obs) \
!         -n=number_of_eofs                 (optional) \
!         -o=output_eof_netcdf_file         (optional) \
!         -fo=output_netcdf_file            (optional) \
!         -mi=missing_value                 (optional) \
!         -use_eps=tolerance                (optional) \
!         -comp_min_norm                    (optional) \
!         -double                           (optional) \
!         -bigfile                          (optional) \
!         -hdf5                             (optional) \
!         -tlimited                         (optional)
!
! By default
! ----------
!
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh is assumed to be
!         regular or Gaussian
!        *-r=*
!         if the input netcdf_variable is from the NEMO or ORCA model (e.g. if **-n=** argument
!         is not set to ``n``) the resolution is assumed to be ``r2``
!        *-b=*
!         if **-n=** is not set to ``n``\ , the dimensions of the grid-mesh, *nlon_orca* and
!         *nlat_orca*, are determined from the **-r=** argument. However, you may override
!         this choice by default with the **-b=** argument
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-a=*
!         the *type_of_analysis* is set to ``scp``. This means that the eigenvectors
!         and eigenvalues are computed from the sums of squares and cross-products matrix
!         between the observed variables
!        *-c=*
!         an *input_climatology_netcdf_file* is not needed if
!         the *type_of_analysis* is set to ``scp``
!        *-d=*
!         the *type_of_distance* is set to ``dist2``. This means that distances and scalar products
!         in the EOF analysis are computed with the diagonal metric associated with the 2-D grid-mesh
!        *-alg=*
!         the *algorithm* option is set to ``var``\ . This means that eigenvectors are estimated in the space domain
!         first in the preliminary steps of the procedure
!        *-n=*
!         *number_of_eofs* is set to ``1`` and an one-component EOF model is computed
!         and stored in the output NetCDF file *output_eof_netcdf_file*
!        *-o=*
!         the *output_eof_netcdf_file* is named ``eof_``\ *netcdf_variable*\ ``.nc``
!        *-fo=*
!         by default, an *output_netcdf_file* with a copy of the input *netcdf_variable* in which
!         missing values replaced by estimates from the selected EOF model is not created
!        *-mi=*
!         the *missing_value* attribute in the output NetCDF file is set to ``1.e+20``
!        *-use_eps=*
!         the numerical rank is determined when solving each regression problem
!        *-comp_min_norm*
!         a minimal norm solution is not computed when solving deficient linear least squares problems
!        *-double*
!         the results of the EOF analysis are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         NetCDF classical format file are created. If **-bigfile** is activated,
!         the output NetCDF files are 64-bit offset format files
!        *-hdf5*
!         NetCDF classical format file are created. If **-hdf5** is activated,
!         the output NetCDF files are NetCDF-4/HDF5 format files
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable for which
!      an EOF analysis must be computed and the **-f=**\ *input_netcdf_file* argument
!      specifies that this NetCDF variable must be extracted from the NetCDF
!      file, *input_netcdf_file*.
!
!   #) The argument **-m=**\ *input_mesh_mask_netcdf_file* specifies the land-sea mask
!      to apply to the *netcdf_variable* for transforming this tridimensional NetCDF variable as a
!      rectangular matrix before computing the EOF analysis. The scale factors
!      associated with the 2-D grid-mesh of this NetCDF variable (needed if **-d=**\ ``dist2``\  is
!      specified when calling the procedure) are also read from the *input_mesh_mask_netcdf_file*.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing, the geographical
!      domain used in the EOF analysis is determined from the attributes of the input mesh mask
!      NetCDF variable named *grid_type*\ mask (e.g. lon1_Eastern_limit, lon2_Western_limit, lat1_Southern_limit
!      and lat2_Northern_limit ) which is read from the input NetCDF file *input_mesh_mask_netcdf_file*\ .
!      If these attributes are missing, the whole geographical domain associated with the *netcdf_variable*
!      is used in the EOF analysis.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices or generating an
!      appropriate mesh-mask before using comp_eof_miss_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used to estimate eigenvectors and
!      principal component time series.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF files will have ``ntime`` = *time2* - *time1* + ``1`` observations.
!
!   #) It is assumed that the specified input *netcdf_variable* have a scalar
!      missing or *_FillValue* attribute and that missing values in the data are
!      identified by the values of this missing or *_FillValue* attribute.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the
!      NetCDF variable is from an experiment with the NEMO or ORCA model.
!      In this case, the duplicate points from the ORCA grid are removed 
!      before the EOF analysis, as far as possible, and, in particular,
!      if the 2-D grid-mesh of the input NetCDF variable covers the whole globe. On output,
!      the duplicate points are restored when writing the EOFs (e.g. the eigenvectors), if the
!      geographical domain of the input NetCDF variable is the whole globe.
!
!      If **-g=** is set to ``n``, it is assumed that the 2-D grid-mesh is regular or Gaussian
!      and as such has no duplicate points.
!
!      The **-g=** argument is also used to determine the name of the NetCDF variables which
!      contain the 2-D mesh-mask and the scale factors in the *input_mesh_mask_netcdf_file* (e.g.
!      these variables are named *grid_type*\ ``mask``, ``e1``\ *grid_type* and ``e2``\ *grid_type*\ ,
!      respectively). This *input_mesh_mask_netcdf_file* may be created by :ref:`comp_clim_miss_3d` if the
!      2-D grid-mesh is regular or gaussian.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` (e.g. if the NetCDF variable
!      is from an experiment with the NEMO or ORCA model), the **-r=** argument gives the 
!      resolution used. If:
!
!      * **-r=**\ ``r2``, the NetCDF variable is from an experiment with the ORCA R2 model
!
!      * **-r=**\ ``r4``, the NetCDF variable is from an experiment with the ORCA R4 model.
!
!   #) If the NetCDF variable is from an experiment with the NEMO or ORCA model,
!      but the resolution is not R2 or R4, the dimensions of the ORCA grid
!      must be specified explicitly with the **-b=** argument.
!
!   #) The **-a=** argument specifies if the observed variables are centered or standardized
!      with an input climatology (specified with the **-c=** argument) before the EOF analysis. If:
!
!      * **-a=**\ ``scp``, the EOF analysis is done on the raw data
!
!      * **-a=**\ ``cov``, the EOF analysis is done on the anomalies
!
!      * **-a=**\ ``cor``, the EOF analysis is done on the standardized anomalies.
!
!   #) The *input_climatology_netcdf_file* specified with the **-c=** argument is needed
!      only if **-a=**\ ``cov`` or **-a=**\ ``cor``.
!
!   #) If **-a=**\ ``cov`` or **-a=**\ ``cor``, the selected time period must agree with the climatology.
!      This means that the first selected time observation (*time1* if the **-t=** argument
!      is present) must correspond to the first day, month, season of the climatology
!      specified with the **-c=** argument.
!
!   #) The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*), the mask
!      (in the *input_mesh_mask_netcdf_file*), the scale factors (in the *input_mesh_mask_netcdf_file*),
!      and the climatology (in the *input_climatology_netcdf_file*) must agree.
!
!   #) The **-n=**\ *number_of_eofs* argument specifies the order of the EOF model
!      which must be estimated from the gappy data
!      and stored in the output NetCDF file specified by the **-o=** argument.
!
!   #) The **-d=** argument specifies the metric and scalar product used in the EOF analysis. If:
!
!      * **-d=**\ ``dist2``, the EOF analysis is done with the diagonal distance
!        associated with the horizontal 2-D grid-mesh (e.g. each grid point is weighted accordingly to
!        the surface associated with it)
!
!      * **-d=**\ ``ident``, the EOF analysis is done with the identity metric : the
!        Euclidean distance and the usual scalar product is used in the EOF analysis.
!
!   #) The **-alg=** argument specifies how the preliminary estimates of the eigenvectors
!      and associated principal components are calculated. If:
!
!      * **-alg=**\ ``var``, eigenvectors are estimated in the space domain first
!        and time coefficients are obtained through a regression analysis in a second step
!
!      * **-alg=**\ ``obs``, eigenvectors are estimated in the time domain first and space
!        coefficients are obtained through a regression analysis in a second step.
!
!      In both cases, the final results are normalized with a partial SVD analysis as
!      described above. Both algorithms are also parallelized if OpenMP is used.
!
!      The default alorithm is **-alg=**\ ``var``\ .
!
!   #) **-use_eps=**\ *tolerance* is a real value used to determine the effective rank
!      of the coefficient matrix for each regression problem solved in comp_eof_miss_3d.
!      Tolerance must be set to the relative precision of the elements of the input data matrix.
!      If each element is correct to, say, 5 digits then **-use_eps=**\ ``0.00001``
!      should be used.
!      Tolerance must not be greater or equal to ``1`` or less or equal than ``0``\ ,
!      otherwise the numerical rank is determined.
!      If tolerance is absent, the numerical rank is determined for each
!      regression problem solved in comp_eof_miss_3d.
!
!   #) The **-comp_min_norm** argument specifies that a minimal norm solution must
!      be computed for solving deficient linear least squares problems.
!      By default, a minimal norm solution is not computed when solving deficient
!      linear least squares problems in the regression steps of the procedure
!
!   #) The **-fo=**\ *output_netcdf_file* argument specifies that the missing values
!      must be estimated from the computed EOF model (and climatology) and that
!      the full data matrix repacked as a tridimensional variable must be stored
!      in the NetCDF file *output_netcdf_file* on output of the procedure.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* and *output_netcdf_eof_file* will be 64-bit
!      offset format files instead of NetCDF classic format files. However, this argument is recognized in
!      the procedure only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* and *output_netcdf_eof_file* will be
!      NetCDF-4/HDF5 format files instead of NetCDF classic format files. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator
!      associated with the netcdf_variables in the output NetCDF files.
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on EOF analysis with missing values in the climate literature, see
!
!      * "Space/Time structure of monsoons interannual variability",
!        by Terray, P.,
!        Journal of Climate, Vol. 8, 2595-2619, 1995. doi: `10.1175/1520-0442(1995)008<2595:STSOMI>2.0.CO;2`_
!
!      * "Detecting Climatic Signals from Ship's Datasets",
!        by Terray, P.,
!        Proceedings of the International Workshop on Digitization and
!        Preparation of Historical Surface Marine Data and Metadata,
!        15-17 September. 1997, Toledo, Spain; 83-88 p.;
!        H.F. Diaz and S.D. Woodruff, Eds., WMO/TD-No.957, MMROA Report No. 43, 1999. http://icoads.noaa.gov/mmroa43_toledo.pdf
!
!      * "Application of Weighted Empirical Orthogonal Function Analysis to ship's datasets",
!        by Terray, P.,
!        Compte-Rendu de la IVme journe Statistique IPSL (Classification et Analyse spatiale).
!        NAI n23. pp. 11-28. 2002. ISSN 1626-8334. https://www.lmd.polytechnique.fr/nai/nai_23.pdf
!
!      * "Analysis of incomplete climate data: Estimation of mean values and covariance matrices
!        and imputation of missing values",
!        by Schneider, T.,
!        Journal of Climate, Vol. 14, 853-871, 2001. doi: `10.1175/1520-0442(2001)014<0853:AOICDE>2.0.CO;2`_
!
!      * "Statistical Analysis in Climate Research",
!        by von Storch, H., and Zwiers, F.W., Cambridge University press,
!        Cambridge, UK, Chapter 13, 484 pp., 2002. ISBN: `9780521012300`_
!
! .. _`10.1175/1520-0442(1995)008<2595:STSOMI>2.0.CO;2` : https://doi.org/10.1175/1520-0442(1995)008<2595:STSOMI>2.0.CO;2
! .. _`10.1175/1520-0442(2001)014<0853:AOICDE>2.0.CO;2` : https://doi.org/10.1175/1520-0442(2001)014<0853:AOICDE>2.0.CO;2
! .. _`9780521012300` : http://www.cambridge.org/gb/academic/subjects/earth-and-environmental-science/climatology-and-climate-change/statistical-analysis-climate-research?format=PB
!
! Outputs
! -------
!
!   comp_eof_miss_3d creates an output NetCDF file that contains the principal component time
!   series, the eigenvectors and the eigenvalues of the *number_of_eofs*\ -EOF model estimated from
!   a dataset with missing values. The number of principal components, eigenvectors and eigenvalues
!   computed and stored in the output NetCDF dataset is determined by the **-n=**\ *number_of_eofs* argument.
!   The output NetCDF dataset contains the following NetCDF variables 
!   (in the description below, nlat and nlon are the length of the dimensions of the input
!   NetCDF variable):
!
!   #) *netcdf_variable*\ _eof\ ``(number_of_eofs,nlat,nlon)`` : the approximations of the selected eigenvectors
!      of the sums of squares and cross-products (**-a=**\ ``scp``), covariance (**-a=**\ ``cov``) or correlation
!      (**-a=**\ ``cor``) matrix between the observed variables. The eigenvectors are sorted by descending
!      order of the associated eigenvalues. The eigenvectors are scaled such that they give estimates of the scalar 
!      products (**-a=**\ ``scp``), covariances (**-a=**\ ``cov``) or correlations (**-a=**\ ``cor``) 
!      between the original observed variables and the associated principal component time series.
!      The eigenvectors are packed in a tridimensional variable whose first and second dimensions
!      are exactly the same as those associated with the input NetCDF variable *netcdf_variable*
!      even if you restrict the geographical domain with the **-x=** and **-y=** arguments. However,
!      outside the selected domain, the output NetCDF variable is filled with missing values. If
!      this is a problem, you can use :ref:`comp_norm_miss_3d` for restricting the geographical
!      domain in the input dataset before using comp_eof_miss_3d.
!
!   #) *netcdf_variable*\ _pc\ ``(ntime,number_of_eofs)`` : the principal component time series
!      sorted by descending order of the eigenvalues .
!
!      The principal component time series are always standardized to unit variance.
!
!   #) *netcdf_variable*\_sing\ ``(number_of_eofs)`` : the singular values of the *A*\ *B* matrix product
!      in decreasing order. Up to a constant scaling factor (equal to the square root of ``1/ntime``),
!      these singular values are estimates of the square roots
!      of the eigenvalues of the sums of squares and cross-products (**-a=**\ ``scp``) or covariance
!      (**-a=**\ ``cov``) or correlation (**-a=**\ ``cor``) matrix between the observed variables.
!
!      These eigenvalues are estimates to the variance described by the principal component time series.
!
!   #) *netcdf_variable*\_var\ ``(number_of_eofs)`` : estimates of the proportion of variance explained by each
!      principal component time series computed as the ratio between the estimates of variance described by the principal
!      component time series and the sum of the variances of the observed variables.
!
!   Optionally, comp_eof_miss_3d can also create an output NetCDF file that contains a copy of the input
!   NetCDF variable with missing values filled with the help of the computed *number_of_eofs*\ -EOF model
!   adjusted to the gappy data. You must use the **-fo=**\ *output_netcdf_file* argument in order to create
!   this second output datat set. This optional output NetCDF dataset contains the following NetCDF variable 
!   (in the description below, nlat and nlon are the length of the dimensions of the input NetCDF variable):
!
!   #) *netcdf_variable*\ ``(ntime,nlat,nlon)`` : a copy of the input NetCDF variable with missing values
!      filled with estimates computed with the help of the selected EOF model.
!
! Examples
! --------
!
!   #)   For estimating a 1-EOF model from the NetCDF file
!        ``hadcrut2v.nc``, which includes a NetCDF 
!        variable ``temanom``, and store the results in a NetCDF file named
!        ``eof1_hadcrut2v.nc``, use the following
!        command (the analysis is done on the raw data) :
!
!        ::
!
!          $ comp_eof_miss_3d \
!            -f=hadcrut2v.nc  \
!            -v=temanom \
!            -m=mesh_mask_hadcrut2v.nc \
!            -g=n \
!            -a=scp \
!            -n=1 \
!            -o=eof1_hadcrut2v.nc
!
!   #)   For estimating a 3-EOF model from the NetCDF file
!        ``hadcrut2v.nc``, which includes a NetCDF 
!        variable ``temanom``, and store the results in a NetCDF file named
!        ``eof3_hadcrut2v.nc`` and, in addition, reconstructing a filled dataset
!        with the help of this 3-EOF model approximation of your gappy data, use
!        the following command :
!
!        ::
!
!          $ comp_eof_miss_3d \
!            -f=hadcrut2v.nc  \
!            -v=temanom \
!            -m=mesh_mask_hadcrut2v.nc \
!            -g=n \
!            -a=scp \
!            -n=3 \
!            -o=eof3_hadcrut2v.nc  \
!            -fo=hadcrut2v_eof3.nc
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,           only : nf90_share
    use Statpack,         only : omp_limit
    use omp_lib,          only : omp_get_num_procs, omp_get_max_threads
#endif
!
#ifdef _MATMUL
    use Statpack,         only : matmul=>matmul2
#endif
!
#ifdef _TRANSPOSE
    use Utilities,        only : transpose=>transpose2
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,           only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,           only : nf90_hdf5
#endif
!
    use netcdf,           only : nf90_max_var_dims, nf90_max_name, nf90_clobber, nf90_real4, nf90_real8,   &
                                 nf90_sizehint_default, nf90_open, nf90_nowrite,                           &
                                 nf90_noerr, nf90_strerror, nf90_get_att, nf90_get_var, nf90_create,       &
                                 nf90_set_fill, nf90_nofill, nf90_unlimited, nf90_def_dim, nf90_def_var,   &
                                 nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_enddef, nf90_get_var, &
                                 nf90_close, nf90_global, nf90_inquire_dimension, nf90_inq_varid,          &
                                 nf90_inquire_variable, nf90_int, nf90_put_var
!
    use Statpack,         only : i1b, i4b, stnd, extd, lgl, true, false, zero, one,                          & 
                                 kint, kfix, kexp, tolower, case_change, string_to_val, lamch,               &
                                 allocate_error, merror, arth, comp_eof_miss3, comp_pc_miss, product_svd_cmp
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, nlon_def2, nlat_def2, nlon_def4, nlat_def4,   &
                                 def_missing, get_listarg, get_variable_dims,                          &
                                 check_lon_domain, select_grid_points_2d, duplicate_orca_points_2d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    integer(i4b), parameter ::                                      &
!
!  maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED WHEN COMPUTING A
!  PARTIAL SVD OF THE DATA MATRIX WITH THE INVERSE ITERATION METHOD.
!
                                   maxiter=2
!
    character(len=*), parameter :: name_proc='comp_eof_miss_3d : ',     &
!
                                   cmethod(3)=(/"scp","cov","cor"/),    &          
!
                                   cdistance(2)=(/"dist2","ident"/),    &   
!
                                   cresolution(2)=(/"r2","r4"/),        &        
!
                                   calgorithm(2)=(/"var","obs"/)         
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_clim, ncid_mask, ncid_eof, ncid_out,               &
                     varid, maskid, meanid, stdid, e1id, e2id,                                   &
                     eofid, pcid, singid, expvarid, varoutid, var_natts,                         &
                     i, i1, i2, j, j1, j2, k, nlon, nlat, nt, nm, nvar, neof,                    &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3,                    &
                     method, distance, rec, offset_rec, nlon_def, nlat_def, alg,                 &
                     output_type, output_mode
    integer(i4b), dimension(3)                 :: coor_id2, startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i4b), dimension(:), allocatable    :: n_row
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:,:),   allocatable  :: temp, mat, mean, std, right_vec, left_vec
    real(stnd), dimension(:),     allocatable  :: sing_val, scal, var_row
    real(stnd)                                 :: missing, missing2, tmp, eps, var_tot
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: replace_data, use_tlimited, comp_min_norm,     &
                                                  failure, use_x, use_y, create_hdf5
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
    logical(lgl), dimension(:,:), allocatable  :: mask
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, clim_name, mask_name,    &
                                                  eof_name, ficout_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
    character                                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
    integer(i4b)  :: ncid2, varid2
!
    logical  :: test_par
#endif
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute an EOF analysis from a tridimensional variable '
        write (prtunit,*) ' with missing values extracted from a NetCDF dataset.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_eof_miss_3d   -f=input_netcdf_file'
        write (prtunit,*) '                    -v=netcdf_variable'
        write (prtunit,*) '                    -m=input_mesh_mask_netcdf_file'
        write (prtunit,*) '                    -g=grid_type                     (optional : n, t, u, v, w, f)'
        write (prtunit,*) '                    -r=resolution                    (optional : r2, r4)'
        write (prtunit,*) '                    -b=nlon_orca, nlat_orca          (optional)'
        write (prtunit,*) '                    -x=lon1,lon2                     (optional)'
        write (prtunit,*) '                    -y=lat1,lat2                     (optional)'
        write (prtunit,*) '                    -t=time1,time2                   (optional)'
        write (prtunit,*) '                    -a=type_of_analysis              (optional : scp, cov, cor)'
        write (prtunit,*) '                    -c=input_climatology_netcdf_file (optional)'
        write (prtunit,*) '                    -d=type_of_distance              (optional : dist2, ident)'
        write (prtunit,*) '                    -alg=algorithm                   (optional : var, obs)'
        write (prtunit,*) '                    -n=number_of_eofs                (optional)'
        write (prtunit,*) '                    -o=output_eof_netcdf_file        (optional)'
        write (prtunit,*) '                    -fo=output_netcdf_file           (optional)'
        write (prtunit,*) '                    -use_eps=                        (optional)'
        write (prtunit,*) '                    -comp_min_norm                   (optional)'
        write (prtunit,*) '                    -mi=missing_value                (optional)'
        write (prtunit,*) '                    -double                          (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                    -bigfile                         (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                    -hdf5                            (optional)'
#endif
        write (prtunit,*) '                    -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-g=n'
        write (prtunit,*) '-r=r2'
        write (prtunit,*) '-b= by default, nlon_orca and nlat_orca are determined from the -r='
        write (prtunit,*) '    argument'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-a=scp'
        write (prtunit,*) '-c= is not needed if -a=scp'
        write (prtunit,*) '-d=dist2'
        write (prtunit,*) '-alg=var, preliminary eigenvectors are estimated from the space domain'
        write (prtunit,*) '-n=1, a 1-component EOF model is estimated and stored'
        write (prtunit,*) '-o=eof_netcdf_variable.nc'
        write (prtunit,*) '-fo= the output_netcdf_file is not created'
        write (prtunit,*) '-mi= the missing_value is equal to 1.e+20'
        write (prtunit,*) '-use_eps= the numerical rank is determined for solving each'
        write (prtunit,*) '          regression problems'
        write (prtunit,*) '-comp_min_norm : a minimal norm solution is not computed'
        write (prtunit,*) '                 when solving deficient linear least squares problems'
        write (prtunit,*) '-double   : the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : NetCDF classical format files are created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, NetCDF classical format file are created'
#endif
        write (prtunit,*) '-tlimited : the time dimension is defined as unlimited'
        write (prtunit,*)
#ifdef _OPENMP
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>23 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name    = " "
    var_name    = " "
    mask_name   = " "
    clim_name   = " "
    eof_name    = " "
    ficout_name = " "
!
    history = "comp_eof_miss_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
    output_type = nf90_real4
    nlat_def    = nlat_def2
    nlon_def    = nlon_def2
    method      = 1
    distance    = 1
    alg         = 1
    neof        = 1
!
    missing2  = def_missing
    eps       = one
!
    grid_type = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x         = false
    use_y         = false
    replace_data  = false
    comp_min_norm = false
    create_hdf5   = false
    use_tlimited  = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-r=')
!
!               DETERMINE THE RESOLUTION FOR THE ORCA GRID.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cresolution(1) )
                        nlat_def  = nlat_def2
                        nlon_def  = nlon_def2
                    case ( cresolution(2) )
                        nlat_def  = nlat_def4
                        nlon_def  = nlon_def4
                    case default
                        call merror( name_proc//"Unknown resolution option" )
                end select
!
            case ('-b=')
!
!               GET THE DIMENSIONS OF THE ORCA GRID.
!
                call get_listarg( name_proc, name(4:),                           &
                                  "nlon_orca and nlat_orca are integer values",  &
                                  "Wrong ORCA grid definition ",                 &
                                  nlon_def, nlat_def )
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
            case ('-c=')
!
!               GET THE NAME OF THE NETCDF CLIMATOLOGY DATASET.
!
                clim_name = name(4:)
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case ( cmethod(3) )
                        method = 3
                    case default
                        call merror( name_proc//"Unknown analysis option" )
                end select
!
            case ('-d=')
!
!               DETERMINE THE DISTANCE FOR THE ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cdistance(1) )
                        distance = 1
                    case ( cdistance(2) )
                        distance = 2
                    case default
                        call merror( name_proc//"Unknown distance option" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE EOF ANALYSIS.
!
                eof_name = name(4:)
!
            case ('-n=')
!
!               GET THE NUMBER OF EOFs.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode==kint ) then
                    read( name(4:), fmt ) neof
                end if
                if ( kcode/=kint .or. neof<=0  )                                      &
                call merror( name_proc//"The number of EOFs is a positive integer value" )
!
            case default
!
                select case ( name(1:4) )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing2
!
                    case ('-fo=')
!
!                       GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE NEW DATA.
!
                        ficout_name = name(5:)
!
                        replace_data =  true
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5 ) )
!
                            case ('-alg=')
!
!                               DETERMINE THE ALGORITHM FOR THE COMPUTATION.
!
                                select case ( name(6:) )
                                    case ( calgorithm(1) )
                                        alg = 1
                                    case ( calgorithm(2) )
                                        alg = 2
                                    case default
                                        call merror( name_proc//"Unknown algorithm option" )
                                end select
!
                            case default
!
                                select case ( name(1:9) )
!
                                    case ('-use_eps=')
!
!                                       DETERMINE THE TOLERANCE FOR SOLVING LINEAR LEAST SQUARES PROBLEMS.
!
                                        call string_to_val( name(10:), kcode, fmt )
                                        if ( kcode==kfix .or. kcode==kexp ) then
                                            read( name(10:), fmt ) eps
                                        end if
                                        if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                             eps<=zero                       .or.     &
                                             eps>=one )                               &
                    call merror( name_proc//"The -use_eps= argument must be a real number and verify 0<use_eps<1 " )
!
                                    case default
!
                                        select case ( name )
!
                                            case ('-comp_min_norm')
!
                                                comp_min_norm = true
!
                                            case ('-double')
!
!                                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                                output_type = nf90_real8
!
#ifdef _USE_NETCDF36
                                            case ('-bigfile')
!
!                                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILES.
!
                                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILES.
!
                                        output_mode = nf90_hdf5
                                        create_hdf5 = true
#endif
!
                                            case ('-tlimited')
!
!                                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                                use_tlimited = true
!
                                            case default
!
!                                               WRONG ARGUMENT.
!
                                                call merror( name_proc//"Wrong argument" )
!
                                        end select
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( mask_name==" " )                                                &
    call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( clim_name==" " .and. method>=2 )                                &
    call merror( name_proc//"Climatology NETCDF file is missing" )
!
    if ( eof_name==" " ) eof_name="eof_"//trim(var_name)//".nc"
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>=2 ) then
        k = nf90_sizehint_default
        status = nf90_open( trim(clim_name), nf90_nowrite, ncid_clim, chunksize=k )
        if ( status/=nf90_noerr )                                        &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE NUMBER OF ATTRIBUTES, THE DIMENSION LENGTHS AND IDS FOR THE NETCDF VARIABLE.
!
    call get_variable_dims( name_proc, trim(var_name), ncid, varid, dimids, dimlens, natts=var_natts  )
!
    if ( count(dimlens/=0)/=3 )           &
    call merror( name_proc//"Input NETCDF variable is not tridimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nt     = dimlens(3)
!
    do i = 1, 3
!
!       GET THE NAMES OF THE DIMENSIONS.
!
        status = nf90_inquire_dimension( ncid, dimids(i), name=dim_names(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       GET THE COORDINATE VARIABLES IDS FROM THEIR NAME, IF ANY.
!
        status = nf90_inq_varid( ncid, trim(dim_names(i)), coor_id(i) )
        if ( status==nf90_noerr ) then
!
!           FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!           FOR THE COORDINATE VARIABLES.
!
            status = nf90_inquire_variable( ncid, coor_id(i), xtype=coor_type(i), ndims=k,    &
                                            dimids=dimlens, natts=coor_natts(i) )
            if ( status/=nf90_noerr )                                                         &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            use_coor_variable(i) = k==1 .and. dimlens(1)==dimids(i)
        else
            use_coor_variable(i) = false
        end if
        if ( .not.use_coor_variable(i) ) coor_type(i) = nf90_int
!
    end do
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
    call merror( name_proc//"Shape disagreement with the mask" )
!
!   GET THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN FROM THE MESH-MASK DATASET.
!
    if ( .not.use_x ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
        if ( status==nf90_noerr ) lon1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
        if ( status==nf90_noerr ) lon2 = i
!
    end if
!
    if ( .not.use_y ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
        if ( status==nf90_noerr ) lat1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
        if ( status==nf90_noerr ) lat2 = i
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!   CHECK THE GEOGRAPHICAL DOMAIN.
!   REMOVE DUPLICATE POINTS FROM THE ORCA_R2 GRID.
!
    if ( grid_type/='n' ) then
!
        if ( nlon>nlon_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (longitude)" )
!
        if ( nlat>nlat_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (latitude)" )
!
!       A TRICK TO DISABLE THE ELIMINATION AND DUPLICATION IN THE HIGH LATITUDES OF THE ORCA GRID
!       IF THE RESOLUTION IS NOT R2 OR R4.
!
        if ( ( nlat_def/=nlat_def2 .or. nlon_def/=nlon_def2 ) .and. ( nlat_def/=nlat_def4 .or. nlon_def/=nlon_def4 ) )    &
        nlat_def = nlat_def + 1
!
    end if
!
    call check_lon_domain( grid_type, nlon, nlon_def, lon1, lon2, lon3 )
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   FIND THE MISSING INDICATOR VALUE ASSOCIATED WITH THE NETCDF VARIABLE.
!
    status = nf90_get_att( ncid, varid, "missing_value", missing )
    if ( status/=nf90_noerr ) then
        status = nf90_get_att( ncid, varid, "_FillValue", missing )
        if ( status/=nf90_noerr )      &
        call merror( name_proc//"No missing or _FillValue attributes for the specified variable" )
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, temp )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED IN THE EOF ANALYSIS FROM THE MASK.
!
    mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
!   SELECT THE DOMAIN.
!
    call select_grid_points_2d( grid_type, lon1, lon2, lat1, lat2,  &
                                nlon_def, nlat_def, mask, j1, j2  )
!
    if ( lon1<0 ) lon3 = nlon
!
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   ALLOCATE WORK ARRAY.
!
    allocate( mat(nvar,time3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec     = time1 - 1
    countlec(1:3)  = (/ lon3, lat3, 1_i4b /)
!
!   GET THE DATA.
!
#ifdef _PARALLEL_READ
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    test_par = i>1 .and. j>1 .and. time3*nvar>omp_limit .and. time3>=j
!
    if ( test_par ) then
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
!$OMP PARALLEL PRIVATE(k,status,rec,startlec,ncid2,varid2,temp)       &
!$OMP         ,SHARED(mat,mask,countlec,j1,j2,lon3,                   &
!$OMP                 lat1,lat2,lat3,nvar,offset_rec,time1,time2)
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( temp(lon3,lat3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!       OPEN THE NETCDF DATASET.
!
        status = nf90_open( trim(fic_name), nf90_share, ncid2  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       GET THE VARIABLE ID FROM ITS NAME.
!
        status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
        startlec(1)  = j1
        startlec(2)  = lat1
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = time1, time2
!
            startlec(3) = k
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3),                  &
                                   start=startlec, count=countlec )     
            if ( status/=nf90_noerr )                                                &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           PACK THE DATA.
!
            rec            = k - offset_rec
            mat(:nvar,rec) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        end do
!
!$OMP END DO
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
!$OMP CRITICAL (closenc)
!
!       CLOSE THE NETCDF DATASET.
!
        status = nf90_close( ncid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
!       REALLOCATE WORK ARRAY.
!
        allocate( temp(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
    else
!
#endif
!
        startlec(1:2)  = (/ j1, lat1 /)
!
        do k = time1, time2
!
            startlec(3) = k
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid, varid, temp(:lon3,:lat3),                      &
                                   start=startlec, count=countlec )     
            if ( status/=nf90_noerr )                                                   &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           PACK THE DATA.
!
            rec = k - offset_rec
            mat(:nvar,rec) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        end do
!
#ifdef _PARALLEL_READ
!
    end if
!
#endif
!
!   SELECT THE METHOD OF ANALYSIS.
!
    select case ( method )
!
        case ( 1 )
!
!           EOF ANALYSIS IS DONE ON THE RAW DATA.
!
!
!           MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
            status = nf90_get_att( ncid, varid, "scale_factor", tmp )
            if ( status==nf90_noerr ) then
                where( mat(:nvar,:time3)/=missing )  mat(:nvar,:time3) = mat(:nvar,:time3)*tmp
            end if
!
!           ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
            status = nf90_get_att( ncid, varid, "add_offset", tmp )
            if ( status==nf90_noerr ) then
                where( mat(:nvar,:time3)/=missing )  mat(:nvar,:time3) = mat(:nvar,:time3) + tmp
            end if
!
        case ( 2, 3 )
!
!           EOF ANALYSIS IS DONE ON THE (STANDARDIZED) ANOMALIES.
!
!
!           GET THE MEAN FIELD ID FROM ITS NAME.
!           FIND THE DIMENSION LENGTHS FOR THE NETCDF MEAN VARIABLE.
!
            call get_variable_dims( name_proc, trim(var_name)//"_mean", ncid_clim,    &
                                    meanid, dimids, dimlens )
!
            if ( count(dimlens/=0)/=3 )     &
            call merror( name_proc//"Input climatology is not tridimensional" )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )    &
            call merror( name_proc//"Shape disagreement with the climatology (mean field)" )
!
            nm = dimlens(3)
!
            failure = false
!
            status = nf90_get_att( ncid_clim, nf90_global, "lon1_"//trim(var_name), i )
            if ( status==nf90_noerr ) failure = j1<i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lon2_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = j2>i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lat1_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = lat1<i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lat2_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = lat2>i
!
            if ( failure ) call merror( name_proc//"Missing values in the climatology " )
!
!           ALLOCATE WORK ARRAY FOR THE MEANS.
!
            allocate( mean(nvar,nm), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            startlec(1:2)  = (/ j1, lat1 /)
!
!           GET THE MEANS.
!
            do j = 1, nm
!
                startlec(3)  = j
!
                status = nf90_get_var( ncid_clim, meanid, temp(:lon3,:lat3),               &   
                                       start=startlec, count=countlec)     
                if ( status/=nf90_noerr )                                                  &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               PACK THE MEAN FIELD.
!
                mean(:nvar,j) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
            end do
!
!           FIND THE MISSING INDICATOR VALUE ASSOCIATED WITH THE MEAN VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid_clim, meanid, "missing_value", tmp )
!
            if ( status==nf90_noerr ) then
                if ( any(mean(:nvar,:nm)==tmp)  )                &
                call merror( name_proc//"Missing values in the climatology (mean field)" )
            end if
!
            if ( method==2 ) then
!
!               EOF ANALYSIS IS DONE ON THE ANOMALIES.
!
                k = ( time3/nm )*nm
!
!               COMPUTE ANOMALIES.
!
                do i = 0, k-1, nm
!
                    i1 = i + 1
                    i2 = i + nm
!
                    where( mat(:nvar,i1:i2)/=missing )                                   &
                    mat(:nvar,i1:i2) = mat(:nvar,i1:i2) - mean(:nvar,:nm)
!
                end do
!
                if ( k/=time3 ) then
                    where( mat(:nvar,k+1:time3)/=missing )                               &
                    mat(:nvar,k+1:time3) = mat(:nvar,k+1:time3) - mean(:nvar,1:time3-k)
                end if
!
!               DEALLOCATE WORK ARRAY.
!
                if ( .not. replace_data ) deallocate( mean )
!
!               MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                status = nf90_get_att( ncid, varid, "scale_factor", tmp )
!
                if ( status==nf90_noerr ) then
                    where( mat(:nvar,:time3)/=missing )  mat(:nvar,:time3) = mat(:nvar,:time3)*tmp
                end if
!
            else
!
!               EOF ANALYSIS IS DONE ON THE STANDARDIZED ANOMALIES.
!
!               GET THE STD ID FROM ITS NAME.
!               FIND THE DIMENSION LENGTHS FOR THE NETCDF STD VARIABLE.
!
                call get_variable_dims( name_proc, trim(var_name)//"_std", ncid_clim,    &
                                        stdid, dimids, dimlens )
!
                if ( count(dimlens/=0)/=3 )          &
                call merror( name_proc//"Input climatology is not tridimensional" )
!
                if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                                  &
                call merror( name_proc//"Shape disagreement with the climatology (std field)" )
!
                if ( dimlens(3)/=nm )                                                          &
                call merror( name_proc//"Shape disagreement in the climatology" )
!
!               ALLOCATE WORK ARRAY.
!
                allocate( std(nvar,nm), stat=status )
                if ( status/=0 ) call merror( name_proc//allocate_error )
!
!               GET THE STDS.
!
                do j = 1, nm
!
                    startlec(3)  = j
!
                    status = nf90_get_var( ncid_clim, stdid, temp(:lon3,:lat3),                & 
                                           start=startlec, count=countlec)     
                    if ( status/=nf90_noerr )                                                  &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE STD FIELD.
!
                    std(:nvar,j) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
                end do
!
!               FIND THE MISSING INDICATOR VALUE ASSOCIATED WITH THE STD VARIABLE, IF ANY.
!
                status = nf90_get_att( ncid_clim, stdid, "missing_value", tmp )
                if ( status==nf90_noerr ) then
                    if ( any(std(:nvar,:nm)==tmp)  )                &
                    call merror( name_proc//"Missing values in the climatology (std field)" )
                end if
!
                tmp = lamch('s')
!
                where ( std(:nvar,:nm)>tmp )
                    std(:nvar,:nm) = one/std(:nvar,:nm)
                elsewhere
                    std(:nvar,:nm) = zero
                end where
!
                k = ( time3/nm )*nm
!
!               COMPUTE STANDARDIZED ANOMALIES.
!
                do i = 0, k-1, nm
!
                    i1 = i + 1
                    i2 = i + nm
!
                    where( mat(:nvar,i1:i2)/=missing )                                       &
                    mat(:nvar,i1:i2) = ( mat(:nvar,i1:i2) - mean(:nvar,:nm) )*std(:nvar,:nm)
!
                end do
!
                if ( k/=time3 ) then
                    where( mat(:nvar,k+1:time3)/=missing )  mat(:nvar,k+1:time3) =             &
                    ( mat(:nvar,k+1:time3) - mean(:nvar,:time3-k) )*std(:nvar,:time3-k)
                end if
!
!               DEALLOCATE WORK ARRAYS.
!
                if ( .not. replace_data ) deallocate( mean, std )
!
            end if
!
    end select
!
!   COMPUTE THE DISTANCE.
!
    if ( distance==1 ) then
!
!       THE DISTANCE IS ASSOCIATED WITH THE (HORIZONTAL) GRID.
!
!
!       ALLOCATE WORK ARRAY.
!
        allocate( scal(nvar), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE SCALE FACTORS IDs FROM THEIR NAMES.
!       VERIFY THE DIMENSION LENGTHS FOR THE SCALE FACTOR VARIABLES.
!
        call get_variable_dims( name_proc, "e1"//chr, ncid_mask,    &
                                e1id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the first scale factor" )
!
        call get_variable_dims( name_proc, "e2"//chr, ncid_mask,    &
                                e2id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the second scale factor" )
!
!       GET THE SCALE FACTORS.
!
        status = nf90_get_var( ncid_mask, e1id, temp(:lon3,:lat3), start=(/j1,lat1/), &  
                               count=(/ lon3, lat3 /)  )     
        if ( status/=nf90_noerr )                                                     &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
        scal(:nvar) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        status = nf90_get_var( ncid_mask, e2id, temp(:lon3,:lat3), start=(/j1,lat1/), &
                               count=(/ lon3, lat3 /)  )     
        if ( status/=nf90_noerr )                                                     &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
        scal(:nvar) = scal(:nvar)*pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
!       COMPUTE THE DIAGONAL METRIC ASSOCIATED WITH THE HORIZONTAL GRID.
!
        scal(:nvar) = sqrt( scal(:nvar)*(one/maxval(scal(:nvar))) )
!        scal(:nvar) = sqrt( scal(:nvar) )
!
!       SCALE THE NETCDF VARIABLE.
!
        do i = 1, time3
            where( mat(:nvar,i)/=missing ) mat(:nvar,i) =  mat(:nvar,i)*scal(:nvar)
        end do
    end if
!
!   COMPUTE ROW INERTIA AND NUMBER OF OBSERVATIONS BY ROWS.
!
!   ALLOCATE WORK ARRAYS.
!
    allocate(  var_row(nvar), n_row(nvar), stat=status   ) 
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    var_row(:nvar) = zero
    n_row(:nvar)   = 0
!
    do i = 1, time3
!
        where (mat(:nvar,i)/=missing ) 
            var_row(:nvar)  = var_row(:nvar) + mat(:,i)**2
            n_row(:nvar)    = n_row(:nvar)   + 1
        end where
!
    end do
!
!   TEST MISSING VALUES.
!
    if ( any(n_row==0_i4b) ) call merror(name_proc//": Some grid-points have only missing values")
!
!   ESTIMATE TOTAL INERTIA.
!
    var_tot = sum( var_row(:nvar)/real(n_row(:nvar),stnd) )
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate(  n_row ) 
!
!   CHECK THE neof VALUE.
!
    k = min( nvar, time3 )
    neof = min( neof, k )
!
!   ALLOCATE WORK ARRAYS FOR THE EOF ANALYSIS WITH MISSING VALUES.
!
    allocate(  sing_val(neof), left_vec(nvar,neof),      &
               right_vec(time3,neof), stat=status        ) 
!
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   ESTIMATE neof SINGULAR VALUES, LEFT AND 
!   RIGHT SINGULAR VECTORS OF mat.
!
    if ( alg==1 ) then
!
        call  comp_eof_miss3( mat(:nvar,:time3), missing, 1, failure,                                    &
                              xeigvec=left_vec(:nvar,:neof), maxiter=maxiter      )
!
        call  comp_pc_miss( mat(:nvar,:time3), left_vec(:nvar,:neof), right_vec(:time3,:neof), missing,  &
                            dimvar=1, tol=eps, min_norm=comp_min_norm   )
!
    else
!
        call  comp_eof_miss3( mat(:nvar,:time3), missing, 2, failure,                                    &
                              xeigvec=right_vec(:time3,:neof), maxiter=maxiter      )

!
        call  comp_pc_miss( mat(:nvar,:time3), right_vec(:time3,:neof), left_vec(:nvar,:neof), missing,  &
                            dimvar=2, tol=eps, min_norm=comp_min_norm   )
!
    end if
!
    if ( failure ) call merror(  name_proc//"Error in the EOF analysis with missing values" )
!
!   COMPUTE PARTIAL SVD.
!
    call product_svd_cmp( left_vec(:nvar,:neof), right_vec(:time3,:neof), &
                          sing_val(:neof), failure, sort='d' )
    if ( failure ) call merror(  name_proc//"Error in the partial SVD analysis (normalisation step)" )
!
    if ( distance==1 ) then
!
!       THE DISTANCE IS ASSOCIATED WITH THE (HORIZONTAL) GRID.
!
        scal(:nvar) = one/scal(:nvar)
!
!       UNSCALE THE EOFs.
!
        left_vec(:nvar,:neof) = left_vec(:nvar,:neof)*spread( scal(:nvar), dim=2, ncopies=neof ) 
!
!       UNSCALE THE DATA IF NECESSARY.
!
        if ( replace_data ) then
!
            do i=1, time3
                where( mat(:nvar,i)/=missing ) mat(:nvar,i) = mat(:nvar,i)*scal(:nvar)
            end do
!
        end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( scal )
!
    end if
!
    left_vec(:nvar,:neof) = left_vec(:nvar,:neof)*spread( sing_val(:neof), dim=1, ncopies=nvar ) 
!
    if ( replace_data ) then
!
!       ESTIMATE MISSING VALUES FIRST.
!
        if ( neof==1 ) then
            do i=1, time3
                where( mat(:nvar,i)==missing ) mat(:nvar,i) = left_vec(:nvar,1)*right_vec(i,1)
            end do
        else
            do i=1, time3
                var_row(:nvar) = matmul( left_vec(:nvar,:neof), right_vec(i,:neof) )
                where( mat(:nvar,i)==missing ) mat(:nvar,i) = var_row(:nvar)
            end do
        end if
!
!       TRANSFORMED DATA IF REQUIRED.
!
        select case ( method )
!
            case ( 2 )
!
!               ADD THE MEAN FIELD.
!
                k = ( time3/nm )*nm
!
                do i = 0, k-1, nm
!
                    i1 = i + 1
                    i2 = i + nm
!
                    mat(:nvar,i1:i2) = mat(:nvar,i1:i2) + mean(:nvar,:nm)
!
                end do
!
                if ( k/=time3 ) then
                    mat(:nvar,k+1:time3) = mat(:nvar,k+1:time3) + mean(:nvar,1:time3-k)
                end if
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( mean )
!
            case ( 3 )
!
!               MULTIPLY BY THE STD FIELD AND ADD THE MEAN FIELD.
!
                where ( std(:nvar,:nm)/=zero )  std(:nvar,:nm) = one/std(:nvar,:nm)
!
                k = ( time3/nm )*nm
!
                do i = 0, k-1, nm
!
                    i1 = i + 1
                    i2 = i + nm
!
                    mat(:nvar,i1:i2) = mat(:nvar,i1:i2)*std(:nvar,:nm) + mean(:nvar,:nm)
!
                end do
!
                if ( k/=time3 ) then
                    mat(:nvar,k+1:time3) = mat(:nvar,k+1:time3)*std(:nvar,:time3-k) + mean(:nvar,:time3-k)
                end if
!
!               DEALLOCATE WORK ARRAYS.
!
                deallocate( mean, std )
!
        end select
!
!       CREATE THE NEW NETCDF DATASET FOR THE TRANSFORMED DATA.
!
        if ( create_hdf5 ) then
            status = nf90_create( ficout_name, cmode=output_mode, ncid=ncid_out )
        else
            k = nf90_sizehint_default
           status = nf90_create( ficout_name, cmode=output_mode, ncid=ncid_out, chunksize=k )
        end if
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_set_fill( ncid_out, nf90_nofill, i )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
        dimlens(1)   = nlon
        dimlens(2)   = nlat
        if ( use_tlimited ) then
            dimlens(3)   = time3
        else
            dimlens(3)   = nf90_unlimited
        end if
!
        do i = 1, 3
!
            status = nf90_def_dim( ncid_out, name=trim(dim_names(i)),                         &
                                   len=dimlens(i), dimid=dimids(i) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
        end do
!
!       DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
        do i = 1, 3
!
            status = nf90_def_var( ncid_out, trim(dim_names(i)), coor_type(i),               &
                                   dimids=dimids(i), varid=coor_id2(i) )
            if ( status/=nf90_noerr )                                                        &
            call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        status = nf90_def_var( ncid_out, trim(var_name),  output_type,                   &
                               dimids=dimids(1:3), varid=varoutid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
        do j = 1, 3
!
            if ( .not.use_coor_variable(j) ) cycle
!
            do i = 1, coor_natts(j)
!
                status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
                if ( status/=nf90_noerr )                                                     &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
                status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_out, coor_id2(j) )
                if ( status/=nf90_noerr )                                                     &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
        end do
!
        if ( .not.use_coor_variable(3) ) then
!
            status = nf90_put_att( ncid_out, coor_id2(3), "long_name", "Time axis" )
            if ( status/=nf90_noerr )                                                  &
            call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_out, coor_id2(3), "short_name", "Time" )
            if ( status/=nf90_noerr )                                                  &
            call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
!       COPY ATTRIBUTES FOR THE TRANSFORMED VARIABLE.
!
        do i = 1, var_natts
!
            status = nf90_inq_attname( ncid, varid, i, att_name )
            if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            if ( ( method==1 .and. ( att_name=="add_offset" .or. att_name=="scale_factor") ) .or.  &
                 ( method==2 .and.   att_name=="scale_factor" )                              .or.  &
                 att_name=="missing_value"                                                   .or.  &
                 att_name=="_FillValue"               )   cycle
!
!           COPY USEFUL ATTRIBUTES.
!
            status = nf90_copy_att(  ncid, varid, att_name, ncid_out, varoutid )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        if (  output_type==nf90_real8 ) then
!
            status = nf90_put_att( ncid_out, varoutid, "missing_value", real( missing2, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_out, varoutid, "_FillValue", real( missing2, dp ) )
!
        else
!
            status = nf90_put_att( ncid_out, varoutid, "missing_value", real( missing2, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_out, varoutid, "_FillValue", real( missing2, sp ) )
!
        end if
!
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, nf90_global, "title", trim(var_name)//" with estimated missing values " )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, nf90_global, "algorithm", name_proc//calgorithm(alg) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, nf90_global, "distance", cdistance(distance) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, nf90_global, "method", cmethod(method) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, nf90_global, "grid_type", grid_type )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, nf90_global, "history", trim(history) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
        status = nf90_enddef( ncid_out )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
        do j = 1, 3    
!
            select case ( j )
                case ( 1 )
                    j1 = 1
                    j2 = nlon
                case ( 2 )
                    j1 = 1
                    j2 = nlat
                case ( 3 )
                    j1 = time1
                    j2 = time3
            end select
!
            if ( use_coor_variable(j) ) then
!
                allocate( record(j2), stat=status )
                if ( status/=0 ) call merror( name_proc//allocate_error )
!
                status = nf90_get_var( ncid, coor_id(j), record, start=(/ j1 /), count=(/ j2 /) )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_var( ncid_out, coor_id2(j), record )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
                deallocate( record )
            else
                status = nf90_put_var( ncid_out, coor_id2(j), arth(j1,1_i4b,j2 ) )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
            end if
!
        end do
!
!       WRITE TRANSFORMED VARIABLE.
!
!
        startlec(1:2) = 1
!
        do i = 1, time3
!
            temp(:nlon,:nlat) = unpack( mat(:nvar,i), mask, missing2 )
!
            if ( grid_type/='n' .and. nlon==nlon_def ) then
!
                if ( lat2>=nlat_def-1 .and. nlat==nlat_def ) then
!
!                   ADD DUPLICATE POINTS TO THE ORCA GRID IF THE WHOLE GLOBE IS USED.
!
                    call duplicate_orca_points_2d( grid_type, mask, temp )
                else
!
!                   ADD DUPLICATE LONGITUDES LINES TO THE ORCA GRID IF THE WHOLE LONGITUDE
!                   DOMAIN IS USED.
!
                    temp(1,:nlat)    = temp(nlon-1,:nlat)
                    temp(nlon,:nlat) = temp(2,:nlat)
                end if
!
            end if
!
            startlec(3) = i
!
            status = nf90_put_var( ncid_out, varoutid, temp, start=startlec )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
!       CLOSE THE NETCDF DATASETS.
!
        status = nf90_close( ncid_out )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE EOF ANALYSIS.
!
    if ( create_hdf5 ) then
        status = nf90_create( eof_name, cmode=output_mode, ncid=ncid_eof )
    else
        k = nf90_sizehint_default
        status = nf90_create( eof_name, cmode=output_mode, ncid=ncid_eof, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_eof, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    dimlens(3)   = neof
    if ( use_tlimited ) then
        dimlens(4)   = time3
    else
        dimlens(4)   = nf90_unlimited
    end if
    dim_names(4) = dim_names(3)
    dim_names(3) = "pc"
!
    do i = 1, 4
!
        status = nf90_def_dim( ncid_eof, name=trim(dim_names(i)),                         &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 2
!
        status = nf90_def_var( ncid_eof, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_eof, trim(dim_names(4)), coor_type(3),                    &
                           dimids=dimids(4), varid=coor_id2(3) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_eof, trim(var_name)//"_eof", output_type,                 &
                           dimids=dimids(1:3), varid=eofid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_eof, trim(var_name)//"_pc", output_type,                  &
                           dimids=dimids(3:4), varid=pcid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_eof, trim(var_name)//"_sing", nf90_real8,                 &
                           dimids=dimids(3), varid=singid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_eof, trim(var_name)//"_var", output_type,                    &
                           dimids=dimids(3), varid=expvarid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_eof, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    if ( .not.use_coor_variable(3) ) then
!
        status = nf90_put_att( ncid_eof, coor_id2(3), "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_eof, coor_id2(3), "short_name", "Time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_put_att( ncid_eof, eofid, "long_name", "Empirical Orthogonal Functions" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, eofid, "short_name", "EOFs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_eof, eofid, "missing_value", real( missing2, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_eof, eofid, "_FillValue", real( missing2, dp ) )
!
    else
!
        status = nf90_put_att( ncid_eof, eofid, "missing_value", real( missing2, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_eof, eofid, "_FillValue", real( missing2, sp ) )
!
    end if
!
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, pcid, "long_name", "Principal Components" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, pcid, "short_name", "PCs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, singid, "long_name", "Singular Values" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, singid, "short_name", "SingVals" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, expvarid, "long_name", "Explained Variances" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, expvarid, "short_name", "ExpVars" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_eof, nf90_global, "title",                &
                     trim(var_name)//"EOF analysis with missing values" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, nf90_global, "distance", cdistance(distance) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, nf90_global, "algorithm", name_proc//calgorithm(alg) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_eof, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_eof )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW  NETCDF DATASET.
!
    do j = 1, 3    
!
        select case ( j )
            case ( 1 )
                j1 = 1
                j2 = nlon
            case ( 2 )
                j1 = 1
                j2 = nlat
            case ( 3 )
                j1 = time1
                j2 = time3
        end select
!
        if ( use_coor_variable(j) ) then
!
            allocate( record(j2), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            status = nf90_get_var( ncid, coor_id(j), record, start=(/ j1 /), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_eof, coor_id2(j), record )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            deallocate( record )
        else
            status = nf90_put_var( ncid_eof, coor_id2(j), arth(j1,1_i4b,j2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
    tmp = sqrt( real(time3, stnd) )
!
    startlec(1:2) = 1
!
!   WRITE EOFS.
!
    do i = 1, neof
!
        left_vec(:nvar,i) = left_vec(:nvar,i)/tmp
        temp(:nlon,:nlat) = unpack( left_vec(:nvar,i), mask, missing2 )
!
        if ( grid_type/='n' .and. nlon==nlon_def ) then
!
            if ( lat2>=nlat_def-1 .and. nlat==nlat_def ) then
!
!               ADD DUPLICATE POINTS TO THE ORCA GRID IF THE WHOLE GLOBE IS USED.
!
                call duplicate_orca_points_2d( grid_type, mask, temp )
            else
!
!               ADD DUPLICATE LONGITUDES LINES TO THE ORCA GRID IF THE WHOLE LONGITUDE
!               DOMAIN IS USED.
!
                temp(1,:nlat)    = temp(nlon-1,:nlat)
                temp(nlon,:nlat) = temp(2,:nlat)
            end if
!
        end if
!
        startlec(3) = i
!
        status = nf90_put_var( ncid_eof, eofid, temp, start=startlec )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   WRITE EOF AMPLITUDE TIME SERIES.
!
    right_vec(:time3,:neof) = right_vec(:time3,:neof)*tmp
    status = nf90_put_var( ncid_eof, pcid, transpose(right_vec(:time3,:neof)) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE SINGULAR VALUES.
!
    sing_val(:neof) = sing_val(:neof)/tmp
    status = nf90_put_var( ncid_eof, singid, sing_val(:neof) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE ESTIMATES OF EXPLAINED VARIANCES.
!
    sing_val(:neof)  = (sing_val(:neof)**2)/var_tot
    status = nf90_put_var( ncid_eof, expvarid, sing_val(:neof) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( temp, mask, mat, left_vec, right_vec, sing_val, var_row )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>=2 ) then
        status = nf90_close( ncid_clim )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_eof )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM comp_eof_miss_3d
! ===============================
!
end program comp_eof_miss_3d
