program comp_freq_func_1d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_freq_func_1d:
!
! =================
! comp_freq_func_1d
! =================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 02/01/2018
!
! Purpose
! =======
!
! Estimate the frequency response function (e.g. the transfer function) of a symmetric linear filter
! (e.g. a high, low or band-pass Lanczos or Hamming filter in this version of comp_freq_func_1d ) as computed
! by :ref:`comp_lanczos_filter_3d` or :ref:`comp_symlin_filter_3d` (or the corresponding procedures for
! unidimensional or fourdimensional NetCDF variables).
!
! For more details on the frequency response function of a frequency filter, see [Bloomfield]_.
!
! The frequency response function is computed at ``NFREQ`` frequencies where ``NFREQ`` is the value 
! given at the **-nf=** argument.
!
! The ``NFREQ`` frequencies are regularly sampled between ``0`` and the Nyquist frequency if the optional
! argument **-fourfreq** is not used or are the ``NFREQ`` Fourier frequencies ``2*pi*j/NFREQ`` for ``j=0`` to ``NFREQ-1``
! if this argument is used.
!
! The frequency response function is written in a NetCDF dataset on output for later use.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_freq_func_1d \
!         -nf=number_of_frequencies \
!         -pl=minimum_period \
!         -ph=maximum_period \
!         -nfc=number_of_filter_coefficients (optional) \
!         -o=output_netcdf_file              (optional) \
!         -mi=missing_value                  (optional) \
!         -hamming                           (optional) \
!         -win=window_choice                 (optional : 0.5 > 1.) \
!         -fourfreq                          (optional) \
!         -notestf                           (optional) \
!         -double                            (optional) \
!         -hdf5                              (optional) \
!         -tlimited                          (optional)
!
! By default
! ----------
!
!        *-nfc=*
!         the *number_of_filter_coefficients* is determined in order to optimize 
!         the frequency response function of the selected filter.
!        *-o=*
!         the *output_netcdf_file* is named ``filter_freq_func.nc``
!        *-mi=*
!         the *missing_value* for the output variable is equal to ``1.e+20``
!        *-hamming*
!         a Lanczos window filter is used. If the **-hamming** argument is specified a Hamming window filter is
!         used instead
!        *-win=*
!         a Hamming window (e.g. **-win=**\ ``0.54``) is convolved with the filter response by default if
!         the **-hamming** argument is used, meaning that a Hamming window filter is estimated. If the **-hamming**
!         argument is not used, this argument has no effect
!        *-fourfreq*
!         the frequency response function is not computed at the Fourier frequencies. If the *-fourfreq*
!         argument is used the frequency response function is estimated at the Fourier frequencies
!        *-notestf*
!         Normally, the value of the **-nfc=** argument must be chosen such that 
!         the full transition bands about the cutoff frequencies ``1/PH`` and ``1/PL``
!         (where ``PH`` and ``PL`` are the values of the **-ph=** and **-pl=**
!         arguments, respectively) of the selected filter are inside the (``0`` ``0.5``) 
!         frequency interval. By using the **-notestf** argument you can get ride of this limitation
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------

!   #) The **-nf=** argument specifies the number of frequencies at which the frequency response function
!      must be evaluated.
!      The **-nf=** argument is a strictly positive integer.
!
!   #) The **-pl=** argument specifies the minimum period of oscillation of the resulting filtered time series. 
!      The *minimum_period* is expressed in number of time observations. 
!
!      Do not use the **-pl=** argument or use **-pl=**\ ``0`` for high-pass filtering 
!      frequencies corresponding to periods shorter than **-ph=**\ ``PH`` .
!
!      The **-pl=** argument is a positive integer equal to ``0`` or greater than ``2``.
!
!   #) The **-ph=** argument specifies the maximum period of oscillation of the resulting filtered time series. 
!      The *maximum_period* is expressed in number of time observations. 
!      Do not use the **-ph=** argument or use **-ph=**\ ``0`` for low-pass filtering 
!      frequencies corresponding to periods longer than **-pl=**\ ``PL`` .
!      For example, **-pl=**\ ``6``  (or ``18``) and **-ph=**\ ``32``  (or ``96``) select periods
!      between ``1.5`` and ``8``  years for quarterly (monthly) time series.
!
!      The **-ph=** argument is a positive integer equal to ``0`` or greater than ``2`` and 
!      less than the length of the multichannel time series or the *periodicity* if the **-p=**
!      argument is used.
!
!      The **-ph=** argument must also be greater or equal to the **-pl=** argument if 
!      both are specified.
!
!   #) Setting **-pl=** and **-ph=** to the same value ``P`` is allowed. 
!      In this case, an -ideal- band-pass filter with peak response near one
!      at the single period ``P`` is computed and applied to the time series.
!
!   #) Setting both **-pl=**\ ``0`` and **-ph=**\ ``0`` is not allowed since in that case,
!      no frequency filtering is done.
!
!   #) The **-nfc=** argument specifies the desired number of symmetric linear filter coefficients.
!      If **-nfc=** is not specified, an optimal value is chosen in order to obtain a good frequency
!      response function for the selected filter.
!
!      However, if **-nfc=** is set to ``K``, the first and last ``(K-1)/2`` time observations in the 
!      resulting filtered time series will be affected by end effects. Thus, the user must choose the number of 
!      filter terms, ``K``, as a compromise between:
!
!           1)  A sharp cutoff, that is, ``1/K`` small; and
!
!           2)  Minimizing the number of data points lost or affected by the filtering operations
!               (since ``(K-1)/2`` data points will be lost or affected from each end of the series).
!
!      Finally, the **-nfc=** argument must be greater or equal to ``3`` and odd.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output variable in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* is set
!      to ``1.e+20``.
!
!   #) If the **-hamming** argument is specified, a Hamming window filter is used instead of
!      Lanczos window filter.
!
!   #) The **-win=** argument controls the form of the window which will be convolved with the filter
!      if a Hamming window filter is requested with the **-hamming** argument.
!      By default, a Hamming window is used (e.g. **-win=**\ ``0.54``).
!
!      Set **-win=**\ ``0.5`` for using a Hanning window or **-win=**\ ``1.``. for a rectangular
!      window  (e.g. "ideal" filter).
!
!      This argument has an effect only if the **-hamming** argument is also specified.
!      The **-win=** argument is a real number greater or equal to ``O.5`` and less or equal to ``1``.
!
!   #)  If the **-fourfreq** argument is used the frequency response function
!       is evaluated at the Fourier frequencies ``2*pi*j/NFREQ`` for ``j=0`` to ``NFREQ-1``,
!       where ``NFREQ`` is the value of the **-nf=** argument.
!
!   #) The **-notestf** argument allows to bypass some of the restrictions on the 
!      number of filter coefficients as specified with the **-nfc=** argument.
!
!      By default, the value of the **-nfc=** argument must be chosen such that the 
!      full transition bands about the cutoff frequencies (``1/PH``  and ``1/PL``) of
!      the selected filter are inside the (``0`` ``0.5``) frequency interval.
!
!      When the **-notestf** argument is specified, only the cutoff frequencies (e.g. ``1/PH``
!      and ``1/PL``) of the selected filter must lie in the (``0`` ``0.5``) frequency interval
!      and not the full transition bands around them.
!
!      This allows to diminish the number of filter coefficients and, thus, to minimize the number
!      of data points lost by the filtering operation (if **-nfc=** is set to ``K``, ``(K-1)/2``  data 
!      points will be "lost" or affected by end effects from each end of the series).
!
!   #) The **-double** argument specifies that, the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
!   #) For more details on Lanczos or symmetric linear filtering and examples of use in the climate literature, see
!
!      * "Fourier analysis of time series- An introduction",
!        by P. Bloomfield, John Wiley and Sons, New York, Chapter 6, 1976. http://as.wiley.com/WileyCDA/WileyTitle/productCd-0471889482.html
!
!      * "Lanczos filtering in one and two dimensions",
!        by C. Duchon, 1016-1022, Journal of applied meteorology, vol. 18, 1979. `10.1175/1520-0450(1979)018<1016:LFIOAT>2.0.CO;2`_
!
! .. _`10.1175/1520-0450(1979)018<1016:LFIOAT>2.0.CO;2` : https://doi.org/10.1175/1520-0450(1979)018<1016:LFIOAT>2.0.CO;2
!
! Outputs
! -------
!
!   comp_freq_func_1d creates an output NetCDF file that contains the frequency response function of the selected filter.
!   The output NetCDF dataset contains the following NetCDF variable (in the description below, NFREQ is the
!   *number_of_frequencies* specified for the **-nf=** argument) :
!
!   #) transfert_function\ ``(NFREQ)`` : the frequency response function of the selected filter at the selected frequencies.
!
! Examples
! --------
!
!   #)   For estimating the frequency response function of a band-passed Lanczos filter (in the biennial time scale
!        if we assume monthly time series since **-pl=**\ ``18``  and **-ph=**\ ``30`` are specified) at ``200`` frequencies
!        regularly sampled between ``0`` and the Nyquist frequency, use the following commands :
!
!        ::
!
!          $ comp_freq_func_1d  \
!            -nf=200 \
!            -pl=18  \
!            -ph=30  \
!            -o=freq_func.nc
!
!   #)   For estimating the frequency response function of the same band-passed Lanczos filter, but at the ``200`` Fourier
!        frequencies between ``0`` and the Nyquist frequency, use the following commands :
!
!        ::
!
!          $ comp_freq_func_1d  \
!            -nf=200    \
!            -pl=18     \
!            -ph=30     \
!            -fourfreq  \
!            -o=freq_func.nc
!
!-
! ================================================================================================
!
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_name, nf90_clobber, nf90_real4, nf90_real8, nf90_noerr, nf90_strerror,      &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,               &
                         nf90_def_var, nf90_put_att, nf90_enddef, nf90_global, nf90_put_var, nf90_close
!
    use Statpack, only : i1b, i4b, stnd, lgl, true, false, zero, one, half, kfix, kexp, kint, tolower,     & 
                         case_change, string_to_val, allocate_error,                                       &
                         merror, lp_coef, lp_coef2, hp_coef, hp_coef2, bd_coef, bd_coef2,                  &
                         pk_coef, freq_func
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing
!
!
! STRONG TYPING IMPOSED
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_freq_func_1d : '
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid_filt, filtid, periodid, coorid,    &
                     i, i1, i2, nfreq, pl, ph, nfc, method, output_type,   &
                     output_mode, dimid, dimlen
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),   allocatable    :: coef, freqr, freqi, period
    real(stnd)                                 :: missing, tmp, freq, win
!
    logical(lgl)                               :: use_tlimited, notest_freq, use_four_freq, use_hamming
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: filt_name, dim_name
    character(len=14)                          :: fmt
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute the frequency response function of a symmetric linear filter'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_freq_func_1d  '
        write (prtunit,*) '        -nf=number_of_frequencies'
        write (prtunit,*) '        -pl=minimum_period'
        write (prtunit,*) '        -ph=maximum_period'
        write (prtunit,*) '        -nfc=number_of_filter_coefficients      (optional)'
        write (prtunit,*) '        -o=output_netcdf_file                   (optional)'
        write (prtunit,*) '        -mi=missing_value                       (optional)'
        write (prtunit,*) '        -hamming                                (optional)'
        write (prtunit,*) '        -win=window_choice                      (optional : 0.5 > 1.)'
        write (prtunit,*) '        -fourfreq                               (optional)'
        write (prtunit,*) '        -notestf                                (optional)'
        write (prtunit,*) '        -double                                 (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                                   (optional)'
#endif
        write (prtunit,*) '        -tlimited                               (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-nfc=            : by default, the number of filter coefficients is determine to optimize'
        write (prtunit,*) '                   the frequency response function of the selected filter'
        write (prtunit,*) '-o=filter_freq_func.nc'
        write (prtunit,*) '-mi=1.e+20       : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-win=0.54        : by default, a Hamming window is convolved with the filter response if'
        write (prtunit,*) '                   -hamming is specified'
        write (prtunit,*) '-hamming         : by default, a Lanczos window filter is computed.'
        write (prtunit,*) '-fourfreq        : by default, the transfert function is not computed  at the Fourier'
        write (prtunit,*) '                   frequencies'
        write (prtunit,*) '-notestf         : by default, the value of the -nfc= argument must be choosen such that the'
        write (prtunit,*) '                   full transition bands about the cutoff frequencies of the selected filter'
        write (prtunit,*) '                   are inside the (0 0.5) interval'
        write (prtunit,*) '-double          : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5            : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited        : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>13 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    filt_name  = " "
!
    history = "comp_freq_func_1d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
!
    nfreq         = -1
    pl            = -1
    ph            = -1
    nfc           = 0
    win           = 0.54_stnd
!
    missing = def_missing
!
    use_hamming      = false
    use_four_freq    = false
    notest_freq      = false
    use_tlimited     = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                filt_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-nf=')
!
!                       GET THE NUMBER OF FREQUENCIES.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) nfreq
                        end if
                        if ( kcode/=kint .or. nfreq<=0 )                           &
                        call merror( name_proc//"The -nf= argument is a strictly positive integer value" )
!
                    case ('-pl=')
!
!                       GET THE MINIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) pl
                        end if
                        if ( kcode/=kint .or. pl<0 .or. pl==1 .or. pl==2 )                           &
                        call merror( name_proc//"The -pl= argument is an integer equal to 0 or greater than 2" )
!
                    case ('-ph=')
!
!                       GET THE MAXIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) ph
                        end if
                        if ( kcode/=kint .or. ph<0 .or. ph==1 .or. ph==2  )                           &
                        call merror( name_proc//"The -ph= argument is an integer equal to 0 or greater than 2" )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument is a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5) )
!
                            case ('-nfc=')
!
!                               DETERMINE THE NUMBER OF FILTER COEFFICIENTS.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) nfc
                                end if
                               if ( kcode/=kint .or. nfc<3 .or. (nfc/2)*2==nfc  )                  &
                               call merror( name_proc//"The -nfc= argument is an odd integer greater than 2" )
!
                            case ('-win=')
!
!                               DETERMINE THE FORM OF THE HAMMING WINDOW.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) win
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     win<half                        .or.     &
                                     win>one )                                &
      call merror( name_proc//"The -win= argument must be a real number and verify 0.5<=win<=1 " )
!
                            case default
!
                                select case ( name(1:) )
!
                                    case ('-hamming')
!
!                                       COMPUTE A HAMMING WINDOW FILTER.
!
                                        use_hamming = true
!
                                    case ('-fourfreq')
!
!                                       COMPUTE THE TRANSFERT FUNCTION AT THE FOURIER FREQUENCIES.
!
                                        use_four_freq = true
!
                                    case ('-notestf')
!
!                                       BYPASS THE TESTS ON THE CUTOFF FREQUENCIES.
!
                                        notest_freq = true
!
                                    case ('-double')
!
!                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                        output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
#endif
!
                                    case ('-tlimited')
!
!                                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                                        use_tlimited = true
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( filt_name==" " ) filt_name="filter_freq_func.nc"
!
    if ( nfreq==-1 )                                              &
    call merror( name_proc//"The -nf= argument is missing" )
!
    if ( pl==-1 )                                                 &
    call merror( name_proc//"The -pl= argument is missing" )
!
    if ( ph==-1 )                                                 &
    call merror( name_proc//"The -ph= argument is missing" )
!
    if ( max(pl,ph)==0 )                                          &
    call merror( name_proc//"Setting both -pl=0 and -ph=0  is not allowed" )
!
    if ( pl==ph .and. use_hamming )                               &
    call merror( name_proc//"Setting -pl= equal to -ph= with -hamming argument is not allowed" )
!
!   CHECK PERIODS ORDERING TO SELECT FREQUENCIES.
!
    if ( ph<pl .and. ph/=0 )     &
    call merror( name_proc//"The -ph= argument must be greater or equal to the -pl= argument" )
!
    if ( ph==0 ) then
!
!       LOW-PASS FILTERING OF THE TIME SERIES.
!
        method = 1
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS.
!
            tmp = one/real( pl, stnd )
            i   = ceiling( one/(half-tmp) )
            nfc = max( i, pl+1 )
!
        end if
!
    else if ( pl==0 ) then
!
!       HIGH-PASS FILTERING OF THE TIME SERIES.
!
        method = 2
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS.
!
            tmp = one/real( ph, stnd )
            i   = ceiling( one/(half-tmp) )
            nfc = max( i, ph+1 )
!
        end if
!
    else if ( pl/=ph ) then
!
!       BAND-PASS FILTERING OF THE TIME SERIES.
!
        method = 3
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS, LANCZOS FILTERING IS ASSUMED.
!
            tmp = one/real( pl, stnd )
            i1  = ceiling( one/(half-tmp) )
!
            tmp = tmp - one/real( ph, stnd )
            i2  = ceiling(  2.6/tmp )
!
            nfc = max( i1, i2, ph+1 )
!
        else if ( .not.use_hamming ) then
!
!           CHECK THAT nfc IS LARGE ENOUGH TO ENSURE UNIT RESPONSE AT BAND CENTER.
!
            tmp = one/real( pl, stnd ) - one/real( ph, stnd )   
!
            if ( real( nfc+1, stnd ) < (2.6/tmp) )         &
            call merror( name_proc//"The -nfc= argument is too small to ensure unit response at band center"  )
!
        end if
!
    else
!
!       BAND-PASS FILTERING OF THE TIME SERIES WITH A PEAK RESPONSE NEAR ONE AT THE PERIOD pl=ph.
!
        method = 4
!
        freq = one/real( pl, stnd )
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS.
!
            tmp = 2.3*max( real( pl, stnd ), one/(half-freq) )
            nfc = ceiling( tmp )
!
        end if
!
    end if
!
!   CHECK IF THE NUMBER OF FILTER COEFFICIENTS IS ODD.
!
    if ( (nfc/2)*2==nfc  ) nfc = nfc + 1
!
!   ALLOCATE WORK ARRAYS FOR THE FILTER COEFFICIENTS, THE VECTOR OF FREQUENCIES AND THE TRANSFERT FUNCTION.
!
    allocate( freqr(nfreq), freqi(nfreq), period(nfreq), coef(nfc), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   COMPUTE THE SYMMETRIC LINEAR FILTER COEFFICIENTS.
!
    if ( use_hamming ) then
!
!       A HAMMING WINDOW FILTER IS REQUESTED.
!
        select case( method )
!
            case( 1_i4b )
!
!               LOW-PASS FILTERING OF THE TIME SERIES.
!
                coef(:nfc) = lp_coef2( pl, nfc, win=win, notest_fc=notest_freq )
!
            case( 2_i4b )
!
!               HIGH-PASS FILTERING OF THE TIME SERIES.
!
                coef(:nfc) = hp_coef2( ph, nfc, win=win, notest_fc=notest_freq )
!
            case( 3_i4b )
!
!               BAND-PASS FILTERING OF THE TIME SERIES.
!
                coef(:nfc) = bd_coef2( pl, ph, nfc, win=win, notest_fc=notest_freq )
!
        end select
!
    else
!
!       A LANCZOS WINDOW FILTER IS REQUESTED.
!
        select case( method )
!
            case( 1_i4b )
!
!               LOW-PASS FILTERING OF THE TIME SERIES.
!
                coef(:nfc) = lp_coef( pl, nfc, notest_fc=notest_freq )
!
            case( 2_i4b )
!
!               HIGH-PASS FILTERING OF THE TIME SERIES.
!
                coef(:nfc) = hp_coef( ph, nfc, notest_fc=notest_freq )
!
            case( 3_i4b )
!
!               BAND-PASS FILTERING OF THE TIME SERIES.
!
                coef(:nfc) = bd_coef( pl, ph, nfc, notest_fc=notest_freq )
!
            case( 4_i4b )
!
!               BAND-PASS FILTERING OF THE TIME SERIES WITH A PEAK RESPONSE NEAR ONE AT THE PERIOD pl=ph .
!
                coef(:nfc) = pk_coef( freq, nfc, notest_freq=notest_freq )
!
        end select
!
    end if
!
!   COMPUTE THE TRANSFERT FUNCTION OF THE DESIGNED SYMMETRIC LINEAR FILTER.
!
    call freq_func( nfreq, coef(:nfc), freqr(:nfreq), four_freq=use_four_freq, freq=freqi(:nfreq) )
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    status = nf90_create( filt_name, cmode=output_mode, ncid=ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_filt, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    if ( use_tlimited ) then
        dimlen = nfreq
    else
        dimlen = nf90_unlimited
    end if
!
    dim_name = "frequency"
!
    status = nf90_def_dim( ncid_filt, name=trim(dim_name), len=dimlen, dimid=dimid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    status = nf90_def_var( ncid_filt, trim(dim_name), output_type, dimids=dimid, varid=coorid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_filt, "periodicity",  output_type, dimids=dimid, varid=periodid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_filt, "transfert_function",  output_type, dimids=dimid, varid=filtid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    status = nf90_put_att( ncid_filt, coorid, "long_name", "Frequency axis" )
    if ( status/=nf90_noerr )  call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, coorid, "short_name", "Frequency" )
    if ( status/=nf90_noerr )  call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, coorid, "units", "cycles per unit of time" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, periodid, "long_name", "Periodicity axis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, periodid, "short_name", "Periodicity" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, periodid, "units", "number of time observations" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   COPY ATTRIBUTES FOR THE FILTERED VARIABLE.
!
    if (  output_type==nf90_real8 ) then
!
       status = nf90_put_att( ncid_filt, periodid, "missing_value", real( missing, dp ) )
       if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
       status = nf90_put_att( ncid_filt, periodid, "_FillValue", real( missing, dp ) )
       if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
       status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, dp ) )
       if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
       status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, coorid, "valid_min", real( zero, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, coorid, "valid_max", real( half, dp ) )
!
    else
!
       status = nf90_put_att( ncid_filt, periodid, "missing_value", real( missing, sp ) )
       if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
       status = nf90_put_att( ncid_filt, periodid, "_FillValue", real( missing, sp ) )
       if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, coorid, "valid_min", real( zero, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, coorid, "valid_max", real( half, sp ) )
!
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_filt, nf90_global, "title", "Transfer function of symmetric linear filter" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "minimum_period", pl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "maximum_period", ph )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "number_of_filter_coefficients", nfc )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_hamming ) then
!
        name = "Hamming"
!
        status = nf90_put_att( ncid_filt, nf90_global, "window_parameter", win )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        name = "Lanczos"
!
    end if
!
    status = nf90_put_att( ncid_filt, nf90_global, "Window", trim( name ) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_four_freq ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_filt, nf90_global, "Fourier_frequencies", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
!   WRITE FREQUENCY AND PERIOD AXES.
!
    status = nf90_put_var( ncid_filt, coorid, freqi(:nfreq) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    period(1)       = missing
    period(2:nfreq) = one/freqi(2:nfreq)
!
    status = nf90_put_var( ncid_filt, periodid, period(:nfreq) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE TRANSFERT FUNCTION.
!
    status = nf90_put_var( ncid_filt, filtid, freqr(:nfreq) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS .
!
    deallocate( coef, freqi, freqr, period )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_freq_func_1d
! ================================
!
end program comp_freq_func_1d
