program comp_index_1d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_index_1d:
!
! =============
! comp_index_1d
! =============
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 02/01/2018
!
! Purpose
! =======
!
!  Compute an index time series from two time series extracted from two unidimensional
!  variables in one or two NetCDF datasets and stored the new index time
!  series in a NetCDF dataset.
!
!  The two original time series can be computed by :ref:`comp_serie_3d`, :ref:`comp_serie_miss_3d`
!  or similar procedures for fourdimensional NetCDF variables.
!
!  Different operations are available for estimating the index time series from the
!  original time series (see the description of the **-op=** argument below).
!
!  Different operations are also available for transforming or normalizing the index
!  time series after its estimation from the two original time series (see the description
!  of the **-a=** argument below).
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_index_1d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -v2=netcdf_variable2 \
!         -f2=input_netcdf_file2                   (optional) \
!         -t=timea1,timea2                         (optional) \
!         -t2=timeb1,timeb2                        (optional) \
!         -a=type_of_transformation                (optional : scp, cov, cor) \
!         -p=periodicity                           (optional) \
!         -o=output_netcdf_file                    (optional) \
!         -n=output_netcdf_variable                (optional) \
!         -op=type_of_operation                    (optional : +, -, *, /) \
!         -sm=smoothing_factor                     (optional) \
!         -mi=missing_value                        (optional) \
!         -3d                                      (optional) \
!         -double                                  (optional) \
!         -hdf5                                    (optional) \
!         -tlimited                                (optional)
!
! By default
! ----------
!
!        *-f2=*
!         the *input_netcdf_file2* is the same as *input_netcdf_file*. This means that
!         the *netcdf_variable2* is extracted from the same NetCDF dataset as *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-t2=*
!         the whole time period associated with the *netcdf_variable2*
!        *-a=*
!         the *type_of_transformation* is set to ``scp``. This means that the index
!         time series is computed from the raw original time series and is not normalized or transformed
!        *-p=*
!         the *periodicity* is set to ``1``. This means that the index time series
!         does not have a seasonal cycle.
!        *-o=*
!         the *output_netcdf_file* is named ``index_``\ *netcdf_variable*\ ``.``\ *netcdf_variable2*\ ``.nc``
!        *-n=*
!         the *output_netcdf_variable* is named
!         `netcdf_variable_netcdf_variable2_index`
!        *-op=*
!         the *type_of_operation* is set to ``-``. This means that the index
!         time series is computed as the differences between the two original time series
!        *-sm=*
!         no smoothing is applied to the index time series
!        *-mi=*
!         the *missing_value* in the output NetCDF file is set to ``1.e+20``
!        *-3d*
!         the *output_netcdf_variable* is defined as an unidimensional NetCDF variable. However,
!         if **-3d** is activated, the *output_netcdf_variable* is defined as an tridimensional NetCDF variable
!         but with two dummy dimensions (e.g. with a length equal to ``1`` )
!        *-double*
!         the data are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         data are stored as double-precision floating point numbers 
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* specifies the first unidimensional time series for computing
!      the index and the **-f=**\ *input_netcdf_file* argument specifies that this NetCDF variable must
!      be extracted from the NetCDF file *input_netcdf_file*.
!
!   #) The **-v2=**\ *netcdf_variable2* specifies the second unidimensional time series for computing
!      the index and the **-f2=**\ *input_netcdf_file2* argument specifies that this NetCDF variable must
!      be extracted from the NetCDF file *input_netcdf_fil2e*.
!      If the **-f2=** argument is absent, it is assumed that this *netcdf_variable2* is also in the NetCDF
!      dataset specifies by the **-f=** argument.
!
!   #) If the **-t=**\ *timea1,timea2* argument is missing, the whole time period
!      associated with the *netcdf_variable* is used to compute the first time series.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *timea2* - *timea1* + ``1`` time observations.
!
!   #) If the **-t=**\ *timeb1,timeb2* argument is missing, the whole time period
!      associated with the *netcdf_variable2* is used to compute the second time series.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *timeb2* - *timeb1* + ``1`` time observations.
!
!   #) The selected number of time observations by the **-t=** and **-t2=** arguments must be equal.
!      This means that the following equality must hold
!
!               *timea2* - *timea1* + ``1`` = *timeb2* - *timeb1* + ``1``
!
!      otherwise, an error message will be issued and the program will stop.
!
!   #) The **-a=**\ ``type_of_transformation`` argument specifies how the index time series is normalized or transformed. If:
!
!      * **-a=**\ ``scp``, the raw index time series is stored
!
!      * **-a=**\ ``cov``, the anomalies of the raw index time series are computed and 
!        stored in the output NetCDF file
!
!      * **-a=**\ ``cor``, the standardized anomalies of the raw index time series are 
!        computed and stored in the output NetCDF file.
!
!      The default is **-a=**\ ``scp`` meaning that the raw index time series is stored.
!
!   #) If the **-p=** argument is specified and **-a=**\ ``cov`` or **-a=**\ ``cor``, the anomalies or the
!      standardized anomalies are computed from the raw index time series by taken into account the *periodicity*
!      of the data as specified by the **-p=** argument.
!
!   #) The **-op=**\ ``type_of_operation`` argument specifies how the index time series is computed from
!      the two original time series. If:
!
!      * **-op=**\ ``-``, the default, the index is computed as the difference between the two original time series
!
!      * **-op=**\ `+``, the index is computed as the sum of the two original time series
!
!      * **-op=**\ ``*``, the index is computed as the product of the two original time series
!
!      * **-op=**\ ``/``, the index is computed as the quotient of the two original time series.
!
!   #) **-sm=**\ *smoothing_factor* means that the (transformed) index time series must be smoothed with a moving average of
!      approximately ``2``\ *\ *smoothing_factor*\ +\ ``1`` terms. *smoothing_factor* must be an integer greater
!      than ``0``. However, if missing values are present in the time original time series or generated in the computation
!      of the index, smoothing is not allowed.
!
!   #) The **-n=**\ *output_netcdf_variable* argument specifies the NetCDF variable which will contains 
!      the computed time series in the output NetCDF file, *output_netcdf_file*, specified by the
!      **-o=**\ *output_netcdf_file* argument.
!
!   #) Missing values are allowed in both NetCDF variables associated with the two original time series. However, if missing
!      values are present, the **-sm=** argument is not allowed.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the NetCDF variable in the *output_netcdf_file*.
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``.
!
!   #) The **-3d** argument specify that the index time series must be stored as
!      a tridimensional NetCDF variable with two dummy dimensions in the output NetCDF file.
!      By default, the index time series is stored as an unidimensional NetCDF variable.
!
!   #) The **-double** argument specify that, the index time series is stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the index time series is stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (eg ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
! Outputs
! -------
!
!   comp_index_1d creates an output NetCDF file that contains the computed index time series.
!   The output NetCDF dataset contains the following NetCDF variable (in the description below,
!   ntime is the number of time steps selected with the **-t=** and **-t2=** arguments):
!
!   #) *output_netcdf_variable*\ ``(ntime)`` : the computed index time series defined as an unidimensional variable.
!
!   or if the **-3d** argument has been specified :
!
!   #) *output_netcdf_variable*\ ``(ntime,1,1)`` : the computed index time series defined as a tridimensional variable
!      with two dummy dimensions.
!
!
! Examples
! --------
!
!   #)   For computing an index time series named ``siod`` as the difference between two unidimensional NetCDF variables
!        both called ``sst`` and extracted, respectively, from the files ``HadISST1_2m_187001_200702_swiosst_nt67.nc``
!        and ``HadISST1_2m_187001_200702_seiosst_nt67.nc``, and store the result in a file named
!        ``HadISST1_2m_187001_200702_sdisst_nt67.nc``, use the following command :
!
!        ::
!
!          $ comp_index_1d \
!            -f=HadISST1_2m_187001_200702_swiosst_nt67.nc -v=sst \
!            -f2=HadISST1_2m_187001_200702_seiosst_nt67.nc -v2=sst \
!            -o=HadISST1_2m_187001_200702_sdisst_nt67.nc -n=siod
!                                                                              
!-
! ================================================================================================
!
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,   only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_var_dims, nf90_max_name, nf90_real4, nf90_real8,              &
                         nf90_open, nf90_noerr, nf90_get_var,                                   &
                         nf90_get_att, nf90_create, nf90_clobber, nf90_set_fill, nf90_nofill,   &
                         nf90_strerror, nf90_unlimited, nf90_def_dim, nf90_def_var,             &
                         nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_inquire_attribute, &
                         nf90_enddef, nf90_put_var, nf90_close, nf90_nowrite, nf90_global
!
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, zero, kfix, kexp,            & 
                         kint, tolower, case_change, string_to_val,                           &
                         allocate_error, merror, arth, comp_mvs_grp, comp_anoma_grp_miss,     &
                         comp_smooth
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing, get_listarg,   &
                                 get_variable_dims, get_variable_dimcoords

!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
!
    character(len=*), parameter :: name_proc='comp_index_1d : ',       &
!
                                   cmethod(3)=(/"scp","cov","cor"/),    &          
!
                                   coperation(4)=(/"-","+","*","/"/)
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid2, ncid_index,                               &
                     varid, varid2, indexid, timeid,                                      &
                     i, nt, nperiod,                                                      &
                     timea1, timea2, timeb1, timeb2, time3,                               &
                     operation, method, smooth_factor, output_mode, output_type
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i4b), allocatable, dimension(:)    :: time_list, period
!
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),     allocatable  :: y, x1, x2, n_period
    real(stnd)                                 :: missing, tmp
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: is_missing, use_tlimited, use_3d, create_hdf5
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, fic2_name, var2_name,   &
                                                  out_name, index_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute an index time series from two time series extracted'
        write (prtunit,*) ' from two unidimensional variables in NetCDF datasets and stored'
        write (prtunit,*) ' the new index time series in a NetCDF dataset'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_index_1d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -v2=netcdf_variable'
        write (prtunit,*) '        -f2=input_netcdf_file2             (optional)'
        write (prtunit,*) '        -t=timea1,timea2                   (optional)'
        write (prtunit,*) '        -t2=timeb1,timeb2                  (optional)'
        write (prtunit,*) '        -o=output_netcdf_file              (optional)'
        write (prtunit,*) '        -n=output_netcdf_variable          (optional)'
        write (prtunit,*) '        -op=type_of_operation              (optional : +,-,*,/)'
        write (prtunit,*) '        -a=type_of_transformation          (optional : scp, cov, cor)'
        write (prtunit,*) '        -p=periodicity                     (optional)'
        write (prtunit,*) '        -sm=smoothing_factor               (optional)'
        write (prtunit,*) '        -mi=missing_value                  (optional)'
        write (prtunit,*) '        -3d                                (optional)'
        write (prtunit,*) '        -double                            (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                              (optional)'
#endif
        write (prtunit,*) '        -tlimited                          (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-f2= the input_netcdf_file2 is the same as input_netcdf_file'
        write (prtunit,*) '-t=  the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-t2= the whole time period  associated with the netcdf_variable2'
        write (prtunit,*) '-o=index_netcdf_variable.netcdf_variable2.nc'
        write (prtunit,*) '-n=netcdf_variable_netcdf_variable2_index'
        write (prtunit,*) '-op=-, the index is defiend as the difference between the two time series'
        write (prtunit,*) '-a=scp, the index is not normalized'
        write (prtunit,*) '-p=1'
        write (prtunit,*) '-sm= no smoothing is applied to the time series'
        write (prtunit,*) '-mi= the missing_value in the output NetCDF dataset is equal to 1.e+20'
        write (prtunit,*) '-3d       : the output_netcdf_variable is an unidimensional variable'
        write (prtunit,*) '-double   : the results are stored as single floating numbers'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>17 ) then
        call merror( name_proc//"Wrong number of line arguments" )
    end if
!
    fic_name   = " "
    var_name   = " "
    fic2_name  = " "
    var2_name  = " "
    out_name   = " "
    index_name = " "
!
    history = "comp_index_1d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
    operation     = 1
    method        = 1
    smooth_factor = 0
    nperiod       = 1
!
    missing = def_missing
!
    timea1     = 1
    timea2     = huge( timea2 )
!
    timeb1    = 1
    timeb2    = huge( timeb2 )
!
    use_3d       = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE FIRST NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE FIRST DATASET.
!
                var_name = name(4:)
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION .
!
                call get_listarg( name_proc, name(4:),                              &
                                  "timea1 and timea2 are integer values",           &
                                  "Wrong period definition for the first variable", &
                                  timea1, timea2 )
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF TRANSFORMATION.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case ( cmethod(3) )
                        method = 3
                    case default
                        call merror( name_proc//"Unknown transformation option" )
                end select
!
            case ('-p=')
!
!               GET THE PERIODICITY.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The periodicity is an integer value" )
!
                read( name(4:), fmt ) nperiod
!
                if ( nperiod<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE INDEX.
!
                out_name = name(4:)
!
            case ('-n=')
!
!               GET THE NAME OF THE INDEX TIME SERIES.
!
                index_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-f2=')
!
!                       GET THE NAME OF THE NETCDF DATASET FOR THE SECOND VARIABLE.
!
                        fic2_name = name(5:)
!
                    case ('-v2=')
!
!                       GET THE NAME OF THE SECOND NETCDF VARIABLE IN THE DATASET.
!
                        var2_name = name(5:)
!
                    case ('-t2=')
!
!                       GET THE PERIOD DEFINITION FOR THE INDICE VARIABLE.
!
                        call get_listarg( name_proc, name(5:),                            &
                                          "timeb1 and timeb2 are integer values",         &
                                          "Wrong period definition for second variable",  &
                                          timeb1, timeb2 )
                    case ('-op=')
!
!                       DETERMINE HOW TO COMPUTE THE INDEX.
!
                        select case ( name(5:) )
                            case ( coperation(1) )
                                operation = 1
                            case ( coperation(2) )
                                operation     = 2
                            case ( coperation(3) )
                                operation     = 3
                            case ( coperation(4) )
                                operation     = 4
                            case default
                                call merror( name_proc//"Unknown operation option" )
                        end select
!
                    case ('-sm=')
!
!                       DETERMINE THE SMOOTHING FACTOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) smooth_factor
                        end if
                        if ( smooth_factor<=0 )        &
         call merror( name_proc//"The -sm= argument is a strictly positive integer value " )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-3d')
!
!                               DEFINE THE OUTPUT VARIABLE AS A 3D NETCDF VARIABLE.
!
                                use_3d = true
!
                            case ('-double')
!
!                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"First input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"First input NETCDF variable is missing" )
!
    if ( var2_name==" " )                                                 &
    call merror( name_proc//"Second input NETCDF variable is missing" )
!
!   CHECK SECOND INPUT FILE NAME AND DEFINE DEFAULT IF NECESSARY.
!
    if ( fic2_name==" " )  fic2_name = fic_name
!
!   CHECK OUTPUT FILE NAME AND DEFINE DEFAULT IF NECESSARY.
!
    if ( out_name==" " )                                                  &
    out_name="index_"//trim(var_name)//"."//trim(var2_name)//".nc"
!
!   CHECK OUTPUT VARIABLE NAME AND DEFINE DEFAULT IF NECESSARY.
!
    if ( index_name==" " )                                                  &
    index_name=trim(var_name)//"_"//trim(var2_name)//"_index"
!
!   OPEN THE NETCDF DATASETS.
!
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( fic2_name==fic_name )  then
        ncid2 =ncid
    else
        status = nf90_open( trim(fic2_name), nf90_nowrite, ncid2 )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                    &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names, &
                                  dimlens_out=dimlens, coor_id_out=coor_id,           &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type, &
                                  use_coor_variable_out=use_coor_variable             )
!
    if ( count(dimlens/=0)/=1 )           &
    call merror( name_proc//"First input NETCDF variable is not unidimensional" )
!
    nt = dimlens(1)
!
!   CHECK THE TIME DEFINITION.
!
    timea1 = max( timea1, 1 )
    timea2 = min( timea2, nt )
    time3  = timea2 - timea1 + 1
    if ( time3<nperiod  ) call merror( name_proc//"Wrong time definition for the first NETCDF variable" )
!
!   GET THE SECOND VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS FOR THE SECOND NETCDF VARIABLE.
!
    call get_variable_dims( name_proc, trim(var2_name), ncid2,    &
                            varid2, dimids, dimlens )
!
    if ( count(dimlens/=0)/=1 )           &
    call merror( name_proc//"Second input NETCDF variable is not unidimensional" )
!
    nt = dimlens(1)
!
!   CHECK THE TIME DEFINITION.
!
    timeb1 = max( timeb1, 1 )
    timeb2 = min( timeb2, nt )
    i      = timeb2 - timeb1 + 1
    if ( i<nperiod  )                 &
    call merror( name_proc//"Wrong time definition for the second NETCDF variable" )
!
    if ( time3/=i )  call merror( name_proc//"-t= and -t2= options mismatch" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( x1(time3), x2(time3), y(time3), stat=status   )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   READ THE FIRST NETCDF VARIABLE.
!
    status = nf90_get_var( ncid, varid, x1(:time3), start=(/ timea1 /), count=(/ time3 /) )  
!
!   FIND THE MISSING INDICATOR VALUE ASSOCIATED WITH THE FIRST NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "missing_value", tmp )
    if ( status/=nf90_noerr ) then
        is_missing = false
    else
        is_missing = any( x1(:time3)==tmp )
    end if
!
    if ( is_missing ) then
        where( x1(:time3)==tmp ) x1(:time3)= missing
    end if
!
!   READ THE SECOND NETCDF VARIABLE.
!
    status = nf90_get_var( ncid2, varid2, x2(:time3), start=(/ timeb1 /), count=(/ time3 /) )  
!
!   FIND THE MISSING INDICATOR VALUE ASSOCIATED WITH THE FIRST NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid2, varid2, "missing_value", tmp )
!
    if ( status==nf90_noerr ) then
!
        if ( any( x2(:time3)==tmp ) ) then
            is_missing = true
            where(x2(:time3)==tmp) x2(:time3) = missing
        end if
!
    end if
!
!   COMPUTE THE INDEX TIME SERIES.
!
    if ( is_missing ) then
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid, varid, "scale_factor", tmp )
        if ( status==nf90_noerr )       &
        where(x1(:time3)/=missing) x1(:time3) = x1(:time3)*tmp
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid, varid, "add_offset", tmp )
        if ( status==nf90_noerr )       &
        where(x1(:time3)/=missing) x1(:time3) = x1(:time3) + tmp
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE SECOND NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid2, varid2, "scale_factor", tmp )
        if ( status==nf90_noerr )       &
        where(x2(:time3)/=missing) x2(:time3) = x2(:time3)*tmp
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE SECOND NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid2, varid2, "add_offset", tmp )
        if ( status==nf90_noerr )       &
        where(x2(:time3)/=missing) x2(:time3) = x2(:time3) + tmp
!
        select case ( operation )
!
           case ( 1 )
               where( x1(:time3)==missing .or. x2(:time3)==missing )
                   y(:time3) = missing
               elsewhere
                   y(:time3) = x1(:time3) - x2(:time3)
               endwhere
           case ( 2 )
               where( x1(:time3)==missing .or. x2(:time3)==missing )
                   y(:time3) = missing
               elsewhere
                   y(:time3) = x1(:time3) + x2(:time3)
               endwhere
           case ( 3 )
               where( x1(:time3)==missing .or. x2(:time3)==missing )
                   y(:time3) = missing
               elsewhere
                   y(:time3) = x1(:time3) * x2(:time3)
               endwhere
           case ( 4 )
               where( x1(:time3)==missing .or. x2(:time3)==missing .or. x2(:time3)==zero )
                   y(:time3) = missing
               elsewhere
                   y(:time3) = x1(:time3)/x2(:time3)
               endwhere
!
        end select
!
    else
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid, varid, "scale_factor", tmp )
        if ( status==nf90_noerr ) x1(:time3) = x1(:time3)*tmp
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid, varid, "add_offset", tmp )
        if ( status==nf90_noerr )  x1(:time3) = x1(:time3) + tmp
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE SECOND NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid2, varid2, "scale_factor", tmp )
        if ( status==nf90_noerr ) x2(:time3) = x2(:time3)*tmp
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE SECOND NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid2, varid2, "add_offset", tmp )
        if ( status==nf90_noerr )  x2(:time3) = x2(:time3) + tmp
!
        select case ( operation )
!
           case ( 1 )
               y(:time3) = x1(:time3) - x2(:time3)
           case ( 2 )
               y(:time3) = x1(:time3) + x2(:time3)
           case ( 3 )
               y(:time3) = x1(:time3) * x2(:time3)
           case ( 4 )
               where( x2(:time3)/=zero )
                   y(:time3) = x1(:time3)/x2(:time3)
               elsewhere
                   y(:time3) = missing
               endwhere
!
        end select
!
   end if
!
!   COMPUTE ANOMALIES OR STANDARDIZED ANOMALIES, IF REQUIRED.
!
    if ( method>1 ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( n_period(nperiod), period(nperiod),     &
                  time_list(time3), stat=status          )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        period(:nperiod)   = (/ (i,i=1,nperiod) /)
        time_list(:time3) = reshape( period(:nperiod), (/ time3 /), pad=period(:nperiod) )
!
!       COMPUTE UNIVARIATE STATISTICS FOR THE INDEX TIME SERIES.
!
        call comp_mvs_grp( y(:time3), true, true, nperiod, time_list(:time3),       &
                           x1(:nperiod), x2(:nperiod), n_period(:nperiod), xmiss=missing )
!
        if ( any( x1(:nperiod)==missing ) )        &
        call merror( name_proc//"Anomalies cannot be computed, climatology is missing " )
!
        if ( method==2 ) then
!
!           COMPUTE ANOMALIES.
!
            call comp_anoma_grp_miss( y(:time3), nperiod, time_list(:time3),       &
                                      missing, x1(:nperiod) )
        else
!
            if ( any( x2(:nperiod)==missing .or. x2(:nperiod)==zero ) )        &
            call merror( name_proc//"Anomalies cannot be standardized, std is missing or zero " )
!
!           COMPUTE STANDARDIZED ANOMALIES.
!
            call comp_anoma_grp_miss( y(:time3), nperiod, time_list(:time3),       &
                                      missing, x1(:nperiod), xstd_grp=x2(:nperiod) )
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( n_period, period, time_list )
!
    end if
!
!   SMOOTH THE INDEX TIME SERIES, IF REQUIRED.
!
    if ( smooth_factor/=0 ) then
!
        if ( any( y(:time3)==missing ) )        &
        call merror( name_proc//"The index time series can not be smoothed (missing values)" )
!
        call comp_smooth( y(:time3), smooth_factor )
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR STORING THE NEW INDEX.
!
    status = nf90_create( out_name, cmode=output_mode, ncid=ncid_index )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_index, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    if ( use_tlimited ) then
        dimlens(3)   = time3
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    dim_names(3) = dim_names(1)
!
    if ( use_3d ) then
!
!       DEFINE DUMMY DIMENSIONS FOR THE OUTPUT NETCDF VARIABLE.
!
        dim_names(1:2) = (/ "x_a", "y_a" /)
!
        dimlens(1:2) = 1
!
        do i = 1, 2
!
            status = nf90_def_dim( ncid_index, trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end if
!
    status = nf90_def_dim( ncid_index, name=trim(dim_names(3)), len=dimlens(3), dimid=dimids(3) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    status = nf90_def_var( ncid_index, trim(dim_names(3)), coor_type(1),                &
                           dimids=dimids(3), varid=timeid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_3d ) then
        status = nf90_def_var( ncid_index, trim(index_name), output_type, dimids=dimids(1:3), varid=indexid )
    else
        status = nf90_def_var( ncid_index, trim(index_name), output_type, dimids=dimids(3), varid=indexid )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
!   ASSIGN ATTRIBUTES TO THE TIME COORDINATE VARIABLE.
!
    if ( use_coor_variable(1) ) then
!
        do i = 1, coor_natts(1)
!
            status = nf90_inq_attname( ncid, coor_id(1), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE TIME COORDINATE VARIABLE OF THE FIRST VARIABLE.
!
            status = nf90_copy_att(  ncid, coor_id(1), att_name, ncid_index, timeid )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    else
!
        status = nf90_put_att( ncid_index, timeid, "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_index, timeid, "short_name", "Time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   ASSIGN ATTRIBUTES TO THE INDEX VARIABLE.
!
    name = trim(var_name)//" "//coperation(operation)//" "//trim(var2_name)
!
    status = nf90_put_att( ncid_index, indexid, "long_name", "Index of "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_index, indexid, "short_name", "Index"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   IF THE INDEX IS COMPUTED AS A SUM OR A DIFFERENCE, WE ASSUME THAT THE UNIT IS THE SAME
!   IN THE ORIGINAL NETCDF VARIABLES.
!
    if ( operation<=2 ) then
!
        status = nf90_inquire_attribute( ncid, varid, "units" )
        if ( status==nf90_noerr ) then
            status = nf90_copy_att(  ncid, varid, "units", ncid_index, indexid )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end if
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_index, indexid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_index, indexid, "_FillValue", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        status = nf90_put_att( ncid_index, indexid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_index, indexid, "_FillValue", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_index, nf90_global, "title", "Time series of "//trim(name) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_index, nf90_global, "operation", coperation(operation) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_index, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_index, nf90_global, "smoothing_factor", smooth_factor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_index, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_index )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
!
!   WRITE TIME VARIABLE FOR THE NEW NETCDF DATASET.
!
    if ( use_coor_variable(1) ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( record(time3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        status = nf90_get_var( ncid, coor_id(1), record, start=(/ timea1 /) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_index, timeid, record )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( record )
!
    else
!
        status = nf90_put_var( ncid_index, timeid, arth(timea1,1_i4b,time3) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   WRITE THE INDEX TIME SERIES.
!
    status = nf90_put_var( ncid_index, indexid, y(:time3) )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( y, x1, x2 )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( fic2_name/=fic_name ) then
        status = nf90_close( ncid2 )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_index )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_index_1d
! ============================
!
end program comp_index_1d
