program comp_invert_eof_4d
!! #############################################################################
!!
!! Copyright 2016 IRD
!!
!! This file is part of ncstat.
!!
!! ncstat is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! ncstat is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the ncstat/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_invert_eof_4d:
!
! ==================
! comp_invert_eof_4d
! ==================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 13/09/2018
!
! Purpose
! =======
!
! Approximate a fourdimensional NetCDF variable (or parts of it) from 
! its Empirical Orthogonal Function (EOF) decomposition or from results of a
! Singular Value Decomposition (SVD) analysis.
!
! Using as input a NetCDF file produced by :ref:`comp_eof_4d` or :ref:`comp_svd_3d`,
! this procedure computes an approximation of a fourdimensional NetCDF variable,
! packed as a *ntime* by *nv* rectangular matrix, **X**\ , of observed variables (e.g. the
! selected cells of the 3-D grid-mesh associated with the tridimensional NetCDF variable),
! of the form:
!
!           **X** = **A**\ **B** + **E**
!
! where
!
!  * **A** is an *ntime* by *k* matrix of *k* selected principal component time series
!  * **B** is the *k* by *nv* matrix of the *k* associated eigenvectors (stored rowwise)
!  * **E** is an *ntime* by *nv* matrix of residuals.
!
! If the selected principal components are the first *k* when the principal components
! are sorted by descending order of the eigenvalues, the matrix product **A**\ **B**
! is a least-squares solution to the problem of minimizing of the sum of all the
! squared elements of **E**\ . In other words, the first *k* principal components
! of **X** are the best linear predictors of the observed variables among all possible
! sets of *k* variables.
!
! This type of approximation can also be computed from the results
! of a previous SVD analysis if the argument **-svd** is activated, however in this case
! the computed approximation is not necessarily optimal in the least square sense.
!
! If the NetCDF variable is tridimensional use :ref:`comp_invert_eof_3d` instead of comp_invert_eof_4d.
!
! An output NetCDF dataset containing the matrix product **A**\ **B** repacked as
! a fourdimensional variable is created.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_invert_eof_4d \
!         -f=input_eof_netcdf_file \
!         -v=netcdf_variable \
!         -se=selected_eofs                 (optional) \
!         -x=lon1,lon2                      (optional) \
!         -y=lat1,lat2                      (optional) \
!         -z=level1,level2                  (optional) \
!         -t=time1,time2                    (optional) \
!         -l=selected_time_period           (optional) \
!         -a=type_of_analysis               (optional : scp, cov, cor) \
!         -c=input_climatology_netcdf_file  (optional) \
!         -o=output_netcdf_file             (optional) \
!         -mi=missing_value                 (optional) \
!         -svd                              (optional) \
!         -double                           (optional) \
!         -bigfile                          (optional) \
!         -hdf5                             (optional) \
!         -tlimited                         (optional)
!
! By default
! ----------
!
!        *-se=*
!         the selected principal components are those stored in the *input_eof_netcdf_file*\ ;
!         in other words, the NetCDF variable *netcdf_variable* is approximated with the number of
!         EOFs (or SVDs) stored in the NetCDF file *eof_input_netcdf_file*
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-z=*
!         the whole vertical resolution associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-l=*
!         the whole time period as determined by the **-t=** argument
!        *-a=*
!         the *type_of_analysis* is set to ``scp``. This means that the eigenvectors (or singular vectors)
!         and eigenvalues (or singular values) have been computed from the sums of squares and cross-products
!         matrix between the observed variables if an EOF (or SVD) model is used
!        *-c=*
!         this argument is not used if the *type_of_analysis* is set to ``scp``
!        *-o=*
!         the *output_netcdf_file* is named  ``approx_``\ *netcdf_variable*\ ``.nc``
!        *-mi=*
!         the *missing_value* attribute in the output NetCDF file is set to ``1.e+20``
!        *-svd*
!         the *input_eof_netcdf_file* is assumed to be produced by :ref:`comp_eof_4d`\ .
!         However, if **-svd** is activated, a file produced by :ref:`comp_svd_3d` is assumed,
!         this means that the approximation is done from a set of singular vectors and 
!         singular variables of a previous SVD analysis
!        *-double*
!         the results of the EOF analysis are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable which must be
!      approximated from its EOF analysis. The EOF model is extracted from the NetCDF file
!      *input_eof_netcdf_file* specified by the **-f=** argument. This NetCDF file must have
!      exactly the same format as the files produced by :ref:`comp_eof_4d` or :ref:`comp_svd_3d`\ .
!
!   #) The **-se=** argument allows the user to select the EOFs (or SVDs) which must be included
!      in the approximation model. The EOFs (or SVDs) list may be given in two formats:
!
!      * **-se=**\ ``1,3,...,nn`` allows to include eof1,eof3,... and eofnn in the EOF (or SVD) model
!
!      * **-se=**\ ``1:4`` allows to include from eof1 to eof4 in the EOF (or SVD) model.
!
!      The two forms of the **-se=** argument may be combined and repeated any number of times.
!      Duplicate EOF (or SVD) numbers are not allowed.
!      If the **-se=** argument is not specified, the NetCDF variable is approximated with the number
!      of EOFs (or SVDs) stored in the *input_eof_netcdf_file*\ .
!
!   #) If the **-x=**\ *lon1,lon2*, **-y=**\ *lat1,lat2* and **-z=**\ *level1,level2* arguments
!      are missing the whole geographical domain associated with the *netcdf_variable* is
!      approximated by the selected EOF (or SVD) model (as specified by the **-se=** argument).
!
!      The longitude, latitude or level range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_4d` for transforming geographical coordinates as indices before
!      using comp_invert_eof_4d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, the whole time period
!      associated with the *netcdf_variable* is approximated by the selected EOF (or SVD) model.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` observations
!      if the **-l=** argument is missing.
!
!   #) The **-l=** argument lists the indices of the time steps which must be included in
!      the output file. The indices of the time steps are counted from the start of the (selected)
!      time period (e.g. *time1* in the **-t=**\ *time1,time2* argument or ``1`` if this argument is not used).
!      The list may be specified in two formats:
!
!      * **-l=**\ *n1,n2,...nn* allows to select for *n1*, *n2*, ... and *nn* time steps
!
!      * **-l=**\ *n1:n2* allows to select time steps from *n1* to *n2*\ .
!
!      Be careful with time period limits, when specifying the **-l=** argument list.
!
!      The two forms of the **-l=** argument may be combined and repeated any number of times.
!      Duplicate time steps are not allowed.
!
!   #) The **-a=** argument specifies if the observed variables have been centered or standardized
!      with an input climatology (specified with the **-c=** argument) before the EOF (or SVD) analysis:
!
!      * **-a=**\ ``scp`` means that the EOF (or SVD) analysis was done on the raw data
!
!      * **-a=**\ ``cov`` means that the EOF (or SVD) analysis was done on the anomalies
!
!      * **-a=**\ ``cor`` means that the EOF (or SVD) analysis was done on the standardized anomalies.
!
!      In all cases, the raw data are approximated if the **-a=** argument is used.
!
!   #) The *input_climatology_netcdf_file* is needed only if **-a=**\ ``cov`` or **-a=**\ ``cor``.
!
!   #) If **-a=**\ ``cov`` or **-a=**\ ``cor``, the selected time period must agree with the climatology.
!      This means that the first selected time observation (*time1* if the **-t=** argument
!      is present) must correspond to the first day, month, season of the climatology
!      specified with the **-c=** argument.
!
!   #) The **-svd** argument specifies that the *input_eof_netcdf_file* is produced by :ref:`comp_svd_3d`
!      instead of :ref:`comp_eof_4d`\ . This means that the approximation will be done from the
!      singular vectors and singular variables of a previous SVD analysis stored in :ref:`comp_svd_3d`\ .
!
!   #) The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*) and the
!      climatology (in the *input_climatology_netcdf_file*) must agree.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the netcdf variable *netcdf_variable* in the NetCDF file *output_netcdf_file*.
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``.
!
!   #) The **-double** argument specify that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on EOF or SVD analysis in the climate literature, see
!
!      * "A manual for EOF and SVD analyses of climate data",
!        by Bjornsson, H., and Venegas, S.A., McGill University, CCGCR Report No. 97-1,
!        Montral, Qubec, 52pp, 1997. https://www.jsg.utexas.edu/fu/files/EOFSVD.pdf
!
!      * "Statistical Analysis in Climate Research",
!        by von Storch, H., and Zwiers, F.W., Cambridge University press,
!        Cambridge, UK, Chapter 13, 484 pp., 2002. ISBN: `9780521012300`_
!
! .. _9780521012300: http://www.cambridge.org/gb/academic/subjects/earth-and-environmental-science/climatology-and-climate-change/statistical-analysis-climate-research?format=PB
!
! Outputs
! -------
!
!   comp_invert_eof_4d creates an output NetCDF file that contains a least squares approximation
!   of the input NetCDF variable computed from selected eigenvectors and principal components of
!   an EOF analysis or from singular vectors and variables of a SVD analysis.
!
!   This least squares approximation is repacked as a fourdimensional NetCDF variable with the same
!   dimensions as the EOFs (or SVDs) of the input NetCDF variable specified in the **-v=** argument.
!
!   This output NetCDF dataset contains the following NetCDF variable 
!   (in the description below, nlev, nlat and nlon are the length of the dimensions of the input NetCDF
!   variable; ntime is the number of time steps selected with the **-l=** and **-t=** arguments):
!
!   #) *netcdf_variable*\ ``(ntime,nlev,nlat,nlon)`` : a least squares approximation of the input
!      NetCDF variable computed with the help of the selected EOF (or SVD) model.
!
! Examples
! --------
!
!   #)   For computing a 10-EOF approximation of a NetCDF variable named
!        ``votemper`` from the NetCDF file ``eof_votemper.t.nc``
!        produced by :ref:`comp_eof_4d` and store the results in a NetCDF file named
!        ``ST7_1m_0101_20012_grid_T_votemper_10pc.nc``, use the following
!        command :
!
!        ::
!
!          $ comp_invert_eof_4d \
!            -f=eof_votemper.t.nc \
!            -v=votemper \
!            -a=scp \
!            -n=10 \
!            -o=ST7_1m_0101_20012_grid_T_votemper_10pc.nc
!
!-
! ================================================================================================
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _MATMUL
    use Statpack,         only : matmul=>matmul2
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,    only : nf90_max_var_dims, nf90_max_name, nf90_clobber, nf90_real4, nf90_real8, nf90_int, nf90_global, &
                          nf90_sizehint_default, nf90_nofill, nf90_unlimited, nf90_nowrite, nf90_noerr, nf90_open,       &
                          nf90_strerror, nf90_inq_varid, nf90_inquire_variable, nf90_inquire_dimension, nf90_get_att,    &
                          nf90_get_var, nf90_create, nf90_set_fill, nf90_def_dim, nf90_def_var, nf90_inq_attname,        &
                          nf90_copy_att, nf90_put_att, nf90_inquire_attribute, nf90_enddef, nf90_put_var, nf90_close
!
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, zero,              & 
                         kfix, kexp, tolower, case_change, string_to_val,           &
                         allocate_error, merror, arth
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,                &
                                 get_listarg, get_int_list, get_variable_dims,       &
                                 get_variable_dimcoords
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
!
    character(len=*), parameter :: name_proc='comp_invert_eof_4d : ',   &
!
                                   cmethod(3)=(/"scp","cov","cor"/)
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_clim, ncid_approx,                    &
                     eofid, pcid, expvarid, meanid, stdid, approxid,                &
                     i, j, j1, j2, k, nlon, nlat, nlevel, nt, nvar, neof,           &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3,       &
                     level1, level2, level3, offset, method, period,                &
                     ntime_list, neof_list, output_type, output_mode
    integer(i4b), dimension(4)                 :: coor_id2
    integer(i4b), dimension(5)                 :: startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i4b), dimension(:),    pointer     :: time_list, eof_list
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),     allocatable  :: temp2
    real(stnd), dimension(:,:),   allocatable  :: mean, std, pc, eof
    real(stnd), dimension(:,:,:), allocatable  :: temp
    real(stnd)                                 :: missing, tmp, totvar
    real(extd), dimension(:),     allocatable  :: record, record2
!
    logical(lgl)                                :: use_tlimited, missing_att, create_hdf5, use_svd
    logical(lgl), dimension(nf90_max_var_dims)  :: use_coor_variable
    logical(lgl), dimension(:,:,:), allocatable :: mask
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, clim_name, approx_name, att_name
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Approximate a NetCDF variable (or parts of it) from its EOF (or SVD) decomposition'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_invert_eof_4d  -f=input_netcdf_file'
        write (prtunit,*) '                     -v=netcdf_variable'
        write (prtunit,*) '                     -se=selected_eofs                (optional)'
        write (prtunit,*) '                     -x=lon1,lon2                     (optional)'
        write (prtunit,*) '                     -y=lat1,lat2                     (optional)'
        write (prtunit,*) '                     -z=level1,level2                 (optional)'
        write (prtunit,*) '                     -t=time1,time2                   (optional)'
        write (prtunit,*) '                     -l=selected_time_period          (optional)'
        write (prtunit,*) '                     -a=type_of_transformation        (optional : scp, cov, cor)'
        write (prtunit,*) '                     -c=input_climatology_netcdf_file (optional)'
        write (prtunit,*) '                     -o=output_netcdf_file            (optional)'
        write (prtunit,*) '                     -mi=missing_value                (optional)'
        write (prtunit,*) '                     -svd                             (optional)'
        write (prtunit,*) '                     -double                          (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                     -bigfile                         (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                     -hdf5                            (optional)'
#endif
        write (prtunit,*) '                     -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-se= by default, the number of EOFs in the input_netcdf_file'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-z= the whole vertical resolution associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-l= the whole time period associated with the netcdf_variable'
        write (prtunit,*) '-a=scp'
        write (prtunit,*) '-o=approx_netcdf_variable.nc'
        write (prtunit,*) '-mi= the missing_value is equal to 1.e+20'
        write (prtunit,*) '-svd      : by default, the approximation is computed from an EOF analysis'
        write (prtunit,*) '-double   : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>17 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name     = " "
    var_name     = " "
    clim_name    = " "
    approx_name  = " "
!
    history = "comp_invert_eof_4d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
    output_type = nf90_real4
    method      = 1
    ntime_list  = 0
    neof_list   = 0
!
    missing     = def_missing
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    level1    = 1
    level2    = huge( level2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_svd      = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET FROM WHICH TO EXTRACT THE EOF MODEL.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE WHICH MUST BE APPROXIMATED.
!
                var_name = name(4:)
!
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
            case ('-z=')
!
!               GET THE LEVEL DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "level1 and level2 are integer values", &
                                  "Wrong level definition",               &
                                  level1, level2 )
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-c=')
!
!               GET THE NAME OF THE NETCDF CLIMATOLOGY DATASET.
!
                clim_name = name(4:)
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF TRANSFORMATION.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case ( cmethod(3) )
                        method = 3
                    case default
                        call merror( name_proc//"Unknown type of analysis" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RECONSTRUCTED VARIABLE.
!
                approx_name = name(4:)
!
            case ('-l=')
!
                if ( ntime_list/=0 ) deallocate( time_list )
!
                call get_int_list(  name_proc, name, 4_i4b, ntime_list, time_list )
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-se=')
!
                        if ( neof_list/=0 ) deallocate( eof_list )
!
                        call get_int_list(  name_proc, name, 5_i4b, neof_list , eof_list )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case ('-svd')
!
!                               COMPUTE THE PROJECTION FROM SINGULAR VECTORS.
!
                                use_svd = true
!
                            case ('-double')
!
!                               STORE DATA AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input EOF NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( clim_name==" " .and. method>=2 )                                &
    call merror( name_proc//"Climatology NETCDF file is missing" )
!
    if ( approx_name==" " )                                               &
    approx_name="approx_"//trim(var_name)//".nc"
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( clim_name/=" " ) then
        k = nf90_sizehint_default
        status = nf90_open( trim(clim_name), nf90_nowrite, ncid_clim, chunksize=k )
        if ( status/=nf90_noerr )                                        &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   GET THE EOF VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    if ( use_svd ) then
        name = trim(var_name)//"_svd"
    else
        name = trim(var_name)//"_eof"
    end if
!
    call get_variable_dimcoords(  name_proc, trim(name), ncid,                        &
                                  eofid,  dimids_out=dimids, dim_names_out=dim_names, &
                                  dimlens_out=dimlens, coor_id_out=coor_id,           &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type, &
                                  use_coor_variable_out=use_coor_variable             )
    if ( count(dimlens/=0)/=4 )           &
    call merror( name_proc//"Input NETCDF EOF (or SVD) variable is not fourdimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nlevel = dimlens(3)
    neof   = dimlens(4)
!
!   GET THE EXPLAINED VARIANCE VARIABLE ID FROM ITS NAME.
!
    if ( use_svd ) then
        name = trim(var_name)//"_SV_EXPVARs"
    else
        name = trim(var_name)//"_var"
    end if
!
    status = nf90_inq_varid( ncid, trim(name), expvarid )
    if ( status/=nf90_noerr )                                              &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE EXPLAINED VARIANCE VARIABLE.
!
    status = nf90_inquire_variable( ncid, expvarid, ndims=i, dimids=dimlens )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( i/=1 .or. dimlens(1)/=dimids(4) )      &
    call merror( name_proc//"Wrong EOF (or SVD) model in the input NETCDF file" )
!
!   GET THE PC VARIABLE ID FROM ITS NAME.
!
    if ( use_svd ) then
        name = trim(var_name)//"_sv"
    else
        name = trim(var_name)//"_pc"
    end if
!
    status = nf90_inq_varid( ncid, trim(name), pcid )
    if ( status/=nf90_noerr )                                                &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE PC VARIABLE.
!
    status = nf90_inquire_variable( ncid, pcid, ndims=i, dimids=dimlens)
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( i/=2 ) call merror( name_proc//"Input NETCDF PC (or SV) variable is not bidimensional" )
    if ( dimlens(1)/=dimids(4) )           &
    call merror( name_proc//"Wrong EOF (or SVD) model in the input NETCDF file" )
!
    status = nf90_inquire_dimension( ncid, dimlens(2), name=dim_names(4), len=nt )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE TIME VARIABLE ID FROM ITS NAME.
!
    status = nf90_inq_varid( ncid, trim(dim_names(4)), coor_id(4) )
    use_coor_variable(4) = status==nf90_noerr
!
    if ( use_coor_variable(4) ) then
!
!       FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES
!       FOR THE TIME VARIABLE.
!
        status = nf90_inquire_variable( ncid, coor_id(4), xtype=coor_type(4), ndims=i,    &
                                        dimids=dimids, natts=coor_natts(4) )
        if ( status/=nf90_noerr )                                                         &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( i/=1 .or. dimlens(2)/=dimids(1) ) use_coor_variable(4) = false
!
    end if
!
    if ( .not.use_coor_variable(4) ) then
        coor_type(4) = nf90_int
    end if
!
!   CHECK THE EOF LIST.
!
    if ( neof_list==0 ) then
!
        neof_list = neof
!
!       ALLOCATE VARIABLE FOR THE EOF  LIST IF NOT DONE BEFORE.
!
        allocate( eof_list(neof_list), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        eof_list(1:neof_list) = arth( 1_i4b, 1_i4b, neof_list )
!
    else
!
        if ( any( eof_list(1:neof_list)>neof ) )          &
        call merror( name_proc//"Wrong -se= argument for the selection of EOFs (or SVDs)" )
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( lon1<0 ) then
        lon2 = max( lon2, 1 )
        lon3 = nlon + lon1 - lon2 + 1
    else
        lon1 = max( lon1, 1 )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
    end if
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    if ( lon1<0 ) then
       lon3 = nlon
       j1   = 1
       j2   = nlon
    else
       j1 = lon1
       j2 = lon2
    end if
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE LEVEL DEFINITION.
!
    level1 = max( level1, 1 )
    level2 = min( level2, nlevel )
    level3 = level2 - level1 + 1
    if ( level3<=0  ) call merror( name_proc//"Wrong vertical definition" )
!
!   CHECK THE TIME DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   CHECK THE TIME COORDINATES LIST.
!
    if ( ntime_list==0 ) then
!
        ntime_list = time3
!
!       ALLOCATE VARIABLE FOR THE TIME COORDINATES LIST IF NOT DONE BEFORE.
!
        allocate( time_list(ntime_list), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        time_list(1:ntime_list) = arth( 1_i4b, 1_i4b, ntime_list )
!
    else
!
        if ( any( time_list(1:ntime_list)>time3 ) )          &
        call merror( name_proc//"-t= and -l= options mismatch" )
!
    end if
!
!   FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE INDICE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, eofid, "missing_value", tmp )
    missing_att = status==nf90_noerr
!
    if ( .not.missing_att ) then
        status = nf90_get_att( ncid, eofid, "_FillValue", tmp )
        missing_att = status==nf90_noerr
    end if
!
    if ( .not.missing_att ) then
        call merror( name_proc//"No missing_value or _FillValue attributes for Input NETCDF EOF variable" )
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat,nlevel), mask(nlon,nlat,nlevel), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   NOW, CONSTRUCT THE MASK ASSOCIATED WITH THE EOF NETCDF VARIABLE.
!
!   GET THE FIRST EOF MODE.
!
    startlec(1:5) = (/ j1, lat1, level1, 1_i4b, time1 /)
    countlec(1:5) = (/ lon3, lat3, level3, 1_i4b, time3 /)
!
    status = nf90_get_var( ncid, eofid, temp(:lon3,:lat3,:level3) ,                & 
                           start=startlec(:4), count=countlec(:4) )  
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED IN THE EOF ANALYSIS FROM THE FIRST EOF MODE.
!
    mask(j1:j2,lat1:lat2,level1:level2) = temp(:lon3,:lat3,:level3)/=tmp
!
!   SELECT THE DOMAIN.
!
    if ( lon1>0 ) then
        mask(1:lon1-1,:nlat,:nlevel)    = false
        mask(lon2+1:nlon,:nlat,:nlevel) = false
    else
        mask(lon2+1:nlon+lon1,:nlat,:nlevel) = false
    end if
!
    mask(:nlon,1:lat1-1,:nlevel)    = false
    mask(:nlon,lat2+1:nlat,:nlevel) = false
!
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   ALLOCATE WORK ARRAYS.
!
    k = max( nvar, time3 )
    allocate( eof(nvar,neof_list), pc(neof_list,ntime_list), temp2(k), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    totvar = zero
!
    do i = 1, neof_list
!
        k           = eof_list(i)
        startlec(4) = k
!
!       GET AND PACK THE EOFS.
!
        status = nf90_get_var( ncid, eofid, temp(:lon3,:lat3,:level3),            &
                               start=startlec(:4), count=countlec(:4)     )  
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        eof(:nvar, i) = pack( temp(:lon3,:lat3,:lon3), mask(j1:j2,lat1:lat2,level1:level2) )
!
!       GET AND PACK THE PCS.
!
        status = nf90_get_var( ncid, pcid, temp2(:time3),                   &
                               start=startlec(4:5), count=countlec(4:5)    )  
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        pc(i,:ntime_list) = temp2(time_list(1:ntime_list))
!
!       GET AND SUM EXPLAINED VARIANCES.
!
        status = nf90_get_var( ncid, expvarid, temp2(1:1),                 &
                               start=startlec(4:4), count=countlec(4:4)    )  
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        totvar = totvar + temp2(1)
!
    end do
!
!   UNSTANDARDIZED THE SINGULAR VARIABLES IF AN SVD APPROXIMATION IS REQUIRED.
!
    if ( use_svd ) then
!
        name = trim(var_name)//"_SV_STDs"
!
        status = nf90_inq_varid( ncid, trim(name), stdid )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_get_var( ncid, stdid, temp2(:neof) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       SCALE THE SINGULAR VARIABLES.
!
        do i = 1, neof_list
!
            j   = eof_list(i)
            tmp = temp2(j)
!
            pc(i,:) = pc(i,:)*tmp
!
        end do
!
    end if
!
!   READ CLIMATOLOGY AND STANDARD-DEVIATIONS IF REQUIRED.
!
    if ( method>1 ) then
!
!       GET THE MEAN FIELD ID FROM ITS NAME.
!       FIND THE DIMENSION LENGTHS FOR THE NETCDF MEAN VARIABLE.
!
        call get_variable_dims( name_proc, trim(var_name)//"_mean", ncid_clim,    &
                                meanid, dimids, dimlens )
!
        if ( count(dimlens/=0)/=4 )     &
        call merror( name_proc//"Input climatology is not fourdimensional" )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
        call merror( name_proc//"Shape disagreement with the climatology (mean field)" )
!
        period = dimlens(4)
!
!       ALLOCATE WORK ARRAY FOR THE MEANS.
!
        allocate( mean(nvar,period), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE MEANS.
!
        do j = 1, period
!
            startlec(4) = j
!
            status = nf90_get_var( ncid_clim, meanid, temp(:lon3,:lat3,:level3),    &   
                                   start=startlec(:4), count=countlec(:4)           )     
            if ( status/=nf90_noerr )                                    &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           PACK THE MEAN FIELD.
!
            mean(:nvar,j) = pack( temp(:lon3,:lat3,:level3), mask(j1:j2,lat1:lat2,level1:level2) )
!
        end do
!
!       FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE MEAN VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid_clim, meanid, "missing_value", tmp )
        missing_att = status==nf90_noerr
!
        if ( .not.missing_att ) then
            status = nf90_get_att( ncid_clim, meanid, "_FillValue", tmp )
            missing_att = status==nf90_noerr
        end if
!
        if ( missing_att ) then
            if ( any( mean(:nvar,:period)==tmp ) )    &
            call merror( name_proc//"Missing values in the input climatology" )
        end if
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE MEAN VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid_clim, meanid, "scale_factor", tmp )
        if ( status==nf90_noerr ) mean(:nvar,:period) = mean(:nvar,:period)*tmp
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE MEAN VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid_clim, meanid, "add_offset", tmp )
        if ( status==nf90_noerr ) mean(:nvar,:period) = mean(:nvar,:period) + tmp
!
        if ( method>2 ) then
!
!           GET THE STD ID FROM ITS NAME.
!           FIND THE DIMENSION LENGTHS FOR THE NETCDF STD VARIABLE.
!
            call get_variable_dims( name_proc, trim(var_name)//"_std", ncid_clim,    &
                                    stdid, dimids, dimlens )
!
            if ( count(dimlens/=0)/=4 )          &
            call merror( name_proc//"Input climatology is not fourdimensional" )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel .or. dimlens(4)/=period )    &
            call merror( name_proc//"Shape disagreement with the climatology (std field)" )
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( std(nvar,period), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           GET THE STDS.
!
            do j = 1, period
!
                startlec(4) = j
!
                status = nf90_get_var( ncid_clim, stdid, temp(:lon3,:lat3,:level3),     & 
                                           start=startlec(:4), count=countlec(:4)       )     
                if ( status/=nf90_noerr )                                          &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               PACK THE STD FIELD.
!
                std(:nvar,j) = pack( temp(:lon3,:lat3,:level3), mask(j1:j2,lat1:lat2,:level1:level2) )
!
            end do
!
!           FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE STD VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid_clim, stdid, "missing_value", tmp )
            missing_att = status==nf90_noerr
!
            if ( .not.missing_att ) then
                status = nf90_get_att( ncid_clim, stdid, "_FillValue", tmp )
                missing_att = status==nf90_noerr
             end if
!
            if ( missing_att ) then
                if ( any( std(:nvar,:period)==tmp ) )    &
                call merror( name_proc//"Missing values in the input standard-deviations field" )
            end if
!
!           FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE STD VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid_clim, stdid, "scale_factor", tmp )
            if ( status==nf90_noerr ) std(:nvar,:period) = std(:nvar,:period)*tmp
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE STD VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid_clim, stdid, "add_offset", tmp )
            if ( status==nf90_noerr ) std(:nvar,:period) = std(:nvar,:period) + tmp
!
        end if
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE APPROXIMATION OF THE NETCDF VARIABLE.
!
    if ( create_hdf5 ) then
        status = nf90_create( trim(approx_name), cmode=output_mode, ncid=ncid_approx )
    else
        k = nf90_sizehint_default
        status = nf90_create( trim(approx_name), cmode=output_mode, ncid=ncid_approx, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_approx, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE CLIMATOLOGY NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    dimlens(3)   = nlevel
    if ( use_tlimited ) then
        dimlens(4)   = ntime_list
    else
        dimlens(4)   = nf90_unlimited
    end if
!
    do i = 1, 4
!
        status = nf90_def_dim( ncid_approx, name=trim(dim_names(i)),                       &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 4
!
        status = nf90_def_var( ncid_approx, trim(dim_names(i)), coor_type(i),              &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                          &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_approx, trim(var_name)//"_approx", output_type,             &
                           dimids=dimids(1:4), varid=approxid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 4
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_approx, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    if ( .not.use_coor_variable(4) ) then
        status = nf90_put_att( ncid_approx, coor_id2(4), "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_approx, coor_id2(4), "short_name", "Time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( use_svd ) then
        name = "SVD"
    else
        name = "EOF"
    end if
!
    status = nf90_put_att( ncid_approx, approxid, "long_name", trim(name)//" approximation of "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_approx, approxid, "short_name", "Reconstructed "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE units ATTRIBUTE, IF POSSIBLE.
!
    if ( clim_name/=" " ) then
!
!       GET THE MEAN VARIABLE ID FROM ITS NAME.
!
        name = trim(var_name)//"_mean"
        status = nf90_inq_varid( ncid_clim, trim(name), meanid )
!
        if ( status==nf90_noerr ) then
!
            status = nf90_inquire_attribute( ncid_clim, meanid, "units"  )
            if ( status==nf90_noerr ) then
!
                status = nf90_copy_att(  ncid_clim, meanid, "units", ncid_approx, approxid )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            end if
!
        end if
!
    end if
!
!   DEFINE missing_value AND _FillValue ATTRIBUTES.
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_approx, approxid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_approx, approxid, "_FillValue", real( missing, dp ) )
!
    else
!
        status = nf90_put_att( ncid_approx, approxid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_approx, approxid, "_FillValue", real( missing, sp ) )
!
    end if
!
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_svd ) then
        name = "SVD model"
    else
        name = "EOF model"
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_approx, nf90_global, "title", trim(var_name)//" Reconstructed from "//trim(name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_approx, nf90_global, "explained_variance", totvar )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_svd ) then
        name = "SVDs_used"
    else
        name = "EOFs_used"
    end if
!
    status = nf90_put_att( ncid_approx, nf90_global, trim(name), neof_list )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_approx, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_approx, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_approx )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW  NETCDF DATASET.
!
    do j = 1, 3
!
        if ( j==1 ) then
            k = nlon
        else if ( j==2 ) then
            k = nlat
        else
            k = nlevel
        end if
!
        if ( use_coor_variable(j) ) then
!
            allocate( record(k), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            status = nf90_get_var( ncid, coor_id(j), record, start=(/ 1 /), count=(/ k /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_approx, coor_id2(j), record )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            deallocate( record )
        else
            status = nf90_put_var( ncid_approx, coor_id2(j), arth(1_i4b,1_i4b,k ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
    offset = time1 - 1
!
    if ( use_coor_variable(4) ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( record(time3), record2(ntime_list),stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        status = nf90_get_var( ncid, coor_id(4), record, start=(/ time1 /), count=(/ time3 /) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        record2(:ntime_list) = record( time_list(1:ntime_list) )
!
        status = nf90_put_var( ncid_approx, coor_id2(4), record2 )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        deallocate( record, record2 )
!
        time_list(1:ntime_list) =  time_list(1:ntime_list) + offset
!
    else
!
        time_list(1:ntime_list) =  time_list(1:ntime_list) + offset
!
        status = nf90_put_var( ncid_approx, coor_id2(4), time_list(1:ntime_list) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    startlec(1:3) = 1_i4b 
    countlec(1:4) = (/ nlon, nlat, nlevel, 1_i4b /)
!
    if ( method>1 ) then
        time_list(1:ntime_list) = mod( time_list(1:ntime_list), period )
        where( time_list(1:ntime_list)==0 )   time_list(1:ntime_list) = period
    end if
!
!   WRITE TRANSFORMED AND SELECTED VALUES.
!
    do i = 1, ntime_list
!
!        EOF APPROXIMATION.
!
         temp2(:nvar) = matmul( eof(:nvar,:neof_list), pc(:neof_list, i) )
!
!        SELECT THE METHOD OF TRANSFORMATION.
!
         if ( method>1 ) then
!
             k = time_list(i)
!
             if ( method==3 ) then
                 temp2(:nvar) = temp2(:nvar)*std(:nvar,k)
             end if
!
             temp2(:nvar) = temp2(:nvar) + mean(:nvar,k)
!
         end if
!
!        WRITE THE RECORD IN THE OUTPUT FILE.
!
         temp(:nlon,:nlat,:nlevel) = unpack( temp2(:nvar), mask(:nlon,:nlat,:nlevel), missing )
!
         startlec(4) = i
!
         status = nf90_put_var( ncid_approx, approxid, temp(:nlon,:nlat,:nlevel),      &
                                start=startlec(:4), count=countlec(:4)         )
         if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    select case ( method )
!
        case ( 1 )
!
            deallocate( eof, pc, temp, temp2, mask, eof_list, time_list )
!
        case ( 2 )
!
            deallocate( eof, pc, temp, temp2, mask,  eof_list, time_list,  mean )
!
        case ( 3 )
!
            deallocate( eof, pc, temp, temp2, mask,  eof_list, time_list,  mean, std )
!
    end select
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( clim_name/=" " ) then
        status = nf90_close( ncid_clim )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_approx )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_invert_eof_4d
! =================================
!
end program comp_invert_eof_4d
