program comp_lanczos_filter_3d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_lanczos_filter_3d:
!
! ======================
! comp_lanczos_filter_3d
! ======================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 13/09/2018
!
! Purpose
! =======
!
! Filter a real multichannel time series in a selected frequency band by Lanczos filtering [Bloomfield]_
! [Duchon]_. The multichannel time series is extracted from a tridimensional variable readed from
! a NetCDF dataset and can also be detrended before Lanczos filtering at the user option.
!
! The number of coefficients used to build the Lanczos filter can be selected and the Lanczos
! filter can be applied to the multichannel time series in the time or frequency domain, also
! at the user option. This gives to the user some control on the desired end-effects of the filter (e.g.
! applying the filter in the frequency domain assumes implicitly that the multichannel time series is part
! of a periodic infinite series whose period is exactly equal to the length of the analyzed time series; on the
! other hand, applying the filter in the time domain implies some loss of data or some distortions of the desired
! response function of the filter at both ends of the filtered time series).
!
! Additionally, the filtering can be done separately for different segments of equal length of the selected
! multichannel time series if this time series is not continuous in time.
!
! The frequency response function (e.g. the transfer function) of the selected Lanczos filter can be computed
! by :ref:`comp_freq_func_1d`. See the references cited below for more details on Lanczos filtering [Bloomfield]_
! [Duchon]_.
!
! This procedure returns the filtered real multichannel time series in a NetCDF dataset. If the NetCDF variable
! is uni- or fourdimensional use :ref:`comp_lanczos_filter_1d` or :ref:`comp_lanczos_filter_4d`, respectively,
! instead of comp_lanczos_filter_3d. It the multichannel time series has a seasonal (or diurnal) cycle,
! use :ref:`comp_clim_3d` and :ref:`comp_norm_3d` or :ref:`comp_stl_3d` in order to estimate and remove
! the harmonic components of the time series before using comp_lanczos_filter_3d.
!
! If you need more control on the filtering parameters, including other windows, use :ref:`comp_symlin_filter_3d`
! instead of comp_lanczos_filter_3d.
!
! This procedure is parallelized if OpenMP is used.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_lanczos_filter_3d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -m=input_mesh_mask_netcdf_file           (optional) \
!         -g=grid_type                             (optional : n, t, u, v, w, f) \
!         -x=lon1,lon2                             (optional) \
!         -y=lat1,lat2                             (optional) \
!         -t=time1,time2                           (optional) \
!         -o=output_netcdf_file                    (optional) \
!         -p=periodicity                           (optional) \
!         -pl=minimum_period                       (optional) \
!         -ph=maximum_period                       (optional) \
!         -tr=trend_removal                        (optional : 0, 1, 2, 3, -1, -2, -3) \
!         -nfc=number_of_filter_coefficients       (optional) \
!         -mi=missing_value                        (optional) \
!         -ngp=number_of_grid_points               (optional) \
!         -notestf                                 (optional) \
!         -usefft                                  (optional) \
!         -double                                  (optional) \
!         -bigfile                                 (optional) \
!         -hdf5                                    (optional) \
!         -tlimited                                (optional)
!
! By default
! ----------
!
!        *-m=*
!         an *input_mesh_mask_netcdf_file* is not used
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input NetCDF variable is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-o=*
!         the *output_netcdf_file* is named ``filt_``\ *netcdf_variable*\ ``.nc``
!        *-p=*
!         the *periodicity* is set to ``time2 - time1 + 1`` , which means
!         that the time series is considered as continuous with only one time segment
!        *-pl=*
!         the *minimum_period* is set to ``0``, which means that no filtering is done
!         for the shorter periods
!        *-ph=*
!         the *maximum_period* is set to ``0``, which means that no filtering is done
!         for the longer periods
!        *-tr=0*
!         the *trend_removal* is set to ``0``, which means that no detrending is done before
!         filtering
!        *-nfc=*
!         the *number_of_filter_coefficients* is determined in order to optimize 
!         the frequency response function of the selected filter
!        *-mi=*
!         the *missing_value* for the output variable is equal to ``1.e+20``
!        *-ngp=*
!         the *number_of_grid_points* is set to the number of grid points in 
!         the selected domain
!        *-notestf*
!         normally, the value of the **-nfc=** argument must be chosen such that 
!         the full transition bands about the cutoff frequencies ``1/PH`` and ``1/PL``
!         (where ``PH`` and ``PL`` are the values of the **-ph=** and **-pl=**
!         arguments, respectively) of the selected filter are inside the (``0`` ``0.5``) 
!         frequency interval. By using the **-notestf** argument you can get ride of this limitation
!        *-usefft*
!         the Lanczos filter is applied in the time domain. When you specify the **-usefft** argument
!         the filter will be applied in the frequency domain, using an FFT algorithm, multiplication and
!         an inverse FFT instead of a convolution in the time domain
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable for which a
!      Lanczos filtering operation must be computed and the **-f=**\ *input_netcdf_file* 
!      argument specifies that this NetCDF variable must be extracted from the NetCDF file
!      *input_netcdf_file*.
!
!   #) The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*) and the NetCDF
!      mesh_mask variable (in the *input_mesh_mask_netcdf_file*) must agree if the **-m=** argument
!      is used.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the
!      NetCDF variable is from an experiment with the NEMO or ORCA model (R2, R4 or R05 resolutions).
!
!      If **-g=** is set to ``n``, it is assumed that the 2-D grid-mesh is regular or Gaussian.
!
!      This argument is also used to determined the name of the NetCDF mesh_mask variable if an 
!      *input_mesh_mask_netcdf_file* is used as specified with the **-m=** argument
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing, the whole
!      geographical domain associated with the *netcdf_variable* is used to select the multi-channel
!      time series.
!
!      The longitude or latitude range must be a vector of two integers specifying
!      the first and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices before
!      using comp_lanczos_filter_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, the whole time period
!      associated with the *netcdf_variable* is used to decompose the time series.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) If the **-p=** argument is specified, the filtering is applied separately for each time segment
!      of length *periodicity* (as determined by the value of the **-p=** argument).
!      The whole selected time period (e.g. *time2* - *time1* + ``1`` ) must also be a multiple of 
!      the *periodicity*.
!
!   #) The **-pl=** argument specifies the minimum period of oscillation of the filtered time series. 
!      The *minimum_period* is expressed in number of time observations. 
!
!      Do not use the **-pl=** argument or use **-pl=**\ ``0`` for high-pass filtering 
!      frequencies corresponding to periods shorter than **-ph=**\ ``PH`` .
!
!      The **-pl=** argument is a positive integer equal to ``0`` or greater than ``2``.
!
!   #) The **-ph=** argument specifies the maximum period of oscillation of the filtered time series. 
!      The *maximum_period* is expressed in number of time observations. 
!      Do not use the **-ph=** argument or use **-ph=**\ ``0`` for low-pass filtering 
!      frequencies corresponding to periods longer than **-pl=**\ ``PL`` .
!      For example, **-pl=**\ ``6``  (or ``18``) and **-ph=**\ ``32``  (or ``96``) select periods
!      between ``1.5`` and ``8``  years for quarterly (monthly) time series.
!
!      The **-ph=** argument is a positive integer equal to ``0`` or greater than ``2`` and 
!      less than the length of the multichannel time series or the *periodicity* if the **-p=**
!      argument is used.
!
!      The **-ph=** argument must also be greater or equal to the **-pl=** argument if 
!      both are specified.
!
!   #) Setting **-pl=** and **-ph=** to the same value ``P`` is allowed. 
!      In this case, an -ideal- band-pass filter with peak response near one
!      at the single period ``P`` is computed and applied to the multichannel time series.
!
!   #) Setting both **-pl=**\ ``0`` and **-ph=**\ ``0`` is also allowed. In that case,
!      no frequencies filtering is done, but the data may be detrended if the **-tr=** argument 
!      is used with a value of ``1``, ``2`` or ``3``.
!
!   #) The **-tr=** argument specifies pre- and post-filtering processing of the multichannel time 
!      series. If:
!
!      * **-tr=**\ ``+/-1``, the means of the time series are removed before time filtering
!
!      * **-tr=**\ ``+/-2``, the drifts from the time series are removed before time 
!        filtering. The drift for each time series is estimated using the 
!        formula : ``drift = ( tseries(ntime) - tseries(1) )/( ntime - 1 )``
!
!      * **-tr=**\ ``+/-3``, the least-squares lines from the multichannel time series are removed before
!        time filtering.
!
!      If **-tr=**\ ``-1``, ``-2`` or ``-3``, the means, drifts or least-squares lines are reintroduced
!      post-filtering, respectively.
!
!      For other values of the **-tr=** argument, nothing is done before or after 
!      filtering.
!
!      If the **-p=** argument is present, the pre-filtering and post-filtering processing is applied to 
!      each time segment, separately.
!
!      The **-tr=** argument must be an integer and the default value for the **-tr=** argument is ``0``.
!
!   #) The **-nfc=** argument specifies the desired number of symmetric linear filter coefficients 
!      for the filtering of the multichannel time series. If **-nfc=** is not specified, an optimal 
!      value is chosen in order to obtain a good frequency response function for the selected filter.
!
!      However, if **-nfc=** is set to ``K``, the first and last ``(K-1)/2`` time observations in the 
!      output NetCDF file will be affected by end effects. Thus, the user must choose the number of 
!      filter terms, ``K``, as a compromise between:
!
!           1)  A sharp cutoff, that is, ``1/K`` small; and
!
!           2)  Minimizing the number of data points lost or affected by the filtering operation
!               (since ``(K-1)/2`` data points will be lost or affected from each end of the series).
!
!      Finally, the **-nfc=** argument must be greater or equal to ``3`` and odd.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output variable in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* is set
!      to ``1.e+20``.
!
!   #) The **-ngp=** argument can be used if you have memory problems when running
!      comp_lanczos_filter_3d on very large datasets. By default, the *number_of_grid_points*
!      is set to the number of cells in the selected domain. In case of memory problems, you can
!      use the **-ngp=** argument with a lower value. This will reduce the memory used by the operator.
!
!   #) The **-notestf** argument allows to bypass some of the restrictions on the 
!      number of filter coefficients as specified with the **-nfc=** argument.
!
!      By default, the value of the **-nfc=** argument must be chosen such that the 
!      full transition bands about the cutoff frequencies (``1/PH``  and ``1/PL``) of
!      the selected filter are inside the (``0`` ``0.5``) frequency interval.
!
!      When the **-notestf** argument is specified, only the cutoff frequencies (e.g. ``1/PH``
!      and ``1/PL``) of the selected filter must lie in the (``0`` ``0.5``) frequency interval
!      and not the full transition bands around them.
!
!      This allows to diminish the number of filter coefficients and, thus, to minimize the number
!      of data points lost by the filtering operation (if **-nfc=** is set to ``K``, ``(K-1)/2``  data 
!      points will be "lost" or affected by end effects from each end of the series).
!
!   #) The **-usefft** argument specifies that the Lanczos filter must be applied in the 
!      frequency domain by using a Fast Fourier Transform and the convolution theorem.
!      Moreover, if the **-usefft** argument is specified, the values at the ends of 
!      the output filtered series are computed implicitly by assuming that the input series
!      is part of a periodic sequence.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It the multichannel time series has a seasonal or diurnal cycle, use :ref:`comp_stl_3d` or :ref:`comp_clim_3d`
!      to remove the pure harmonic components from the time series before filtering.
!
!   #) It is assumed that the data has no missing values excepted those associated with a constant land-sea
!      mask.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
!   #) For more details on Lanczos filtering and examples of use in the climate literature, see
!
!      * "Fourier analysis of time series- An introduction",
!        by Bloomfield, P., John Wiley and Sons, New York, Chapter 6, 1976. http://as.wiley.com/WileyCDA/WileyTitle/productCd-0471889482.html
!
!      * "Lanczos filtering in one and two dimensions",
!        by Duchon, C., Journal of applied meteorology, vol. 18, 1016-1022, 1979. doi: `10.1175/1520-0450(1979)018<1016:LFIOAT>2.0.CO;2`_
!
! .. _`10.1175/1520-0450(1979)018<1016:LFIOAT>2.0.CO;2` : https://doi.org/10.1175/1520-0450(1979)018<1016:LFIOAT>2.0.CO;2
!
! Outputs
! -------
!
!   comp_lanczos_filter_3d creates an output NetCDF file that contains the filtered time series estimated from the
!   multichannel time series associated with the input NetCDF variable. The output NetCDF data set contains the
!   following NetCDF variable (in the description below, nlat and nlon are the length of the spatial dimensions
!   of the input NetCDF variable and ntime is the selected number of time observations) :
!
!   #) *netcdf_variable*\ _filt\ ``(ntime,nlat,nlon)`` : the filtered time series for each of the time series 
!      of the 2-D grid-mesh associated with the input NetCDF variable.
!
!   The filtered multichannel time series is packed in a tridimensional variable whose first and second dimensions
!   are exactly the same as those associated with the input NetCDF variable *netcdf_variable*
!   even if you restrict the geographical domain with the **-x=** and **-y=** arguments. However,
!   outside the selected domain, the output NetCDF variable is filled with missing values.
!
! Examples
! --------
!
!   #)   For filtering a real multichannel monthly time series between ``18`` and ``24`` months (e.g. biennial time scale)
!        from a tridimensional NetCDF variable called ``mslp`` extracted from the
!        file ``mslp.monthly.mean_ncep2.nc``, which includes monthly time series, and store
!        the results in the NetCDF file ``tbo_mslp_ncep2.nc``, use the following
!        command :
!
!        ::
!
!          $ comp_lanczos_filter_3d \
!            -f=mslp.monthly.mean_ncep2.nc \
!            -v=mslp \
!            -m=mesh_mask_mslp_ncep2.nc \
!            -pl=18 \
!            -ph=30 \
!            -o=tbo_mslp_ncep2.nc
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env, only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,    only : nf90_share
#endif
!
#ifdef _OPENMP
    use Statpack,  only : omp_limit
    use omp_lib,   only : omp_get_num_procs, omp_get_max_threads
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,      &
                         nf90_sizehint_default, nf90_open, nf90_close,                                &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,         &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,       &
                         nf90_def_var, nf90_inq_varid, nf90_inq_attname, nf90_copy_att, nf90_put_att, &
                         nf90_enddef, nf90_global, nf90_put_var
!
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, one, half, kfix, kexp, kint, tolower,             & 
                         merror, allocate_error, arth, case_change, string_to_val, detrend, lp_coef, hp_coef,      &
                         bd_coef, pk_coef, symlin_filter2, init_fft, end_fft
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,                                  &
                                 get_listarg, get_variable_dims, get_variable_dimcoords
!
!
! STRONG TYPING IMPOSED
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_lanczos_filter_3d : '
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_mask, ncid_filt,                                 &
                     varid, maskid, filtid,                                                    &
                     i, i0, i1, i2, imax, j, j1, j2, k, l, nlon, nlat, nt, nvar, var_natts,    &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3, ngp,             &
                     pl, ph, nfc, trend, method, offset_rec, output_type, output_mode,         &
                     period, nperiod
    integer(i4b), dimension(3)                 :: coor_id2, startlec, countlec, startwri
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),     allocatable  :: coef
    real(stnd), dimension(:,:),   allocatable  :: temp, mat
    real(stnd)                                 :: missing, freq, scalfac, offset, tmp
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: use_tlimited, use_x, use_y, use_mask, detrend_only, use_scalfac, &
                                                  use_offset, use_period, notest_freq, use_fft, create_hdf5
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
    logical(lgl), dimension(:,:), allocatable  :: mask
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, mask_name, filt_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
    character                                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
!
    integer(i4b)  :: ncid2, varid2, rec
!
#endif
!
#ifdef _OPENMP
!
    logical  :: test_par
!
#endif
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Filter a real tridimensional variable extracted from a NetCDF dataset'
        write (prtunit,*) ' in a selected frequency band by Lanczos filtering'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_lanczos_filter_3d '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -m=input_mesh_mask_netcdf_file          (optional)'
        write (prtunit,*) '        -g=grid_type                            (optional : n, t, u, v, w, f)'
        write (prtunit,*) '        -x=lon1,lon2                            (optional)'
        write (prtunit,*) '        -y=lat1,lat2                            (optional)'
        write (prtunit,*) '        -t=time1,time2                          (optional)'
        write (prtunit,*) '        -o=output_netcdf_file                   (optional)'
        write (prtunit,*) '        -p=periodicity                          (optional)'
        write (prtunit,*) '        -pl=minimum_period                      (optional)'
        write (prtunit,*) '        -ph=maximum_period                      (optional)'
        write (prtunit,*) '        -tr=trend_removal                       (optional : 0,1,2,3,-1,-2,-3)'
        write (prtunit,*) '        -nfc=number_of_filter_coefficients      (optional)'
        write (prtunit,*) '        -mi=missing_value                       (optional)'
        write (prtunit,*) '        -notestf                                (optional)'
        write (prtunit,*) '        -usefft                                 (optional)'
        write (prtunit,*) '        -ngp=number_of_grid_points              (optional)'
        write (prtunit,*) '        -double                                 (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '        -bigfile                                (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                                   (optional)'
#endif
        write (prtunit,*) '        -tlimited                               (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-m= by default, an input_mesh_mask_netcdf_file is not used'
        write (prtunit,*) '-g= the grid type is set to n'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-o=filt_netcdf_variable.nc'
        write (prtunit,*) '-p=time2-time1+1 : by default, the periodicity argument is not used'
        write (prtunit,*) '-pl=0            : by default no filtering is done'
        write (prtunit,*) '-ph=0            : by default no filtering is done'
        write (prtunit,*) '-tr=0            : by default, no trend removal is done'
        write (prtunit,*) '-nfc=            : by default, the number of filter coefficients is determined to optimize'
        write (prtunit,*) '                   the frequency response function of the selected filter'
        write (prtunit,*) '-mi=1.e+20       : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-ngp=            : by default, the number_of_grid_points is set to the selected domain'
        write (prtunit,*) '-notestf         : by default, the value of the -nfc= argument must be choosen such that the'
        write (prtunit,*) '                   full transition bands about the cutoff frequencies of the selected filter'
        write (prtunit,*) '                   are inside the (0 0.5) interval'
        write (prtunit,*) '-usefft          : by default, the filter is applied in the time domain'
        write (prtunit,*) '-double          : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile         : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5            : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited        : by default, the time dimension is defined as unlimited'
#ifdef _OPENMP
        write (prtunit,*)
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>21 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name   = " "
    var_name   = " "
    mask_name  = " "
    filt_name  = " "
!
    history = "comp_lanczos_filter_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
!
    pl            = 0
    ph            = 0
    trend         = 0
    nfc           = 0
!
    missing = def_missing
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    period    = huge( period )
    nperiod   = 1
!
    ngp       = huge( ngp )
!
    use_mask     = false
    use_x        = false
    use_y        = false
    detrend_only = false
    notest_freq  = false
    use_fft      = false
    create_hdf5  = false
    use_tlimited = false
    use_period   = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-p=')
!
!               GET THE PERIODICITY.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The periodicity is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
                use_period = true
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
                use_mask = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                filt_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-pl=')
!
!                       GET THE MINIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) pl
                        end if
                        if ( kcode/=kint .or. pl<0 .or. pl==1 .or. pl==2 )                           &
                        call merror( name_proc//"The -pl= argument is an integer equal to 0 or greater than 2" )
!
                    case ('-ph=')
!
!                       GET THE MAXIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) ph
                        end if
                        if ( kcode/=kint .or. ph<0 .or. ph==1 .or. ph==2  )                           &
                        call merror( name_proc//"The -ph= argument is an integer equal to 0 or greater than 2" )
!
                    case ('-tr=')
!
!                       DETERMINE PRE- AND POST-FILTERING COMPUTATIONS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) trend
                        else
                            call merror( name_proc//"The -tr= argument is an integer value" )
                        end if
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument is a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5) )
!
                            case ('-ngp=')
!
!                               GET THE NUMBER OF GRID POINTS IN ORDER TO REDUCE MEMORY USAGE.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) ngp
                                end if
!
                                if ( kcode/=kint .or. ngp<=0 )      &
                                call merror( name_proc//"The -ngp= argument is an integer > 0 " )
!
                            case ('-nfc=')
!
!                               DETERMINE THE NUMBER OF FILTER COEFFICIENTS.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) nfc
                                end if
                               if ( kcode/=kint .or. nfc<3 .or. (nfc/2)*2==nfc  )                  &
                               call merror( name_proc//"The -nfc= argument is an odd integer value greater than 2" )
!
                            case default
!
                                select case ( name(1:) )
!
                                    case ('-notestf')
!
!                                       BYPASS THE TESTS ON THE CUTOFF FREQUENCIES.
!
                                        notest_freq = true
!
                                    case ('-usefft')
!
!                                       APPLY THE FILTER IN THE FREQUENCY DOMAIN.
!
                                        use_fft = true
!
                                    case ('-double')
!
!                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                        output_type = nf90_real8
!
#ifdef _USE_NETCDF36
                                    case ('-bigfile')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
                                        create_hdf5 = true
#endif
!
                                    case ('-tlimited')
!
!                                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                                        use_tlimited = true
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( filt_name==" " ) filt_name="filt_"//trim(var_name)//".nc"
!
!   CHECK PERIODS ORDERING TO SELECT FREQUENCIES.
!
    if ( ph<pl .and. ph/=0 )     &
    call merror( name_proc//"The -ph= argument must be greater or equal to the -pl= argument" )
!
    if ( max(pl,ph)==0 ) then
!
!       DETREND THE TIME SERIES.
!
        method = 0
!
        if ( trend>=1 .and. trend<=3 ) detrend_only = true
!
    else if ( ph==0 ) then
!
!       LOW-PASS FILTERING OF THE TIME SERIES.
!
        method = 1
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS.
!
            tmp = one/real( pl, stnd )
            i   = ceiling( one/(half-tmp) )
            nfc = max( i, pl+1 )
!
        end if
!
    else if ( pl==0 ) then
!
!       HIGH-PASS FILTERING OF THE TIME SERIES.
!
        method = 2
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS.
!
            tmp = one/real( ph, stnd )
            i   = ceiling( one/(half-tmp) )
            nfc = max( i, ph+1 )
!
        end if
!
    else if ( pl/=ph ) then
!
!       BAND-PASS FILTERING OF THE TIME SERIES.
!
        method = 3
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS.
!
            tmp = one/real( pl, stnd )
            i1  = ceiling( one/(half-tmp) )
!
            tmp = tmp - one/real( ph, stnd )
            i2  = ceiling(  2.6/tmp )
!
            nfc = max( i1, i2, ph+1 )
!
        else
!
!           CHECK THAT nfc IS LARGE ENOUGH TO ENSURE UNIT RESPONSE AT BAND CENTER.
!
            tmp = one/real( pl, stnd ) - one/real( ph, stnd )   
!
            if ( real( nfc+1, stnd ) < (2.6/tmp) )         &
            call merror( name_proc//"The -nfc= argument is too small to ensure unit response at band center"  )
!
        end if
!
    else
!
!       BAND-PASS FILTERING OF THE TIME SERIES WITH A PEAK RESPONSE NEAR ONE AT THE PERIOD pl=ph .
!
        method = 4
!
        freq = one/real( pl, stnd )
!
        if ( nfc==0 ) then
!
!           SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS.
!
            tmp = 2.3*max( real( pl, stnd ), one/(half-freq) )
            nfc = ceiling( tmp )
!
        end if
!
    end if
!
!   CHECK IF THE NUMBER OF FILTER COEFFICIENTS IS ODD.
!
    if ( (nfc/2)*2==nfc  ) nfc = nfc + 1
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr )                                                &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_mask ) then
        status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                          &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names,       &
                                  dimlens_out=dimlens, coor_id_out=coor_id,                 &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type,       &
                                  use_coor_variable_out=use_coor_variable, natts=var_natts  )
!
    if ( count(dimlens/=0)/=3 )           &
    call merror( name_proc//"Input NETCDF variable is not tridimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nt     = dimlens(3)
!
    if ( use_mask ) then
!
        chr = merge( 't', grid_type, grid_type=='w' )
!
!       GET THE MASK ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
        call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
        call merror( name_proc//"Shape disagreement with the mask" )
!
!       GET THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN FROM THE MESH-MASK DATASET.
!
        if ( .not.use_x ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
            if ( status==nf90_noerr ) lon1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
            if ( status==nf90_noerr ) lon2 = i
!
        end if
!
        if ( .not.use_y ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
            if ( status==nf90_noerr ) lat1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
            if ( status==nf90_noerr ) lat2 = i
!
        end if
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( lon1<0 ) then
        lon2 = max( lon2, 1 )
        lon3 = nlon + lon1 - lon2 + 1
    else
        lon1 = max( lon1, 1 )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
    end if
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    if ( lon1<0 ) then
       lon3 = nlon
       j1   = 1
       j2   = nlon
    else
       j1 = lon1
       j2 = lon2
    end if
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE TIME DEFINITION AND PERIODICITY DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
!
    if ( time3<4  ) call merror( name_proc//"Wrong time definition or not enough time observations" )
!
    if ( use_period ) then
!
        if ( period<4  ) call merror( name_proc//"Not enough time observations in the periods" )
!
        nperiod = time3/period
        if ( nperiod*period/=time3 )    &
        call merror( name_proc//"The selected time period is not a multiple of the periodicity" )
!
    else
!
        period = time3
!
    end if
!
    if ( period<ph)      &
    call merror( name_proc//"Periods longer than signal length cannot be observed" )
!
    if ( period<nfc)     &
    call merror( name_proc//"The number of filter coefficients must be less or equal to signal length" )
!
!   ALLOCATE WORK ARRAYS.
!
    if ( detrend_only ) then
        allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
    else
        allocate( temp(nlon,nlat), mask(nlon,nlat), coef(nfc), stat=status )
    end if
!
   if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   COMPUTE THE SYMMETRIC LINEAR FILTER COEFFICIENTS IF NEEDED.
!
    select case( method )
!
        case( 1_i4b )
!
!           LOW-PASS FILTERING OF THE TIME SERIES.
!
            coef(:nfc) = lp_coef( pl, nfc, notest_fc=notest_freq )
!
        case( 2_i4b )
!
!           HIGH-PASS FILTERING OF THE TIME SERIES.
!
            coef(:nfc) = hp_coef( ph, nfc, notest_fc=notest_freq )
!
        case( 3_i4b )
!
!           BAND-PASS FILTERING OF THE TIME SERIES.
!
            coef(:nfc) = bd_coef( pl, ph, nfc, notest_fc=notest_freq )
!
        case( 4_i4b )
!
!           BAND-PASS FILTERING OF THE TIME SERIES WITH A PEAK RESPONSE NEAR ONE AT THE PERIOD pl=ph.
!
            coef(:nfc) = pk_coef( freq, nfc, notest_freq=notest_freq )
!
    end select
!
    if ( use_mask ) then
!
!       GET THE MASK.
!
        status = nf90_get_var( ncid_mask, maskid, temp )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DETERMINE THE GRID POINTS USED IN THE CORRELATION ANALYSIS FROM THE MASK.
!
        mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
    else
!
!       CONSTRUCT A DUMMY MASK.
!
        mask(:nlon,:nlat) = true
!
    end if
!
!   SELECT THE DOMAIN.
!
    if ( lon1>0 ) then
        mask(1:lon1-1,:nlat)    = false
        mask(lon2+1:nlon,:nlat) = false
    else
        mask(lon2+1:nlon+lon1,:nlat) = false
    end if
!
    mask(:nlon,1:lat1-1)    = false
    mask(:nlon,lat2+1:nlat) = false
!
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
    ngp = min( ngp, nvar )
!
!   GET THE SCALE FACTOR IF ANY.
!
    status = nf90_get_att( ncid, varid, "scale_factor", scalfac )
    use_scalfac = status==nf90_noerr 
!
!   GET THE OFFSET IF ANY.
!
    status = nf90_get_att( ncid, varid, "add_offset", offset )
    use_offset = status==nf90_noerr 
!
!   ALLOCATE WORK ARRAY FOR THE MULTICHANNEL TIME SERIES.
!
    allocate( mat(nvar,period), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    if ( create_hdf5 ) then
        status = nf90_create( filt_name, cmode=output_mode, ncid=ncid_filt )
    else
        k = nf90_sizehint_default
        status = nf90_create( filt_name, cmode=output_mode, ncid=ncid_filt, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_filt, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
!    dimlens(1)   = nlon
!    dimlens(2)   = nlat
    if ( use_tlimited ) then
        dimlens(3)   = time3
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    do i = 1, 3
!
        status = nf90_def_dim( ncid_filt, name=trim(dim_names(i)),                         &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 3
!
        status = nf90_def_var( ncid_filt, trim(dim_names(i)), coor_type(i),               &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_filt, trim(var_name)//"_filt",  output_type,              &
                           dimids=dimids(1:3), varid=filtid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_filt, coor_id2(j) )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    if ( .not.use_coor_variable(3) ) then
!
        status = nf90_put_att( ncid_filt, coor_id2(3), "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                          &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, coor_id2(3), "short_name", "Time" )
        if ( status/=nf90_noerr )                                                          &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   COPY ATTRIBUTES FOR THE FILTERED VARIABLE.
!
    do i = 1, var_natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( att_name=="add_offset"         .or.  &
             att_name=="scale_factor"       .or.  &
             att_name=="missing_value"      .or.  &
             att_name=="_FillValue"               )   cycle
!
!       COPY USEFUL ATTRIBUTES.
!
        status = nf90_copy_att(  ncid, varid, att_name, ncid_filt, filtid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, dp ) )
!
    else
!
        status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, sp ) )
!
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_filt, nf90_global, "title", " Lanczos filtering of "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "minimum_period", pl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "maximum_period", ph )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "trend_removal", trend )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "number_of_filter_coefficients", nfc )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_fft ) then
        name = "time domain"
    else
        name = "frequency domain"
    end if
!
    status = nf90_put_att( ncid_filt, nf90_global, "algorithm", trim(name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "number_of_time_segments", nperiod )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3    
!
        select case ( j )
!
            case ( 1 )
                i1 = 1
                i2 = nlon
!
            case ( 2 )
                i1 = 1
                i2 = nlat
!
            case ( 3 )
                i1 = time1
                i2 = time3
!
        end select
!
        if ( use_coor_variable(j) ) then
!
            allocate( record(i2), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            status = nf90_get_var( ncid, coor_id(j), record, start=(/ i1 /), count=(/ i2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_filt, coor_id2(j), record )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            deallocate( record )
        else
            status = nf90_put_var( ncid_filt, coor_id2(j), arth(i1,1_i4b,i2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec = time1 - 1
!
    countlec(1:3) = (/ lon3, lat3, 1_i4b /)
    startwri(1:2) = 1
!
    imax = nvar/ngp
    if ( (imax*ngp)/=nvar ) imax = imax + 1
!
    if ( use_fft ) then
!
!       INITIALIZE THE FFT SUBROUTINE IF REQUIRED.
!
        call init_fft( (/ nvar, period /), dim=2_i4b )
!
    end if
!
!   COMPUTE AND WRITE FILTERED VARIABLE.
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
#endif
!
#ifdef _PARALLEL_READ
!
    test_par = i>1 .and. j>1 .and. time3*nvar>omp_limit .and. time3>=j .and. nperiod==1
!
    if ( test_par ) then
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
!$OMP PARALLEL PRIVATE(k,status,rec,startlec,ncid2,varid2,temp)      &
!$OMP         ,SHARED(mat,mask,varid,countlec,j1,j2,lon3,            &
!$OMP                 lat1,lat2,lat3,nvar,offset_rec,time1,time2)
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( temp(lon3,lat3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
        status = nf90_open( trim(fic_name), nf90_share, ncid2  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       GET THE VARIABLE ID FROM ITS NAME.
!
        status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!       FIRST GET THE DATA.
!
        startlec(1)  = j1
        startlec(2)  = lat1
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = time1, time2
!
            startlec(3) = k
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3),                 &
                                   start=startlec, count=countlec )     
            if ( status/=nf90_noerr )                                                &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           PACK THE DATA.
!
            rec            = k - offset_rec
            mat(:nvar,rec) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        end do
!
!$OMP END DO
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
!$OMP CRITICAL (closenc)
!
        status = nf90_close( ncid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
!       REALLOCATE WORK ARRAY.
!
        allocate( temp(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
        if ( use_scalfac ) mat(:nvar,:time3) = mat(:nvar,:time3)*scalfac
!
!       ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
        if ( use_offset ) mat(:nvar,:time3) = mat(:nvar,:time3) + offset
!
        if ( detrend_only ) then
!
!           DETREND THE MULTICHANNEL TIME SERIES, ONLY.
!
            call detrend( mat(:nvar,:time3), trend )
!
        else if ( method/=0 ) then
!
!           FREQUENCIES FILTERING OF THE MULTICHANNEL TIME SERIES IS REQUIRED.
!
            test_par = imax>1
!
!$OMP PARALLEL IF(test_par)               &
!$OMP         ,PRIVATE(i,i0,i1,i2)        &
!$OMP         ,SHARED(mat,coef,trend,use_fft,nvar,imax,ngp,time3,nfc)
!
!
!$OMP DO SCHEDULE(STATIC) 
!
            do i = 1, imax
!
                i0 = (i-1)*ngp
                i1 = i0 + 1
                i2 = min( nvar, i0+ngp )
!
                call symlin_filter2( mat(i1:i2,:time3), coef(:nfc), trend=trend, usefft=use_fft, initfft=false )
!
            end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
        end if
!
!       WRITE THE FILTERED TIME SERIES.
!
        do k = 1, time3
!
            temp(:nlon,:nlat) = unpack( mat(:nvar,k), mask, missing )
!
            startwri(3) = k
!
            status = nf90_put_var( ncid_filt, filtid, temp, start=startwri )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    else
!
#endif
!
#ifdef _OPENMP
        test_par = i>1 .and. j>1 .and. imax>1 .and. period*nvar>omp_limit
#endif
!
        startlec(1:2)  = (/ j1, lat1 /)
!
!       FILTER THE TIME SERIES BY PERIODS.
!
        do i = 1, nperiod
!
!           FIRST GET THE DATA.
!
            l = offset_rec + (i-1)*period
!
            do k = 1, period
!
                l           = l + 1
                startlec(3) = l
!
!               FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
                status = nf90_get_var( ncid, varid, temp(:lon3,:lat3), start=startlec, count=countlec )
                if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               PACK THE DATA.
!
                mat(:nvar,k) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
            end do
!
!           MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
            if ( use_scalfac ) mat(:nvar,:period) = mat(:nvar,:period)*scalfac
!
!           ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
            if ( use_offset ) mat(:nvar,:period) = mat(:nvar,:period) + offset
!
            if ( detrend_only ) then
!
!               DETREND THE MULTICHANNEL TIME SERIES, ONLY.
!
                call detrend( mat(:nvar,:period), trend )
!
            else if ( method/=0 ) then
!
!               FREQUENCIES FILTERING OF THE MULTICHANNEL TIME SERIES IS REQUIRED.
!
!$OMP PARALLEL IF(test_par)              &
!$OMP         ,PRIVATE(k,i0,i1,i2)       &
!$OMP         ,SHARED(mat,coef,trend,use_fft,nvar,imax,ngp,period,nfc)
!
!$OMP DO SCHEDULE(STATIC) 
!
                do k = 1, imax
!
                    i0 = (k-1)*ngp
                    i1 = i0 + 1
                    i2 = min( nvar, i0+ngp )
!
                    call symlin_filter2( mat(i1:i2,:period), coef(:nfc), trend=trend, usefft=use_fft, initfft=false )
!
                end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
            end if
!
!           WRITE THE FILTERED TIME SEGMENT.
!
            l = (i-1)*period
!
            do k = 1, period
!
                temp(:nlon,:nlat) = unpack( mat(:nvar,k), mask, missing )
!
                l           = l + 1
                startwri(3) = l
!
                status = nf90_put_var( ncid_filt, filtid, temp, start=startwri )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
        end do
!
#ifdef _PARALLEL_READ
!
    end if
!
#endif
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( use_fft ) then
        call end_fft( )
    end if
!
    if ( detrend_only ) then
        deallocate( temp, mask, mat )
    else
        deallocate( temp, mask, mat, coef )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_mask ) then
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_lanczos_filter_3d
! =====================================
!
end program comp_lanczos_filter_3d
