program comp_mask_4d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_mask_4d:
!
! ============
! comp_mask_4d
! ============
!
!
! Authors
! =======
!
! Eric Maisonnave (CERFACS) and Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 13/09/2018
!
! Purpose
! =======
!
! Create a mesh-mask variable by specifying geographical bounds in indices or geographical units
! (e.g. degrees, meters or pressure)  and/or by modifying an input mesh-mask read from a NetCDF
! dataset and store the resulting mesh-mask variable in a NetCDF dataset if the **-o=** argument
! (described below) is specified.
!
! If an output NetCDF dataset is not specified with the **-o=** argument, the geographical bounds
! of the selected domain are printed as indices on output of the procedure. Such indices can then
! be used for specifying the **-x=**,  **-y=** and **-z=** arguments for the selected domain in
! other NCSTAT procedures such as use :ref:`comp_serie_4d`.
!
! If you want to create a mesh-mask variable for a tridimensional NetCDF variable, use :ref:`comp_mask_3d` 
! instead of comp_mask_4d.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_mask_4d \
!         -f=input_netcdf_file \
!         -vlon=longitude_variable \
!         -vlat=latitude_variable \
!         -vlev=level_variable \
!         -g=grid_type                      (optional : n, t, u, v, w, f) \
!         -x=lon1,lon2                      (optional) \
!         -y=lat1,lat2                      (optional) \
!         -z=level1,level2                  (optional) \
!         -a=type_of_coordinates            (optional : indices, units) \
!         -imsk=in_mask_value               (optional : ident, 0, 1) \
!         -omsk=out_mask_value              (optional : ident, 0, 1) \
!         -fmsk=input_mask_netcdf_file      (optional) \
!         -vmsk=input_mask_netcdf_variable  (optional) \
!         -val=mask_value                   (optional) \
!         -rel=mask_relation                (optional : eq, gt, ge, lt, le) \
!         -vout=output_mask_netcdf_variable (optional) \
!         -o=output_mask_netcdf_file        (optional) \
!         -noscalfac                        (optional) \
!         -hdf5                             (optional)
!
! By default
! ----------
!
!        *-g=*
!         the *grid_type* is set to ``n``,  which means that the 2-D grid-mesh associated with
!         the input *latitude_variable* and *longitude_variable* is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude range associated with the *longitude_variable*
!        *-y=*
!         the whole latitude range associated with the *latitude_variable*
!        *-z=*
!         the whole vertical range associated with the *level_variable*
!        *-a=*
!         the *type_of_coordinates* is set to ``indices``. This means that the
!         geographical bounds specified in the **-x=** and  **-y=** arguments
!         are given as indices not in geographical units ( e.g. degrees of longitude
!         and latitude or meters for example)
!        *-imsk=*
!         the *in_mask_value* is set to ``1``. This means that points inside the
!         specified domain are selected
!        *-omsk=*
!         the *out_mask_value* is set to ``0``. This means that points outside the
!         specified domain are not selected
!        *-fmsk=*
!         an *input_mask_netcdf_file* is not used
!        *-vmsk=*
!         an *input_mask_netcdf_variable* is not used
!        *-val=*
!         the *mask_value* is set to ``1.``
!        *-rel=*
!         the *mask_relation* is set to ``eq``
!        *-vout=*
!         the *output_mask_netcdf_variable* is named *grid_type*//``mask`` if
!         an output NetCDF file is created
!        *-o=*
!         an *output_mask_netcdf_file* is not created
!        *-noscalfac*
!         if the **-fmsk=** argument is present, the scale factors NetCDF variables,
!         if they exist in the *input_mask_netcdf_file*, are automatically copied
!         to the output NetCDF file. If the **-noscalfac** argument is specified,
!         the scale factors NetCDF variables are not copied in the output NetCDF file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!
! Remarks
! -------
!
!   #) The **-vlon=**, **-vlat=** and **-vlev=** arguments specify the coordinate variables from which 
!      the mesh-mask variable will be constructed and the **-f=**\ *input_netcdf_file* argument 
!      specifies that these coordinate NetCDF variables must be extracted from the NetCDF file
!      *input_netcdf_file*. The longitude and latitude coordinate NetCDF variables may be one
!      or two dimensional arrays.
!
!   #) If the **-x=**\ *lon1,lon2*, **-y=**\ *lat1,lat2* and **-z=**\ *level1,level2* arguments are
!      missing the whole geographical domain associated with the *longitude_variable* and *latitude_variable*,
!      and the whole vertical resolution associated with the *level_variable* are used
!      to construct the mesh-mask.
!
!      If **-a=**\ ``indices`` (the default value), the longitude (specified with the **-x=** argument),
!      latitude (specified with the **-y=** argument) or level (specified with the **-z=** argument) range
!      must be a vector of two integers specifying the first and last selected indices along each dimension.
!      The indices are relative to ``1``\ . Negative values for *lon1* are not allowed.
!
!      If **-a=**\ ``units``, the longitude or latitude range must be a vector of two 
!      integers specifying the longitude and latitude limits of the domain in degrees of longitude
!      or latitude, respectively. In that case, negative values are allowed for *lon1*, *lon2*,
!      *lat1* and *lat2*. Moreover, the longitudes *lon1* and *lon2* are shifted to be in the interval
!      between ``-180`` and ``180`` degrees, assuming that the grid is periodic in longitude.
!      The vertical range must be a vector of two integers specifying the first and last levels
!      of the domain in the unit of the NetCDF variable given in the **-vlev=** argument.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the input
!      3-D grid-mesh associated with the input coordinate variables is from an experiment with
!      the ORCA model (R2, R4 or R05 resolutions).
!
!      This argument is also used to determine the name of the *output_mask_netcdf_variable*
!      in the output NetCDF file if the **-vout=** argument is not specified.
!      In that case, the *output_mask_netcdf_variable* will be named *grid_type*//``mask``.
!
!   #) The **-imsk=** argument specifies the value (``0``, ``1`` or ``ident``) to be given to grid points
!      inside of the specified domain (as specified by the **-x=**, **-y=** and **-z=** arguments). 
!      By default, values inside the specified domain are set to ``1`` (e.g. **-imsk=**\ ``1``). 
!
!      If **-imsk=**\ ``ident``, the **-vmsk=** and **-fmsk=** arguments must also be given and the values
!      inside the selected domain are read from the *input_mask_netcdf_variable* and are modified according
!      to the logical relation determined by the **-rel=** and **-val=** arguments.
!
!   #) The **-omsk=** argument is identical to **-imsk=** argument, but concerns points 
!      outside the specified domain (as specified by the  **-x=**, **-y=** and **-z=** arguments).
!      By default, values outside the specified domain are set to ``0`` (**-omsk=**\ ``0``). 
!
!      If **-omsk=**\ ``ident``, the **-vmsk=** and **-fmsk=** arguments must also be given and the values
!      outside the selected domain are read from the *input_mask_netcdf_variable* and are modified according
!      to the logical relation determined by the **-rel=** and **-val=** arguments.
!
!   #) If **-imsk=**\ ``ident`` or **-omsk=**\ ``ident``, the **-fmsk=** and **-vmsk=** arguments
!      must be specified.
!      In that case, the output mesh-mask variable *output_mask_netcdf_variable* is first
!      initialized from the *input_mask_netcdf_variable* (as specified by the **-vmsk=**
!      and **-fmsk=** arguments) inside the domain if **-imsk=**\ ``ident`` or outside the domain if
!      **-omsk=**\ ``ident``, and then modified in these regions according to the logical relationship:
!
!      * **output_mask**\ (i,j,k) = ``1`` if **input_mask**\ (i,j,k) .\ *mask_relation*\ . *mask_value*
!        is ``true``
!
!      * **output_mask**\ (i,j,k) = ``0`` otherwise
!
!      where *mask_relation* is determined from the **-rel=** argument and *mask_value* 
!      from the **-val=** argument.
!
!      By default, *mask_relation* is set to ``eq`` and *mask_value* is set to ``1.``\  . *mask_value* must
!      be specified as a real number.
!
!      The **-fmsk=** argument must be present, if the **-vmsk=** argument is specified, otherwise
!      the program will stop.
!
!      If the **-fmsk=** and **-vmsk=** arguments are not specified, the output mesh-mask variable
!      is computed by applying the *in_mask_value* value (as specified by the **-imsk=** argument)
!      inside the specified domain and the *out_mask_value* (as specified by the **-omsk=** argument)
!      outside the specified domain.
!
!   #) The geographical shapes of the *input_mask_netcdf_variable* (in the *input_mask_netcdf_file*) must
!      agree with the input coordinate variables *longitude_variable*, *latitude_variable* and *level_variable*
!      (in the *input_netcdf_file*) if an *input_mask_netcdf_file* is used.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_mask_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
! Outputs
! -------
!
!   comp_mask_4d creates an output NetCDF file if the **-o=** argument is specified. This NetCDF dataset
!   will contain the constructed mesh-mask (e.g. with values equal to ``0`` or ``1``) in a NetCDF
!   variable called *output_mask_netcdf_variable* whose geographical dimensions are the same as those
!   associated with the 3-D grid associated with the *longitude_variable*, *latitude_variable* and *level_variable*
!   specified with the **-vlon=**, **-vlat=** and **-vlev=** arguments (in the description below, nlat, nlon and nlev
!   are the lengths of the geographical dimensions of the input 3-D grid) :
!
!   #) *output_mask_netcdf_variable*\ ``(nlev,nlat,nlon)`` : the constructed mesh-mask (e.g. with values
!      equal to ``0`` or ``1``) on the 3-D grid-mesh associated with the input coordinate variables.
!
!   Optionally, if an *input_mask_netcdf_file* is specified with the **-fmsk=** argument and this
!   NetCDF file contains the scale factors associated with the 3-D grid associated with the input
!   *longitude_variable*, *latitude_variable* and *level_variable* NetCDF variables, the output NetCDF
!   file will also contain the following variables (if the **-noscalfac** argument is not specified) :
!
!   #) *netcdf_variable*\ _e1\ *grid_type*\ ``(nlat,nlon)`` : the first scale factor associated with the
!      2-D grid-mesh of the input coordinate variables.
!
!   #) *netcdf_variable*\ _e2\ *grid_type*\ ``(nlat,nlon)`` : the second scale factor associated with the
!      2-D grid-mesh of the input coordinate variables.
!
!   #) *netcdf_variable*\ _e3\ *grid_type*\ ``(nlev,1,1)`` : the third scale factor associated with the
!      3-D grid-mesh of the input coordinate variables.
!
!   Multiplying the first two scale factors together gives the surface of each cell in the 3-D grid-mesh
!   associated with the *output_mask_netcdf_variable*.
!
!-
! ================================================================================================
!
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,   only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_var_dims, nf90_max_name, nf90_nowrite, nf90_noerr, nf90_clobber,  &
                         nf90_open, nf90_get_var, nf90_strerror, nf90_inq_varid, nf90_create,       &
                         nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_def_var, nf90_put_att,      &
                         nf90_inquire_variable, nf90_inq_attname, nf90_copy_att, nf90_global,       &
                         nf90_enddef, nf90_close, nf90_real4, nf90_real8, nf90_put_var
!
    use Statpack, only : i1b, i4b, stnd, lgl, zero, one, c90, c180, c360,        &
                         true, false, allocate_error, merror,                    &
                         kfix, kexp, tolower, case_change, string_to_val
!
    use NCSTAT_Utilities, only : prtunit, nchar, get_listarg, get_variable_dims
!
!
! STRONG TYPING IMPOSED
! =====================
!
    implicit none
!
!
! PARAMETERS
! ==========
!
!
    real(stnd), parameter   :: ident = -one
!
    character(len=*), parameter :: name_proc='comp_mask_4d : ',               &
!
                                   mmethod(3)=(/"ident", "0    ","1    "/),   &
!
                                   crelation(5)=(/"eq","gt","ge","lt","le"/), &
!
                                   coordinates(2)=(/"indices", "units  " /)     
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_fmsk, ncid_mask,                     &
                     lonid, latid, zid, vmskid, maskid,                            &
                     e1id, e2id, e3id, e1id_out, e2id_out, e3id_out,               &
                     nlon, nlat, nlevel, ndim_lon, ndim_lat, ndim_z, i, j,         &
                     lon1, lon2, lon3, lat1, lat2, lat3, level1, level2, level3,   &
                     ilon1, ilon2, ilat1, ilat2, ilevel1, ilevel2,                 &
                     mskrel, coord, natts, output_mode
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, dimids2
!
    integer(i1b)  :: kcode
!
    real(stnd)                                :: mask_in, mask_out, mskval
    real(stnd), allocatable, dimension(:)     :: temp1
    real(stnd), allocatable, dimension(:,:)   :: lon, lat, temp2
    real(stnd), allocatable, dimension(:,:,:) :: mask, mask2
!
    logical(lgl)                              :: use_units, use_x, use_y, use_z, copy_scalfac, create_hdf5
!
    character(len=14)                         :: fmt
    character(len=nchar)                      :: name, history
    character(len=nf90_max_name)              :: fic_name, lon_name, lat_name, mask_name, var_name, &
                                                 fmsk_name, vmsk_name, att_name, z_name
    character(len=nf90_max_name), dimension(3) :: dim_names
    character                                  :: chr, grid_type
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Create a mesh-mask variable by specifying geographical bounds'
        write (prtunit,*) ' in indices or geographical units and/or modifying an input mesh-mask.'
        write (prtunit,*) ' The resulting mesh-mask variable can be stored in a NetCDF dataset'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_mask_4d -f=input_netcdf_file'
        write (prtunit,*) '              -vlon=longitude_variable'
        write (prtunit,*) '              -vlat=latitude_variable'
        write (prtunit,*) '              -vlev=level_variable'
        write (prtunit,*) '              -g=grid_type                      (optional : n, t, u, v, w, f)'
        write (prtunit,*) '              -x=lon1,lon2                      (optional)'
        write (prtunit,*) '              -y=lat1,lat2                      (optional)'
        write (prtunit,*) '              -z=level1,level2                  (optional)'
        write (prtunit,*) '              -a=type_of_coordinates            (optional : indices, units)'
        write (prtunit,*) '              -imsk=in_mask_value               (optional : ident, 0, 1)'
        write (prtunit,*) '              -omsk=out_mask_value              (optional : ident, 0, 1)'
        write (prtunit,*) '              -fmsk=input_mask_netcdf_file      (optional)'
        write (prtunit,*) '              -vmsk=input_mask_netcdf_variable  (optional)'
        write (prtunit,*) '              -val=mask_value                   (optional)'
        write (prtunit,*) '              -rel=mask_relation                (optional : eq,gt,ge,lt,le)'
        write (prtunit,*) '              -vout=output_mask_netcdf_variable (optional)'
        write (prtunit,*) '              -o=output_mask_netcdf_file        (optional)'
        write (prtunit,*) '              -noscalfac                        (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '              -hdf5                             (optional)'
#endif
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-g= the grid type is set to n'
        write (prtunit,*) '-x= the whole longitude domain associated with the longitude_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the latitude_variable'
        write (prtunit,*) '-z= the whole vertical resolution associated with the level_variable'
        write (prtunit,*) '-a=indices : by default, the geographical bounds must be given as indices'
        write (prtunit,*) '-imsk=1 : grid points inside the specified domain are selected'
        write (prtunit,*) '-omsk=0 : grid points outside the specified domain are not selected'
        write (prtunit,*) '-fmsk= an input_mask_netcdf_file is not used'
        write (prtunit,*) '-vmsk= an input_mask_netcdf_variable is not used'
        write (prtunit,*) '-val=1.'
        write (prtunit,*) '-rel=eq'
        write (prtunit,*) '-vout= the output_mask_netcdf_variable is named grid_type//mask'
        write (prtunit,*) '-o= an output_mask_netcdf_file is not created'
        write (prtunit,*) '-noscalfac : if present in the input_mask_netcdf_file, the'
        write (prtunit,*) '             scale factors are copied in the output_mask_netcdf_file'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5      : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<4 .or. narg>19 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name  = " "
    lon_name  = " "
    lat_name  = " "
    z_name    = " "
    fmsk_name = " "
    vmsk_name = " "
    var_name  = " "
    mask_name = " "
!
    history = "comp_mask_4d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
!
    grid_type   = "n"
!
    coord       = 1
    mskrel      = 1
    mskval      = one
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    level1    = 1
    level2    = huge( level2 )
!
    use_x        = false
    use_y        = false
    use_z        = false
    use_units    = false
    copy_scalfac = true
    create_hdf5  = false
!
    mask_in   =  one
    mask_out  =  zero
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (x index).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (y index).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-z=')
!
!               GET THE VERTICAL DEFINITION (z index).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "level1 and level2 are integer values", &
                                  "Wrong domain definition (vertical)",   &
                                  level1, level2 )
!
                use_z = true
!
            case ('-a=')
!
!               DETERMINE THE TYPE OF COORDINATES.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( coordinates(1) )
                        coord  = 1
                    case ( coordinates(2) )
                        coord  = 2
                        use_units = true
                    case default
                        call merror( name_proc//"Unknown coordinate option" )
                end select
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR THE CLIMATOLOGY.
!
                mask_name = name(4:)
!
            case default
!
                call case_change( name(4:5), tolower )
!
                select case ( name(1:5 ) )
!
                    case ('-val=')
!
!                       GET THE MASK VALUE FOR THE INPUT MESH-MASK FILE.
!
                        call string_to_val( name(6:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -val= argument must be a real number " )
                        read( name(6:), fmt ) mskval
!
                    case ('-rel=')
!
!                       DETERMINE THE MASK RELATION FOR THE INPUT MESH-MASK FILE.
!
                        call case_change( name(6:), tolower )
!
                        select case ( name(6:) )
                           case ( crelation(1) )
                               mskrel = 1
                           case ( crelation(2) )
                               mskrel = 2
                           case ( crelation(3) )
                               mskrel = 3
                           case ( crelation(4) )
                               mskrel = 4
                           case ( crelation(5) )
                               mskrel = 5
                           case default
                               call merror( name_proc//"Unknown mask_relation option" )
                        end select
!
                    case default
!
                        select case ( name(1:6 ) )
!
                            case ('-vlon=')
!
!                               GET THE NAME OF THE INPUT LONGITUDE VARIABLE.
!
                                lon_name = name(7:)
!
                            case ('-vlat=')
!
!                               GET THE NAME OF THE INPUT LATITUDE VARIABLE.
!
                                lat_name = name(7:)
!
                            case ('-vlev=')
!
!                               GET THE NAME OF THE INPUT VERTICAL (DEPTH OR HEIGHT) VARIABLE.
!
                                z_name = name(7:)
!
                            case ('-fmsk=')
!
!                               GET THE NAME OF THE INPUT MESH-MASK FILE.
!
                                fmsk_name = name(7:)
!
                            case ('-vmsk=')
!
!                               GET THE NAME OF THE INPUT MESH-MASK VARIABLE.
!
                                vmsk_name = name(7:)
!
                            case ('-vout=')
!
!                               GET THE NAME OF THE OUTPUT MESH-MASK VARIABLE.
!
                                var_name = name(7:)
!
                            case ('-imsk=')
!
!                               GET MASKING METHOD INSIDE DOMAIN LIMITS.
!
                                call case_change( name(7:), tolower )
!
                                select case ( name(7:) )
!
                                    case ( mmethod(1) )
!
!                                       NO CHANGES.
!
                                        mask_in   =  ident
!
                                    case ( mmethod(2) )
!
!                                       FILL SELECTED REGION WITH 0.
!
                                        mask_in   =  zero
!
                                    case ( mmethod(3) )
!
!                                       FILL SELECTED REGION WITH 1.
!
                                        mask_in   =  one
!
                                    case default
                                        call merror( name_proc//"Unknown masking value (-imsk=)" )
                                end select
!
                            case ('-omsk=')
!
!                               GET MASKING METHOD OUTSIDE DOMAIN LIMITS.
!
                                call case_change( name(7:), tolower )
!
                                select case ( name(7:) )
!
                                    case ( mmethod(1) )
!
!                                       NO CHANGES OUTSIDE SELECTED REGION.
!
                                        mask_out   =  ident
!
                                    case ( mmethod(2) )
!
!                                       FILL OUTSIDE SELECTED REGION WITH 0.
!
                                        mask_out   =  zero
!
                                    case ( mmethod(3) )
!
!                                       FILL OUTSIDE SELECTED REGION WITH 1.
!
                                        mask_out   =  one
!
                                    case default
                                        call merror( name_proc//"Unknown masking value(-omsk=)" )
!
                                end select
!
                            case default
!
                                call case_change( name(8:), tolower )
!
                                select case ( name(1:) )
!
                                    case ('-noscalfac')
!
!                                       STORE SCALE FACTORS.
!
                                        copy_scalfac = false
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
                                        create_hdf5 = true
#endif
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS AND SET DEFAULT VALUES.
!
    if ( fic_name==" " ) call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( lon_name==" " ) call merror( name_proc//"Input longitude NETCDF variable is missing" )
!
    if ( lat_name==" " ) call merror( name_proc//"Input latitude NETCDF variable is missing" )
!
    if ( z_name==" " ) call merror( name_proc//"Input vertical NETCDF variable is missing" )
!
    if ( mask_out==ident .or. mask_in==ident ) then
        if ( vmsk_name==" " ) call merror( name_proc//"Input NETCDF mask_variable is missing" )
    end if
!
    if ( fmsk_name==" " ) copy_scalfac = false
!
    if ( vmsk_name/=" " ) then
        if ( fmsk_name==" " ) call merror( name_proc//"Input NETCDF mask_file is missing" )
    end if
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
    if ( var_name==" " ) var_name = chr//"mask"
!
!   OPEN THE NETCDF DATASET.
!
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE LONGITUDE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE LONGITUDE NETCDF VARIABLE.
!
    call get_variable_dims( name_proc, trim(lon_name), ncid, lonid, dimids, dimlens )
    ndim_lon = count( dimlens/=0 )
!
    if ( ndim_lon<=2 ) then
!
        nlon = dimlens(1)
        if ( ndim_lon==2 ) nlat = dimlens(2)
!
    else
!
        call merror( name_proc//"Input longitude NETCDF variable is not one or twodimensional" )
!
    end if
!
!   GET THE LATITUDE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE LATITUDE NETCDF VARIABLE.
!
    call get_variable_dims( name_proc, trim(lat_name), ncid, latid, dimids, dimlens )
    ndim_lat = count( dimlens/=0 )
!
    if ( ndim_lat<=2 ) then
!
        if ( ndim_lon==2 ) then
!
            if ( ndim_lat==1 ) then
                if ( nlat/=dimlens(1)  )       &
                call merror( name_proc//"Shape disagreement between the longitude and latitude variables" )
            else
                if ( nlon/=dimlens(1) .or. nlat/=dimlens(2)  )       &
                call merror( name_proc//"Shape disagreement between the longitude and latitude variables" )
            end if
!
        else if ( ndim_lat==1 ) then
!
            nlat = dimlens(1)
!
        else
!
            if ( nlon/=dimlens(1) )       &
            call merror( name_proc//"Shape disagreement between the longitude and latitude variables" )
            nlat = dimlens(2)
!
        end if
!
    else
!
        call merror( name_proc//"Input latitude NETCDF variable is not one or twodimensional" )
!
    end if
!
!   GET THE VERTICAL VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE VERTICAL NETCDF VARIABLE.
!
    call get_variable_dims( name_proc, trim(z_name), ncid, zid, dimids, dimlens )
    ndim_z = count( dimlens/=0 )
!
    if ( ndim_z==3 ) then
!
        nlevel = dimlens(3)
        if ( dimlens(1)/=1 .or. dimlens(2)/=1 ) &
        call merror( name_proc//"Two first dimensions of input vertical NETCDF variable must be dummy" )
!
    else if ( ndim_z==1) then
!
        nlevel = dimlens(1)
!
    else
!
        call merror( name_proc//"Input vertical NETCDF variable is not tridimensional" )
!
    end if
!
!
!   ALLOCATE WORK ARRAYS.
!
    i = nlevel
    if ( ndim_lon==1 ) i = max( nlon, i )
    if ( ndim_lat==1 ) i = max( nlat, i )
!
    allocate( mask(nlon, nlat, nlevel), temp2(nlon, nlat), temp1(i),  stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    if ( use_units ) then
!
!       THE GEOGRAPHICAL DOMAIN IS SPECIFIED IN DEGREES.
!
!       CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
        if ( .not.use_x ) then
            lon1 = -180
            lon2 = 180
        end if
!
        if ( .not.use_y ) then
            lat1 = -90
            lat2 = 90
        end if
!
        lat1 = max( lat1, -90_i4b )
        lat2 = min( lat2, 90_i4b )
!
        if ( lat1>lat2 ) call merror( name_proc//"Wrong domain definition (latitude)" )
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( lon(nlon, nlat), lat(nlon, nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET LONGITUDE VARIABLE.
!
        if ( ndim_lon==1 ) then
!
!           REGULAR GRIDS.
!
            status = nf90_get_var( ncid, lonid, temp1(:nlon), count=(/ nlon /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            lon(:nlon,:nlat) = spread ( temp1(:nlon), dim=2, ncopies=nlat )
!
        else
!
!           IRREGULAR GRIDS.
!
            status = nf90_get_var( ncid, lonid, lon(:nlon,:nlat), count=(/ nlon, nlat /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
!       SHIFT LONGITUDES BETWEEN -180 AND 180 DEGREES.
!
        where ( lon > c180 )     &
        lon = lon - aint ( ( lon + c180 ) / c360, stnd ) * c360
!
        where ( lon < -c180 )    &
        lon = lon - aint ( ( lon - c180 ) / c360, stnd  ) * c360
!
        if ( lon1 > c180 ) then
            lon1 = lon1 - aint ( ( lon1 + c180 ) / c360, stnd  ) * c360
        else if ( lon1 < -c180 ) then
            lon1 = lon1 - aint ( ( lon1 - c180 ) / c360, stnd  ) * c360
        end if
!
        if ( lon2 > c180 ) then
            lon2 = lon2 - aint ( ( lon2 + c180 ) / c360, stnd  ) * c360
        else if ( lon2 < -c180 ) then
            lon2 = lon2 - aint ( ( lon2 - c180 ) / c360, stnd  ) * c360
        end if
!
!       GET LATITUDE VARIABLE.
!
        if ( ndim_lat==1 ) then
!
!           REGULAR GRIDS.
!
            status = nf90_get_var( ncid, latid, temp1(:nlat), count=(/ nlat /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            lat(:nlon,:nlat) = spread ( temp1(:nlat), dim=1, ncopies=nlon )
!
        else
!
!           IRREGULAR GRIDS.
!
            status = nf90_get_var( ncid, latid, lat(:nlon,:nlat), count=(/ nlon, nlat /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
        if ( any( abs(lat(:nlon,:nlat))>c90 ) )  &
        call merror( name_proc//" Wrong latitude variable : abs(latitude)>90 " )
!
!       CREATE GEOGRAPHICAL MASK USING LONGITUDES / LATITUDES BOUNDS.
!
        if ( lon1<=lon2 ) then
!
            where ( lon>=lon1 .and. lon<=lon2  .and. &
                    lat>=lat1 .and. lat<=lat2      )
                temp2(:nlon,:nlat) = mask_in
            elsewhere
                temp2(:nlon,:nlat) = mask_out
            end where
!
        else
!
            where ( ( ( lon>=lon1  .and. lon<=c180  )  .or.   &
                      ( lon>=-c180 .and. lon<=lon2 ) ) .and.  &
                    ( lat>=lat1 .and. lat<=lat2     )     )
                temp2(:nlon,:nlat) = mask_in
            elsewhere
                temp2(:nlon,:nlat) = mask_out
            end where
!
        endif
!
!       COPY HORIZONTAL MASK ALONG VERTICAL DIMENSION.
!
        mask(:nlon,:nlat,:nlevel) = spread ( temp2(:nlon,:nlat), dim=3, ncopies=nlevel )
!
!       GET DEPTH OR HEIGHT VARIABLE.
!
        if ( ndim_z==1 ) then
!
            status = nf90_get_var( ncid, zid, temp1(:nlevel), count=(/ nlevel /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            status = nf90_get_var( ncid, zid, temp1(:nlevel), count=(/ 1, 1, nlevel /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
!       ERASE GEOGRAPHICAL MASK OUTSIDE LEVEL BOUNDS.
!
        if ( use_z ) then
!
            i      = min( level1, level2 )
            level2 = max( level1, level2 )
            level1 = i
!
            do i = 1, nlevel
!
                if ( temp1(i)<level1 .or. temp1(i)>level2 )    &
                mask(:nlon,:nlat,i) = mask_out
!
            end do
!
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( lat, lon )
!
    else
!
!       THE GEOGRAPHICAL DOMAIN IS SPECIFIED IN INDICES.
!
!       CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
        lat1 = max( lat1, 1 )
        lat2 = min( lat2, nlat )
        lat3 = lat2 - lat1 + 1
        if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
        if ( lon1<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
        if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
        level1 = max( level1, 1 )
        level2 = min( level2, nlevel )
        level3 = level2 - level1 + 1
        if ( level3<=0  )  call merror( name_proc//"Wrong level definition" )
!
!       CREATE GEOGRAPHICAL MASK USING INDICES.
!
        mask(:nlon,:nlat,:nlevel)               = mask_out
        mask(lon1:lon2,lat1:lat2,level1:level2) = mask_in
!
    end if
!
    if ( fmsk_name/=" " ) then
!
!       OPEN THE INPUT NETCDF MESH-MASK DATASET.
!
        status = nf90_open( trim(fmsk_name), nf90_nowrite, ncid_fmsk )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( copy_scalfac ) then
!
            status = nf90_inq_varid ( ncid_fmsk, "e1"//chr, e1id )
            if ( status/=nf90_noerr ) copy_scalfac = false
!
            status = nf90_inq_varid ( ncid_fmsk, "e2"//chr, e2id )
            if ( status/=nf90_noerr ) copy_scalfac = false
!
            status = nf90_inq_varid ( ncid_fmsk, "e3"//chr, e3id )
            if ( status/=nf90_noerr ) copy_scalfac = false
!
        end if
!
        if ( copy_scalfac ) then
!
!           GET THE SCALE FACTORS IDs FROM THEIR NAMES.
!           VERIFY THE DIMENSION LENGTHS FOR THE SCALE FACTOR VARIABLES.
!
            call get_variable_dims( name_proc, "e1"//chr, ncid_fmsk,    &
                                    e1id, dimids, dimlens )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
            call merror( name_proc//"Shape disagreement with the first input scale factor" )
!
            call get_variable_dims( name_proc, "e2"//chr, ncid_fmsk,    &
                                    e2id, dimids, dimlens )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
            call merror( name_proc//"Shape disagreement with the second input scale factor" )
!
            call get_variable_dims( name_proc, "e3"//chr, ncid_fmsk,    &
                                    e3id, dimids, dimlens )
!
            if ( dimlens(1)/=1 .or. dimlens(2)/=1 .or. dimlens(3)/=nlevel )               &
            call merror( name_proc//"Shape disagreement with the third input scale factor" )
!
        end if
!
        if ( mask_out==ident.or.mask_in==ident ) then
!
!           CONSTRUCT MASK FROM THE INPUT MESH-MASK.
!
!           GET THE MASK ID FROM ITS NAME.
!           VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
            call get_variable_dims( name_proc, trim(vmsk_name), ncid_fmsk,    &
                                    vmskid, dimids, dimlens )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
            call merror( name_proc//"Shape disagreement with the input mask" )
!
!           ALLOCATE WORK ARRAY.
!
            allocate( mask2(nlon, nlat, nlevel), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           GET THE MASK.
!
            status = nf90_get_var( ncid_fmsk, vmskid, mask2(:nlon,:nlat,:nlevel)  )     
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           DETERMINE THE GRID POINTS USED FROM THE MASK, MASK RELATION AND MASK VALUE.
!
            select case ( mskrel )
!
                case ( 1 )
!
                    where( mask(:nlon,:nlat,:nlevel)==ident )   &
                    mask(:nlon,:nlat,:nlevel) = merge( one, zero, mask2(:nlon,:nlat,:nlevel)==mskval )
!
                case ( 2 )
!
                    where( mask(:nlon,:nlat,:nlevel)==ident )   &
                    mask(:nlon,:nlat,:nlevel) = merge( one, zero, mask2(:nlon,:nlat,:nlevel)>mskval )
!
                case ( 3 )
!
                    where( mask(:nlon,:nlat,:nlevel)==ident )   &
                    mask(:nlon,:nlat,:nlevel) = merge( one, zero, mask2(:nlon,:nlat,:nlevel)>=mskval )
!
                case ( 4 )
!
                    where( mask(:nlon,:nlat,:nlevel)==ident )   &
                    mask(:nlon,:nlat,:nlevel) = merge( one, zero, mask2(:nlon,:nlat,:nlevel)<mskval )
!
                case ( 5 )
!
                    where( mask(:nlon,:nlat,:nlevel)==ident )   &
                    mask(:nlon,:nlat,:nlevel) = merge( one, zero, mask2(:nlon,:nlat,:nlevel)<=mskval )
!
            end select
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( mask2 )
!
        end if
!
    end if
!
!   COMPUTE INDICE BOUNDS.
!
    do ilon1=1,nlon
        if ( any( mask(ilon1,:nlat,:nlevel)==one ) ) exit
    end do
!
    do ilon2=nlon,1,-1
        if ( any( mask(ilon2,:nlat,:nlevel)==one ) ) exit
    end do
!
    do ilat1=1,nlat
        if ( any( mask(:nlon,ilat1,:nlevel)==one ) ) exit
    end do
!
    do ilat2=nlat,1,-1
        if ( any( mask(:nlon,ilat2,:nlevel)==one ) ) exit
    end do
!
    do ilevel1=1,nlevel
        if ( any( mask(:nlon,:nlat,ilevel1)==one ) ) exit
    end do
!
    do ilevel2=nlevel,1,-1
        if ( any( mask(:nlon,:nlat,ilevel2)==one ) ) exit
    end do
!
    if ( ilat1>ilat2 .or. ilon1>ilon2 .or. ilevel1>ilevel2 )    &
    call merror( name_proc//"Selected domain(mask) is empty" )
!
    if ( ilon1==1 .and. ilon2==nlon ) then
!
        do ilon2=1,nlon
            if ( all( mask(ilon2,ilat1:ilat2,ilevel1:ilevel2)==zero ) ) exit
        end do
!
        do ilon1=nlon,1,-1
            if ( all( mask(ilon1,ilat1:ilat2,ilevel1:ilevel2)==zero ) ) exit
        end do
!
        if ( ilon1==0 ) then
!
            ilon1 = 1
            ilon2 = nlon
!
        else if ( ilon2<=ilon1 ) then
!
            if ( all( mask(ilon2:ilon1,ilat1:ilat2,ilevel1:ilevel2)==zero ) ) then
                ilon1 = ilon1 - nlon
                ilon2 = ilon2 - 1
            else
                ilon1 = 1
                ilon2 = nlon
            end if
!
        else
!
           call merror( name_proc//"Selected domain(mask) is empty" )
!
        end if
!
    end if
!
    if ( mask_name==" " ) then
!
!       PRINT THE INDICES ON OUTPUT.
!
        write (prtunit,*)
        write (prtunit,*) 'COORDINATE INDICES FROM '//trim(lon_name)//", " &
                         //trim(lat_name)//" AND "//trim(z_name)//" READ FROM "//trim(fic_name)
        write (prtunit,*)
        write (prtunit,*) 'lon1= ',ilon1
        write (prtunit,*) 'lon2= ',ilon2
        write (prtunit,*) 'lat1= ',ilat1
        write (prtunit,*) 'lat2= ',ilat2
        write (prtunit,*) 'level1= ',ilevel1
        write (prtunit,*) 'level2= ',ilevel2
        write (prtunit,*)
!
    else
!
!       CREATE THE NEW NETCDF DATASET FOR THE OUTPUT MESH MASK.
!
        status = nf90_create( trim(mask_name), cmode=nf90_clobber, ncid=ncid_mask )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_set_fill( ncid_mask, nf90_nofill, i )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       DEFINE THE DIMENSIONS FOR THE OUTPUT MASK NETCDF DATASET.
!
        dimlens(1)   = nlon
        dimlens(2)   = nlat
        dimlens(3)   = nlevel
!
        dim_names(1) = trim(lon_name)
        dim_names(2) = trim(lat_name)
        dim_names(3) = trim(z_name)
!
        do i = 1, 3
!
            status = nf90_def_dim( ncid_mask, name=trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        if ( copy_scalfac ) then
!
!           DEFINE DUMMY DIMENSIONS FOR THE THIRD SCALE FACTOR.
!
            dim_names(:2) = (/ "x_a", "y_a" /)
!
            do i = 1, 2
!
                status = nf90_def_dim( ncid_mask, name=trim(dim_names(i)), len=1, dimid=dimids2(i) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
            dimids2(3) = dimids(3)
!
        end if
!
!       DEFINE THE VARIABLES FOR THE OUTPUT MASK NETCDF DATASET.
!
        status = nf90_def_var( ncid_mask, trim(var_name), nf90_real4, dimids=dimids(:3), varid=maskid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( copy_scalfac ) then
!
            status = nf90_def_var( ncid_mask, "e1"//chr, nf90_real8, dimids=dimids(:2), varid=e1id_out )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_def_var( ncid_mask, "e2"//chr, nf90_real8, dimids=dimids(:2), varid=e2id_out )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_def_var( ncid_mask, "e3"//chr, nf90_real8, dimids=dimids2(:3), varid=e3id_out )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
!       ADD valid_min and valid_max ATTRIBUTES.
!
        status = nf90_put_att( ncid_mask, maskid, "valid_min", zero )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "valid_max", one )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       ADD LONGITUDE / LATITUDE BOUNDS ATTRIBUTES.
!
        status = nf90_put_att( ncid_mask, maskid, "lon1_Eastern_limit", ilon1 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lon2_Western_limit", ilon2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lat1_Southern_limit", ilat1 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lat2_Northern_limit", ilat2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lev1_lower_limit", ilevel1 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lev2_upper_limit", ilevel2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( copy_scalfac ) then 
!
            dimids(1) = e1id
            dimids(2) = e2id
            dimids(3) = e3id
!
            dimlens(1) = e1id_out
            dimlens(2) = e2id_out
            dimlens(3) = e3id_out
!
!           COPY SCALE FACTORS ATTRIBUTES.
!
            do j=1,3
!
                status = nf90_inquire_variable( ncid_fmsk, dimids(j), natts=natts )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                do i=1, natts
!
                    status = nf90_inq_attname( ncid_fmsk, dimids(j), i, att_name )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   COPY ATTRIBUTES.
!
                    status = nf90_copy_att(  ncid_fmsk, dimids(j), att_name, ncid_mask, dimlens(j) )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                end do
!
            end do
!
        end if
!
!       SET GLOBAL ATTRIBUTES.
!
        status = nf90_put_att( ncid_mask, nf90_global, "title", trim(var_name)//" geographical mask" )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, nf90_global, "history", trim(history) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       LEAVE DEFINE MODE FOR THE MESH-MASK NETCDF DATASET.
!
        status = nf90_enddef( ncid_mask )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       WRITE VARIABLES FOR THE MESH-MASK NETCDF DATASET.
!
        status = nf90_put_var( ncid_mask, maskid, mask(:nlon,:nlat,:nlevel) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( copy_scalfac ) then
!
!           COPY THE FIRST AND SECOND SCALE FACTORS.
!
            do j=1,2
!
                status = nf90_get_var( ncid_fmsk, dimids(j), temp2(:nlon,:nlat), count=(/ nlon, nlat /)  )     
                if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_var( ncid_mask, dimlens(j), temp2(:nlon,:nlat) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
!           COPY THE THIRD SCALE FACTOR.
!
            status = nf90_get_var( ncid_fmsk, e3id, temp1(:nlevel), count=(/ 1, 1, nlevel /)  )     
            if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_mask, e3id_out, temp1(:nlevel), count=(/ 1, 1, nlevel/) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( mask, temp1, temp2 )
!
!       CLOSE THE NETCDF DATASET.
!
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( fmsk_name/=" " ) then
        status = nf90_close( ncid_fmsk )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!
! END OF PROGRAM comp_mask_4d
! ===========================
!
end program comp_mask_4d
