program comp_norm_3d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_norm_3d:
!
! ============
! comp_norm_3d
! ============
!
! Authors
! =======
!
! Eric Maisonnave (CERFACS) and Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 19/11/2017
!
! Purpose
! =======
!
! Select, transform and normalize time series from a tridimensional NetCDF variable
! extracted from a NetCDF dataset.
!
! The procedure allows a large variety of transformations on the input tridimensional
! NetCDF variable such as:
!
! #) removing and applying scale_factor and add_offset attributes if they are present
!
! #) changing the missing_value attribute (with the **-mi=** argument)
!
! #) applying a given mesh-mask given in input of the procedure (with the **-m=** argument)
!
! #) selecting only specific time steps in the output file (with the **-t=**\ , **-l=** and **-p=** arguments)
!
! #) centering or standardizing the time series associated with selected cells of the 
!    2-D grid-mesh associated with the input tridimensional NetCDF variable (with the **-a=** argument)
!    with the help of an input climatology (specified with the **-c=** argument)
!
! #) reducing the spatial dimensions of the output NetCDF dataset (with the **-compact** argument).
!
! An output NetCDF file containing the transformed tridimensional variable
! is created.
!
! If your data contains missing values use :ref:`comp_norm_miss_3d` instead of comp_norm_3d
! to transform your dataset.
!
! Finally, if the NetCDF variable is fourdimensional use :ref:`comp_norm_4d` instead of comp_norm_3d.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_norm_3d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -m=input_mesh_mask_netcdf_file    (optional) \
!         -g=grid_type                      (optional : n, t, u, v, w, f) \
!         -r=resolution                     (optional : r2, r4) \
!         -b=nlon_orca, nlat_orca           (optional) \
!         -x=lon1,lon2                      (optional) \
!         -y=lat1,lat2                      (optional) \
!         -t=time1,time2                    (optional) \
!         -c=input_climatology_netcdf_file  (optional) \
!         -a=type_of_transformation         (optional : scp, cov, cor, spa, tim) \
!         -d=type_of_distance               (optional : dist2, ident) \
!         -o=output_netcdf_file             (optional) \
!         -p=periodicity                    (optional) \
!         -l=selected_time_period           (optional) \
!         -cv=climatology_netcdf_variable   (optional) \
!         -mi=missing_value                 (optional) \
!         -double                           (optional) \
!         -bigfile                          (optional) \
!         -hdf5                             (optional) \
!         -compact                          (optional) \
!         -tlimited                         (optional)
!
! By default
! ----------
!
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input NetCDF variable is assumed to be regular or Gaussian
!        *-r=*
!         if the input *netcdf_variable* is from the NEMO or ORCA model (e.g. if **-g=** argument
!         is not set to ``n``) the resolution is assumed to be ``r2``
!        *-b=*
!         if **-g=** is not set to ``n``\ , the dimensions of the 2-D grid-mesh, *nlon_orca* and
!         *nlat_orca*, are determined from the **-r=** argument. However, you may override
!         this choice by default with the **-b=** argument
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-a=*
!         the *type_of_transformation* is set to ``scp``. This means that the time series
!         in the 2-D grid-mesh associated with the input *netcdf_variable* are written as
!         raw data without any centering or standardization
!        *-d=*
!         the *type_of_distance* is set to ``dist2``.
!        *-o=*
!         the *output_netcdf_file* is named ``norm_``\ *netcdf_variable*\ ``.nc``
!        *-p=*
!         the *periodicity* is equal to the periodicity of the climatology if **-a=**\ ``cov``
!         or  **-a=**\ ``cor``\   or to ``time2``\  - ``time1``\  + ``1``\   if **-a=**\ ``scp``
!        *-l=*
!         the whole time period as specified by the **-t=**\ ``time1,time2`` argument
!        *-cv=*
!         this argument have the same value as the **-v=** argument
!        *-mi=*
!         the *missing_value* in the output NetCDF file is set to ``1.e+20``
!        *-double*
!         the data are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         data are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-compact*
!         the output NetCDF file is not compacted
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable which must be transformed
!      and the **-f=**\ *input_netcdf_file* argument specifies that this NetCDF variable must
!      be extracted from the NetCDF file, *input_netcdf_file*.
!
!   #) The optional argument **-m=**\ *input_mesh_mask_netcdf_file* specifies the land-sea mask
!      to apply to *netcdf_variable* for transforming this tridimensional NetCDF variable. By default,
!      it is assumed that each cell in the 2-D grid-mesh associated with the input tridimensional
!      NetCDF variable is a valid time series which must be written in the output NetCDF file.
!
!      The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*) and the mask
!      (in the *input_mesh_mask_netcdf_file*) must agree if an *input_mesh_mask_netcdf_file* is used.
!
!      Refer to :ref:`comp_clim_3d` or :ref:`comp_mask_3d` for creating a valid
!      *input_mesh_mask_netcdf_file* NetCDF file for regular or gaussian grids before
!      using comp_norm_3d.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2*  arguments are missing the whole
!      geographical domain associated with the *netcdf_variable* is used.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``.
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices before
!      using comp_norm_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, data in the whole time period
!      associated with the *netcdf_variable* is taken into account.
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the NetCDF variable
!      is from an experiment with the ORCA model. In this case, the duplicate points from the ORCA grid
!      are removed before the transformation, as far as possible, and, in particular,
!      if the whole globe is used as the geographical domain. On output, the duplicate
!      points are restored when writing the output file, if and only if, the whole globe is used
!      as the geographical domain.
!      If **-g=** is set to ``n``\ , it is assumed that the grid has no duplicate points.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` (i.e. if the NetCDF variable is from
!      an experiment with the ORCA model), the **-r=** argument gives the resolution used. If:
!
!      * **-r=**\ ``r2`` the NetCDF variable is from an experiment with the ORCA R2 model
!
!      * **-r=**\ ``r4`` the NetCDF variable is from an experiment with the ORCA R4 model.
!
!   #) If the NetCDF variable is from an experiment with the ORCA model,
!      but the resolution is not R2 or R4, the dimensions of the ORCA grid
!      must be specified explicitly with the **-b=** argument.
!
!   #) The **-a=** argument specifies if the data are centered or standardized
!      with an input climatology (specified with the **-c=** argument):
!
!      * **-a=**\ ``scp`` means that the raw data are output
!
!      * **-a=**\ ``cov`` means that the anomalies are output
!
!      * **-a=**\ ``cor`` means that the standardized anomalies are output
!
!      * **-a=**\ ``spa`` means that the standardized anomalies are output, but the anomalies are
!        standardized by the standard-deviation averaged over the specified domain
!        for each selected time step
!
!      * **-a=**\ ``tim`` means that the standardized anomalies are output, but the anomalies are
!        standardized by the standard-deviation averaged over the selected time steps
!        for each grid-point.
!
!   #) The *input_climatology_netcdf_file* specified with the **-c=** argument is needed only
!      if **-a=**\ ``cov``, **-a=**\ ``cor``, **-a=**\ ``spa`` or **-a=**\ ``tim``.
!
!   #) If **-a=**\ ``cov``, **-a=**\ ``cor``, **-a=**\ ``spa`` or **-a=**\ ``tim``, the selected time
!      period must agree with the climatology. This means that the first selected time observation (*time1* 
!      if the **-t=** argument is present) must correspond to the first day, month, season
!      of the climatology specified with the **-c=** argument.
!
!   #) The geographical shapes of the *netcdf_variable* (*input_netcdf_file*), the mask
!      (*input_mesh_mask_netcdf_file*), the scale factors (*input_mesh_mask_netcdf_file*),
!      and the climatology (*input_climatology_netcdf_file*) must agree.
!
!   #) The **-d=**\ *type_of_distance* argument is used only if **-a=**\ ``spa`` is specified. If:
!
!      * **-d=**\ ``dist2``, the anomalies are standardized by a weighted
!        standard-deviation. The sum of squares associated with a
!        grid-point is weighted accordingly to the surface associated with that
!        grid-point when computing the standard-deviation over the domain
!
!      * **-d=**\ ``ident`` means that the anomalies are standardized by a simple arithmetic
!        standard-deviation.
!
!   #) The **-l=** argument selects the indices of the time steps which must be included in
!      the output NetCDF file. The indices of the time steps are counted from the start of the (selected)
!      time period (e.g. *time1* in the **-t=**\ *time1,time2* argument or ``1`` if this argument is missing).
!      The argument list can be specified in two forms:
!
!      * **-l=**\ *n1,n2,...nn* allows to standardize and select for *n1*, *n2*, ... and *nn* time steps.
!
!        If *periodicity* is defined (with **-p=** option or if **-a=** is set to ``cov``, ``cor`` or ``spa``),
!        *n1*, *n2*, ... *nn* time steps are selected for each period separately (see second example below)
!
!      * **-l=**\ *n1:n2* allows to standardize and select time steps from *n1* to *n2* (or from *n1* to *n2*
!        for each period separately, if *periodicity* is defined with **-p=** option or if **-a=** is set to ``cov``,
!        ``cor`` or ``spa``).
!
!      The two forms of the **-l=** argument may be combined and repeated any number of times.
!      Duplicate time steps are not allowed.
!
!      Be careful with time period limits when specifying the **-l=** argument.
!
!   #) If the **-p=** argument is specified and **-a=**\ ``cov``, **-a=**\ ``cor``, **-a=**\ ``spa`` or **-a=**\ ``tim``,
!      the periodicity deduced from the climatology (given by the **-c=** argument) overrides the **-p=** argument.
!
!   #) If the variable used to compute the climatology has not the same name has the variable
!      specified by the **-v=** argument, use the **-cv=** argument to specify the variable name for
!      the climatology.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the *netcdf_variable* (specified by the **-v=** argument) in the *output_netcdf_file*.
!      *missing_value* must be a real number outside of the range of the *netcdf_variable*.
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``.
!
!   #) The **-double** argument specifies that the output NetCDF variable
!      must be stored as double-precision floating point numbers instead of
!      single-precision floating point numbers in the *output_netcdf_file*.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 CPP or _USE_NETCDF4 macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) If the **-compact** argument is specified and if a domain is selected (with the **-x=** and **-y=**
!      arguments) then only data for the selected domain will be output. By default, the whole
!      grid is stored (with missing values outside the selected domain).
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations.
!
!   #) It is assumed that the data has no missing values. If it is the case, use :ref:`comp_norm_miss_3d`
!      instead of comp_norm_3d.
!
! Outputs
! -------
!
!   comp_norm_3d creates an output NetCDF file that contains the coordinate NetCDF variables
!   of the input NetCDF dataset *input_netcdf_file* and the transformed NetCDF variable.
!   This NetCDF variable will have the same dimensions and name as the input NetCDF variable
!   in the file *input_netcdf_file* (in the description below, nlat and nlon are the length
!   of the spatial dimensions of the input NetCDF variable) :
!
!   #) *netcdf_variable*\ ``(ntime,nlat,nlon)`` : the transformed NetCDF variable as specified by the
!      **-m=**, **-a=** and **-mi=** arguments.
!
!   By default, the whole grid associated with the input NetCDF variable is stored (with missing values outside
!   the selected domain). Note, however, that if the argument **-compact** is used the geographical dimensions of
!   the output NetCDF variable will be reduced to the selected domain as specified by the **-x=** and **-y=** arguments
!   (e.g. in this case ``nlat``\ =\ *lat2*\ -\ *lat1*\ +\ ``1`` and ``nlon``\ =\ *lon2*\ -\ *lon1*\ +\ ``1``  ). The number
!   of time steps written in the output NetCDF file ( e.g. ``ntime``\ ) is determined from the **-t=**, **-l=** and **-p=**
!   arguments.
!
! Examples
! --------
!
!   #)   For computing time series of (monthly) anomalies
!        from a NetCDF variable ``sosstsst`` stored in a
!        file ``ST7_1m_00101_20012_grid_T_sosstsst.nc``,
!        apply a specific mask to the resulting time series and, finally, store the results
!        in the NetCDF file ``anoma_ST7_1m_00101_20012_grid_T_sosstsst.nc``, use the following 
!        command (note that the output file is compacted):
!
!        ::
!
!          $ comp_norm_3d \
!            -f=ST7_1m_00101_20012_grid_T_sosstsst.nc \
!            -v=sosstsst \
!            -g=t \
!            -m=meshmask.indopacific.nc \
!            -a=cov \
!            -c=clim_ST7_1m_00101_20012_grid_T_sosstsst.nc \
!            -o=anoma_ST7_1m_00101_20012_grid_T_sosstsst.nc \
!            -compact
!
!   #)   For selecting the first 120 days of each year (with a 365 days calendar) from
!        the daily NetCDF file ``ST7_1d_00101_20012_grid_T_sosstsst.nc``, which includes a NetCDF
!        variable ``sosstsst``, and store the results in the NetCDF file
!        ``select_ST7_1d_00101_20012_grid_T_sosstsst.nc``, use the following
!        command :
!
!        ::
!
!          $ comp_norm_3d \
!            -f=ST7_1d_00101_20012_grid_T_sosstsst.nc \
!            -v=sosstsst \
!            -m=meshmask.orca2.nc \
!            -g=t \
!            -p=365 \
!            -l=1:120 \
!            -o=select_ST7_1d_00101_20012_grid_T_sosstsst.nc
!
!-
! ================================================================================================
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,           only : nf90_max_var_dims, nf90_max_name, nf90_real4, nf90_real8,                     &
                                 nf90_open, nf90_nowrite, nf90_noerr, nf90_sizehint_default,                   &
                                 nf90_strerror, nf90_inquire_dimension,                                        &
                                 nf90_get_att, nf90_get_var, nf90_create, nf90_clobber,                        &
                                 nf90_set_fill, nf90_nofill, nf90_unlimited, nf90_def_dim, nf90_def_var,       &
                                 nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_global, nf90_enddef,      &
                                 nf90_put_var, nf90_close, nf90_share, nf90_inq_dimid
!
    use Statpack, only         : i1b, i4b, stnd, extd, lgl, true, false, zero, one, kfix, kexp,      & 
                                 kint, tolower, case_change, string_to_val,                          &
                                 allocate_error, merror, arth
!
    use NCSTAT_Utilities, only : prtunit, nchar, nlon_def2, nlat_def2, nlon_def4, nlat_def4,           &
                                 sp, dp, def_missing, get_listarg, get_int_list, get_variable_dims,    &
                                 get_variable_dimcoords, check_lon_domain, select_grid_points_2d,      &
                                 duplicate_orca_points_2d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_norm_3d : ',                     &
!
                                   cmethod(5)=(/"scp","cov","cor","spa","tim"/),    &          
!
                                   cdistance(2)=(/"dist2","ident"/),                &   
!
                                   cresolution(2)=(/"r2","r4"/)
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_clim, ncid_mask, ncid_norm,               &
                     varid, maskid, meanid, stdid, nobsid, e1id, e2id, normid,          &
                     i, i1, i2, j, j1, j2, k, l, nlon, nlat, nt, nvar,                  &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3,           &
                     lon1b, lon2b, lat1b, lat2b,                                        &
                     ioffset, method, distance, nlon_def, nlat_def,                     &
                     ntime_list, period, nperiod, output_type, output_mode, var_natts
    integer(i4b), dimension(3)                 :: startlec, startlec2, countlec, coor_id2
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i4b), dimension(:),    pointer     :: time_list
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:,:),   allocatable  :: temp, mean, std
    real(stnd), dimension(:),     allocatable  :: dat
    real(stnd)                                 :: missing, tmp, offset, scalfac
    real(extd), dimension(:),     allocatable  :: record, record2
!
    logical(lgl)                               :: use_tlimited, scale_factor, add_offset, create_hdf5,     &
                                                  use_x, use_y, use_compact, use_mask, failure, transform
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
    logical(lgl), dimension(:,:), allocatable  :: mask
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, clim_name, mask_name,    &
                                                  norm_name, att_name, clim_var_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                          :: fmt
    character                                  :: chr, grid_type
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Select and normalize time series from a tridimensional variable'
        write (prtunit,*) ' extracted from a NetCDF dataset'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_norm_3d  -f=input_netcdf_file'
        write (prtunit,*) '               -v=netcdf_variable'
        write (prtunit,*) '               -m=input_mesh_mask_netcdf_file   (optional)'
        write (prtunit,*) '               -g=grid_type                     (optional : n, t, u, v, w, f)'
        write (prtunit,*) '               -r=resolution                    (optional : r2, r4)'
        write (prtunit,*) '               -b=nlon_orca, nlat_orca          (optional)'
        write (prtunit,*) '               -x=lon1,lon2                     (optional)'
        write (prtunit,*) '               -y=lat1,lat2                     (optional)'
        write (prtunit,*) '               -t=time1,time2                   (optional)'
        write (prtunit,*) '               -c=input_climatology_netcdf_file (optional)'
        write (prtunit,*) '               -a=type_of_transformation        (optional : scp, cov, cor, spa, tim)'
        write (prtunit,*) '               -d=type_of_distance              (optional : dist2, ident)'
        write (prtunit,*) '               -o=output_netcdf_file            (optional)'
        write (prtunit,*) '               -p=periodicity                   (optional)'
        write (prtunit,*) '               -l=selected_time_period          (optional)'
        write (prtunit,*) '               -cv=climatology_netcdf_variable  (optional)'
        write (prtunit,*) '               -mi=missing_value                (optional)'
        write (prtunit,*) '               -double                          (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '               -bigfile                         (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '               -hdf5                            (optional)'
#endif
        write (prtunit,*) '               -compact                         (optional)'
        write (prtunit,*) '               -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-g=n'
        write (prtunit,*) '-r=r2'
        write (prtunit,*) '-b= by default, nlon_orca and nlat_orca are determined from the -r='
        write (prtunit,*) '    argument'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-a=scp'
        write (prtunit,*) '-d=dist2'
        write (prtunit,*) '-o=norm_netcdf_variable.nc'
        write (prtunit,*) '-p=(climatology periodicity) or time2-time1+1 if -a=scp'
        write (prtunit,*) '-l= the whole time period associated with the netcdf_variable'
        write (prtunit,*) '-cv= by default, the same value as the -v= argument'
        write (prtunit,*) '-mi= by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-double   : by default, the data are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : by default, a netCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-compact  : by default, the output netCDF is not compacted'
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>22 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name       = " "
    var_name       = " "
    mask_name      = " "
    clim_name      = " "
    clim_var_name  = " "
    norm_name      = " "
!
    history = "comp_norm_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
    output_type = nf90_real4
    nlat_def    = nlat_def2
    nlon_def    = nlon_def2
    method      = 1
    distance    = 1
    period      = 0
    nperiod     = 0
    ntime_list  = 0
!
    missing     = def_missing
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x        = false
    use_y        = false
    use_compact  = false
    use_mask     = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-r=')
!
!               DETERMINE THE RESOLUTION FOR THE ORCA GRID.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cresolution(1) )
                        nlat_def  = nlat_def2
                        nlon_def  = nlon_def2
                    case ( cresolution(2) )
                        nlat_def  = nlat_def4
                        nlon_def  = nlon_def4
                    case default
                        call merror( name_proc//"Unknown resolution option" )
                end select
!
            case ('-b=')
!
!               GET THE DIMENSIONS OF THE ORCA GRID.
!
                call get_listarg( name_proc, name(4:),                           &
                                  "nlon_orca and nlat_orca are integer values",  &
                                  "Wrong ORCA grid definition ",                 &
                                  nlon_def, nlat_def )
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
                use_mask  = true
!
            case ('-c=')
!
!               GET THE NAME OF THE NETCDF CLIMATOLOGY DATASET.
!
                clim_name = name(4:)
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF TRANSFORMATION.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case ( cmethod(3) )
                        method = 3
                    case ( cmethod(4) )
                        method = 4
                    case ( cmethod(5) )
                        method = 5
                    case default
                        call merror( name_proc//"Unknown transformation option" )
                end select
!
            case ('-d=')
!
!               DETERMINE THE WEIGHTS FOR COMPUTING THE STD.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cdistance(1) )
                        distance = 1
                    case ( cdistance(2) )
                        distance = 2
                    case default
                        call merror( name_proc//"Unknown distance option" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE (STANDARDIZED) VARIABLE.
!
                norm_name = name(4:)
!
            case ('-p=')
!
!               GET THE PERIODICITY.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The periodicity is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
            case ('-l=')
!
                if ( ntime_list/=0 ) deallocate( time_list )
!
                call get_int_list(  name_proc, name, 4, ntime_list, time_list )
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-cv=')
!
!                       GET THE NAME OF THE NETCDF VARIABLE FOR THE CLIMATOLOGY.
!
                        clim_var_name = name(5:)
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case ('-double')
!
!                               STORE DATA AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
                            case ('-compact')
!
!                               COMPACT THE OUTPUT FILE.
!
                                use_compact = true
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( distance==1 .and. method==4 .and. .not.use_mask )               &
    call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( clim_name==" " .and. method>=2 )                                &
    call merror( name_proc//"Climatology NETCDF file is missing" )
!
    if ( norm_name==" " ) norm_name="norm_"//trim(var_name)//".nc"
!
    if ( clim_var_name==" " )   clim_var_name = var_name
!
!   OPEN THE NETCDF DATASETS.
!
    if ( ntime_list==0 ) then
        i = nf90_nowrite
    else
        i = nf90_share
    end if
!
    k = nf90_sizehint_default
!
    status = nf90_open( trim(fic_name), i, ncid, chunksize=k )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_mask ) then
        status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( method>=2 ) then
        status = nf90_open( trim(clim_name), nf90_nowrite, ncid_clim )
        if ( status/=nf90_noerr )                                        &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES,
!   AND THE NUMBER OF ATTRIBUTES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                    &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names, &
                                  dimlens_out=dimlens, coor_id_out=coor_id,           &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type, &
                                  use_coor_variable_out=use_coor_variable,            &
                                  natts=var_natts             )
!
    if ( count(dimlens/=0)/=3 )           &
    call merror( name_proc//"Input NETCDF variable is not tridimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nt     = dimlens(3)
!
    if ( use_mask ) then
!
        chr = merge( 't', grid_type, grid_type=='w' )
!
!       GET THE MASK ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
        call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
        call merror( name_proc//"Shape disagreement with the mask" )
!
        if ( .not.use_x ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
            if ( status==nf90_noerr ) lon1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
            if ( status==nf90_noerr ) lon2 = i
!
        end if
!
        if ( .not.use_y ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
            if ( status==nf90_noerr ) lat1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
            if ( status==nf90_noerr ) lat2 = i
!
        end if
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( grid_type/='n' ) then
!
        if ( nlon>nlon_def )               &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (longitude)" )
!
        if ( nlat>nlat_def )               &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (latitude)" )
!
!       A TRICK TO DISABLE THE ELIMINATION AND DUPLICATION IN THE HIGH LATITUDES OF THE ORCA GRID
!       IF THE RESOLUTION IS NOT R2 OR R4.
!
        if ( ( nlat_def/=nlat_def2 .or. nlon_def/=nlon_def2 ) .and. ( nlat_def/=nlat_def4 .or. nlon_def/=nlon_def4 ) )    &
        nlat_def = nlat_def + 1
!
    end if
!
!   CHECK THE GEOGRAPHICAL DOMAIN.
!   REMOVE DUPLICATE POINTS FROM THE ORCA_R2 GRID.
!
    call check_lon_domain( grid_type, nlon, nlon_def, lon1, lon2, lon3 )
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE TIME DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    if ( method==1 ) then
!
        if ( period==0 ) period = time3
!
    else
!
!       GET THE PERIOD ID FROM ITS NAME.
!       FIND THE DIMENSION LENGTHS FOR THE PERIOD VARIABLE.
!
        call get_variable_dims( name_proc, "period", ncid_clim, i, dimids, dimlens )
!
        if ( count(dimlens/=0)/=1 )     &
        call merror( name_proc//" Wrong climatology file specification " )
!
        period = dimlens(1)
!
    end if
!
    if ( period>time3 )  call merror( name_proc//"-t=, -p= (or -c=) options mismatch" )
!
    tmp     = real( time3, stnd)/real( period, stnd)
    nperiod = ceiling( tmp )
!
!   CHECK THE TIME COORDINATES LIST.
!
    if ( ntime_list==0 ) then
!
        ntime_list = period
!
!       ALLOCATE VARIABLE FOR THE TIME COORDINATES LIST IF NOT DONE BEFORE.
!
        allocate( time_list(ntime_list), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        time_list(1:ntime_list) = arth( 1_i4b, 1_i4b, ntime_list )
!
    else
!
        if ( any( time_list(1:ntime_list)>period ) )          &
        call merror( name_proc//"-p= (or -c=), -t= and -l= options mismatch" )
!
    end if
!
!   FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "scale_factor", scalfac )
    if ( status/=nf90_noerr ) then
        scale_factor = false
        scalfac      = one
    else
        scale_factor = true
    end if
!
!   FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "add_offset", offset )
    if ( status/=nf90_noerr ) then
        add_offset = false
        offset     = zero
    else
        add_offset = true
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    if ( use_mask ) then
!
!       GET THE MASK.
!
        status = nf90_get_var( ncid_mask, maskid, temp )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DETERMINE THE GRID POINTS USED IN THE NORM ANALYSIS FROM THE MASK.
!
        mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
    else
!
        mask(:nlon,:nlat) = true
!
    end if
!
    call select_grid_points_2d( grid_type, lon1, lon2, lat1, lat2,  &
                                nlon_def, nlat_def, mask, j1, j2    )
!
    if ( lon1<0 ) lon3 = nlon
!   
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   ALLOCATE WORK ARRAY.
!
    allocate( dat(max(nvar,ntime_list)), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   READ CLIMATOLOGY AND STANDARD-DEVIATIONS IF REQUIRED.
!
    startlec(1:2)  = (/ j1, lat1 /)
    countlec(1:3)  = (/ lon3, lat3, 1_i4b /)
!
    if ( method>1 ) then
!
!       GET THE MEAN FIELD ID FROM ITS NAME.
!       FIND THE DIMENSION LENGTHS FOR THE NETCDF MEAN VARIABLE.
!
        call get_variable_dims( name_proc, trim(clim_var_name)//"_mean", ncid_clim,    &
                                meanid, dimids, dimlens )
!
        if ( count(dimlens/=0)/=3 )     &
        call merror( name_proc//"Input climatology is not tridimensional (mean field)" )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=period )    &
        call merror( name_proc//"Shape disagreement with the climatology (mean field)" )
!
        failure = false
!
        status = nf90_get_att( ncid_clim, nf90_global, "lon1_"//trim(clim_var_name), i )
        if ( status==nf90_noerr ) failure = j1<i
!
        status = nf90_get_att( ncid_clim, nf90_global, "lon2_"//trim(clim_var_name), i )
        if ( status==nf90_noerr .and. .not.failure ) failure = j2>i
!
        status = nf90_get_att( ncid_clim, nf90_global, "lat1_"//trim(clim_var_name), i )
        if ( status==nf90_noerr .and. .not.failure ) failure = lat1<i
!
        status = nf90_get_att( ncid_clim, nf90_global, "lat2_"//trim(clim_var_name), i )
        if ( status==nf90_noerr .and. .not.failure ) failure = lat2>i
!
        if ( failure ) call merror( name_proc//"Missing values in the climatology " )
!
!       ALLOCATE WORK ARRAY FOR THE MEANS.
!
        allocate( mean(nvar,ntime_list), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE MEANS.
!
        do j = 1, ntime_list
!
            startlec(3) = time_list(j)
!
            status = nf90_get_var( ncid_clim, meanid, temp(:lon3,:lat3),    &   
                                   start=startlec,                          &
                                   count=countlec                           )     
            if ( status/=nf90_noerr )                                            &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           PACK THE MEAN FIELD .
!
            mean(:nvar,j) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        end do
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE MEAN VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid_clim, meanid, "scale_factor", tmp )
        if ( status==nf90_noerr ) mean(:nvar,:ntime_list) = mean(:nvar,:ntime_list)*tmp
!
!       FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE MEAN VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid_clim, meanid, "add_offset", tmp )
        if ( status==nf90_noerr ) mean(:nvar,:ntime_list) = mean(:nvar,:ntime_list) + tmp
!
        if ( method>2 ) then
!
!           GET THE STD ID FROM ITS NAME.
!           FIND THE DIMENSION LENGTHS FOR THE NETCDF STD VARIABLE.
!
            call get_variable_dims( name_proc, trim(clim_var_name)//"_std", ncid_clim,    &
                                    stdid, dimids, dimlens )
!
            if ( count(dimlens/=0)/=3 )          &
            call merror( name_proc//"Input climatology is not tridimensional (std field)" )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=period )    &
            call merror( name_proc//"Shape disagreement with the climatology (std field)" )
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( std(nvar,ntime_list), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           GET THE STDS.
!
            do j = 1, ntime_list
!
                startlec(3) = time_list(j)
!
                status = nf90_get_var( ncid_clim, stdid, temp(:lon3,:lat3),     & 
                                       start=startlec,                          &
                                       count=countlec                           )     
                if ( status/=nf90_noerr )                                          &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               PACK THE STD FIELD.
!
                std(:nvar,j) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
            end do
!
!           FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE STD VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid_clim, stdid, "scale_factor", tmp )
            if ( status==nf90_noerr ) std(:nvar,:ntime_list) = std(:nvar,:ntime_list)*tmp
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE STD VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid_clim, stdid, "add_offset", tmp )
            if ( status==nf90_noerr ) std(:nvar,:ntime_list) = std(:nvar,:ntime_list) + tmp
!
            select case( method )
!
                case( 3 )
!
!                   STANDARDIZED VALUES ARE REQUESTED.
!
                    where ( std(:nvar,:ntime_list)/=zero )                 &
                    std(:nvar,:ntime_list) = one/std(:nvar,:ntime_list)
!
                case( 4 ) 
!
!                   STANDARDIZED VALUES BY THE MEAN OF STANDARD-DEVIATIONS 
!                   AVERAGED OVER THE DOMAIN ARE REQUESTED.
!
                    if ( distance==1 ) then
!
                       if ( .not.use_mask )         &
                       call merror( name_proc//"Input mesh-mask NETCDF file is missing" )

!
!                       A WEIGHTED MEAN OF STANDARD-DEVIATIONS OVER THE DOMAIN IS REQUESTED.
!
!                       THE DISTANCE (WEIGHTS) IS ASSOCIATED WITH THE (HORIZONTAL) GRID.
!
!                       GET THE SCALE FACTORS IDs FROM THEIR NAMES.
!                       VERIFY THE DIMENSION LENGTHS FOR THE SCALE FACTOR VARIABLES.
!
                        call get_variable_dims( name_proc, "e1"//chr, ncid_mask,    &
                                                e1id, dimids, dimlens )
!
                        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
                        call merror( name_proc//"Shape disagreement with the first scale factor" )
!
                        call get_variable_dims( name_proc, "e2"//chr, ncid_mask,    &
                                                e2id, dimids, dimlens )
!
                        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
                        call merror( name_proc//"Shape disagreement with the second scale factor" )
!
!                       GET THE SCALE FACTORS.
!
                        status = nf90_get_var( ncid_mask, e1id, temp(:lon3,:lat3),       &  
                                               start=startlec(1:2), count=countlec(1:2)  )     
                        if ( status/=nf90_noerr )                                                &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
                        dat(:nvar) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
                        status = nf90_get_var( ncid_mask, e2id, temp(:lon3,:lat3),        &
                                                start=startlec(1:2), count=countlec(1:2)  )     
                        if ( status/=nf90_noerr )                                                     &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
                        dat(:nvar) = dat(:nvar)*pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
                        tmp = sum( dat(:nvar) )
!
                        std(:nvar,:ntime_list) =                                                   &
                        (std(:nvar,:ntime_list)**2)*spread(dat(:nvar),dim=2,ncopies=ntime_list ) 
!
                    else
!
                        tmp = real( nvar, stnd )
!
                        std(:nvar,:ntime_list) =  std(:nvar,:ntime_list)**2
!
                    end if
!
                    dat(:ntime_list) = sqrt( sum( std(:nvar,:ntime_list) , dim=1 )/tmp )
                    where ( dat(:ntime_list)/=zero )  dat(:ntime_list ) = one/dat(:ntime_list)
                    std(:nvar,:ntime_list) = spread( dat(:ntime_list), dim=1, ncopies=nvar )
!
                case( 5 )
!
!                   STANDARDIZED VALUES BY THE MEAN OF STANDARD-DEVIATIONS 
!                   AVERAGED OVER THE TIME STEPS ARE REQUESTED.
!
!                   GET THE NOBS FIELD ID FROM ITS NAME.
!                   FIND THE DIMENSION LENGTH FOR THE NETCDF NOBS VARIABLE.
!
                    call get_variable_dims( name_proc, trim(clim_var_name)//"_nobs", ncid_clim,    &
                                            nobsid, dimids, dimlens )
!
                    if ( count(dimlens/=0)/=1 )     &
                    call merror( name_proc//"NOBS field in climatology file is not a vector" )
!
                    if ( dimlens(1)/=period )                                                 &
                    call merror( name_proc//"Shape disagreement with the climatology (nobs field)" )
!
                   do j = 1, ntime_list
!
                       status = nf90_get_var( ncid_clim, nobsid, dat(j), start=(/ time_list(j) /)  )   
                      if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                   end do
!
                   std(:nvar,:ntime_list) = (std(:nvar,:ntime_list)**2)*spread( dat(:ntime_list),dim=1,ncopies=nvar )
                   tmp = one/sum( dat(:ntime_list) )
                   dat(:nvar) = sqrt( sum( std(:nvar,:ntime_list), dim=2 )*tmp )
                   where ( dat(:nvar)/=zero )  dat(:nvar) = one/dat(:nvar)
                   std(:nvar,:ntime_list) = spread( dat(:nvar), dim=2, ncopies=ntime_list )
!
            end select
!
        end if
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE SELECTED AND TRANSFORMED DATA.
!
    if ( create_hdf5 ) then
!
        status = nf90_create( trim(norm_name), cmode=output_mode, ncid=ncid_norm )
!
    else
!
        k = nf90_sizehint_default
        status = nf90_create( trim(norm_name), cmode=output_mode, ncid=ncid_norm, chunksize=k )
!
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_norm, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    if ( use_compact ) then
        startlec2(1:2) = (/ j1, lat1 /)
        dimlens(1:2)   = (/ lon3, lat3 /)
    else
        startlec2(1:2) = 1
        dimlens(1:2)   = (/ nlon, nlat /)
    end if
!
    if ( use_tlimited ) then
        dimlens(3)   = nperiod*ntime_list
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    do i = 1, 3
!
        status = nf90_def_dim( ncid_norm, name=trim(dim_names(i)),                         &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 3
        status = nf90_def_var( ncid_norm, trim(dim_names(i)), coor_type(i),               &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end do
!
    status = nf90_def_var( ncid_norm, trim(var_name), output_type,                        &
                           dimids=dimids(1:3), varid=normid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_norm, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    if ( .not.use_coor_variable(3) ) then
!
        status = nf90_put_att( ncid_norm, coor_id2(3), "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_norm, coor_id2(3), "short_name", "Time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   COPY ATTRIBUTES FOR THE TRANSFORMED VARIABLE.
!
    do i = 1, var_natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( att_name=="add_offset"                .or.  &
             att_name=="scale_factor"              .or.  &
             att_name=="missing_value"             .or.  &
             att_name=="_FillValue"                .or.  &
             ( method>2  .and.   att_name=="units" )    )  cycle
!
!       COPY USEFUL ATTRIBUTES.
!
        status = nf90_copy_att(  ncid, varid, att_name, ncid_norm, normid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_norm, normid, "missing_value", real( missing, dp )  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_norm, normid, "_FillValue", real( missing, dp )  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        status = nf90_put_att( ncid_norm, normid, "missing_value", real( missing, sp )  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_norm, normid, "_FillValue", real( missing, sp )  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_norm, nf90_global, "title", trim(var_name)//" NORM analysis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_norm, nf90_global, "distance", cdistance(distance) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_norm, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_norm, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_norm, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_norm )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    dimlens(3) = time3
!
!   ALLOCATE WORK ARRAYS.
!
    j = maxval( dimlens(1:3) )
    allocate( record(j), record2(time3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 2    
!
        i2 = dimlens(j)
!
        if ( use_coor_variable(j) ) then
!
            i1 = startlec2(j)
!
            status = nf90_get_var( ncid, coor_id(j), record(:i2), start= (/ i1 /), count=(/ i2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_norm, coor_id2(j), record(:i2), count=(/ i2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            status = nf90_put_var( ncid_norm, coor_id2(j), arth(1_i4b,1_i4b,i2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end do
!
    if ( use_coor_variable(3) ) then
!
!       COPY TIME VARIABLE FROM THE INPUT NETCDF DATASET.
!
        status = nf90_get_var( ncid, coor_id(3), record(:time3), start=(/ time1 /), count=(/ time3 /) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    else
!
!       CREATE TIME VARIABLE FOR THE NEW NETCDF DATASET.
!
        record(:time3) = arth( time1, 1_i4b, time3 )
!
    end if
!
!   WRITE TRANSFORMED AND SELECTED VALUES.
!
    ioffset = time1 - 1_i4b
!
    if ( use_compact ) then
!
        lon1b = j1
        lon2b = j2
        lat1b = lat1
        lat2b = lat2
!
    else
!
        lon1b = 1
        lon2b = nlon
        lat1b = 1
        lat2b = nlat
!
    end if
!
!    lon1b = startlec2(1)
!    lon2b = dimlens(1) - lon1b + 1
!    lat1b = startlec2(2)
!    lat2b = dimlens(2) - lat1b + 1
!
    startlec2(1:2) = 1
    dimlens(3)     = 1
!
    temp(:nlon,:nlat) = missing
    transform = scale_factor .or. add_offset .or. ( method>=2 .and. method<=5 ) .or. use_mask
!
    do i = 0, nperiod-1
!
        l = i*ntime_list
!
        do k = 1, ntime_list
!
            i1 = time_list(k)
!
            if ( i1>time3 ) exit
!
            i2           = k + l
            record2(i2)  = record(i1)
            i1           = i1 + ioffset
            startlec(3)  = i1
            startlec2(3) = i2
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE.
!
            status = nf90_get_var( ncid, varid, temp(j1:j2,lat1:lat2), start=startlec, count=countlec )     
            if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            if ( transform ) then
!
!               PACK THE DATA.
!
                dat(:nvar) = pack( temp(j1:j2,lat1:lat2), mask(j1:j2,lat1:lat2) )
!
!               MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                if ( scale_factor ) dat(:nvar) = dat(:nvar)*scalfac
!
!               ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
                if ( add_offset ) dat(:nvar) = dat(:nvar) + offset
!
!               SELECT THE METHOD OF TRANSFORMATION.
!
                select case ( method )
!
                    case ( 2 )
!
                        dat(:nvar) = dat(:nvar) - mean(:nvar,k)
!
                    case ( 3:5 )
!
                        dat(:nvar) = (dat(:nvar) - mean(:nvar,k))*std(:nvar,k)
!
                end select
!
!               WRITE THE RECORD IN THE OUTPUT FILE.
!
                temp(j1:j2,lat1:lat2) = unpack( dat(:nvar), mask(j1:j2,lat1:lat2), missing )
!
            end if
!
            if ( grid_type/='n' .and. nlon==nlon_def ) then
!
                if ( lat2>=nlat_def-1 .and. nlat==nlat_def ) then
!
!                   ADD DUPLICATE POINTS TO THE ORCA GRID IF THE WHOLE GLOBE IS USED.
!
                    call duplicate_orca_points_2d( grid_type, mask, temp )
!
                else
!
!                   ADD DUPLICATE LONGITUDES LINES TO THE ORCA GRID IF THE WHOLE LONGITUDE
!                   DOMAIN IS USED.
!
                    temp(1,:nlat)    = temp(nlon-1,:nlat)
                    temp(nlon,:nlat) = temp(2,:nlat)
!
                end if
!
            end if
!
            status = nf90_put_var( ncid_norm, normid, temp(lon1b:lon2b,lat1b:lat2b),   &
                                   start=startlec2, count=dimlens )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        time_list(:ntime_list) = time_list(:ntime_list) + period
!
    end do
!
!   CREATE TIME VARIABLE FOR THE NEW NETCDF DATASET.
!
    status = nf90_put_var( ncid_norm, coor_id2(3), record2(:i2), count=(/ i2 /) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS.
!
    select case ( method )
!
        case ( 1 )
!
            deallocate( temp, mask, dat, time_list, record, record2 )
!
        case ( 2 )
!
            deallocate( temp, mask, dat, time_list, record, record2, mean )
!
        case ( 3:5 )
!
            deallocate( temp, mask, dat, time_list, record, record2, mean, std )
!
    end select
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_mask ) then
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( method >= 2 ) then
        status = nf90_close( ncid_clim )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_norm )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_norm_3d
! ===========================
!
end program comp_norm_3d
