program comp_section_4d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_section_4d:
!
! ===============
! comp_section_4d
! ===============
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 05/01/2018
!
! Purpose
! =======
!
! Compute a longitude-level-time, latitude-level-time or longitude-latitude-time section
! (see the description of the **-s=** argument below) from a fourdimensional variable extracted from
! a NetCDF dataset.
!
! Different options are available for computing the cross-section from raw data, anomalies after
! the removal of a climatology or standardized anomalies (see the description of the **-a=** argument
! below). Different options are also available for averaging the grid-point time series in the selected
! domain (see the description of the **-d=** argument below).
!
! The computed section is stored in an output NetCDF dataset.
!
! If the NetCDF variable is tridimensional use :ref:`comp_section_3d` instead 
! of comp_section_4d.
!
! This procedure is parallelized if OpenMP is used and the NCSTAT software has been built
! with the _PARALLEL_READ CPP flag. Moreover, this procedure computes the section with only 
! one pass through the data and an out-of-core strategy which is highly efficient on huge datasets.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_section_4d \
!         -f=input_netcdf_file \
!         -v=input_netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -o=output_netcdf_file \
!         -n=output_netcdf_variable \
!         -s=section                           (optional : lon, lat,dep) \
!         -g=grid_type                         (optional : n, t, u, v, w, f) \
!         -r=resolution                        (optional : r2, r4) \
!         -b=nlon_orca, nlat_orca, nlevel_orca (optional) \
!         -x=lon1,lon2                         (optional) \
!         -y=lat1,lat2                         (optional) \
!         -z=level1,level2                     (optional) \
!         -t=time1,time2                       (optional) \
!         -c=input_climatology_netcdf_file     (optional) \
!         -a=type_of_analysis                  (optional : scp, cov, cor) \
!         -d=type_of_distance                  (optional : dist2, dist3, ident) \
!         -sm=smoothing_factor                 (optional) \
!         -mi=missing_value                    (optional) \
!         -double                              (optional) \
!         -bigfile                             (optional) \
!         -hdf5                                (optional) \
!         -tlimited                            (optional)
!
! By default
! ----------
!
!        *-s=*
!         by default a latitude-time section is computed, which is equivalent to use ``lon``
!         for the  **-s=** argument
!        *-g=*
!         the *grid_type* is set to ``n``, which means that the 3-D grid-mesh associated with
!         the input NetCDF variable is assumed to be regular or Gaussian
!        *-r=*
!         if the input netcdf_variable is from the NEMO or ORCA model (e.g. if **-g=** argument
!         is not set to ``n``) the resolution is assumed to be ``r2``
!        *-b=*
!         if **-g=** is not set to ``n``\ , the dimensions of the 3-D grid-mesh, *nlon_orca*, 
!         *nlat_orca* and *nlevel_orca* are determined from the **-r=** argument. However, you
!         may override this choice by default with the **-b=** argument
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-z=*
!         the whole vertical resolution associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-a=*
!         the *type_of_analysis* is set to ``scp``. This means that the section
!         is computed from the raw data
!        *-d=*
!         the *type_of_distance* is set to ``dist3``. This means that the section is computing as a weighted average
!         and that the weight associated with each grid-point time series in the selected domain is proportional to
!         the volume or the weight associated with it
!        *-sm=*
!         no time smoothing is applied to the section
!        *-mi=*
!         by default, the *missing_value* in the output NetCDF variable is set to ``1.e+20``
!        *-double*
!         the section is stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         section is stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable from which the section must
!      be computed and the **-f=**\ *input_netcdf_file* argument specifies that this NetCDF variable must
!      be extracted from the NetCDF file, *input_netcdf_file*.
!
!   #) If the **-x=**\ *lon1,lon2*, **-y=**\ *lat1,lat2* and **-z=**\ *level1,level2* arguments
!      are missing the whole geographical domain associated with the *netcdf_variable* is
!      used for computing the section.
!      The longitude, latitude or level range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_4d` for transforming geographical coordinates as indices before
!      using comp_section_4d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used for computing the section.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) It is assumed that the data has no missing values excepted those associated with a constant
!      land-sea mask.
!
!   #) The **-s=** argument determines how the cross-section is computed. If:
!
!      * **-s=**\ ``lon``, a latitude-level-time section is computed
!
!      * **-s=**\ ``lat``, a longitude-level-time section is computed
!
!      * **-s=**\ ``dep``, a longitude-latitude-time section is computed.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f``, it is assumed that the NetCDF variable
!      is from an experiment with the NEMO or ORCA model.
!      In this case, the duplicate points from the ORCA grid are removed 
!      before computing the section, as far as possible, and, in particular,
!      if the 3-D grid-mesh of the input NetCDF variable covers the whole globe.
!
!      If **-g=** is set to ``n``, it is assumed that the 3-D grid-mesh is regular or Gaussian
!      and as such has no duplicate points.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` (e.g. if the NetCDF variable
!      is from an experiment with the NEMO or ORCA model), the **-r=** argument gives the 
!      resolution used. If:
!
!      * **-r=**\ ``r2``, the NetCDF variable is from an experiment with the ORCA R2 model
!
!      * **-r=**\ ``r4``, the NetCDF variable is from an experiment with the ORCA R4 model.
!
!   #) If the NetCDF variable is from an experiment with the ORCA model,
!      but the resolution is not R2 or R4, the dimensions of the ORCA grid
!      must be specified explicitly with the **-b=** argument.
!
!   #) The **-a=** argument specifies if the grid-point time series are centered or standardized
!      with an input climatology (specified with the **-c=** argument) before computing the cross-section
!      for the selected domain. If:
!
!      * **-a=**\ ``scp``, the section is computed from the raw data
!
!      * **-a=**\ ``cov``, the section is computed from the anomalies
!
!      * **-a=**\ ``cor``, the section is computed from the standardized anomalies.
!
!   #) The *input_climatology_netcdf_file* specified with the **-c=** argument is needed
!      only if **-a=**\ ``cov`` or **-a=**\ ``cor``.
!
!   #) If **-a=**\ ``cov`` or **-a=**\ ``cor``, the selected time period must agree with the climatology.
!      This means that the first selected time observation (*time1* if the **-t=** argument
!      is present) must correspond to the first day, month, season of the climatology
!      specified with the **-c=** argument.
!
!   #) The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*),
!      the mask (*input_mesh_mask_netcdf_file*), the scale factors (in the *input_mesh_mask_netcdf_file*),
!      and the climatology (in the *input_climatology_netcdf_file*) must agree.
!
!   #) The **-d=** argument specifies the weighting method for computing the cross-section. If:
!
!      * **-d=**\ ``dist3``, the section is computed with the diagonal distance associated with the
!        whole 3-D grid-mesh (each grid point is weighted accordingly to the surface associated with it)
!
!      * **-d=**\ ``dist2``, the section is computed with the diagonal distance associated with the
!        horizontal 2-D grid-mesh (each grid point is weighted accordingly to the volume or weight  associated
!        with it)
!
!      * **-d=**\ ``ident``, the section is computed with the identity metric (each grid point has
!        the same weight when computing the cross-section).
!
!   #) **-sm=**\ *smoothing_factor* means that the section must be smoothed in time with a moving average of
!      approximately ``2``\ *\ *smoothing_factor*\ +\ ``1`` terms. *smoothing_factor* must be an integer greater
!      than ``0``.
!
!   #) The **-n=**\ *output_netcdf_variable* argument specifies the NetCDF variable which will contains 
!      the computed section in the output NetCDF file, *output_netcdf_file*, specified by the
!      **-o=**\  argument.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the *output_netcdf_variable* in the *output_netcdf_file*.
!      If the **-mi=** argument is not specified, *missing_value* is set to ``1.e+20`` in the output NetCDF
!      dataset.
!
!   #) The **-double** argument specify that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations.
!
! Outputs
! -------
!
!   comp_section_4d creates an output NetCDF file that contains the computed cross-section.
!   The output NetCDF dataset contains the following NetCDF variable (in the description below,
!   lat3, lon3 and lev3 are, respectively, equal to ``lat2 - lat1 + 1``, ``lon2 - lon1 + 1`` and
!   ``lev2 - lev1 + 1`` ) :
!
!   If **-s=**\ ``lon`` is specified:
!
!   #) *output_netcdf_variable*\ ``(ntime,lev3,lat3)`` : the computed latitude-level-time section.
!
!   If **-s=**\ ``lat`` is specified:
!
!   #) *output_netcdf_variable*\ ``(ntime,lev3,lon3)`` : the computed longitude-level-time section.
!
!   If **-s=**\ ``dep`` is specified:
!
!   #) *output_netcdf_variable*\ ``(ntime,lat3,lon3)`` : the computed longitude-latitude-time section.
!
! Examples
! --------
!
!   #)   For computing a monthly longitude-level-time section from the
!        file ``ST7_1m_0101_20012_grid_T_votemper.nc``, which includes a NetCDF
!        variable ``votemper``, and store the results in a NetCDF variable named
!        ``temp`` in the file ``section_temp.orca2.nc``, use the following command
!        (note that the cross-section is computed from monthly anomalies) :
!
!        ::
!
!          $ comp_section_4d \
!            -f=ST7_1m_0101_20012_grid_T_votemper.nc \
!            -v=votemper \
!            -m=meshmask.orca2.nc \
!            -o=section_temp.orca2.nc \
!            -n=temp \
!            -s=lat \
!            -g=t \
!            -c=clim_votemper_grid_T.nc \
!            -a=cov \
!            -d=dist3
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,    only : nf90_share
    use Statpack,  only : omp_limit
    use omp_lib,   only : omp_get_num_procs, omp_get_max_threads
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,           only : nf90_max_var_dims, nf90_max_name, nf90_real4, nf90_real8, nf90_int,          &
                                 nf90_sizehint_default, nf90_nowrite, nf90_open, nf90_noerr, nf90_strerror,   &
                                 nf90_inq_varid, nf90_inquire_variable, nf90_inquire_dimension, nf90_get_att, &
                                 nf90_get_var, nf90_create, nf90_clobber, nf90_set_fill, nf90_nofill,         &
                                 nf90_unlimited, nf90_def_dim, nf90_def_var, nf90_inq_attname, nf90_copy_att, &
                                 nf90_put_att, nf90_global, nf90_enddef, nf90_put_var, nf90_close
!
    use Statpack,         only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, kfix, kexp,      &
                                 kint, tolower, case_change, string_to_val,                          &
                                 allocate_error, merror, arth, comp_smooth
!
    use NCSTAT_Utilities, only : prtunit, nchar, nlon_def2, nlat_def2, ndepth_def2, nlon_def4, nlat_def4,  &
                                 ndepth_def4, sp, dp, def_missing, get_listarg, get_variable_dims,         &
                                 check_lon_domain, select_grid_points_3d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_section_4d : ',            &
!
                                   cmethod(3)=(/"scp","cov","cor"/),          &          
!
                                   cdistance(3)=(/"dist2","dist3","ident"/),  &       
!
                                   cresolution(2)=(/"r2","r4"/),                &        
!
                                   csection(3)=(/"lon","lat","dep"/)    
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_clim, ncid_mask, ncid_out,                        &
                     varid, maskid, meanid, stdid, e1id, e2id, e3id, sectionid,                 &
                     i, j, j1, j2, k, nlon, nlat, nlevel, nt, nm,                               &
                     lon1, lon2, lon3, lat1, lat2, lat3,                                        &
                     time1, time2, time3, level1, level2, level3, dim_section,                  &
                     method, distance, rec, offset_rec, natts, nlon_def, nlat_def, nlevel_def,  &
                     smooth_factor, output_type, output_mode
    integer(i4b), dimension(3)                 :: coor_natts, coor_type, coor_id, coor_id2
    integer(i4b), dimension(4)                 :: startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:,:,:,:), allocatable  :: mean, std
    real(stnd), dimension(:,:,:),   allocatable  :: temp, section, scal
    real(stnd), dimension(:,:),     allocatable  :: scal_section, temp2
    real(stnd), dimension(:),       allocatable  :: temp3
    real(stnd)                                   :: tmp, missing
    real(extd), dimension(:),       allocatable  :: record
!
    logical(lgl)                                 :: present_missing, use_tlimited, create_hdf5,    &
                                                    use_x, use_y, use_z, failure
    logical(lgl), dimension(3)                   :: use_coor_variable
    logical(lgl), dimension(:,:,:), allocatable  :: mask
    logical(lgl), dimension(:,:),   allocatable  :: mask_section
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, clim_name, mask_name,    &
                                                  out_name, section_name, att_name 
    character(len=nf90_max_name), dimension(4) :: dim_names
    character(len=14)                          :: fmt
    character                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
    integer(i4b) :: ncid2, varid2
!
    logical  :: test_par
!
#endif
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute cross-sections from a fourdimensional variable '
        write (prtunit,*) ' extracted from a NetCDF dataset.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_section_4d -f=input_netcdf_file'
        write (prtunit,*) '                 -v=input_netcdf_variable'
        write (prtunit,*) '                 -m=input_mesh_mask_netcdf_file'
        write (prtunit,*) '                 -o=output_netcdf_file'
        write (prtunit,*) '                 -n=output_netcdf_variable'
        write (prtunit,*) '                 -s=section                           (optional : lon,lat,dep)'
        write (prtunit,*) '                 -g=grid_type                         (optional : n, t, u, v, w, f)'
        write (prtunit,*) '                 -r=resolution                        (optional : r2, r4)'
        write (prtunit,*) '                 -b=nlon_orca, nlat_orca, nlevel_orca (optional)'
        write (prtunit,*) '                 -x=lon1,lon2                         (optional)'
        write (prtunit,*) '                 -y=lat1,lat2                         (optional)'
        write (prtunit,*) '                 -z=level1,level2                     (optional)'
        write (prtunit,*) '                 -t=time1,time2                       (optional)'
        write (prtunit,*) '                 -c=input_climatology_netcdf_file     (optional)'
        write (prtunit,*) '                 -a=type_of_analysis                  (optional : scp, cov, cor)'
        write (prtunit,*) '                 -d=type_of_distance                  (optional : dist2, dist3, ident)'
        write (prtunit,*) '                 -sm=smoothing_factor                 (optional)'
        write (prtunit,*) '                 -mi=missing_value                    (optional)'
        write (prtunit,*) '                 -double                              (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                 -bigfile                             (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                 -hdf5                                (optional)'
#endif
        write (prtunit,*) '                 -tlimited                            (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-s=lon , This means that a latitude-level-time section is computed'
        write (prtunit,*) '-g=n'
        write (prtunit,*) '-r=r2'
        write (prtunit,*) '-b= nlon_orca, nlat_orca and nlevel_orca are determined from the -r=argument'
        write (prtunit,*) '-x= the whole longitude domain associated with the input_netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the input_netcdf_variable'
        write (prtunit,*) '-z= the whole vertical resolution associated with the input_netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the input_netcdf_variable'
        write (prtunit,*) '-a=scp'
        write (prtunit,*) '-d=dist3'
        write (prtunit,*) '-sm=  no smoothing is applied to the time section'
        write (prtunit,*) '-mi= the missing_value is equal to 1.e+20'
        write (prtunit,*) '-double   : the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : a netCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : the time dimension is defined as unlimited'
        write (prtunit,*)
#ifdef _PARALLEL_READ
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<5 .or. narg>22 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name     = " "
    var_name     = " "
    mask_name    = " "
    clim_name    = " "
    out_name     = " "
    section_name = " "
!
    history = "comp_section_4d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
    nlat_def      = nlat_def2
    nlon_def      = nlon_def2
    nlevel_def    = ndepth_def2
    dim_section   = 1
    method        = 1
    distance      = 2
    smooth_factor = 0
!
    missing       = def_missing
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    level1    = 1
    level2    = huge( level2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x        = false
    use_y        = false
    use_z        = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-s=')
!
!               DETERMINE THE SECTION.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( csection(1) )
                        dim_section = 1
                    case ( csection(2) )
                        dim_section = 2
                    case ( csection(3) )
                        dim_section = 3
                    case default
                        call merror( name_proc//"Unknown section option" )
                end select
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-r=')
!
!               DETERMINE THE RESOLUTION FOR THE ORCA GRID.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cresolution(1) )
                        nlat_def   = nlat_def2
                        nlon_def   = nlon_def2
                        nlevel_def = ndepth_def2
                    case ( cresolution(2) )
                        nlat_def   = nlat_def4
                        nlon_def   = nlon_def4
                        nlevel_def = ndepth_def4
                    case default
                        call merror( name_proc//"Unknown resolution option" )
                end select
!
            case ('-b=')
!
!               GET THE DIMENSIONS OF THE ORCA GRID.
!
                call get_listarg( name_proc, name(4:),                                        &
                                  "nlon_orca, nlat_orca and nlevel_orca are integer values",  &
                                  "Wrong ORCA grid definition ",                              &
                                  nlon_def, nlat_def, nlevel_def )
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-z=')
!
!               GET THE LEVEL DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "level1 and level2 are integer values", &
                                  "Wrong level definition",               &
                                  level1, level2 )
!
                use_z = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
            case ('-c=')
!
!               GET THE NAME OF THE NETCDF CLIMATOLOGY DATASET.
!
                clim_name = name(4:)
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case ( cmethod(3) )
                        method = 3
                    case default
                        call merror( name_proc//"Unknown analysis option" )
                end select
!
            case ('-d=')
!
!               DETERMINE THE DISTANCE FOR THE ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cdistance(1) )
                        distance = 1
                    case ( cdistance(2) )
                        distance = 2
                    case ( cdistance(3) )
                        distance = 3
                    case default
                        call merror( name_proc//"Unknown distance option" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE SECTION.
!
               out_name = name(4:)
!
            case ('-n=')
!
!               GET THE NAME OF THE NEW SECTION.
!
                section_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-sm=')
!
!                       DETERMINE THE SMOOTHING FACTOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) smooth_factor
                        end if
                        if ( smooth_factor<=0 )        &
         call merror( name_proc//"The -sm= argument is a strictly positive integer value " )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case ('-double')
!
!                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " ) call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " ) call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( mask_name==" " ) call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( clim_name==" " .and. method>=2 )                               &
    call merror( name_proc//"Climatology NETCDF file is missing" )
!
    if ( out_name==" " ) call merror( name_proc//"Output NETCDF file is missing" )
!
    if ( section_name==" " ) call merror( name_proc//"Output NETCDF variable is missing" )
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>=2 ) then
!
        k = nf90_sizehint_default
        status = nf90_open( trim(clim_name), nf90_nowrite, ncid_clim, chunksize=k )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!
    status = nf90_inq_varid( ncid, trim(var_name), varid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE EXTERNAL TYPE, THE DIMENSION LENGTHS AND NAMES AND THE NUMBER OF ATTRIBUTS
!   FOR THE NETCDF VARIABLE.
!
    status = nf90_inquire_variable( ncid, varid, ndims=i, dimids=dimids, natts=natts )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( i/=4 ) call merror( name_proc//"Input NETCDF variable is not fourdimensional" )
!
    do i = 1, 4
!
        status = nf90_inquire_dimension( ncid, dimids(i), name=dim_names(i), len=dimlens(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nlevel = dimlens(3)
    nt     = dimlens(4)
!
    j = 0
!
    do i = 1, 4
!
        if ( i==dim_section ) cycle
!
        j = j + 1
!
!       GET THE COORDINATE VARIABLES IDS FROM THEIR NAME, IF ANY.
!
        status = nf90_inq_varid( ncid, trim(dim_names(i)), coor_id(j) )
        if ( status==nf90_noerr ) then
!
!           FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!           FOR THE COORDINATE VARIABLES.
!
            status = nf90_inquire_variable( ncid, coor_id(j), xtype=coor_type(j), ndims=k,    &
                                            dimids=dimlens, natts=coor_natts(j) )
            if ( status/=nf90_noerr )                                                         &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            use_coor_variable(j) = k==1 .and. dimlens(1)==dimids(i)
        else
            use_coor_variable(j) = false
        end if
!
        if ( .not.use_coor_variable(j) ) coor_type(j) = nf90_int
!
    end do
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
    call merror( name_proc//"Shape disagreement with the mask" )
!
    if ( .not.use_x ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
        if ( status==nf90_noerr ) lon1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
        if ( status==nf90_noerr ) lon2 = i
!
    end if
!
    if ( .not.use_y ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
        if ( status==nf90_noerr ) lat1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
        if ( status==nf90_noerr ) lat2 = i
!
    end if
!
    if ( .not.use_z ) then
!
        status = nf90_get_att( ncid_mask, maskid, "level1_First_level", i )
        if ( status==nf90_noerr ) level1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "level2_Last_level", i )
        if ( status==nf90_noerr ) level2 = i
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!   CHECK THE GEOGRAPHICAL DOMAIN.
!   REMOVE DUPLICATE POINTS FROM THE ORCA GRID.
!
    if ( grid_type/='n' ) then
!
        if ( nlon>nlon_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (longitude)" )
!
        if ( nlat>nlat_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (latitude)" )
!
        if ( nlevel>nlevel_def )    &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (level)" )
!
!       A TRICK TO DISABLE THE ELIMINATION AND DUPLICATION IN THE HIGH LATITUDES OF THE ORCA GRID
!       IF THE RESOLUTION IS NOT R2 OR R4.
!
        if ( ( nlat_def/=nlat_def2 .or. nlon_def/=nlon_def2 ) .and. ( nlat_def/=nlat_def4 .or. nlon_def/=nlon_def4 ) )    &
        nlat_def = nlat_def + 1
!
    end if
!
    call check_lon_domain( grid_type, nlon, nlon_def, lon1, lon2, lon3 )
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE LEVEL DEFINITION.
!
    level1 = max( level1, 1 )
    level2 = min( level2, nlevel )
    level3 = level2 - level1 + 1
    if ( level3<=0  ) call merror( name_proc//"Wrong level definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat,nlevel), mask(nlon,nlat,nlevel), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, temp )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED IN THE ANALYSIS FROM THE MASK.
!
    mask(:nlon,:nlat,:nlevel) = temp(:nlon,:nlat,:nlevel)==one
!
!   SELECT THE DOMAIN.
!
    j1 = lon1
    j2 = lon2
!
    call select_grid_points_3d( grid_type, j1, j2, lat1, lat2, level1, level2, &
                                nlon_def, nlat_def, mask, lon1, lon2  )
!
    if ( j1<0 ) lon3 = nlon
!
!   ALLOCATE WORK ARRAYS.
!
    select case ( dim_section )
        case ( 1 )
            j1 = lat3
            j2 = level3
            j  = lon3
        case ( 2 )
            j1 = lon3
            j2 = level3
            j  = lat3
        case ( 3 )
            j1 = lon3
            j2 = lat3
            j  = level3
    end select
!
    allocate( scal(lon3,lat3,level3), mask_section(j1,j2), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    mask_section(:j1,:j2) = any( mask(lon1:lon2,lat1:lat2,level1:level2), dim=dim_section )
    i = count( mask_section(:j1,:j2) )
    if ( i<=0  )  call merror( name_proc//"Wrong section definition : section is empty" )
    present_missing = i/=j1*j2
!
!   COMPUTE THE DISTANCE.
!
    if ( distance<=2 ) then
!
!       THE DISTANCE IS ASSOCIATED WITH THE (HORIZONTAL) GRID.
!
!
!       ALLOCATE WORK ARRAY.
!
        allocate( temp2(lon3,lat3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE FIRST AND SECOND SCALE FACTORS IDs FROM THEIR NAMES.
!       VERIFY THE DIMENSION LENGTHS FOR THE SCALE FACTOR VARIABLES.
!
        call get_variable_dims( name_proc, "e1"//chr, ncid_mask,    &
                                e1id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the first scale factor" )
!
        call get_variable_dims( name_proc, "e2"//chr, ncid_mask,    &
                                e2id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the second scale factor" )
!
!       GET THE FIRST AND SECOND SCALE FACTORS.
!
        status = nf90_get_var( ncid_mask, e1id, temp2(:lon3,:lat3),      &
                               start=(/lon1,lat1/), count=(/lon3,lat3/)  ) 
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        temp(:lon3,:lat3,1) = temp2(:lon3,:lat3)
!
        status = nf90_get_var( ncid_mask, e2id, temp2(:lon3,:lat3),      &
                               start=(/lon1,lat1/), count=(/lon3,lat3/)  ) 
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       COMPUTE THE DIAGONAL METRIC ASSOCIATED WITH THE HORIZONTAL GRID.
!
        temp2(:lon3,:lat3) = temp2(:lon3,:lat3)*temp(:lon3,:lat3,1)
        scal(:lon3,:lat3,:level3) = spread( temp2, dim=3, ncopies=level3 )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp2 )
!
        if ( distance==2 ) then
!
!           ALLOCATE WORK ARRAY.
!
            allocate( temp3(level3), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           THE DISTANCE IS ASSOCIATED WITH THE FULL 3D GRID.
!
            chr = merge( grid_type, 't', grid_type=='w' .or. grid_type=='n' )
!
!           GET THE THIRD SCALE FACTOR ID FROM ITS NAME.
!           VERIFY THE DIMENSION LENGTHS FOR THE THIRD SCALE FACTOR VARIABLE.
!
            call get_variable_dims( name_proc, "e3"//chr, ncid_mask,                    &
                                    e3id, dimids, dimlens )
!
            if ( dimlens(1)/=1 .or. dimlens(2)/=1 .or. dimlens(3)/=nlevel )             &
            call merror( name_proc//"Shape disagreement with the third scale factor" )
!
!           GET THE THIRD SCALE FACTOR VARIABLE.
!
            status = nf90_get_var( ncid_mask, e3id, temp3(:level3),                     &
                               start=(/1_i4b,1_i4b,level1/), count=(/1_i4b,1_i4b,level3/) )    
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           CHECK THE THIRD SCALE FACTOR.
!
            if ( temp3(1)<zero )                        &
    call merror( name_proc//" First level with -d=dist3 are not allowed for mesh-mask "//trim(mask_name) )
!
            if ( temp3(level3)<zero )                   &
    call merror( name_proc//" Last level with -d=dist3 are not allowed for mesh-mask "//trim(mask_name) )
!
            temp(:lon3,:lat3,:level3) =                   &
            spread( spread( temp3, dim=1, ncopies=lat3), dim=1, ncopies=lon3 )
!
!           COMPUTE THE DIAGONAL METRIC ASSOCIATED WITH THE 3D GRID.
!
            scal(:lon3,:lat3,:level3) = scal(:lon3,:lat3,:level3)*temp(:lon3,:lat3,:level3)
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( temp3 )
!
        end if
!
!       APPLY THE MASK.
!
        where ( .not. mask(lon1:lon2,lat1:lat2,level1:level2) ) scal(:lon3,:lat3,:level3) = zero
!
    else
!
!       THE DISTANCE IS THE IDENTITY.
!
        scal(:lon3,:lat3,:level3) = merge( one, zero, mask(lon1:lon2,lat1:lat2,level1:level2) )
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( section(j1,j2,time3), scal_section(j1,j2), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   COMPUTE THE DIAGONAL METRIC.
!
    scal_section(:j1,:j2) = sum( scal(:lon3,:lat3,:level3), dim=dim_section )
    where ( mask_section(:j1,:j2) ) scal_section(:j1,:j2) = one/scal_section(:j1,:j2)
!
!BUG: due to a bug in the pgf90/95 compiler (version 7.1.1), the following line of code can't be used
!     and has been replaced by the following section of code performing the same task.
!
!    scal(:lon3,:lat3,:level3) =      &
!    scal(:lon3,:lat3,:level3)*spread( scal_section(:j1,:j2), dim=dim_section, ncopies=j )
!
    select case ( dim_section )
!
        case ( 1 )
!
            scal(:lon3,:lat3,:level3) =      &
            scal(:lon3,:lat3,:level3)*spread( scal_section(:lat3,:level3), dim=1, ncopies=lon3 )
!
        case ( 2 )
!
            scal(:lon3,:lat3,:level3) =      &
            scal(:lon3,:lat3,:level3)*spread( scal_section(:lon3,:level3), dim=2, ncopies=lat3 )
!
        case ( 3 )
!
            scal(:lon3,:lat3,:level3) =      &
            scal(:lon3,:lat3,:level3)*spread( scal_section(:lon3,:lat3), dim=3, ncopies=level3 )
!
    end select
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec = time1 - 1
!
    startlec(1:3) = (/ lon1, lat1, level1 /)
    countlec(1:4) = (/ lon3, lat3, level3, 1_i4b /)
!
#ifdef _PARALLEL_READ
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    test_par = i>1 .and. j>1 .and. time3*lon3*lat3*level3>omp_limit
!
    if ( test_par ) then
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
    end if
!
#endif
!
!   SELECT THE METHOD FOR COMPUTING THE SECTION.
!
    select case ( method )
!
        case ( 1 )
!
!           COMPUTE THE SECTION ON THE RAW DATA.
!
#ifdef _PARALLEL_READ
!
            if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(k,status,startlec,ncid2,varid2,temp)                  &
!$OMP         ,SHARED(lon1,lon3,lat1,lat3,level1,level3,time3,offset_rec,    &
!$OMP                 j1,j2,varid,countlec,dim_section,scal,section)
!
!               ALLOCATE WORK ARRAY.
!
                allocate( temp(lon3,lat3,level3), stat=status )
                if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!               OPEN THE NETCDF FILE.
!
                status = nf90_open( trim(fic_name), nf90_share, ncid2  )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               GET THE VARIABLE ID FROM ITS NAME.
!
                status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!               READ THE DATA.
!
                startlec(1)  = lon1
                startlec(2)  = lat1
                startlec(3)  = level1
!
!$OMP DO SCHEDULE(STATIC)
                do k = 1, time3
!
                    startlec(4) = k + offset_rec
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                    status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3,:level3),                   &
                                           start=startlec, count=countlec )     
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   ADJUST FOR THE SCALE FACTORS.
!
                    temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3)*scal(:lon3,:lat3,:level3)
!
!                   COMPUTE THE SECTION.
!
                    section(:j1,:j2,k) = sum( temp(:lon3,:lat3,:level3), dim=dim_section )
!
                end do
!$OMP END DO
!
!               DEALLOCATE WORK ARRAYS.
!
                deallocate( temp )
!
!$OMP CRITICAL (closenc)
!
!               CLOSE THE NETCDF DATASET.
!
                status = nf90_close( ncid2 )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
            else
!
#endif
!
                do k = 1, time3
!
                    startlec(4) = k + offset_rec
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                    status = nf90_get_var( ncid, varid, temp(:lon3,:lat3,:level3),                    &
                                           start=startlec, count=countlec )     
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   ADJUST FOR THE SCALE FACTORS.
!
                    temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3)*scal(:lon3,:lat3,:level3)
!
!                   COMPUTE THE SECTION.
!
                    section(:j1,:j2,k) = sum( temp(:lon3,:lat3,:level3), dim=dim_section )
!
                end do
!
#ifdef _PARALLEL_READ
!
            end if
!
#endif
!
        case ( 2, 3 )
!
!           COMPUTE THE SECTION ON THE (STANDARDIZED) ANOMALIES.
!
!
!           GET THE MEAN FIELD ID FROM ITS NAME.
!           FIND THE DIMENSION LENGTHS FOR THE NETCDF MEAN VARIABLE.
!
            call get_variable_dims( name_proc, trim(var_name)//"_mean", ncid_clim,    &
                                    meanid, dimids, dimlens )
!
            if ( count(dimlens/=0)/=4 )     &
            call merror( name_proc//"Input climatology is not fourdimensional" )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
            call merror( name_proc//"Shape disagreement with the climatology (mean field)" )
!
            nm = dimlens(4)
!
            failure = false
!
            status = nf90_get_att( ncid_clim, nf90_global, "lon1_"//trim(var_name), i )
            if ( status==nf90_noerr ) failure = lon1<i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lon2_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = lon2>i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lat1_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = lat1<i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lat2_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = lat2>i
!
            status = nf90_get_att( ncid_clim, nf90_global, "level1_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = level1<i
!
            status = nf90_get_att( ncid_clim, nf90_global, "level2_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = level2>i
!
            if ( failure ) call merror( name_proc//"Missing values in the climatology " )
!
!           ALLOCATE WORK ARRAY FOR THE MEANS.
!
            allocate( mean(lon3,lat3,level3,nm), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           GET THE MEANS.
!
            status = nf90_get_var( ncid_clim, meanid, mean,                                 &
                     start=(/ lon1, lat1, level1, 1 /), count=(/ lon3, lat3, level3, nm /)  )   
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           ADJUST FOR THE MASK.
!
            do i = 1, nm
                where ( .not. mask(lon1:lon2,lat1:lat2,level1:level2) )     &
                mean(:lon3,:lat3,:level3,i) = zero
            end do
!
            if ( method==2 ) then
!
!               COMPUTE THE SECTION ON THE ANOMALIES.
!
#ifdef _PARALLEL_READ
!
                if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,j,k,status,startlec,ncid2,varid2,temp)                 &
!$OMP         ,SHARED(lon1,lon3,lat1,lat3,level1,level3,time3,nm,offset_rec,    &
!$OMP                 j1,j2,varid,countlec,dim_section,mean,scal,section)
!
!                   ALLOCATE WORK ARRAY.
!
                    allocate( temp(lon3,lat3,level3), stat=status )
                    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!                   OPEN THE NETCDF FILE.
!
                    status = nf90_open( trim(fic_name), nf90_share, ncid2  )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   GET THE VARIABLE ID FROM ITS NAME.
!
                    status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!                   READ THE DATA.
!
                    startlec(1)  = lon1
                    startlec(2)  = lat1
                    startlec(3)  = level1
!
!$OMP DO SCHEDULE(STATIC)
                    do i = 0, time3-1, nm
!
                        do j = 1, nm
!
                            k   = i + j
!
                            if ( k>time3 ) exit
!
                            startlec(4) = k + offset_rec
!
!                           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                            status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3,:level3),                 &
                                                   start=startlec, count=countlec )     
                            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                           COMPUTE ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                            temp(:lon3,:lat3,:level3) =         &
                            (temp(:lon3,:lat3,:level3)-mean(:lon3,:lat3,:level3,j) )*scal(:lon3,:lat3,:level3)
!
!                           COMPUTE THE SECTION.
!
                            section(:j1,:j2,k) = sum( temp(:lon3,:lat3,:level3), dim=dim_section )
!
                        end do
!
                    end do
!$OMP END DO
!
!                   DEALLOCATE WORK ARRAYS.
!
                    deallocate( temp )
!
!$OMP CRITICAL (closenc)
!
!                   CLOSE THE NETCDF DATASET.
!
                    status = nf90_close( ncid2 )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
                else
!
#endif
!
!                   GET THE DATA.
!
                    do i = 0, time3-1, nm
!
                        do j = 1, nm
!
                            k   = i + j
!
                            if ( k>time3 ) exit
!
                            startlec(4) = k + offset_rec
!
!                           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                            status = nf90_get_var( ncid, varid, temp(:lon3,:lat3,:level3),             &
                                                   start=startlec, count=countlec )     
                            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                           COMPUTE ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                            temp(:lon3,:lat3,:level3) =         &
                            (temp(:lon3,:lat3,:level3)-mean(:lon3,:lat3,:level3,j) )*scal(:lon3,:lat3,:level3)
!
!                           COMPUTE THE SECTION.
!
                            section(:j1,:j2,k) = sum( temp(:lon3,:lat3,:level3), dim=dim_section )
!
                        end do
!
                    end do
!
#ifdef _PARALLEL_READ
!
                end if
!
#endif
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( mean )
!
            else
!
!               COMPUTE THE SECTION ON THE STANDARDIZED ANOMALIES.
!
!               GET THE STD ID FROM ITS NAME.
!               FIND THE DIMENSION LENGTHS FOR THE NETCDF STD VARIABLE.
!
                call get_variable_dims( name_proc, trim(var_name)//"_std", ncid_clim,    &
                                        stdid, dimids, dimlens )
!
                if ( count(dimlens/=0)/=4 )          &
                call merror( name_proc//"Input climatology is not fourdimensional" )
!
                if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
                call merror( name_proc//"Shape disagreement with the climatology (std field)" )
!
                if ( dimlens(4)/=nm )     &
                call merror( name_proc//"Shape disagreement in the climatology" )
!
!               ALLOCATE WORK ARRAY.
!
                allocate( std(lon3,lat3,level3,nm), stat=status )
                if ( status/=0 ) call merror( name_proc//allocate_error )
!
!               GET THE STDS.
!
                status = nf90_get_var( ncid_clim, stdid, std,                                 &
                                       start=(/ lon1, lat1, level1, 1 /), count=(/ lon3, lat3, level3, nm /)  )   
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               ADJUST FOR THE SCALE FACTORS AND THE MASK.
!
                do i = 1, nm
!
                    where ( .not. mask(lon1:lon2,lat1:lat2,level1:level2) ) std(:lon3,:lat3,:level3,i) = zero
!
                    where ( std(:lon3,:lat3,:level3,i)/=zero )       &
                    std(:lon3,:lat3,:level3,i) = scal(:lon3,:lat3,:level3)/std(:lon3,:lat3,:level3,i)
!
                end do
!
!                where ( std(:lon3,:lat3,:level3,:nm)/=zero )       &
!                std(:lon3,:lat3,:level3,:nm) = one/std(:lon3,:lat3,:level3,:nm)
!                std(:lon3,:lat3,:level3,:nm) =                     &
!                std(:lon3,:lat3,:level3,:nm)*spread( scal(:lon3,:lat3,:level3), dim=4, ncopies=nm )
!
!               GET THE DATA.
!
#ifdef _PARALLEL_READ
!
                if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,j,k,status,startlec,ncid2,varid2,temp)                 &
!$OMP         ,SHARED(lon1,lon3,lat1,lat3,level1,level3,time3,nm,offset_rec,    &
!$OMP                 j1,j2,varid,countlec,dim_section,mean,std,section)
!
!                   ALLOCATE WORK ARRAY.
!
                    allocate( temp(lon3,lat3,level3), stat=status )
                    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!                   OPEN THE NETCDF FILE.
!
                    status = nf90_open( trim(fic_name), nf90_share, ncid2  )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   GET THE VARIABLE ID FROM ITS NAME.
!
                    status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!                   READ THE DATA.
!
                    startlec(1)  = lon1
                    startlec(2)  = lat1
                    startlec(3)  = level1
!
!$OMP DO SCHEDULE(STATIC)
                    do i = 0, time3-1, nm
!
                        do j = 1, nm
!
                            k   = i + j
!
                            if ( k>time3 ) exit
!
                            startlec(4) = k + offset_rec
!
!                           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                            status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3,:level3),                 &
                                                   start=startlec, count=countlec )     
                            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                           COMPUTE ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                            temp(:lon3,:lat3,:level3) =         &
                           ( temp(:lon3,:lat3,:level3) - mean(:lon3,:lat3,:level3,j) )*std(:lon3,:lat3,:level3,j)
!
!                           COMPUTE THE SECTION.
!
                            section(:j1,:j2,k) = sum( temp(:lon3,:lat3,:level3), dim=dim_section )
!
                        end do
!
                    end do
!$OMP END DO
!
!                   DEALLOCATE WORK ARRAYS.
!
                    deallocate( temp )
!
!$OMP CRITICAL (closenc)
!
!                   CLOSE THE NETCDF DATASET.
!
                    status = nf90_close( ncid2 )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
                else
!
#endif
                    do i = 0, time3-1, nm
!
                        do j = 1, nm
!
                            k   = i + j
!
                            if ( k>time3 ) exit
!
                            startlec(4) = k + offset_rec
!
!                           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                            status = nf90_get_var( ncid, varid, temp(:lon3,:lat3,:level3),                      &
                                                   start=startlec, count=countlec )     
                            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                           COMPUTE STANDARDIZED ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                            temp(:lon3,:lat3,:level3) =           &
                            ( temp(:lon3,:lat3,:level3) - mean(:lon3,:lat3,:level3,j) )*std(:lon3,:lat3,:level3,j)
!
!                           COMPUTE THE SECTION.
!
                            section(:j1,:j2,k) = sum( temp(:lon3,:lat3,:level3), dim=dim_section )
!
                        end do
!
                    end do
!
#ifdef _PARALLEL_READ
!
                end if
!
#endif
!
!               DEALLOCATE WORK ARRAYS.
!
                deallocate( mean, std )
!
            end if
!
    end select
!
    if ( method<=2 ) then
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid, varid, "scale_factor", tmp )
        if ( status==nf90_noerr ) section(:j1,:j2,:time3) = section(:j1,:j2,:time3)*tmp
!
        if ( method==1 ) then
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid, varid, "add_offset", tmp )
            if ( status==nf90_noerr )  section(:j1,:j2,:time3) = section(:j1,:j2,:time3) + tmp

        end if
!
    end if
!
!   SMOOTH THE TIME SECTION IF REQUIRED.
!
    if ( smooth_factor/=0 ) then
        call comp_smooth( section(:j1,:j2,:time3), smooth_factor )
    end if
!
    if ( present_missing ) then
!
        mask_section(:j1,:j2) = .not.mask_section(:j1,:j2)
!
        do rec = 1, time3
!
            where( mask_section(:j1,:j2) ) section(:j1,:j2,rec) = missing
!
        end do
!
    end if
!
!   CREATE THE NEW NETCDF DATASET.
!
    if ( create_hdf5 ) then
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out )
    else
        k = nf90_sizehint_default
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_out, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = lon3
    dimlens(2)   = lat3
    dimlens(3)   = level3
!
    if ( use_tlimited ) then
        dimlens(4)   = time3
    else
        dimlens(4)   = nf90_unlimited
    end if
!
    j = 0
!
    do i = 1, 4
!
        if ( i==dim_section ) cycle
!
        j = j + 1
!
        status = nf90_def_dim( ncid_out, trim(dim_names(i)), len=dimlens(i), dimid=dimids(j) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    j = 0
!
    do i = 1, 4
!
        if ( i==dim_section ) cycle
!
        j = j + 1
!
        status = nf90_def_var( ncid_out, trim(dim_names(i)), coor_type(j),                &
                               dimids=dimids(j), varid=coor_id2(j) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_out, section_name, output_type, dimids=dimids(:3), varid=sectionid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_out, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    if ( .not.use_coor_variable(3) ) then
!
        status = nf90_put_att( ncid_out, coor_id2(3), "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, coor_id2(3), "short_name", "Time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_put_att( ncid_out, sectionid, "long_name", trim(section_name)//" Section" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, sectionid, "short_name", trim(section_name)//" Section" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method<=2 ) then
        status = nf90_copy_att(  ncid, varid, 'units', ncid_out, sectionid )
    end if
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_out, sectionid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, sectionid, "_FillValue", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        status = nf90_put_att( ncid_out, sectionid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, sectionid, "_FillValue", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_out, nf90_global, "title", " Section from "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "distance", cdistance(distance) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "smoothing_factor", smooth_factor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_out )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
!
!   ALLOCATE WORK ARRAY.
!
    j = 0
!
    do i = 1, 4
!
        if ( i==dim_section ) cycle
!
        j = j + 1
!
        select case ( i )
!
            case ( 1 )
                j1 = lon1
                j2 = lon3
!
            case ( 2 )
                j1 = lat1
                j2 = lat3
!
            case ( 3 )
                j1 = level1
                j2 = level3
!
            case ( 4 )
                j1 = time1
                j2 = time3
!
        end select
!
        if ( use_coor_variable(j) ) then
!
            allocate( record(j2), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            status = nf90_get_var( ncid, coor_id(j), record, start=(/ j1 /), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_out, coor_id2(j), record )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            deallocate( record )
!
        else
            status = nf90_put_var( ncid_out, coor_id2(j), arth(j1,1_i4b,j2 ) )
!
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end do
!
    status = nf90_put_var( ncid_out, sectionid, section )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( allocated(temp) ) then
        deallocate( mask, mask_section, section, scal, scal_section, temp )
    else
        deallocate( mask, mask_section, section, scal, scal_section )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>=2 ) then
        status = nf90_close( ncid_clim )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_out )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM comp_section_4d
! ==============================
!
end program comp_section_4d
