program comp_serie_3d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_serie_3d:
!
! =============
! comp_serie_3d
! =============
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 05/01/2018
!
! Purpose
! =======
!
! Compute a time series from a tridimensional variable
! extracted from a NetCDF dataset.
!
! Different options are available for computing the time series from raw data, anomalies
! after the removal of a climatology or standardized anomalies (see the description of the
! **-a=** argument below). Different options are also available for averaging the different
! pointwise time series in the selected domain (see the description of the **-d=** argument below).
!
! The computed time series is stored in an output NetCDF dataset.
!
! If your data contains missing values, use :ref:`comp_serie_miss_3d` instead of comp_serie_3d
! to compute the time series from your gappy dataset.
!
! If the NetCDF variable is fourdimensional, use :ref:`comp_serie_4d` instead 
! of comp_serie_3d.
!
! This procedure is parallelized if OpenMP is used and the NCSTAT software has been built
! with the _PARALLEL_READ CPP macro. Moreover, this procedure computes the time series with only 
! one pass through the data and an out-of-core strategy which is highly efficient on huge datasets.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_serie_3d \
!         -f=input_netcdf_file \
!         -v=input_netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -o=output_netcdf_file \
!         -n=output_netcdf_variable \
!         -g=grid_type                      (optional : n, t, u, v, w, f) \
!         -r=resolution                     (optional : r2, r4) \
!         -b=nlon_orca,nlat_orca            (optional) \
!         -x=lon1,lon2                      (optional) \
!         -y=lat1,lat2                      (optional) \
!         -t=time1,time2                    (optional) \
!         -c=input_climatology_netcdf_file  (optional) \
!         -a=type_of_analysis               (optional : scp, cov, cor) \
!         -d=type_of_distance               (optional : dist2, ident) \
!         -sm=smoothing_factor              (optional) \
!         -3d                               (optional) \
!         -double                           (optional) \
!         -hdf5                             (optional) \
!         -tlimited                         (optional)
!
! By default
! ----------
!
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input NetCDF variable is assumed to be regular or Gaussian
!        *-r=*
!         if the input netcdf_variable is from the NEMO or ORCA model (eg if **-g=** argument
!         is not set to ``n``) the resolution is assumed to be ``r2``
!        *-b=*
!         if **-g=** is not set to ``n``\ , the dimensions of the 2-D grid-mesh, *nlon_orca*, 
!         and *nlat_orca* are determined from the **-r=** argument. However, you
!         may override this choice by default with the **-b=** argument
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-a=*
!         the *type_of_analysis* is set to ``scp``. This means that the averaged time series
!         is computed from the raw data
!        *-d=*
!         the *type_of_distance* is set to ``dist2``. This means that the averaged time series is computing as a weighted average
!         and that the weight associated with each point time series in the selected domain is proportional to the surface
!         associated with it
!        *-sm=*
!         no smoothing is applied to the computed time series
!        *-3d*
!         the *output_netcdf_variable* is defined as an unidimensional NetCDF variable. However,
!         if **-3d** is activated, the *output_netcdf_variable* is defined as an tridimensional NetCDF variable
!         but with two dummy dimensions (e.g. with a length equal to ``1`` )
!        *-double*
!         the time series is stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         time series is stored as double-precision floating point numbers 
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable from which the time series must
!      be computed and the **-f=**\ *input_netcdf_file* argument specifies that this NetCDF variable must
!      be extracted from the NetCDF file *input_netcdf_file*.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing, the whole
!      geographical domain associated with the *netcdf_variable* is used to
!      construct the time series.
!
!      The longitude or latitude range must be a vector of two integers specifying
!      the first and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices before
!      using comp_serie_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, the whole time period
!      associated with the *netcdf_variable* is used to compute the time series.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) It is assumed that the data has no missing values excepted those associated with a constant land-sea
!      mask. If your dataset has missing values, use :ref:`comp_serie_miss_3d` instead of comp_serie_3d.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f``, it is assumed that the
!      NetCDF variable is from an experiment with the NEMO or ORCA model.
!      In this case, the duplicate points from the ORCA grid are removed 
!      before computing the time series, as far as possible, and, in particular,
!      if the 2-D grid-mesh of the input NetCDF variable covers the whole globe.
!
!      If **-g=** is set to ``n``, it is assumed that the 2-D grid-mesh is regular or Gaussian
!      and as such has no duplicate points.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` (e.g. if the NetCDF variable
!      is from an experiment with the NEMO or ORCA model), the **-r=** argument gives the 
!      resolution used. If:
!
!      * **-r=**\ ``r2``, the NetCDF variable is from an experiment with the ORCA R2 model
!
!      * **-r=**\ ``r4``, the NetCDF variable is from an experiment with the ORCA R4 model.
!
!   #) If the NetCDF variable is from an experiment with the NEMO or ORCA model,
!      but the resolution is not R2 or R4, the dimensions of the ORCA grid
!      must be specified explicitly with the **-b=** argument.
!
!   #) The **-a=** argument specifies if the point time series are centered or standardized
!      with an input climatology (specified with the **-c=** argument) before computing the
!      averaged time series for the selected domain. If:
!
!      * **-a=**\ ``scp``, the averaged time series is computed from the raw data
!
!      * **-a=**\ ``cov``, the averaged time series is computed from the anomalies
!
!      * **-a=**\ ``cor``, the averaged time series is computed from the standardized anomalies.
!
!   #) The *input_climatology_netcdf_file* specified with the **-c=** argument is needed
!      only if **-a=**\ ``cov`` or **-a=**\ ``cor``.
!
!   #) If **-a=**\ ``cov`` or **-a=**\ ``cor``, the selected time period must agree with the climatology.
!      This means that the first selected time observation (*time1* if the **-t=** argument
!      is present) must correspond to the first day, month, season of the climatology
!      specified with the **-c=** argument.
!
!   #) The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*), the mask
!      (in the *input_mesh_mask_netcdf_file*), the scale factors (in the *input_mesh_mask_netcdf_file*),
!      and the climatology (in the *input_climatology_netcdf_file*) must agree.
!
!   #) The **-d=** argument specifies the weighting method for computing the averaged time series. If:
!
!      * **-d=**\ ``dist2``, the averaged time series is computed with the diagonal distance associated with the
!        horizontal 2-D grid-mesh (each grid point is weighted accordingly to the surface associated with it)
!
!      * **-d=**\ ``ident``, the time series is computed with the identity metric (each grid point has the same weight
!        when computing the averaged time series).
!
!   #) **-sm=**\ *smoothing_factor* means that the averaged time series must be smoothed with a moving average of
!      approximately ``2``\ *\ *smoothing_factor*\ +\ ``1`` terms. *smoothing_factor* must be an integer greater
!      than ``0``.
!
!   #) The **-n=**\ *output_netcdf_variable* argument specifies the NetCDF variable which will contains 
!      the computed time series in the output NetCDF file, *output_netcdf_file*, specified by the
!      **-o=**\ *output_netcdf_file* argument.
!
!   #) The **-3d** argument specifies that the averaged time series must be stored as
!      a tridimensional NetCDF variable with two dummy dimensions in the output NetCDF file.
!      By default, the time series is stored as an unidimensional NetCDF variable.
!
!   #) The **-double** argument specifies that, the time series is stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the time series is stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
! Outputs
! -------
!
!   comp_serie_3d creates an output NetCDF file that contains the computed time series.
!   The output NetCDF dataset contains the following NetCDF variable (in the description below,
!   ntime is the number of time steps selected with the **-t=** argument) :
!
!   #) *output_netcdf_variable*\ ``(ntime)`` : the averaged time series defined as an unidimensional variable.
!
!   or if the **-3d** argument has been specified :
!
!   #) *output_netcdf_variable*\ ``(ntime,1,1)`` : the averaged time series defined as a tridimensional variable
!      with two dummy dimensions.
!
! Examples
! --------
!
!   #)   For computing a monthly time series from the
!        file ``HadISST1_1m_187001_200702_sst_reg1m.nc``, which includes a NetCDF
!        variable ``sst``, and store the results in a NetCDF variable named
!        ``nino34_sst`` in the file ``HadISST1_1m_187001_200702_nino34sst.nc``,
!        use the following command :
!
!        ::
!
!          $ comp_serie_3d \
!            -f=HadISST1_1m_187001_200702_sst_reg1m.nc \
!            -v=sst \
!            -x=11,60 \
!            -y=86,95 \
!            -m=mask_HadISST1_sst.nc \
!            -n=nino34_sst \
!            -o=HadISST1_1m_187001_200702_nino34sst.nc
!
!   #)   For computing a monthly time series from the
!        file ``F31_1m_000101_011012_sosstsst_grid_T.nc``, which includes a NetCDF
!        variable ``sosstsst`` (from the NEMO model), and store the results in a NetCDF variable
!        ``nino34_sosstsst`` in the file ``F31_1m_000101_011012_nino34sst.nc``,
!        use the following command :
!
!        ::
!
!          $ comp_serie_3d \
!            -f=F31_1m_000101_011012_sosstsst_grid_T.nc \
!            -v=sosstsst \
!            -x=236,335 \
!            -y=240,260 \
!            -m=F31_mesh_mask.nc \
!            -b=722,511  \
!            -g=t  \
!            -n=nino34_sosstsst \
!            -o=F31_1m_000101_011012_nino34sst.nc
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,    only : nf90_share
    use Statpack,  only : omp_limit
    use omp_lib,   only : omp_get_num_procs, omp_get_max_threads
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,           only : nf90_max_var_dims, nf90_max_name, nf90_real4, nf90_real8,                     &
                                 nf90_int, nf90_sizehint_default, nf90_open, nf90_nowrite, nf90_noerr,         &
                                 nf90_strerror, nf90_inq_varid, nf90_inquire_variable, nf90_inquire_dimension, &
                                 nf90_get_att, nf90_get_var, nf90_create, nf90_clobber,                        &
                                 nf90_set_fill, nf90_nofill, nf90_unlimited, nf90_def_dim, nf90_def_var,       &
                                 nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_global, nf90_enddef,      &
                                 nf90_put_var, nf90_close
!
    use Statpack,         only : i1b, i4b, stnd, extd, lgl, true, false, zero, one,       &
                                 kint, tolower, case_change, string_to_val,               &
                                 allocate_error, merror, arth, comp_smooth
!
    use NCSTAT_Utilities, only : prtunit, nchar, nlon_def2, nlat_def2, nlon_def4, nlat_def4,   &
                                 get_listarg, get_variable_dims, check_lon_domain,             &
                                 select_grid_points_2d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_serie_3d : ',        &
!
                                   cmethod(3)=(/"scp","cov","cor"/),    &          
!
                                   cresolution(2)=(/"r2","r4"/),        &        
!
                                   cdistance(2)=(/"dist2","ident"/)         
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_clim, ncid_mask, ncid_out,                        &
                     recordid, varid, maskid, meanid, stdid, e1id, e2id, timeid, serieid,       &
                     i, j, j1, j2, k, nlon, nlat, nt, period, ndom,                             &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3,                   &
                     method, distance, offset_rec, recordtype, recordnatts, natts,              &
                     nlon_def, nlat_def, smooth_factor, output_mode, output_type
    integer(i4b), dimension(3)                 :: startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens
    integer(i1b)                               :: kcode
!
    real(stnd)                                 :: tmp
    real(stnd), dimension(:,:),   allocatable  :: temp, mean, std
    real(stnd), dimension(:),     allocatable  :: serie, scal
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: use_record_variable, use_tlimited,          &
                                                  use_x, use_y, use_3d, create_hdf5, failure
    logical(lgl), dimension(:,:), allocatable  :: mask
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, clim_name, mask_name,    &
                                                  out_name, serie_name, att_name 
    character(len=nf90_max_name), dimension(3) :: dim_names
    character(len=14)                          :: fmt
    character                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
    integer(i4b)                            :: ncid2, varid2
    real(stnd), dimension(:),  allocatable  :: tempvec
!
    logical  :: test_par
!
#endif
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute a time series from a tridimensional variable '
        write (prtunit,*) ' extracted from a NetCDF dataset.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_serie_3d  -f=input_netcdf_file'
        write (prtunit,*) '                -v=input_netcdf_variable'
        write (prtunit,*) '                -m=input_mesh_mask_netcdf_file'
        write (prtunit,*) '                -o=output_netcdf_file'
        write (prtunit,*) '                -n=output_netcdf_variable'
        write (prtunit,*) '                -g=grid_type                     (optional : n, t, u, v, w, f)'
        write (prtunit,*) '                -r=resolution                    (optional : r2, r4)'
        write (prtunit,*) '                -b=nlon_orca,nlat_orca           (optional)'
        write (prtunit,*) '                -x=lon1,lon2                     (optional)'
        write (prtunit,*) '                -y=lat1,lat2                     (optional)'
        write (prtunit,*) '                -t=time1,time2                   (optional)'
        write (prtunit,*) '                -c=input_climatology_netcdf_file (optional)'
        write (prtunit,*) '                -a=type_of_analysis              (optional : scp, cov, cor)'
        write (prtunit,*) '                -d=type_of_distance              (optional : dist2, ident)'
        write (prtunit,*) '                -sm=smoothing_factor             (optional)'
        write (prtunit,*) '                -3d                              (optional)'
        write (prtunit,*) '                -double                          (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '                -hdf5                            (optional)'
#endif
        write (prtunit,*) '                -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-g=n'
        write (prtunit,*) '-r=r2'
        write (prtunit,*) '-b= nlon_orca and nlat_orca are determined from the -r=argument'
        write (prtunit,*) '-x= the whole longitude domain associated with the input_netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the input_netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the input_netcdf_variable'
        write (prtunit,*) '-a=scp'
        write (prtunit,*) '-d=dist2'
        write (prtunit,*) '-sm= no smoothing is applied to the time series'
        write (prtunit,*) '-3d       : the output_netcdf_variable is an unidimensional variable'
        write (prtunit,*) '-double   : the results are stored as single floating numbers'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : the time dimension is defined as unlimited'
        write (prtunit,*)
#ifdef _PARALLEL_READ
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<5 .or. narg>19 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name    = " "
    var_name    = " "
    mask_name   = " "
    clim_name   = " "
    out_name    = " "
    serie_name  = " "
!
    history = "comp_serie_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
    nlat_def      = nlat_def2
    nlon_def      = nlon_def2
    method        = 1
    distance      = 1
    smooth_factor = 0
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x        = false
    use_y        = false
    use_3d       = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-r=')
!
!               DETERMINE THE RESOLUTION FOR THE ORCA GRID.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cresolution(1) )
                        nlat_def   = nlat_def2
                        nlon_def   = nlon_def2
                    case ( cresolution(2) )
                        nlat_def   = nlat_def4
                        nlon_def   = nlon_def4
                    case default
                        call merror( name_proc//"Unknown resolution option" )
                end select
!
            case ('-b=')
!
!               GET THE DIMENSIONS OF THE ORCA GRID.
!
                call get_listarg( name_proc, name(4:),                           &
                                  "nlon_orca and nlat_orca are integer values",  &
                                  "Wrong ORCA grid definition ",                 &
                                  nlon_def, nlat_def )
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
            case ('-c=')
!
!               GET THE NAME OF THE NETCDF CLIMATOLOGY DATASET.
!
                clim_name = name(4:)
!
            case ('-a=')
!
!               DETERMINE THE METHOD OF ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case ( cmethod(3) )
                        method = 3
                    case default
                        call merror( name_proc//"Unknown analysis option" )
                end select
!
            case ('-d=')
!
!               DETERMINE THE DISTANCE FOR THE ANALYSIS.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cdistance(1) )
                        distance = 1
                    case ( cdistance(2) )
                        distance = 2
                    case default
                        call merror( name_proc//"Unknown distance option" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE TIME SERIES.
!
               out_name = name(4:)
!
            case ('-n=')
!
!               GET THE NAME OF THE NEW TIME SERIES.
!
                serie_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-sm=')
!
!                       DETERMINE THE SMOOTHING FACTOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) smooth_factor
                        end if
                        if ( smooth_factor<=0 )        &
         call merror( name_proc//"The -sm= argument is a strictly positive integer value " )
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-3d')
!
!                               DEFINE THE OUTPUT VARIABLE AS A 3D NETCDF VARIABLE.
!
                                use_3d = true
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case ('-double')
!
!                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " ) call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " ) call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( mask_name==" " ) call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( clim_name==" " .and. method>=2 )                               &
    call merror( name_proc//"Climatology NETCDF file is missing" )
!
    if ( out_name==" " ) call merror( name_proc//"Output NETCDF file is missing" )
!
    if ( serie_name==" " ) call merror( name_proc//"Output NETCDF variable is missing" )
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>=2 ) then
        k = nf90_sizehint_default
        status = nf90_open( trim(clim_name), nf90_nowrite, ncid_clim, chunksize=k )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!
    status = nf90_inq_varid( ncid, trim(var_name), varid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!
    status = nf90_inquire_variable( ncid, varid, ndims=i, dimids=dimids, natts=natts )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( i/=3 ) call merror( name_proc//"Input NETCDF variable is not tridimensional" )
!
    do i = 1, 3
!
        status = nf90_inquire_dimension( ncid, dimids(i), name=dim_names(i), len=dimlens(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nt     = dimlens(3)
!
    k = dimids(3)
!
!   GET THE RECORD VARIABLE ID FROM ITS NAME, IF ANY.
!
    status = nf90_inq_varid( ncid, trim(dim_names(3)), recordid )
    if ( status==nf90_noerr ) then
!
!       FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES FOR THE RECORD VARIABLE.
!
        status = nf90_inquire_variable( ncid, recordid, xtype=recordtype,                 &
                                        ndims=i, dimids=dimids, natts=recordnatts )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        use_record_variable = i==1 .and. dimids(1)==k
    else
        use_record_variable = false
    end if
    if ( .not.use_record_variable ) recordtype = nf90_int
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
    call merror( name_proc//"Shape disagreement with the mask" )
!
    if ( .not.use_x ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
        if ( status==nf90_noerr ) lon1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
        if ( status==nf90_noerr ) lon2 = i
!
    end if
!
    if ( .not.use_y ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
        if ( status==nf90_noerr ) lat1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
        if ( status==nf90_noerr ) lat2 = i
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!   CHECK THE GEOGRAPHICAL DOMAIN.
!   REMOVE DUPLICATE POINTS FROM THE ORCA GRID.
!
    if ( grid_type/='n' ) then
!
        if ( nlon>nlon_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (longitude)" )
!
        if ( nlat>nlat_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (latitude)" )
!
!       A TRICK TO DISABLE THE ELIMINATION AND DUPLICATION IN THE HIGH LATITUDES OF THE ORCA GRID
!       IF THE RESOLUTION IS NOT R2 OR R4.
!
        if ( ( nlat_def/=nlat_def2 .or. nlon_def/=nlon_def2 ) .and. ( nlat_def/=nlat_def4 .or. nlon_def/=nlon_def4 ) )    &
        nlat_def = nlat_def + 1
!
    end if
!
    call check_lon_domain( grid_type, nlon, nlon_def, lon1, lon2, lon3 )
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, temp )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED IN THE ANALYSIS FROM THE MASK.
!
    mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
!   SELECT THE DOMAIN.
!
    call select_grid_points_2d( grid_type, lon1, lon2, lat1, lat2,  &
                                nlon_def, nlat_def, mask, j1, j2  )
!
    if ( lon1<0 ) lon3 = nlon
!
    ndom = count( mask )
    if ( ndom<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( scal(ndom), serie(time3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   COMPUTE THE DISTANCE.
!
    if ( distance==1 ) then
!
!       THE DISTANCE IS ASSOCIATED WITH THE (HORIZONTAL) GRID.
!
!
!       GET THE SCALE FACTORS IDs FROM THEIR NAMES.
!       VERIFY THE DIMENSION LENGTHS FOR THE SCALE FACTOR VARIABLES.
!
        call get_variable_dims( name_proc, "e1"//chr, ncid_mask,    &
                                e1id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the first scale factor" )
!
        call get_variable_dims( name_proc, "e2"//chr, ncid_mask,    &
                                e2id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the second scale factor" )
!
!       GET THE SCALE FACTORS.
!
        status = nf90_get_var( ncid_mask, e1id, temp(:lon3,:lat3),             &
                               start=(/j1,lat1/), count=(/lon3,lat3/) )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        scal(:ndom) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        status = nf90_get_var( ncid_mask, e2id, temp(:lon3,:lat3),             &     
                               start=(/j1,lat1/), count=(/lon3,lat3/) )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        scal(:ndom) = scal(:ndom)*pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
!       COMPUTE THE DIAGONAL METRIC ASSOCIATED WITH THE HORIZONTAL GRID.
!
        scal(:ndom) = scal(:ndom)*(one/sum(scal(:ndom)))
!
    else
!
!       THE DISTANCE IS THE IDENTITY.
!
        scal(:ndom) = one/real( ndom, stnd )
!
    end if
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec = time1 - 1
!
    startlec(1:2) = (/ j1, lat1 /)
    countlec(1:3) = (/ lon3, lat3, 1_i4b /)
!
#ifdef _PARALLEL_READ
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    test_par = i>1 .and. j>1 .and. time3*ndom>omp_limit
!
#endif
!
!   SELECT THE METHOD FOR COMPUTING THE TIME SERIES.
!
    select case ( method )
!
        case ( 1 )
!
!           COMPUTE THE TIME SERIES ON THE RAW DATA.
!
!
!           GET THE DATA.
!
#ifdef _PARALLEL_READ
!
            if ( test_par ) then
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( temp )
!
!$OMP PARALLEL PRIVATE(k,status,startlec,ncid2,varid2,temp,tempvec)   &
!$OMP         ,SHARED(j1,j2,lon3,lat1,lat2,lat3,time3,                &
!$OMP                 ndom,offset_rec,countlec,mask,scal,serie)
!
!               ALLOCATE WORK ARRAYS.
!
                allocate( temp(lon3,lat3), tempvec(ndom),  stat=status )
                if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!               OPEN THE NETCDF FILE.
!
                status = nf90_open( trim(fic_name), nf90_share, ncid2 )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               GET THE VARIABLE ID FROM ITS NAME.
!
                status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!               GET THE DATA.
!
                startlec(1)  = j1
                startlec(2)  = lat1
!
!$OMP DO SCHEDULE(STATIC)
                do k = 1, time3
!
                    startlec(3) = k + offset_rec
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                    status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3),                       &
                                           start=startlec, count=countlec )     
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA AND ADJUST FOR THE SCALE FACTORS.
!
                    tempvec(:ndom) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
                    serie(k) = sum( tempvec(:ndom)*scal(:ndom) )
!
                end do
!$OMP END DO
!
!               DEALLOCATE WORK ARRAYS.
!
                deallocate( temp, tempvec )
!
!$OMP CRITICAL (closenc)
!
!               CLOSE THE NETCDF DATASET.
!
                status = nf90_close( ncid2 )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
            else
!
#endif
!
                do k = 1, time3
!
                    startlec(3) = k + offset_rec
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                    status = nf90_get_var( ncid, varid, temp(:lon3,:lat3),                       &
                                       start=startlec, count=countlec )     
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA AND ADJUST FOR THE SCALE FACTORS.
!
                    serie(k) = sum( pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )*scal(:ndom) )
!
                end do
!
#ifdef _PARALLEL_READ
!
            end if
!
#endif
!
        case ( 2, 3 )
!
!           GET THE MEAN FIELD ID FROM ITS NAME.
!           FIND THE DIMENSION LENGTHS FOR THE NETCDF MEAN VARIABLE.
!
            call get_variable_dims( name_proc, trim(var_name)//"_mean", ncid_clim,    &
                                    meanid, dimids, dimlens )
!
            if ( count(dimlens/=0)/=3 )     &
            call merror( name_proc//"Input climatology is not tridimensional" )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )    &
            call merror( name_proc//"Shape disagreement with the climatology (mean field)" )
!
            period = dimlens(3)
!
            failure = false
!
            status = nf90_get_att( ncid_clim, nf90_global, "lon1_"//trim(var_name), i )
            if ( status==nf90_noerr ) failure = j1<i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lon2_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = j2>i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lat1_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = lat1<i
!
            status = nf90_get_att( ncid_clim, nf90_global, "lat2_"//trim(var_name), i )
            if ( status==nf90_noerr .and. .not.failure ) failure = lat2>i
!
            if ( failure ) call merror( name_proc//"Missing values in the climatology " )
!
!           ALLOCATE WORK ARRAY FOR THE MEANS.
!
            allocate( mean(ndom,period), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!           GET THE MEANS.
!
            do j = 1, period
!
                startlec(3) = j
!
                status = nf90_get_var( ncid_clim, meanid, temp(:lon3,:lat3),                    &
                                       start=startlec, count=countlec  )   
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                mean(:ndom,j) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
            end do
!
            if ( method==2 ) then
!
!               COMPUTE THE TIME SERIES ON THE ANOMALIES.
!
!               GET THE DATA.
!
#ifdef _PARALLEL_READ
!
                if ( test_par ) then
!
!                   DEALLOCATE WORK ARRAY.
!
                    deallocate( temp )
!
!$OMP PARALLEL PRIVATE(i,j,k,status,startlec,ncid2,varid2,temp,tempvec)         &
!$OMP         ,SHARED(j1,j2,lon3,lat1,lat2,lat3,time3,period,                   &
!$OMP                 ndom,offset_rec,countlec,mask,mean,scal,serie)
!
!                   ALLOCATE WORK ARRAYS.
!
                    allocate( temp(lon3,lat3), tempvec(ndom),  stat=status )
                    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!                   OPEN THE NETCDF FILE.
!
                    status = nf90_open( trim(fic_name), nf90_share, ncid2  )
                    if ( status/=nf90_noerr )                                         &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   GET THE VARIABLE ID FROM ITS NAME.
!
                    status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!                   GET THE DATA.
!
                    startlec(1)  = j1
                    startlec(2)  = lat1
!
!$OMP DO SCHEDULE(STATIC)
                    do i = 0, time3-1, period
!
                        do j = 1, period
!
                            k   = i + j
!
                            if ( k<=time3 ) then
                                startlec(3) = k + offset_rec
!
!                               FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                                status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3),           &
                                                       start=startlec, count=countlec )     
                                if ( status/=nf90_noerr )                                          &
                                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                               PACK THE DATA, COMPUTE ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                                tempvec(:ndom) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
                                serie(k) =    sum( ( tempvec(:ndom)-mean(:ndom,j) )*scal(:ndom) )
                            end if
!
                        end do
!
                    end do
!$OMP END DO
!
!                   DEALLOCATE WORK ARRAYS.
!
                    deallocate( temp, tempvec )
!
!$OMP CRITICAL (closenc)
!
!                   CLOSE THE NETCDF DATASET.
!
                    status = nf90_close( ncid2 )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
                else
!
#endif
!
                    do i = 0, time3-1, period
!
                        do j = 1, period
!
                            k   = i + j
!
                            if ( k>time3 ) exit
!
                            startlec(3) = k + offset_rec
!
!                           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                            status = nf90_get_var( ncid, varid, temp(:lon3,:lat3),            &
                                                   start=startlec, count=countlec )     
                            if ( status/=nf90_noerr )                                         &
                            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                           PACK THE DATA, COMPUTE ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                            serie(k) =    sum(       &
       ( pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) ) - mean(:ndom,j) )*scal(:ndom) )
!
                        end do
!
                    end do
!
#ifdef _PARALLEL_READ
!
                end if
!
#endif
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( mean )
!
            else
!
!               COMPUTE THE TIME SERIES ON THE STANDARDIZED ANOMALIES.
!
!
!               GET THE STD ID FROM ITS NAME.
!               FIND THE DIMENSION LENGTHS FOR THE NETCDF STD VARIABLE.
!
                call get_variable_dims( name_proc, trim(var_name)//"_std", ncid_clim,    &
                                        stdid, dimids, dimlens )
!
                if ( count(dimlens/=0)/=3 )          &
                call merror( name_proc//"Input climatology is not tridimensional" )
!
                if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                                  &
                call merror( name_proc//"Shape disagreement with the climatology (std field)" )
!
                if ( dimlens(3)/=period )     &
                call merror( name_proc//"Shape disagreement in the climatology" )
!
!               ALLOCATE WORK ARRAY.
!
                allocate( std(ndom,period), stat=status )
                if ( status/=0 ) call merror( name_proc//allocate_error )
!
!               GET THE STDS AND ADJUST FOR THE SCALE FACTORS.
!
                do j = 1, period
!
                    startlec(3) = j
!
                    status = nf90_get_var( ncid_clim, stdid, temp(:lon3,:lat3),           &
                                           start=startlec, count=countlec  )   
                    if ( status/=nf90_noerr )                                             &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                    std(:ndom,j) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
                    where ( std(:ndom,j)/=zero )  std(:ndom,j) = scal(:ndom)/std(:ndom,j)
!
                 end do
!
!               GET THE DATA.
!
#ifdef _PARALLEL_READ
!
                 if ( test_par ) then
!
!                   DEALLOCATE WORK ARRAY.
!
                    deallocate( temp )
!
!$OMP PARALLEL PRIVATE(i,j,k,status,startlec,ncid2,varid2,temp,tempvec)      &
!$OMP         ,SHARED(j1,j2,lon3,lat1,lat2,lat3,time3,period,ndom,           &
!$OMP                 offset_rec,countlec,mask,mean,std,serie)
!
!                   ALLOCATE WORK ARRAYS.
!
                    allocate( temp(lon3,lat3), tempvec(ndom),  stat=status )
                    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!                   OPEN THE NETCDF FILE.
!
                    status = nf90_open( trim(fic_name), nf90_share, ncid2 )
                    if ( status/=nf90_noerr )                                              &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   GET THE VARIABLE ID FROM ITS NAME.
!
                    status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!                   GET THE DATA.
!
                    startlec(1)  = j1
                    startlec(2)  = lat1
!
!$OMP DO SCHEDULE(STATIC)
                    do i = 0, time3-1, period
!
                        do j = 1, period
!
                            k   = i + j
!
                            if ( k<=time3 ) then
!
                                startlec(3) = k + offset_rec
!
!                               FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE .
!
                                status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3),            &
                                                       start=startlec, count=countlec )     
                                if ( status/=nf90_noerr )                                          &
                                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                               PACK THE DATA, COMPUTE STANDARDIZED ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                                tempvec(:ndom) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
                                serie(k) =    sum( ( tempvec(:ndom)-mean(:ndom,j) )*std(:ndom,j) )
!
                            end if
!
                        end do
!
                    end do
!$OMP END DO
!
!                   DEALLOCATE WORK ARRAYS.
!
                    deallocate( temp, tempvec )
!
!$OMP CRITICAL (closenc)
!
!                   CLOSE THE NETCDF DATASET.
!
                    status = nf90_close( ncid2 )
                    if ( status/=nf90_noerr )                  &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
                else
!
#endif
!
                    do i = 0, time3-1, period
!
                        do j = 1, period
!
                            k   = i + j
!
                            if ( k>time3 ) exit
!
                            startlec(3) = k + offset_rec
!
!                           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                            status = nf90_get_var( ncid, varid, temp(:lon3,:lat3),            &
                                                   start=startlec, count=countlec )     
                            if ( status/=nf90_noerr )                                         &
                            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                           PACK THE DATA, COMPUTE STANDARDIZED ANOMALIES AND ADJUST FOR THE SCALE FACTORS.
!
                            serie(k) =  sum(             &
          ( pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) ) - mean(:ndom,j) )*std(:ndom,j) )
!
                        end do
!
                    end do
!
#ifdef _PARALLEL_READ
!
                end if
!
#endif
!
!               DEALLOCATE WORK ARRAYS.
!
                deallocate( mean, std )
!
            end if
!
    end select
!
    if ( method<=2 ) then
!
!       FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncid, varid, "scale_factor", tmp )
        if ( status==nf90_noerr ) serie(:time3) = serie(:time3)*tmp
!
        if ( method==1 ) then
!
!           FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
            status = nf90_get_att( ncid, varid, "add_offset", tmp )
            if ( status==nf90_noerr )  serie(:time3) = serie(:time3) + tmp

        end if
!
    end if
!
!   SMOOTH THE TIME SERIES IF REQUIRED.
!
    if ( smooth_factor/=0 ) then
        call comp_smooth( serie(:time3), smooth_factor )
    end if
!
!   CREATE THE NEW NETCDF DATASET.
!
    status = nf90_create( out_name, cmode=output_mode, ncid=ncid_out )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_out, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSION FOR THE NEW NETCDF DATASET.
!
    if ( use_tlimited ) then
        dimlens(3)   = time3
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    if ( use_3d ) then
!
!       DEFINE DUMMY DIMENSIONS FOR THE OUTPUT NETCDF VARIABLE.
!
        dim_names(1:2) = (/ "x_a", "y_a" /)
!
        dimlens(1:2) = 1
!
        do i = 1, 2
!
            status = nf90_def_dim( ncid_out, trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end if
!
    status = nf90_def_dim( ncid_out, trim(dim_names(3)), len=dimlens(3), dimid=dimids(3) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    status = nf90_def_var( ncid_out, trim(dim_names(3)), recordtype,               &
                           dimids=dimids(3), varid=timeid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_3d ) then
        status = nf90_def_var( ncid_out, serie_name, output_type, dimids=dimids(1:3), varid=serieid )
    else
        status = nf90_def_var( ncid_out, serie_name, output_type, dimids=dimids(3), varid=serieid )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    if ( use_record_variable ) then
!
        do i = 1, recordnatts
!
            status = nf90_inq_attname( ncid, recordid, i, att_name )
            if ( status/=nf90_noerr )                                               &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE RECORD VARIABLE.
!
            status = nf90_copy_att(  ncid, recordid, att_name, ncid_out, timeid )
            if ( status/=nf90_noerr )                                               &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    else
!
        status = nf90_put_att( ncid_out, timeid, "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, timeid, "short_name", "Time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_put_att( ncid_out, serieid, "long_name", trim(serie_name)//" Time Series" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, serieid, "short_name", trim(serie_name)//" Series" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method<=2 ) then
        status = nf90_copy_att(  ncid, varid, 'units', ncid_out, serieid )
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_out, nf90_global, "title", " Time series from "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "distance", cdistance(distance) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "smoothing_factor", smooth_factor )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_out )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW  NETCDF DATASET.
!
    if ( use_record_variable ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( record(time3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        status = nf90_get_var( ncid, recordid, record, start=(/ time1 /) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_out, timeid, record )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( record )
    else
        status = nf90_put_var( ncid_out, timeid, arth(time1,1_i4b,time3 ) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( use_3d ) then
        status = nf90_put_var( ncid_out, serieid, serie,                  &
                               start=(/1,1,1/), count=(/1_i4b,1_i4b,time3/) ) 
    else
        status = nf90_put_var( ncid_out, serieid, serie )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( allocated(temp) ) then
        deallocate( mask, serie, scal, temp )
    else
        deallocate( mask, serie, scal )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>=2 ) then
        status = nf90_close( ncid_clim )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_out )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM comp_serie_3d
! ============================
!
end program comp_serie_3d
