program comp_spectrum_1d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_spectrum_1d:
!
! ================
! comp_spectrum_1d
! ================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 02/01/2018
!
! Purpose
! =======
!
! Compute Fast Fourier Transform (FFT) estimates of the power spectrum of a real time series
! extracted from a uni or bidimensional variable readed from a NetCDF dataset.
!
! At the user option, the time series can be detrended before computing the power spectrum (see
! the **-tr=** and **-tr2=** arguments).
!
! The user can select different data windows for the computations of the Power Spectral Density (PSD)
! estimates (see the **-win=** argument).
!
! In order to obtain stable PSD estimates, the real time series can be segmented and PSD estimates
! on, eventually overlapping, segments can be averaged at the user option. The stability of the PSD estimates
! is increased by this averaging process. That is, the greater the number of segments, the more stable, the resulting
! PSD estimates.
!
! Optionally, theses PSD estimates may also been smoothed in the frequency
! domain by Daniell weights (e.g. a simple moving average; see the **-nsm=** argument).
!
! The PSD estimates are returned in units, which are the square of the data
! if the **-nonormpsd** argument is used  (meaning that the sum of the PSD estimates is equal to the
! variance of the time series) or in PSD units otherwise (meaning that the total area under the power
! spectrum is equal to the variance of the time series).
!
! Finally, again at the user option, statistical confidence interval and equivalent
! number of degrees of freedom of the power spectrum estimates can be estimated.
!
! All these power spectrum statistics are stored in an output NetCDF dataset.
!
! For more information on spectral analysis, see the references below [Bloomfield]_ [Diggle]_.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_spectrum_1d \
!              -f=input_netcdf_file \
!              -v=netcdf_variable \
!              -t=time1,time2                    (optional) \
!              -o=output_netcdf_file             (optional) \
!              -ni=index_for_2d_netcdf_variable  (optional) \
!              -p=period_length                  (optional) \
!              -sl=segment_length                (optional : 16,32,64,128,...) \
!              -tr=trend_removal                 (optional : 0,1,2,3) \
!              -tr2=trend_removal_on_segment     (optional : 0,1,2,3) \
!              -win=window_choice                (optional : 1,2,3,4,5,6) \
!              -tap=tapered_percentage           (optional : 0.0 > 1.) \
!              -nsm=Daniell_filter_length        (optional) \
!              -pro=critical_probability         (optional : 0.0 > 1.) \
!              -mi=missing_value                 (optional) \
!              -nonormpsd                        (optional) \
!              -nooverlap                        (optional) \
!              -conflim                          (optional) \
!              -double                           (optional) \
!              -hdf5                             (optional) \
!              -tlimited                         (optional)
!
! By default
! ----------
!
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-o=*
!         the *output_netcdf_file* is named ``spectrum_``\ *netcdf_variable*\ ``.nc``
!        *-ni=*
!         if the *netcdf_variable* is bidimensional, the first time series is used
!        *-p=*
!         the *period_length* is set to ``time2 - time1 + 1``, which means
!         that the time series is considered as continuous with only one time interval
!        *-sl=*
!         the *segment_length* is set to ``0``, which means that the time series (or each
!         time interval if the **-p=** argument is used) is not segmented to stabilize the
!         the PSD estimates
!        *-tr=*
!         the *trend_removal* is set to ``1``, which means that the mean of the whole time series
!         is removed before the spectral computations
!        *-tr2=*
!         the *trend_removal_on_segment* is set to ``0``, which means that no trend processing is
!         done on each time segment before the spectral computations
!        *-win=*
!         the *window_choice* is set to ``1``, which means that the Bartlett window is used in the
!         spectral computations
!        *-tap=*
!         the *tapered_percentage* is set to ``0.2``, which means that if a split-cosine-bell window
!         is used for the spectral computations (e.g. if the *window_choice* is set to ``6`` with the
!         **-win=** argument), ``20``\ % of the data are tapered
!        *-nsm=*
!         the *Daniell_filter_length* is set to ``0``, which means that the PSD estimates are not
!         smoothed in the frequency domain by a moving average
!        *-pro=*
!         the *critical_probability* is set to ``0.05``, which means that if a confidence interval
!         for the spectrum is requested with the **-conflim** argument, the lower and upper ``95``\ %
!         confidence limit factors are computed
!        *-mi=*
!         the *missing_value* for the output NetCDF variables is set to ``1.e+20``
!        *-nonormpsd*
!         the PSD estimates are normalized in such a way that the total area under
!         the power spectrum is equal to the variance of the time series. If **-nonormpsd**
!         is specified, the PSD estimates are returned in units which are the square of the data
!        *-nooverlap*
!         Normally, if the **-sl=** argument is used, the time segments are overlapped. However,
!         if **-nooverlap** is specified, the time segments are not overlapped
!        *-conflim*
!         Normally, the equivalent number of degrees of freedom and the confidence limit factors
!         of the power spectrum estimates are not computed. However, if **-conflim** is specified,
!         these statistics are computed and stored in the output NetCDF dataset
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable (e.g. the time series)
!      for which a  power spectrum must be computed and the **-f=**\ *input_netcdf_file* 
!      argument specifies that this NetCDF variable must be extracted from the NetCDF file
!      *input_netcdf_file*.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used to estimate the power spectrum.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!      *time2* - *time1* + ``1`` must be an even integer if the **-sl=** argument is not used
!      and of any length if this argument is used.
!
!   #) The **-ni=**\ *index_for_2d_netcdf_variable* argument specifies the index for selecting the
!      time series if the *netcdf_variable* is a 2D NetCDF variable. By default, the first time series
!      is used, which is equivalent to set *index_for_2d_netcdf_variable* to ``1``\ .
!
!   #) If the **-p=** argument is specified, the power spectrum estimates are computed separately for
!      each time period of length *period_length* (as determined by the value of the **-p=** argument).
!
!      The whole selected time period (e.g. *time2* - *time1* + ``1`` ) must be a multiple of 
!      the *period_length*.
!      The *period_length* must also be of even length if the **-sl=** argument is not used.
!
!   #) The **-sl=**\ *segment_length* argument specifies an integer used to segment the time series.
!      The *segment_length* must be a positive even integer less or equal to the length of the time series
!      (e.g. *time2* - *time1* + ``1`` ) or the *period_length* if the **-p=** argument is used.
!      Spectral computations are at ( *segment_length*\ /\ ``2``\  )  + ``1`` Fourier frequencies and suggested
!      values for *segment_length* are ``16``, ``32``, ``64``, ``128``, ...  (e.g. the spectral estimates are taken
!      at frequencies ``(i-1)``\  / *segment_length* for i=1,2, ... , (*segment_length*\ /\ ``2``\ ) + ``1``.
!      By default, *segment_length* is set to ``0``, e.g. the time series is not segmented and the segment length is
!      equal to *time2* - *time1* + ``1`` or *period_length* if the **-p=** argument is used.
!
!   #) The **-tr=** argument specifies pre-filtering processing of the real time series before
!      estimating the power spectrum of the time series. If:
!
!      * **-tr=**\ ``+1``, the mean of the time series is removed before computing the power spectrum
!
!      * **-tr=**\ ``+2``, the drift from the time series is removed before computing the power spectrum.
!        The drift for the time series is estimated using the formula : ``drift = ( tseries(ntime) - tseries(1) )/( ntime - 1 )``
!
!      * **-tr=**\ ``+3``, the least-squares line from the time series is removed before computing the
!        power spectrum.
!
!      For other values of the **-tr=** argument, nothing is done before estimating the spectrum.
!
!      If the **-p=** argument is present, the pre-filtering processing is applied to 
!      each time interval, separately.
!
!      The **-tr=** argument must be an integer and the default value for the **-tr=** argument is ``1``,
!      e.g. the mean of the time series is removed before the spectrum computations.
!
!   #) The **-tr2=** argument specifies pre-filtering processing of the real time series before
!      estimating the power spectrum on each time segment. If:
!
!      * **-tr2=**\ ``+1``, the mean of the time segment is removed before computing the power spectrum
!        on the segment
!
!      * **-tr2=**\ ``+2``, the drift from the time segment is removed before computing the power spectrum
!        on the segment
!
!      * **-tr2=**\ ``+3``, the least-squares line from the time segment is removed before computing the
!        power spectrum on the segment.
!
!      For other values of the **-tr2=** argument, nothing is done before estimating the spectrum on each
!      time segment.
!
!      The **-tr2=** argument must be an integer and the default value for the **-tr2=** argument is ``0``,
!      e.g. nothing is done before the spectral computations on each time segment.
!
!   #) The **-win=**\ *window_choice* argument specifies the form of the data window used in the computations
!      of the power spectrum. If:
!
!      * **-win=**\ ``1`` The Bartlett window is used
!
!      * **-win=**\ ``2`` The square window is used
!
!      * **-win=**\ ``3`` The Welch window is used
!
!      * **-win=**\ ``4`` The Hann window is used
!
!      * **-win=**\ ``5`` The Hamming window is used
!
!      * **-win=**\ ``6`` A split-cosine-bell window is used.
!
!      The *window_choice* must be an integer between ``1`` and ``6``.
!      The default value for the *window_choice* is ``1``, e.g. the Bartlett window is used for the spectral
!      computations.
!
!   #) The **-tap=**\ *tapered_percentage* argument specifies the total percentage of the data to be tapered if
!      the *window_choice* is set to ``6`` with the **-win=** argument.
!      The *tapered_percentage* must be greater than ``0`` and less or equal to ``1``.
!      The default value for the *window_choice* is ``0.2`` (e.g. ``20``\ % of the data are tapered, ``10``\ % at
!      the start of the series and ``10``\ % at the end).
!
!   #) The **-nsm=**\ *Daniell_filter_length* argument specified that the PSD estimates must be computed by smoothing
!      the periodogram with Daniell weights (e.g. a simple moving average).
!      On entry, the **-nsm=** argument gives the length of the Daniell filter to be applied.
!      The *Daniell_filter_length* must be odd, greater than ``2`` and less or equal to ( *segment_length*\ /\ ``2``\  ) + ``1``.
!      By default, the PSD estimates are not smoothed in the frequency domain.
!
!   #) If the **-nooverlap** argument is specified and the **-sl=** argument is also specified,
!      comp_spectrum_1d segments the time series without any overlapping (in each period 
!      if the **-p=** argument is used).
!      By default, comp_spectrum_1d overlaps the segments by one half of their length (which is
!      equal to the value of the **-sl=** argument).
!
!      In both cases, each segment will be FFT'd, and the resulting periodograms
!      will be averaged together to obtain PSD estimates at the
!      ( *segment_length*\ /\ ``2``\  ) + ``1`` frequencies.
!
!      If the **-sl=** argument is not specified, the **-nooverlap** argument has no effect.           
!
!   #) If the **-nonormpsd** argument is specified, the sum of the PSD estimates is equal to
!      the variance of the time series. By default, the PSD estimates are normalized in such
!      a way that the total area under the power spectrum is equal to the variance of the time
!      series.
!
!   #) If the **-conflim** argument is specified, the equivalent number of degrees of freedom 
!      and the lower and upper (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence limit factors of the power
!      spectrum estimates are computed. Multiply the PSD estimates by the lower and upper confidence
!      limit factors to get the lower and upper limits of a (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence
!      interval for the PSD estimates.
!
!      By default, these statistics are not computed.
!
!   #) The **-pro=**\ *critical_probability* argument specifies the critical probability which
!      is used to determine the lower and upper confidence limit factors. *critical_probability*
!      must be greater than ``0`` and less than ``1``. The default value is ``0.05``.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output NetCDF variables in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* is set
!      to ``1.e+20``.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the real time series has no missing values.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
!   #) For more details on power spectrum analysis and examples of use in the climate literature, see
!
!      * "Fourier analysis of time series- An introduction",
!        by Bloomfield, P., John Wiley and Sons, New York, Chapter 6, 1976. http://as.wiley.com/WileyCDA/WileyTitle/productCd-0471889482.html
!
!      * "Time series: a biostatistical introduction",
!        by P.J. Diggle, Clarendon Press, Oxford, Chap. 4, 1990. http://www.oupcanada.com/catalog/9780198522263.html
!
!      * "Impact of intra-daily SST variability on ENSO characteristics in a coupled model"
!        by S. Masson et al., 2012, Climate Dynamics, Vol. 39:681-707, doi: `10.1007/s00382-011-1247-2`_
!
!      * "The role of the intra-daily SST variability in the Indian monsoon variability in a global coupled model"
!        by P. Terray et al., 2012, Climate Dynamics, Vol. 39:729-754, doi: `10.1007/s00382-011-1240-9`_
!
!      * "Southern Hemisphere extra-tropical forcing: A new paradigm for El Nino-Southern Oscillation"
!        by P. Terray, 2011, Climate Dynamics, Vol. 36:2171-2199, doi: `10.1007/s00382-010-0825-z`_
!
! .. _`10.1007/s00382-011-1247-2` : http://dx.doi.org/10.1007/s00382-011-1247-2
! .. _`10.1007/s00382-011-1240-9` : http://dx.doi.org/10.1007/s00382-011-1240-9
! .. _`10.1007/s00382-010-0825-z` : http://dx.doi.org/10.1007/s00382-010-0825-z
!
! Outputs
! -------
!
!   comp_spectrum_1d creates an output NetCDF file that contains the power spectrum of the
!   time series associated with the input NetCDF variable. The output NetCDF dataset contains the
!   following NetCDF variables (in the description below, nfreq is the number of frequencies where
!   the PSD estimates are computed as determined by the **-t=**, **-p=** and **-sl=** arguments and
!   nperiod is ``1`` if the **-p=** argument is not used or equal to ``ntime``\  / *period_length* if
!   it is used, where ``ntime``\  = *time2* - *time1* + ``1`` ) :
!
!   #) frequency\ ``(nfreq)`` : the frequencies at which the PSD estimates are computed.
!
!   #) periodicity\ ``(nfreq)`` : the corresponding periods in number of time observations.
!
!   #) *netcdf_variable*\ _spectrum\ ``(nfreq,nperiod)`` : the corresponding PSD estimates
!      for each time interval of the time series associated with the input NetCDF variable.
!
!   #) edof\ ``(nfreq)`` : the equivalent number of degrees of freedom of the PSD estimates
!      the time series associated with the input NetCDF variable.
!
!      This variable is stored only if the **-conflim** argument has been specified when calling
!      comp_spectrum_1d.
!
!   #) lower_factor\ ``(nfreq)`` : the lower confidence limit factors of the PSD estimates.
!      Multiply the PSD estimates by the corresponding lower confidence limit factors to get the lower
!      limits of the (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence
!      intervals for the PSD estimates.
!
!      This variable is stored only if the **-conflim** argument has been specified when calling
!      comp_spectrum_1d.
!
!   #) upper_factor\ ``(nfreq)`` : the upper confidence limit factors of the PSD estimates.
!      Multiply the PSD estimates by the corresponding upper confidence limit factors to get the upper
!      limits of the (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence
!      intervals for the PSD estimates.
!
!      This variable is stored only if the **-conflim** argument has been specified when calling
!      comp_spectrum_1d.
!
! Examples
! --------
!
!   #)   For computing the power spectrum of a real monthly time series from a NetCDF variable called ``sst``
!        extracted from the file ``sst.monthly.nino34.nc``, which includes a monthly time
!        series, and store the results in the NetCDF file ``spectrum_sst_nino34.nc``, use the following
!        command :
!
!        ::
!
!          $ comp_spectrum_1d \
!            -f=sst.monthly.nino34.nc \
!            -v=sst \
!            -sl=128 \
!            -nonormpsd \
!            -nooverlap \
!            -o=spectrum_sst_nino34.nc
!
!-
! ================================================================================================
!
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,   &
                         nf90_int, nf90_sizehint_default, nf90_open,                               &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,      &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,    &
                         nf90_def_var, nf90_put_att, nf90_enddef,                                  &
                         nf90_global, nf90_put_var, nf90_close
!
    use Statpack, only : i1b, i4b, stnd,  lgl, true, false, zero, one, kfix, kexp, kint, tolower,    & 
                         case_change, string_to_val, val_to_string, arth,                            &
                         allocate_error, merror, power_spctrm2, init_fft, end_fft
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,      &
                                 get_listarg, get_variable_dims
!
!
! STRONG TYPING IMPOSED
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_spectrum_1d : '
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_spect,                        &
                     varid, spectid, periodid, edofid, conlwrid, conuprid,  &
                     i, l, l1, l2, iindex, ndim, nt,                        &
                     time1, time2, time3, nf, nfm1, sl,                     &
                     trend, trend2, win, output_type, output_mode, period,  &
                     nperiod, nsmooth
    integer(i4b), dimension(2)                 :: coor_id, coor_type
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),   allocatable    :: vec, conlwr, conupr, edof
    real(stnd), dimension(:,:), allocatable    :: psvec
    real(stnd)                                 :: missing, tmp, taperp, probtest
!
    logical(lgl)                               :: use_tlimited, use_period, normpsd, overlap, conflim, use_miss
!
    character(len=nchar)                       :: name, name2, history
    character(len=nf90_max_name)               :: fic_name, var_name, spect_name
    character(len=nf90_max_name), dimension(2) :: dim_names
    character(len=14)                          :: fmt
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute a Fast Fourier Transform (FFT) estimate of the power spectrum (PSD)'
        write (prtunit,*) ' of a real time series. The time series is extracted from a uni or bidimensional'
        write (prtunit,*) ' variable readed from a NetCDF dataset'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_spectrum_1d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -t=time1,time2                   (optional)'
        write (prtunit,*) '        -o=output_netcdf_file            (optional)'
        write (prtunit,*) '        -ni=index_for_2d_netcdf_variable (optional)'
        write (prtunit,*) '        -p=period_length                 (optional)'
        write (prtunit,*) '        -sl=segment_length               (optional : 16,32,64,128,...)'
        write (prtunit,*) '        -tr=trend_removal                (optional : 0,1,2,3)'
        write (prtunit,*) '        -tr2=trend_removal_on_segment    (optional : 0,1,2,3)'
        write (prtunit,*) '        -win=window_choice               (optional : 1,2,3,4,5,6)'
        write (prtunit,*) '        -tap=tapered_percentage          (optional : 0. > 1.)'
        write (prtunit,*) '        -nsm=Daniell_filter_length       (optional)'
        write (prtunit,*) '        -pro=critical_probability        (optional : 0. > 1.)'
        write (prtunit,*) '        -mi=missing_value                (optional)'
        write (prtunit,*) '        -nonormpsd                       (optional)'
        write (prtunit,*) '        -nooverlap                       (optional)'
        write (prtunit,*) '        -conflim                         (optional)'
        write (prtunit,*) '        -double                          (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                            (optional)'
#endif
        write (prtunit,*) '        -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-t=the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-o=spectrum_netcdf_variable.nc'
        write (prtunit,*) '-ni=1            : by default, if the netcdf_variable is bidimensional,'
        write (prtunit,*) '                   the first time series is used'
        write (prtunit,*) '-p=time2-time1+1 : by default, the period argument is not used'
        write (prtunit,*) '-sl=0            : by default, the time series is not segmented'
        write (prtunit,*) '-tr=1            : by default, the mean of the time series is removed'
        write (prtunit,*) '-tr2=0           : by default, no preprocessing is done on the segments'
        write (prtunit,*) '-win=1           : by default, the Bartlett window is used'
        write (prtunit,*) '-tap=0.2         : by default, if a split-cosine-bell window is used, 20% of the data are tapered'
        write (prtunit,*) '-nsm=0           : by default, the PSD estimates are not smoothed in the frequency domain'
        write (prtunit,*) '-pro=0.05        : by default, the lower and upper 95% confidence limit factors are computed'
        write (prtunit,*) '-mi=1.e+20       : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-nonormpsd       : by default, the PSD estimates are normalized'
        write (prtunit,*) '-nooverlap       : by default, the segments are overlapped'
        write (prtunit,*) '-conflim         : by default, the number of degrees of freedom and the confidence limit factors'
        write (prtunit,*) '                   are not computed'
        write (prtunit,*) '-double          : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5            : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited        : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>20 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name    = " "
    var_name    = " "
    spect_name  = " "
!
    history = "comp_spectrum_1d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
!
    iindex        = 1
    sl            = 0
    trend         = 1
    trend2        = 0
    win           = 1
    nsmooth       = 0
    taperp        = 0.2_stnd
    probtest      = 0.05_stnd
!
    missing = def_missing
!
    time1 = 1
    time2 = huge( time2 )
!
    period    = huge( period )
    nperiod   = 1
!
    normpsd      = true
    overlap      = true
    conflim      = false
    use_period   = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-p=')
!
!               GET THE PERIOD.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The period is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
                use_period = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                spect_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-ni=')
!
!                       GET THE INDEX FOR A 2D VARIABLE.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) iindex
                        end if
                        if ( kcode/=kint .or. iindex<=0  )                           &
                        call merror( name_proc//"The index for a 2D variable is a positive integer value" )
!
                    case ('-sl=')
!
!                       GET THE LENGTH OF THE SEGMENTS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) sl
                        end if
!
                        if ( kcode/=kint .or. sl<0  )                           &
                        call merror( name_proc//"The -sl= argument is a positive integer value" )
!
!                       CHECK IF sl IS A POSITIVE EVEN INTEGER.
!
                        if ( sl/=2*(sl/2) )     &
                        call merror( name_proc//'The -sl= argument is an even integer !' )
!
                    case ('-tr=')
!
!                       DETERMINE PRE-FILTERING COMPUTATIONS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) trend
                        else
                            call merror( name_proc//"The -tr= argument is an integer value" )
                        end if
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5) )
!
                            case ('-tr2=')
!
!                               DETERMINE PRE-FILTERING COMPUTATIONS ON EACH SEGMENT.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) trend2
                                else
                                    call merror( name_proc//"The -tr2= argument is an integer value" )
                                end if
!
                            case ('-win=')
!
!                               DETERMINE THE FORM OF THE DATA WINDOW.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) win
                                end if
                                if ( kcode/=kint .or.      &
                                     win<1       .or.      &
                                     win>6 )               &
      call merror( name_proc//"The -win= argument must be an integer number and verify 1<=win<=6 " )
!
                            case ('-nsm=')
!
!                               DETERMINE THE LENGTH OF THE DANIELL FILTER TO BE APPLIED.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) nsmooth
                                end if
!
                                if ( kcode/=kint .or.             &
                                     nsmooth<3   .or.             &
                                     nsmooth==(nsmooth/2)*2   )   &
      call merror( name_proc//"The -nsm= argument must be an odd integer and verify 3<=nsm " )
!
                            case ('-tap=')
!
!                               DETERMINE THE TOTAL PERCENTAGE OF THE DATA TO BE TAPERED IF -WIN=6.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) taperp
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     taperp<=zero                    .or.     &
                                     taperp>one )                             &
      call merror( name_proc//"The -tap= argument must be a real number and verify 0.<tap<=1. " )
!
                            case ('-pro=')
!
!                               DETERMINE THE CRITICAL PROBABILITY.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) probtest
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     probtest<=zero                  .or.     &
                                     probtest>=one )                          &
      call merror( name_proc//"The -pro= argument must be a real number and verify 0.<pro<1. " )
!
                            case default
!
                                select case ( name(1:) )
!
                                    case ('-nonormpsd')
!
!                                       DO NOT NORMALIZE THE PSD ESTIMATES.
!
                                        normpsd = false
!
                                    case ('-nooverlap')
!
!                                       DO NOT OVERLAP THE SEGMENTS.
!
                                        overlap = false
!
                                    case ('-conflim')
!
!                                       COMPUTE THE CONFIDENCE LIMIT FACTORS.
!
                                        conflim = true
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
#endif
!
                                    case ('-tlimited')
!
!                                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                                        use_tlimited = true
!
                                    case ('-double')
!
!                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                        output_type = nf90_real8
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( spect_name==" " ) spect_name="spectrum_"//trim(var_name)//".nc"
!
!   OPEN THE NETCDF DATASET.
!
    i = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=i )
    if ( status/=nf90_noerr )                                                &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!
    call get_variable_dims(  name_proc, trim(var_name), ncid,               &
                             varid, dimids_out=dimids, dimlens_out=dimlens  )
!
    ndim = count(dimlens/=0)
    if ( ndim/=1 .and. ndim/=2 )     &
    call merror( name_proc//"Input NETCDF variable is not uni or bidimensional" )
!
    if ( ndim==2 ) then
        if ( iindex>dimlens(1) )     &
        call merror( name_proc//"Selected index for 2D input NETCDF variable is out of range " )
        nt = dimlens(2)
    else 
        nt = dimlens(1)
    end if
!
!   CHECK THE TIME DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
!
    if ( time3<4  )       &
    call merror( name_proc//"Wrong time definition or not enough time observations" )
!
    if ( use_period ) then
!
        if ( period<4  )     &
        call merror( name_proc//"Not enough time observations in the periods" )
!
        nperiod = time3/period
        if ( nperiod*period/=time3 )    &
        call merror( name_proc//"The selected time period is not a multiple of the period" )
!
    else
!
        period = time3
!
    end if
!
!   CHECK SEGMENT LENGTH.
!
    if ( sl==0 ) then
!
        i = 2*(period/2)
        if ( period/=i )       &
    call merror( name_proc//"The period or the number of observations is not an even integer" )
!
        sl = period
!
    else
!
        if ( period<sl)       &
    call merror( name_proc//"Length of segments cannot be greater than the number of observations or the period")
!
    end if
!
!   DETERMINE THE NUMBER OF FREQUENCIES.
!
    nf = ( sl/2 ) + 1
!
    if ( nsmooth>nf )    &
    call merror( name_proc//"The length of the Daniell filter must be less or equal to the number of frequencies")
!
!   ALLOCATE WORK ARRAYS FOR THE TIME SERIES AND THE PSD ESTIMATES.
!
    if ( conflim ) then
        allocate( vec(time3), psvec(nf,nperiod),                   &
                  edof(nf), conlwr(nf), conupr(nf), stat=status    )
    else
        allocate( vec(time3), psvec(nf,nperiod), stat=status )
    end if
!
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE DATA.
!
    if ( ndim==1 ) then
        status = nf90_get_var( ncid, varid, vec(:time3),              &   
                               start=(/ time1 /), count=(/ time3 /)   )
    else 
        status = nf90_get_var( ncid, varid, vec(:time3),                          &   
                               start=(/ iindex, time1 /), count=(/ 1, time3 /)    )
    end if
!
    if ( status/=nf90_noerr )                                     &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE missing_value ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "missing_value", tmp )
    use_miss = status==nf90_noerr
!
    if ( .not.use_miss ) then
!
!       FIND THE _FillValue ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE,
!       IF NO MISSING VALUE ATTRIBUTE.
!
        status = nf90_get_att( ncid, varid, "_FillValue", tmp )
        use_miss = status==nf90_noerr
!
    end if
!
    if ( use_miss ) then
!
        if ( any( vec(:time3)==tmp ) )      &
        call merror( name_proc//"Missing values are not allowed" )
!
    end if
!
!   MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
    status = nf90_get_att( ncid, varid, "scale_factor", tmp )
    if ( status==nf90_noerr ) vec(:time3) = vec(:time3)*tmp
!
!   ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
    status = nf90_get_att( ncid, varid, "add_offset", tmp )
    if ( status==nf90_noerr ) vec(:time3) = vec(:time3) + tmp
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    nfm1 = nf - 1
!
    call init_fft( nfm1 )
!
!   ESTIMATE THE POWER SPECTRUM OF THE REAL TIME SERIES.
!
    l2 = 0
!
    do l=1, nperiod
!
        l1 = l2 + 1
        l2 = l2 + period
!
        if ( l==1 .and. conflim ) then
!
            call power_spctrm2( vec(l1:l2), sl, psvec(:nf,l), overlap=overlap,                     &
                                normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,      &
                                win=win, taperp=taperp, edof=edof(:nf), conlwr=conlwr(:nf),        &
                                conupr=conupr(:nf), probtest=probtest, initfft=false               )
!
        else
!
            call power_spctrm2( vec(l1:l2), sl, psvec(:nf,l), overlap=overlap,                     &
                                normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,      &
                                win=win, taperp=taperp, initfft=false     )
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft( )
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    status = nf90_create( spect_name, cmode=output_mode, ncid=ncid_spect )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_spect, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nf
!
    if ( use_tlimited ) then
        dimlens(2)   = nperiod
    else
        dimlens(2)   = nf90_unlimited
    end if
!
    coor_type(1) = output_type
    coor_type(2) = nf90_int
!
    dim_names(1) = "frequency"
    dim_names(2) = "time"
!
    do i = 1, 2
!
        status = nf90_def_dim( ncid_spect, name=trim(dim_names(i)),                         &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 2
!
        status = nf90_def_var( ncid_spect, trim(dim_names(i)), coor_type(i),              &
                               dimids=dimids(i), varid=coor_id(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_spect, "periodicity",  output_type,   &
                           dimids=dimids(1), varid=periodid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_spect, trim(var_name)//"_spectrum",  output_type,   &
                           dimids=dimids(1:2), varid=spectid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( conflim ) then
!
        status = nf90_def_var( ncid_spect, "edof",  output_type,   &
                               dimids=dimids(1), varid=edofid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_spect, "lower_factor",  output_type,   &
                               dimids=dimids(1), varid=conlwrid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_spect, "upper_factor",  output_type,   &
                               dimids=dimids(1), varid=conuprid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    status = nf90_put_att( ncid_spect, coor_id(1), "long_name", "Frequency axis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id(1), "short_name", "Frequency" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id(1), "units", "cycles per unit of time" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, periodid, "long_name", "Periodicity axis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, periodid, "short_name", "Periodicity" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, periodid, "units", "Number of observations" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id(2), "long_name", "Time axis" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id(2), "short_name", "Time" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( ndim==1 ) then
        name = trim(var_name)
    else
        call val_to_string( iindex, fmt, i )
        name = trim(var_name)//"("//fmt(1:i)//")"
    end if
!
    status = nf90_put_att( ncid_spect, spectid, "long_name", "Power Spectral Density of "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, spectid, "short_name", "Power Spectral Density"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( normpsd ) then
!
        status = nf90_put_att( ncid_spect, spectid, "units", "spectral density units"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        name2 = " "
!
        status = nf90_get_att( ncid, varid, "units", name2 )
        if ( status==nf90_noerr ) then
            status = nf90_put_att( ncid_spect, spectid, "units", "("//trim(name2)//")**2 "  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end if
!
    if ( conflim ) then
!
        status = nf90_put_att( ncid_spect, edofid, "long_name",                              &
                               "Equivalent number of degrees of freedom of PSD estimates"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, edofid, "short_name", "Number of degrees of freedom"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conlwrid, "long_name",                              &
                               "Lower (1-critical_probability)*100% confidence limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conlwrid, "short_name", "Lower limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conuprid, "long_name",                              &
                               "Upper (1-critical_probability)*100% confidence limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conuprid, "short_name", "Upper limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   SET missing_value AND _FillValue ATTRIBUTES.
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_spect, coor_id(1), "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, coor_id(1), "_FillValue", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, spectid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, spectid, "_FillValue", real( missing, dp ) )
!
    else
!
        status = nf90_put_att( ncid_spect, coor_id(1), "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, coor_id(1), "_FillValue", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, spectid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, spectid, "_FillValue", real( missing, sp ) )
!
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   SET GLOBAL ATTRIBUTES.
!
    name2 = " Power spectrum analysis of "//trim(name)
!
    status = nf90_put_att( ncid_spect, nf90_global, "title", trim(name2) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "period_length", period )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "segment_length", sl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "trend_removal", trend )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "trend_removal_segment", trend2 )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "window_choice", win )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( win==6 ) then
!
        status = nf90_put_att( ncid_spect, nf90_global, "tapered_percentage", taperp )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( conflim ) then
!
        status = nf90_put_att( ncid_spect, nf90_global, "critical_probability", probtest )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( normpsd ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_spect, nf90_global, "normalization", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( nsmooth/=0 ) then
!
        i = 1
!
        status = nf90_put_att( ncid_spect, nf90_global, "Daniell_filter_length", nsmooth )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_spect, nf90_global, "smoothing", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( overlap ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_spect, nf90_global, "overlapping", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_spect )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
!   WRITE FREQUENCY AND PERIOD AXES.
!
    tmp = one/real( sl, stnd )
    vec(:nf) = arth( zero, tmp, nf )
!
    status = nf90_put_var( ncid_spect, coor_id(1), vec(:nf) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_var( ncid_spect, coor_id(2), arth(1_i4b,1_i4b,nperiod) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   BUILD UP PERIODICITY VECTOR.
!
    vec(1)    = missing
    vec(2:nf) = one/vec(2:nf)
!
    status = nf90_put_var( ncid_spect, periodid, vec(:nf) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE PSD ESTIMATES.
!
    status = nf90_put_var( ncid_spect, spectid, psvec(:nf,:nperiod) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( conflim ) then
!
!       WRITE THE NUMBER OF DEGREES OF FREEDOM AND THE CONFIDENCE LIMIT FACTORS.
!
        status = nf90_put_var( ncid_spect, edofid, edof(:nf) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_spect, conlwrid, conlwr(:nf) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_spect, conuprid, conupr(:nf) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( conflim ) then
        deallocate( vec, psvec, edof, conlwr, conupr )
    else
        deallocate( vec, psvec )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_spect )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_spectrum_1d
! ===============================
!
end program comp_spectrum_1d
