program comp_stat_4d
!! #############################################################################
!!
!! Copyright 2019 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_stat_4d:
!
! ============
! comp_stat_4d
! ============
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 20/11/2019
!
! Purpose
! =======
!
! Compute univariate statistics from a fourdimensional variable extracted from
! a NetCDF dataset and, optionally, the associated mesh-mask and scale
! factors of the 3-D grid-mesh associated with the input NetCDF variable.
!
! Mean, variance, standard-deviation, skewness, kurtosis, minimum and maximum are computed
! for each point in the time series of the 3-D grid-mesh associated with the input 
! NetCDF variable.
! These univariate statistics may be computed by taking into account
! the periodicity of the data. These statistics are stored in an output NetCDF dataset.
!
! The mean is a simple, but informative, measure, of the central tendency of a variable.
! The standard-deviation and variance are conventional measures of variation of a variable.
! If *X*\ (:) is a vector of *ntime* observations for one grid-point in the time series of the 3-D
! grid-mesh, the **MEAN**\ , **VAR** (e.g. variance) and **STD** (e.g. standard-deviation) statistics
! for this grid-point are estimated by:
!
! * **MEAN** = sum( *X*\ (:) )/\ *ntime*
!
! * **VAR** = sum( [*X*\ (:)-\ **MEAN**\ ]**2 )/(*ntime*\ -``1``)
!
! * **STD** = sqrt(\ **VAR**\ )
!
! Note that the divisor used in calculating variance and standard-deviation is the number of degrees of freedom
! (e.g. the number of observations minus ``1``), this is in contrast with the formulae used in :ref:`comp_clim_4d`.
!
! Skewness measures the deviation of the distribution of a variable from symmetry [vonStorch_Zwiers]_ [Burgers_Stephenson]_
! [White]_ [Masson_etal]_. For a symmetrical distribution, the skewness coefficient is always equal to ``0``, but the converse
! is not true. Skewness is ``0`` for a normal distribution. For unimodal distributions shifted to
! the right (left), the skewness coefficient is positive (negative).
!
! If the argument **-nobias** is specified, the skewness coefficient is estimated as
!
!
! **SKEWNESS** = (*ntime*\ .\ **M3**\ )/[(*ntime*\ -``1``).(*ntime*\ -``2``).\ **STD3**\ ]
!
!
! , otherwise the following (biased) formulae is used:
!
!
! **SKEWNESS** = **M3**\ /(*ntime*\ .\ **STD3**\ )
!
!
! where
!
!  * **M3**
!    is equal to the sum of the deviations of the observations from the mean raised
!    to the third power (e.g. sum( [*X*\ (:)-**MEAN**\ ]**3) where *X*\ (:) is the vector
!    of observations)
!  * **STD3**
!    is the standard deviation raised to the third power
!
! In order to interpret correctly, the skewness of a variable, note that the Standard Error
! (SE) of the skewness coefficient is given by
!
!
! *SE*\ [\ **SKEWNESS**\ ] = sqrt( [``6``\ .\ *ntime*\ .(*ntime*\ -``1``)]/[(*ntime*\ -``2``).(*ntime*\ +\ ``1``).(*ntime*\ +\ ``3``)\ ] )
!
!
! *SE*\ [\ **SKEWNESS**\ ] is not very different from the quantity sqrt( [``6``\ /\ *ntime*\ ] ) when
! the number of observations is sufficiently high.
!
! Moreover, the quantity **SKEWNESS**\ /\ *SE*\ [\ **SKEWNESS**\ ] follows asymptotically a normal (e.g.
! Gaussian) distribution with mean ``0`` and variance equal to ``1`` when the sample *X*\ (:) were
! independent Gaussian observations. With a sample of independent Gaussian observations, a value
! twice the standard error is thus associated with a 5% significance level. However, in climate analysis
! the observations are in general autocorrelated.
!
! Kurtosis measures the flatness or peakedness of the distribution of a variable [vonStorch_Zwiers]_ [Burgers_Stephenson]_
! [White]_ . As computed by comp_stat_3d, the kurtosis coefficient is always greater or equal
! to ``-2`` and is equal to ``0`` for a normal distribution.
! In most cases, if the kurtosis is greater (lower) than ``0`` then the distribution is more
! peaked (flatter) than the normal distribution with the same mean and standard-deviation.
!
! If the argument **-nobias** is specified, the kurtosis coefficient is estimated as
!
!
! **KURTOSIS** = **A** - 3.[(*ntime*\ -``1``).(*ntime*\ -``1``)]/[(*ntime*\ -``2``).(*ntime*\ -``3``)]
!
!
! , otherwise the following (biased) formulae is used:
!
!
! **KURTOSIS** = **M4**\ /(*ntime*\ .\ **STD4**\ ) - ``3``
!
! where
!
!  * **M4**
!    is equal to the sum of the deviations of the observations from the mean raised
!    to the fourth power (e.g. sum( [*X*\ (:)-**MEAN**\ ]**4) where *X*\ (:) is the vector
!    of observations)
!  * **STD4**
!    is the standard deviation raised to the fourth power
!  * **A**
!    is equal to [*ntime*\ .(*ntime*\ +\ ``1``).\ **M4**\ ]/[(*ntime*\ -``1``).(*ntime*\ -``2``).(*ntime*\ -``3``).\ **STD4**\ ]
!
! In order to interpret correctly the kurtosis of a variable, note that the SE
! of the kurtosis coefficient calculated from a sample drawn from a Gaussian distribution
! is given by
!
!
! *SE*\ [\ **KURTOSIS**\ ] = sqrt( **B**\ /\ **C** )
!
! where
!
!  * **B**
!    is equal to ``24``\ .\ *ntime*\ .(*ntime*\ -``1``).(*ntime*\ -``1``)
!  * **C**
!    is equal to (*ntime*\ -``3``).(*ntime*\ -``2``).(*ntime*\ +\ ``3``).(*ntime*\ +\ ``5``)
!
! and the quantity **KURTOSIS**\ /\ *SE*\ [\ **KURTOSIS**\ ] follows also asymptotically a normal
! distribution with mean ``0`` and variance equal to ``1`` when the sample *X*\ (:) were
! independent Gaussian observations.
!
! Extreme departures from the mean will cause very high values of kurtosis. Consequently, the kurtosis coefficient
! can be used to detect outliers. High values of kurtosis can also be a result of one or two extreme observations
! in a sample of observations.
!
! The standard errors of the skewness and kurtosis coefficients are calculated and stored in the output NetCDF file
! if the argument **-stderror** is specified.
!
! Moreover, the two-tailed significance levels of the statistics **SKEWNESS**\ /\ *SE*\ [\ **SKEWNESS**\ ] and
! **KURTOSIS**\ /\ *SE*\ [\ **KURTOSIS**\ ] (under the hypothesis that the sample *X*\ (:) were independent
! Gaussian observations) are calculated and stored in the output NetCDF file if the argument **-prob** is
! specified.
!
! Thus, if you are interested in how well the distribution of the time series in the 3-D grid-mesh
! associated with the input NetCDF variable can be approximated by the normal distribution, the skewness
! and kurtosis coefficients, their standard errors and significance levels, can give you some useful
! information on this issue.
!
! If the NetCDF variable is tridimensional use :ref:`comp_stat_3d` instead 
! of comp_stat_4d.
!
! This procedure is parallelized if OpenMP is used and the NCSTAT software has been built
! with the _PARALLEL_READ CPP macro (however, in this version of comp_stat_4d,
! parellelism is on the number of periods as determined by the **-p=**\ *periodicity* argument.
! This means that the number of processors or threads must not be greater than the *periodicity*
! parameter).
!
! Moreover, this procedure computes all the univariate statistics with only one pass through
! the data and an out-of-core strategy which is highly efficient on huge datasets.
!
! Optionally, a mesh-mask NetCDF dataset may also be created. This dataset will contain
! a presence-absence 3-D mask and scale factor variables which may be used to compute the surface
! or volume of each cell in the 3-D grid-mesh associated with the input NetCDF variable. This
! mesh-mask NetCDF dataset will be used by other NCSTAT procedures such as :ref:`comp_serie_4d`,
! :ref:`comp_eof_4d`, etc.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_stat_4d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -p=periodicity                                  (optional) \
!         -x=lon1,lon2                                    (optional) \
!         -y=lat1,lat2                                    (optional) \
!         -z=level1,level2                                (optional) \
!         -t=time1,time2                                  (optional) \
!         -o=output_statistics_netcdf_file                (optional) \
!         -m=output_mesh_mask_netcdf_file                 (optional) \
!         -vz=name_of_the_vertical_netcdf_variable        (optional) \
!         -z0=value_of_the_highest_level                  (optional) \
!         -sf=method_for_computing_the_third_scale_factor (optional : method1, method2) \
!         -yl=latl1,latl2                                 (optional) \
!         -mi=missing_value                               (optional) \
!         -nobias                                         (optional) \
!         -stderror                                       (optional) \
!         -prob                                           (optional) \
!         -double                                         (optional) \
!         -bigfile                                        (optional) \
!         -hdf5                                           (optional) \
!         -tlimited                                       (optional)
!
! By default
! ----------
!
!        *-p=*
!         the *periodicity* is equal to ``1``
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-z=*
!         the whole vertical resolution associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-o=*
!         the *output_statistics_netcdf_file* is named ``stat_``\ *netcdf_variable*\ ``.nc``
!        *-m=*
!         the *output_mesh_mask_netcdf_file* is not created
!        *-vz=*
!         the variable with the same name as the third dimension, if any
!         (e.g., the associated coordinate variable)
!        *-z0=*
!         a value of ``0`` is assumed for the highest level when computing the third scale factor
!        *-sf=*
!         ``method2`` (e.g. it is assumed that each level or depth 
!         is located in the middle of each layer and 
!         the third scale factor is computed accordingly)
!        *-yl=*
!         it is assumed that the domain is the whole globe when computing the scale factors
!        *-mi=*
!         the *missing_value* for the statistics variables in the output NetCDF file is set to ``1.e+20``
!        *-nobias*
!         biased estimates of kurtosis and skewness coefficients are computed.
!         However, if **-nobias** is activated, unbiased estimates of kurtosis
!         and skewness are computed
!        *-stderror*
!         the standard errors of the kurtosis and skewness coefficients are not computed.
!         However, if **-stderror** is activated, these standard errors are computed
!        *-prob*
!         the significance levels of the kurtosis and skewness coefficients are not computed.
!         However, if **-prob** is activated, these significance levels are computed
!        *-double*
!         the statistic variables are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         statistics are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable for 
!      which statistics must be computed and the **-f=**\ *input_netcdf_file* 
!      argument specifies that this NetCDF variable must be extracted from
!      the NetCDF file *input_netcdf_file*.
!
!   #) The **-p=**\ *periodicity* argument gives the periodicity of the input data.
!      For example, with monthly data **-p=**\ ``12`` should be specified,
!      with yearly data **-p=**\ ``1`` may be used, etc.
!      By default, the *periodicity* is set to ``1``. Note that the
!      output NetCDF file will have *periodicity* time observations.    
!
!   #) If the **-x=**\ *lon1,lon2*, **-y=**\ *lat1,lat2* and **-z=**\ *level1,level2* arguments
!      are missing, statistics are computed for all the points of the 3-D grid-mesh
!      associated with the *netcdf_variable*.
!
!      The longitude, latitude or depth range must be a vector of two integers specifying
!      the first and last selected indices along each dimension.
!      The indices are relative to ``1``.
!      Negative values for *lon1* are not allowed.
!
!      Refer to :ref:`comp_mask_4d` for transforming geographical coordinates as indices before
!      using comp_stat_4d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, the whole time period
!      associated with the *netcdf_variable* is used to compute the statistics.
!
!      The selected time period is a vector of two integers specifying the
!      first and last time observations.
!      The indices are relative to ``1``.
!
!   #) It is assumed that the data has no missing values (excepted missing values associated with
!      a constant land-sea mask as indicated by a *missing_value* or *_FillValue* attribute).
!
!   #) If the **-m=**\ *output_mesh_mask_netcdf_file* argument is present and the
!      **-yl=** argument is missing it is assumed that the whole geographical 
!      domain associated with the NetCDF variable is the earth and that the 3-D grid-mesh
!      is regular or Gaussian when computing the scale factors.
!
!      If the domain is not the whole globe, the **-yl=** argument must be
!      specified, otherwise the first and last columns (elements) of the first two scale
!      factors are wrong.
!
!      The **-yl=** argument specifies the latitude limits of the domain in degrees
!      (*latl1* and *latl2* must be real numbers).
!
!   #) If the **-m=**\ *output_mesh_mask_netcdf_file* argument is present, the third scale factor is computed
!      with the help of the vertical coordinate variable (or the **-vz=**\ *name_of_the_vertical_netcdf_variable*)
!      if this vertical coordinate variable is strictly monotonic.
!
!   #) The **-z0=**\ *value_of_the_highest_level* argument specifies a value for the highest level or depth
!      in order to compute the first/last element of the third scale factor.
!      The default value is ``0``.
!
!   #) The **-sf=**\ *method_for_computing_the_third_scale_factor* argument allows to specify the method
!      for computing the third scale factor, if a mesh-mask NetCDF file is created:
!
!      * **-sf=**\ ``method1`` : the third scale factor is computed as the differences between successive
!        levels (or depths)
!
!      * **-sf=**\ ``method2`` : the third scale factor is computed by assuming that each level or depth
!        is located at the middle of the corresponding layer.
!
!   #) If the **-m=**\ *output_mesh_mask_netcdf_file* argument is present and if some scale factors can
!      not be computed, these scale factors are set to ``1``.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for the
!      variance (VAR), standard-deviation (STD), skewness (SKEW) and kurtosis (KURT)
!      variables in the *output_statistics_netcdf_file*.
!      If the **-mi=** argument is not specified and the NetCDF variable has a *missing_value*
!      or *_FillValue* attribute, the *missing_value* is set to ``1.e+20``.
!      This argument is not used if the NetCDF variable specified in the **-v=** argument
!      has no *missing_value* or *_FillValue* attribute (excepted for the skewness and kurtosis
!      statistics).
!
!   #) If **-nobias** is specified, unbiased estimates of skewness and kurtosis are computed.
!      If the **-nobias** argument is absent, the biased standard estimates are computed.
!
!   #) If **-stderror** is specified, the standard errors of skewness and kurtosis are computed.
!      If the **-stderror** argument is absent, the standard errors are not computed.
!
!   #) If **-prob** is specified, the significance levels of skewness and kurtosis are computed.
!      Moreover, the **-prob** argument implies also the **-stderror** argument, even if this
!      argument is not activated.
!      If the **-prob** argument is absent, the significance levels are not computed.
!
!   #) The **-double** argument specifies that the VAR, STD, SKEW and KURT variables must be stored
!      as double-precision floating point numbers instead of single-precision floating point
!      numbers in the *output_climatology_netcdf_file*.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) At least 4 observations by period, as determined from the **-p=**\ *periodicity* argument,
!      are required, otherwise the program will stop.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on the use of skewness and kurtosis coefficients in the climate literature see
!
!      * "Skewness, Kurtosis and Extreme Values of Northern Hemisphere Geopotential Heights",
!        by White, G., Monthly Weather Review, Vol. 108, 1446-1455, 1980. doi: `10.1175/1520-0493(1980)108<1446:SKAEVO>2.0.CO;2`_
!
!      * "The Normality of El Nino",
!        by Burgers, G., and Stephenson, D.B ,
!        Geophysical Research Letters, 26, 1027-1030, 1999. doi: `10.1029/1999GL900161`_
!
!      * "Impact of intra-daily SST variability on ENSO characteristics in a coupled model"
!        by Masson, S., et al., Climate Dynamics, Vol. 39, 681-707, 2012. doi: `10.1007/s00382-011-1247-2`_
!
!      * "Statistical Analysis in Climate Research",
!        by von Storch, H., and Zwiers, F.W., Cambridge University press,
!        Cambridge, UK, Chapter 2, 484 pp., 2002. ISBN: `9780521012300`_
!
! .. _`10.1175/1520-0493(1980)108<1446:SKAEVO>2.0.CO;2` : https://doi.org/10.1175/1520-0493(1980)108<1446:SKAEVO>2.0.CO;2
! .. _`10.1029/1999GL900161` : http://dx.doi.org/10.1029/1999GL900161
! .. _`10.1007/s00382-011-1247-2` : http://dx.doi.org/10.1007/s00382-011-1247-2
! .. _`9780521012300` : http://www.cambridge.org/gb/academic/subjects/earth-and-environmental-science/climatology-and-climate-change/statistical-analysis-climate-research?format=PB
!
! Outputs
! -------
!
!   comp_stat_4d creates an output NetCDF file that contains the univariate statistics and number
!   of observations of the input NetCDF variable, taking into account eventually the periodicity of the
!   data as determined by the **-p=**\ *periodicity* argument. The output NetCDF dataset contains the
!   following NetCDF variables (in the description below, nlon, nlat, and nlev are the length of the dimensions
!   of the input NetCDF variable) :
!
!   #) *netcdf_variable*\ _mean\ ``(periodicity,nlev,nlat,nlon)`` : the means for each point in
!      the time series of the 3-D grid-mesh associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _var\ ``(periodicity,nlev,nlat,nlon)`` : the variances for each point in
!      the time series of the 3-D grid-mesh associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _std\ ``(periodicity,nlev,nlat,nlon)`` : the standard-deviations for each point 
!      in the time series of the 3-D grid-mesh associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _skew\ ``(periodicity,nlev,nlat,nlon)`` : the skewness for each point 
!      in the time series of the 3-D grid-mesh associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _kurt\ ``(periodicity,nlev,nlat,nlon)`` : the kurtosis for each point 
!      in the time series of the 3-D grid-mesh associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _min\ ``(periodicity,nlev,nlat,nlon)`` : the minima for each point 
!      in the time series of the 3-D grid-mesh associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _max\ ``(periodicity,nlev,nlat,nlon)`` : the maxima for each point 
!      in the time series of the 3-D grid-mesh associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _skew_prob\ ``(periodicity,nlev,nlat,nlon)`` : the significance levels of the skewness
!      coefficients.
!
!      This variable is stored only if the  **-prob** argument has been specified when calling
!      comp_stat_4d.
!
!   #) *netcdf_variable*\ _kurt_prob\ ``(periodicity,nlev,nlat,nlon)`` : the significance levels of the kurtosis
!      coefficients.
!
!      This variable is stored only if the  **-prob** argument has been specified when calling
!      comp_stat_4d.
!
!   #) *netcdf_variable*\ _nobs\ ``(periodicity)`` : the number of observations used to compute
!      the statistics.
!
!   #) *netcdf_variable*\ _skew_se\ ``(periodicity)`` : the standard-errors of the skewness coefficients.
!
!      This variable is stored only if the **-stderror** or **-prob** arguments have been specified when calling
!      comp_stat_4d.
!
!   #) *netcdf_variable*\ _kurt_se\ ``(periodicity)`` : the standard-errors of the kurtosis coefficients.
!
!      This variable is stored only if the **-stderror** or **-prob** arguments have been specified when calling
!      comp_stat_4d.
!
!   All the statistics and associated probabilities, excepted the number of observations and the standard-errors
!   of the skewness and kurtosis coefficients are packed in fourdimensional variables whose first, second and third
!   dimensions are exactly the same as those associated with the input NetCDF variable *netcdf_variable*
!   even if you restrict the geographical domain with the **-x=**\ , **-y=** and **-z=** arguments. However,
!   outside the selected domain, these output NetCDF variables are filled with missing values.
!
!   Optionally, comp_stat_4d can also create an output mesh-mask NetCDF file that contains
!   the following NetCDF variables :
!
!   #) *netcdf_variable*\ _nmask\ ``(nlev,nlat,nlon)`` : a presence-absence or height-land-sea 3-D mask
!      associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _e1n\ ``(nlat,nlon)`` : the first scale factor associated with the
!      3-D grid-mesh of the input NetCDF variable.
!
!   #) *netcdf_variable*\ _e2n\ ``(nlat,nlon)`` : the second scale factor associated with the
!      3-D grid-mesh of the input NetCDF variable.
!
!   #) *netcdf_variable*\ _e3n\ ``(nlev,1,1)`` : the third scale factor associated with the
!      3-D grid-mesh of the input NetCDF variable.
!
!   Multiplying the two first scale factors together gives the surface of each cell in the 2-D grid-mesh
!   associated with the input NetCDF variable. Multiplying the three scale factors together gives
!   the volume (or a quantity proportional to the weight if the unit of the vertical coordinate
!   variable is in hPa) of each parcel in the 3-D grid-mesh associated with 
!   the input NetCDF variable.
!
! Examples
! --------
!
!   #)   For computing monthly statistics from the NetCDF
!        file ``ST7_1m_00101_20012_grid_T_votemper.nc``, which includes a NetCDF
!        variable ``votemper``, and store the results in a NetCDF file named
!        ``stat_ST7_1m_00101_20012_grid_T_votemper.nc``, use the following
!        command :
!
!        ::
!
!          $ comp_stat_4d \
!            -f=ST7_1m_0101_20012_grid_T_votemper.nc  \
!            -v=votemper \
!            -p=12 \
!            -o=stat_ST7_1m_00101_20012_grid_T_votemper.nc
!
!   #)   For computing monthly unbiased univariate statistics from the NetCDF
!        file ``vwnd.mon.mean.nc``, which includes a NetCDF variable ``vwnd``, and
!        store the results in a NetCDF file named ``stat_vwnd.nc`` and, in addition,
!        generate an associated mesh_mask_NetCDF_file named ``mesh_mask_wind_ncep2.nc``,
!        use the following command :
!
!        ::
!
!          $ comp_stat_4d \
!            -f=vwnd.mon.mean.nc \
!            -v=vwnd  \
!            -p=12 \
!            -nobias \
!            -m=mesh_mask_wind_ncep2.nc
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,    only : nf90_share
    use omp_lib,   only : omp_get_num_procs, omp_get_max_threads
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_var_dims, nf90_max_name, nf90_clobber, nf90_real4, nf90_real8,          &
                         nf90_int, nf90_short, nf90_byte, nf90_sizehint_default, nf90_open, nf90_nowrite, &
                         nf90_noerr, nf90_strerror, nf90_inq_varid, nf90_inquire_variable,                &
                         nf90_inquire_dimension, nf90_get_var, nf90_close, nf90_get_att,                  &
                         nf90_create, nf90_def_dim, nf90_def_var, nf90_inq_attname, nf90_copy_att,        &
                         nf90_put_att, nf90_inquire_attribute, nf90_rename_att, nf90_enddef,              &
                         nf90_put_var, nf90_create, nf90_set_fill, nf90_enddef, nf90_close,               &
                         nf90_unlimited, nf90_global, nf90_nofill
!
    use Statpack, only : i1b, i4b, stnd, extd, lgl, zero, half, one, two, three, five, six, c24, c90, c180,    &
                         c360, twopi, pio2, true, false, allocate_error, merror, arth, comp_unistat,           &
                         probn, kfix, kexp, kint, kchr, tolower, case_change, string_to_val
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, radius, def_missing_scal,                  &
                                 def_missing, get_listarg
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_stat_4d : ',                                           &   
!
     cmethod(2)=(/"method1","method2"/),                                                                  &
!
     cstat1(7)=(/"_mean","_var ","_std ","_skew","_kurt","_min ","_max "/) ,                              & 
!
     cstat2(7)=(/"Means of ","Vars of  ","STDs of  ","Skews of ","Kurts of ","Mins of  ","Maxs of  "/),   & 
!
     cstat3(2)=(/" of skewness"," of kurtosis"/)
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_stat, ncid_mask,                            & 
                     varid, nobsid, maskid, e1id, e2id, e3id, periodid,                   &
                     nlon, nlat, nlevel, nt, nperiod, natts, xtype, xtype_att,            &
                     i, j, j2, k, nyears,                                                 &
                     lon1, lon2, lon3, lat1, lat2, lat3,                                  &
                     time1, time2, time3, time2p1, level1, level2, level3, method,        &
                     output_type, output_mode
    integer(i4b),              dimension(3)                 :: dimids2, se_id, prob_id
    integer(i4b),              dimension(4)                 :: coor_natts, coor_type,             &
                                                               coor_id, coor_id2, startlec
    integer(i4b),              dimension(7)                 :: stat_id, stat_type
    integer(i4b),              dimension(nf90_max_var_dims) :: dimids, dimlens
    integer(i4b), allocatable, dimension(:)                 :: n_period
!
    integer(i1b)  :: kcode
!
    real(stnd), allocatable, dimension(:,:,:,:,:) :: stat
    real(stnd), allocatable, dimension(:,:,:,:)   :: data
    real(stnd), allocatable, dimension(:,:,:)     :: mask
    real(stnd), allocatable, dimension(:)         :: temp1, temp2
    real(stnd)                                    :: z0, tmp, tmp2, missing, missing2, latl1, latl2
!
    real(extd), allocatable, dimension(:)       :: record
!
    logical(lgl)                                :: first, last, nobias, stderror, prob, use_yl,   &
                                                   use_tlimited, use_missing, use_fillvalue,      &
                                                   create_hdf5
    logical(lgl), dimension(4)                  :: use_coor_variable
    logical(lgl), allocatable, dimension(:,:,:) :: maskl
!
    character(len=14)                          :: fmt
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, stat_name,      &
                                                  mask_name, att_name, z_name
    character(len=nf90_max_name), dimension(4) :: dim_names
!
#ifdef _PARALLEL_READ
!
    integer(i4b)  :: ncid2, varid2
!
    logical       :: test_par
!
#endif
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute univariate statistics from a fourdimensional variable '
        write (prtunit,*) ' extracted from a NetCDF dataset.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_stat_4d -f=input_netcdf_file'
        write (prtunit,*) '              -v=netcdf_variable'
        write (prtunit,*) '              -p=periodicity                                  (optional)'
        write (prtunit,*) '              -x=lon1,lon2                                    (optional)'
        write (prtunit,*) '              -y=lat1,lat2                                    (optional)'
        write (prtunit,*) '              -z=level1,level2                                (optional)'
        write (prtunit,*) '              -t=time1,time2                                  (optional)'
        write (prtunit,*) '              -o=output_statistics_netcdf_file                (optional)'
        write (prtunit,*) '              -m=output_mesh_mask_netcdf_file                 (optional)'
        write (prtunit,*) '              -vz=name_of_the_vertical_netcdf_variable        (optional)'
        write (prtunit,*) '              -z0=value_of_the_highest_level                  (optional)'
        write (prtunit,*) '              -sf=method_for_computing_the_third_scale_factor (optional : method1, method2)'
        write (prtunit,*) '              -yl=latl1,latl2                                 (optional)'
        write (prtunit,*) '              -mi=missing_value                               (optional)'
        write (prtunit,*) '              -nobias                                         (optional)'
        write (prtunit,*) '              -stderror                                       (optional)'
        write (prtunit,*) '              -prob                                           (optional)'
        write (prtunit,*) '              -double                                         (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '              -bigfile                                        (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '              -hdf5                                           (optional)'
#endif
        write (prtunit,*) '              -tlimited                                       (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-p= the periodicity is set to 1'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-z= the whole vertical resolution associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-o=stat_netcdf_variable.nc'
        write (prtunit,*) '-m= the mesh-mask netcdf file is not created'
        write (prtunit,*) '-vz= the variable with the same name as the third dimension, if any'
        write (prtunit,*) '-z0= a value of zero is assumed for the highest level'
        write (prtunit,*) '-sf=method2 (it is assumed that each level (depth) is in the middle of each layer)'
        write (prtunit,*) '-yl= it is assumed that the domain is the whole globe'
        write (prtunit,*) '-mi= the missing_value for the statistic variables is equal to 1.e+20'
        write (prtunit,*) '-nobias   : by default, biased estimates of kurtosis and skewness'
        write (prtunit,*) '            coefficients are computed'                            
        write (prtunit,*) '-stderror : by default, standard-errors of kurtosis and skewness'
        write (prtunit,*) '            coefficients are not computed'                            
        write (prtunit,*) '-prob     : by default, significance levels of kurtosis and skewness'
        write (prtunit,*) '            coefficients are not computed'                            
        write (prtunit,*) '-double   : by default, the statistics are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
#ifdef _PARALLEL_READ
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>21 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name  = " "
    var_name  = " "
    stat_name = " "
    mask_name = " "
    z_name    = " "
!
    history   = "comp_stat_4d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    nperiod     = 1
    method      = 2
    output_mode = nf90_clobber
    output_type = nf90_real4
!
    missing2    = def_missing
    z0          = zero
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    level1    = 1
    level2    = huge( level2 )
    time1     = 1
    time2     = huge( time2 )
!
    latl1     = zero
    latl2     = zero
!
    nobias       = false
    stderror     = false
    prob         = false
    use_yl       = false
    create_hdf5  = false
    use_tlimited = false
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-p=')
!
!               GET THE PERIODICITY OF THE DATA.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode==kint ) then
                    read( name(4:), fmt ) nperiod
                end if
!
                if ( kcode/=kint .or. nperiod<=0 )      &
                call merror( name_proc//"The periodicity is an integer > 0 " )
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
            case ('-z=')
!
!               GET THE VERTICAL DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "level1 and level2 are integer values", &
                                  "Wrong level definition",               &
                                  level1, level2 )
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR THE STATISTICS.
!
                stat_name = name(4:)
!
            case ('-m=')
!
!               GET THE NAME OF THE OPTIONAL MASK NETCDF DATASET.
!
                mask_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-vz=')
!
!                       GET THE NAME OF THE NETCDF VARIABLE ASSOCIATED WITH
!                       THE VERTICAL (DEPTH OR HEIGHT) DIMENSION.
!
                         z_name = name(5:)
!
                    case ('-z0=')
!
!                       GET THE VALUE ASSOCIATED WITH THE HIGHEST LEVEL.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kchr ) then
                            call merror( name_proc//"The -z0= argument is a numerical value" )
                        else
                            read( name(5:), fmt ) z0
                        end if
!
                    case ('-sf=')
!
!                       DETERMINE THE METHOD FOR COMPUTING THE THIRD SCALE FACTOR.
!
                        call case_change( name(5:), tolower )
!
                        select case ( name(5:) )
                            case ( cmethod(1) )
                                method = 1
                            case ( cmethod(2) )
                                method = 2
                            case default
                                call merror( name_proc//"Unknown option for computing the third scale factor" )
                        end select
!
                    case ('-yl=')
!
!                       GET THE DOMAIN LIMITS.
!
                        call get_listarg( name_proc, name(5:),                              &
                                          "latl1 and latl2 must be real numbers",           &
                                          "Wrong domain definition limits (latl1 latl2)",   &
                                          latl1, latl2 )
!
                        if ( abs(latl1)>c90 .or. abs(latl2)>c90 )        &
                 call merror( name_proc//" abs(latl1)>90. or abs(latl2)>90. in -yl= argument" )
!
                        use_yl = true
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR FOR THE STATISTIC VARIABLES.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing2
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case ('-nobias')
!
!                               COMPUTE UNBIASED ESTIMATES OF SKEWNESS AND KURTOSIS.
!
                                nobias = true
!
                            case ('-stderror')
!
!                               COMPUTE STANDARD ERRORS ESTIMATES OF SKEWNESS AND KURTOSIS.
!
                                stderror = true
!
                            case ('-prob')
!
!                               COMPUTE SIGNIFICANCE LEVELS OF SKEWNESS AND KURTOSIS.
!
                                prob = true
!
                            case ('-double')
!
!                               STORE STATISTICS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " ) call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " ) call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( stat_name==" " ) stat_name="stat_"//trim(var_name)//".nc"
!
    if ( prob ) stderror = true
!
!   OPEN THE NETCDF DATASET.
!
    k = nf90_sizehint_default
!
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!
    status = nf90_inq_varid( ncid, trim(var_name), varid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE EXTERNAL TYPE, THE DIMENSION LENGTHS AND NAMES, 
!   THE NUMBER OF ATTRIBUTS FOR THE NETCDF VARIABLE.
!
    status = nf90_inquire_variable( ncid, varid, xtype=xtype, ndims=i, dimids=dimids, natts=natts )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( i/=4 ) call merror( name_proc//"Input NETCDF variable is not fourdimensional" )
!
    do i = 1, 4
!
        status = nf90_inquire_dimension( ncid, dimids(i), name=dim_names(i), len=dimlens(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nlevel = dimlens(3)
    nt     = dimlens(4)
!
    if ( z_name/=" " )  dim_names(3) = z_name
!
    do i = 1, 4
!
!       GET THE COORDINATE VARIABLES IDS FROM THEIR NAME, IF ANY.
!
        status = nf90_inq_varid( ncid, trim(dim_names(i)), coor_id(i) )
        if ( status==nf90_noerr ) then
!
!           FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!           FOR THE COORDINATE VARIABLES.
!
            status = nf90_inquire_variable( ncid, coor_id(i), xtype=coor_type(i), ndims=k,    &
                                            dimids=dimlens, natts=coor_natts(i) )
            if ( status/=nf90_noerr )                                                         &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            use_coor_variable(i) = k==1 .and. dimlens(1)==dimids(i)
!
        else
!
            use_coor_variable(i) = false
!
        end if
!
        if ( .not.use_coor_variable(i) ) coor_type(i) = nf90_int
!
    end do
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
    if ( lon1<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
    lon2 = min( lon2, nlon )
    lon3 = lon2 - lon1 + 1
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
!   CHECK THE VERTICAL DEFINITION.
!
    level1 = max( level1, 1 )
    level2 = min( level2, nlevel )
    level3 = level2 - level1 + 1
    if ( level3<=0  )  call merror( name_proc//"Wrong level definition" )
!
!   CHECK THE TIME DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0 ) call merror( name_proc//"Wrong time definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    tmp    = time3/nperiod
    nyears = ceiling( tmp )
    if ( nyears<4 ) call merror( name_proc//"Less than four observations by period" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( data(lon3,lat3,level3,nperiod),        &
              stat(lon3,lat3,level3,7,nperiod),      &
              n_period(nperiod),                     &
              stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
#ifdef _PARALLEL_READ
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    test_par = i>1 .and. j>1 .and. nperiod>1
!
    if ( test_par ) then
!
        j2  = time1 - 1
!
!$OMP PARALLEL PRIVATE(k,j,i,status,startlec,ncid2,varid2)    &
!$OMP         ,SHARED(data,stat,n_period,nobias,j2,           &
!$OMP                 lon1,lon3,lat1,lat3,level1,level3,      &
!$OMP                 nyears,nperiod,time2)
!
!$OMP CRITICAL (opennc)
!
!       OPEN THE NETCDF DATASET.
!
        status = nf90_open( trim(fic_name), nf90_share, ncid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       GET THE VARIABLE ID FROM ITS NAME.
!
        status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
        startlec(1)  = lon1
        startlec(2)  = lat1
        startlec(3)  = level1
!
!$OMP DO SCHEDULE(STATIC)
        do j = 1, nperiod
!
            k            = j + j2
            startlec(4)  = k
!
!           FILL THE WORK ARRAY WITH THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid2, varid2, data(:lon3,:lat3,:level3,j:j), start=startlec )     
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COMPUTE UNIVARIATE STATISTICS FOR THE NETCDF VARIABLE.
!
            call comp_unistat( data(:lon3,:lat3,:level3,j:j), true, false, stat(:lon3,:lat3,:level3,:7,j) )
!
            do i = 2, nyears
!
                k = k + nperiod
!
                if ( k<=time2 ) then
!
                    startlec(4)  = k
!
!                   FILL THE WORK ARRAY WITH THE NETCDF VARIABLE.
!
                    status = nf90_get_var( ncid2, varid2, data(:lon3,:lat3,:level3,j:j), start=startlec )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   COMPUTE UNIVARIATE STATISTICS FOR THE NETCDF VARIABLE.
!
                    call comp_unistat( data(:lon3,:lat3,:level3,j:j), false, false, stat(:lon3,:lat3,:level3,:7,j) )
!
                end if
!
            end do
!
!           FINISH COMPUTATIONS OF UNIVARIATE STATISTICS FOR THE NETCDF VARIABLE.
!
            call comp_unistat( data(:lon3,:lat3,:level3,1:0), false, true, stat(:lon3,:lat3,:level3,:7,j),    &
                               xnobs=n_period(j), nobias=nobias )
        end do
!$OMP END DO
!
!$OMP CRITICAL (closenc)
!
      status = nf90_close( ncid2 )
      if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
    else
!
#endif
!
!       FILL THE WORK ARRAY WITH THE NETCDF VARIABLE.
!
        startlec(1:4)  = (/ lon1, lat1, level1, time1 /)
!
        status = nf90_get_var( ncid, varid, data(:lon3,:lat3,:level3,:nperiod), start=startlec )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        do j = 1, nperiod
!
!           COMPUTE UNIVARIATE STATISTICS FOR THE NETCDF VARIABLE.
!
            call comp_unistat( data(:lon3,:lat3,:level3,j:j), true, false, stat(:lon3,:lat3,:level3,:7,j) )
!
        end do
!
        k       = time1
        time2p1 = time2 + 1
!
        do i = 2, nyears
!
            k            = k + nperiod
            j2           = min( time2p1-k,nperiod  )
            startlec(4)  = k
!
!           FILL THE WORK ARRAY WITH THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid, varid, data(:lon3,:lat3,:level3,1:j2), start=startlec )     
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            do j = 1, j2
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE NETCDF VARIABLE.
!
                call comp_unistat( data(:lon3,:lat3,:level3,j:j), false, false, stat(:lon3,:lat3,:level3,:7,j) )
!
            end do
!
        end do
!
        do j = 1, nperiod
!
!           FINISH COMPUTATIONS OF UNIVARIATE STATISTICS FOR THE NETCDF VARIABLE.
!
            call comp_unistat( data(:lon3,:lat3,:level3,1:0), false, true, stat(:lon3,:lat3,:level3,:7,j),    &
                               xnobs=n_period(j), nobias=nobias )
        end do
!
#ifdef _PARALLEL_READ
!
    end if
!
#endif
!
!   MULTIPLY THE VARIANCES BY THE SCALE FACTOR IF REQUIRED.
!
    status = nf90_get_att( ncid, varid, "scale_factor", tmp )
    if ( status==nf90_noerr )  stat(:lon3,:lat3,:level3,2,:nperiod) = stat(:lon3,:lat3,:level3,2,:nperiod)*tmp
!
!   ADJUST SKEWNESS AND KURTOSIS IF REQUIRED.
!
    if (.not.nobias ) then
!
        do i = 1, nperiod
!
            tmp  = real( n_period(i), stnd )
            tmp2 = ( tmp - one )/tmp
!
            where ( stat(:lon3,:lat3,:level3,2,i)>zero )
!
                stat(:lon3,:lat3,:level3,4,i) = ((sqrt(tmp2))**3)*stat(:lon3,:lat3,:level3,4,i)
                stat(:lon3,:lat3,:level3,5,i) = (tmp2**2)*( stat(:lon3,:lat3,:level3,5,i) + three ) - three
!
            elsewhere
!
                stat(:lon3,:lat3,:level3,4,i) = missing2
                stat(:lon3,:lat3,:level3,5,i) = missing2
!
            end where
!
        end do
!
    else
!
        where ( stat(:lon3,:lat3,:level3,2,:nperiod)<=zero )
!
            stat(:lon3,:lat3,:level3,4,:nperiod) = missing2
            stat(:lon3,:lat3,:level3,5,:nperiod) = missing2
!
        end where
!
    end if
!
!   FIND THE MISSING INDICATOR VALUE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "missing_value", missing )
!
    use_missing = status==nf90_noerr
!
    if (.not.use_missing ) then
!
!       TRY THE _FillValue ATTRIBUTE IF THE MISSING ATTRIBUTE IS ABSENT.
!
        status = nf90_get_att( ncid, varid, "_FillValue", missing )
        use_fillvalue = status==nf90_noerr
!
    else
!
        use_fillvalue = false
!
    end if
!
    if ( use_missing .or. use_fillvalue ) then
!
!       ALLOCATE LOGICAL MASK ARRAY.
!
        allocate( maskl(lon3,lat3,level3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       APPLY THE LAND-SEA MASK TO THE STATISTIC ARRAY.
!
        maskl(:lon3,:lat3,:level3) = data(:lon3,:lat3,:level3,1)==missing
!
        if ( any(maskl(:lon3,:lat3,:level3)) ) then
!
            do i = 1, nperiod
!
                where( maskl(:lon3,:lat3,:level3) )
!
                    stat(:lon3,:lat3,:level3,1,i) = missing
                    stat(:lon3,:lat3,:level3,2,i) = missing2
                    stat(:lon3,:lat3,:level3,3,i) = missing2
                    stat(:lon3,:lat3,:level3,4,i) = missing2
                    stat(:lon3,:lat3,:level3,5,i) = missing2
                    stat(:lon3,:lat3,:level3,6,i) = missing
                    stat(:lon3,:lat3,:level3,7,i) = missing
!
                end where
!
            end do
!
!           DEALLOCATE LOGICAL MASK ARRAY, IF NOT NEEDED LATER.
!
            if ( mask_name==" " ) deallocate( maskl )
!
        else
!
!           DEALLOCATE LOGICAL MASK ARRAY.
!
            deallocate( maskl )
!
        end if
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE STATISTICS.
!
    if ( create_hdf5 ) then
        status = nf90_create( trim(stat_name), cmode=output_mode, ncid=ncid_stat )
    else
        k = nf90_sizehint_default
        status = nf90_create( trim(stat_name), cmode=output_mode, ncid=ncid_stat, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    dimlens(3)   = nlevel
    if ( use_tlimited ) then
        dimlens(4)   = nperiod
    else
        dimlens(4)   = nf90_unlimited
    end if
!
    do i = 1, 4
!
        status = nf90_def_dim( ncid_stat, name=trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 4
!
        status = nf90_def_var( ncid_stat, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                          &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    stat_type(1)   = xtype
    stat_type(2:5) = output_type
    stat_type(6:7) = xtype
!
    do i = 1, 7
!
        status = nf90_def_var( ncid_stat, trim(var_name)//trim(cstat1(i)),  stat_type(i),   &
                               dimids=dimids(:4), varid=stat_id(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_stat, trim(var_name)//"_nobs", nf90_int, dimids=dimids(4), varid=nobsid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( stderror ) then
!
        do j = 1, 2
!
            j2 = j + 3
!
            status = nf90_def_var( ncid_stat, trim(var_name)//trim(cstat1(j2))//"_se", output_type,   &
                                   dimids=dimids(4), varid=se_id(j) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            if ( prob ) then
!
                status = nf90_def_var( ncid_stat, trim(var_name)//trim(cstat1(j2))//"_prob", output_type,    &
                                       dimids=dimids(:4), varid=prob_id(j) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end if
!
        end do
!
    end if
!
    status = nf90_def_var( ncid_stat, "period", nf90_int, dimids=dimids(4), varid=periodid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 4
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_stat, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    status = nf90_put_att( ncid_stat, periodid, "long_name", "Period axis" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, periodid, "short_name", "Period" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    do i = 1, natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       COPY AND MODIFY USEFUL ATTRIBUTES.
!
        select case ( att_name )
!
            case ( 'missing_value' )
!
!               COPY missing_value ATTRIBUTE, IF REQUIRED.
!
                do j = 1, 7
!
                    if ( j>=2 .and. j<=5 ) cycle
!
                    status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(j) )
                    if ( status/=nf90_noerr )                                 &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                    status = nf90_inquire_attribute(  ncid, varid, att_name, xtype=xtype_att )
                    if ( status/=nf90_noerr )                                             &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                    if ( xtype==xtype_att ) then
!
                        status = nf90_rename_att( ncid_stat, stat_id(j), att_name, "_FillValue" )
                        if ( status/=nf90_noerr )                                 &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                        status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(j) )
                        if ( status/=nf90_noerr )                                 &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                    end if
!
                end do
!
            case ( '_FillValue' )
!
!               COPY fill_value ATTRIBUTE, IF REQUIRED.
!
                if ( use_fillvalue ) then
!
                    do j = 1, 7
!
                        if ( j>=2 .and. j<=5 ) cycle
!
                        status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(j) )
                        if ( status/=nf90_noerr )                                 &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                        status = nf90_rename_att( ncid_stat, stat_id(j), att_name, "missing_value" )
                        if ( status/=nf90_noerr )                                 &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                        status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(j) )
                        if ( status/=nf90_noerr )                                 &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                    end do
!
                end if
!
            case ( 'units' )
!
!               COPY units ATTRIBUTE.
!
                do j = 1, 7
!
                    if ( j==4 .or. j==5 ) cycle
!
                    if ( j==2 ) then
!
                        name = " "
                        status = nf90_get_att( ncid, varid, att_name, name )
!
                        if ( status==nf90_noerr ) then
                           status = nf90_put_att( ncid_stat, stat_id(j), att_name, "("//trim(name)//")**2 "  )
                            if ( status/=nf90_noerr )                   &
                            call merror( name_proc//trim( nf90_strerror( status ) ) )
                        end if
!
                    else
!
                        status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(j) )
                        if ( status/=nf90_noerr )                                 &
                        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                    end if
!
                end do
!
            case ( 'scale_factor' )
!
!               COPY scale_factor ATTRIBUTE.
!
                do j = 1, 7
!
                    if ( j==4 .or. j==5 ) cycle
!
                    status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(j) )
                    if ( status/=nf90_noerr )                                 &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                end do
!
            case ( 'long_name', 'short_name' )
!
#ifdef _USE_NETCDF36
!
! Due to a bug in the NETCDF 3.5 library, the following commented section of code can't be used.
! However, if you are using NETCDF 3.5.1 or higher, you can use this section safely.
!
                status = nf90_inquire_attribute( ncid, varid, att_name, len=j )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               CHECK LENGTH OF STRING ATTRIBUTE.
!
                if ( j <= nchar ) then
                    status = nf90_get_att( ncid, varid, att_name, name )
                    if ( status/=nf90_noerr )                                             &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
                    name(j+1:nchar) = " "
                else
                    name = var_name
                end if
#else
!
! This section of code must be used if you are using NETCDF 3.5.
!
                name(1:nchar) = " "
                status = nf90_get_att( ncid, varid, att_name, name )
                if ( status/=nf90_noerr )                                             &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
#endif
!
!               MODIFY USEFUL ATTRIBUTES.
!
                do j = 1, 7
!
                    status = nf90_put_att( ncid_stat, stat_id(j), att_name, cstat2(j)//trim(name)  )
                    if ( status/=nf90_noerr )                                             &
                    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                end do
!
            case ( 'add_offset' )
!
!               COPY add_offset ATTRIBUTE.
!
                status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(1) )
                if ( status/=nf90_noerr )                                             &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(6) )
                if ( status/=nf90_noerr )                                             &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_copy_att(  ncid, varid, att_name, ncid_stat, stat_id(7) )
                if ( status/=nf90_noerr )                                             &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end select
!
    end do
!
    if ( use_missing .or. use_fillvalue ) then
        k = 2
    else
        k = 4
    end if
!
    if ( output_type==nf90_real4 ) then
!
        do j = k, 5
!
            status = nf90_put_att( ncid_stat, stat_id(j), "missing_value", real( missing2, sp )  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, stat_id(j), "_FillValue", real( missing2, sp )  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        if ( prob ) then
!
            do j = 1, 2
!
                status = nf90_put_att( ncid_stat, prob_id(j), "missing_value", real( missing2, sp )  )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_stat, prob_id(j), "_FillValue", real( missing2, sp )  )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_stat, prob_id(j), "valid_min", real( zero, sp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_stat, prob_id(j), "valid_max", real( one, sp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
        end if
!
    else
!
        do j = k, 5
!
            status = nf90_put_att( ncid_stat, stat_id(j), "missing_value", real( missing2, dp )  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, stat_id(j), "_FillValue", real( missing2, dp )  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        if ( prob ) then
!
            do j = 1, 2
!
                status = nf90_put_att( ncid_stat, prob_id(j), "missing_value", real( missing2, dp )  )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_stat, prob_id(j), "_FillValue", real( missing2, dp )  )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_stat, prob_id(j), "valid_min", real( zero, dp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_stat, prob_id(j), "valid_max", real( one, dp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
        end if
!
    end if
!
    status = nf90_put_att( ncid_stat, nobsid, "long_name", "Number of observations" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nobsid, "short_name", "Nobs" )
    if ( status/=nf90_noerr ) call merror( name_proc// trim( nf90_strerror( status ) ) )
!
    if ( stderror ) then
!
        do j = 1, 2
!
            status = nf90_put_att( ncid_stat, se_id(j), "long_name", "Standard error"//cstat3(j) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, se_id(j), "short_name", "SE"//cstat3(j) )
            if ( status/=nf90_noerr ) call merror( name_proc// trim( nf90_strerror( status ) ) )
!
        end do
!
        if ( prob ) then
!
            status = nf90_put_att( ncid_stat, prob_id(1), "long_name", "Prob > !SKEW! under h0: SKEWh0=0"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, prob_id(1), "short_name", "Probability"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, prob_id(2), "long_name", "Prob > !KURT! under h0: KURTh0=0"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, prob_id(2), "short_name", "Probability"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_stat, nf90_global, "title", trim(var_name)//" Statistics" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "lon1_"//trim(var_name), lon1 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "lon2_"//trim(var_name), lon2 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "lat1_"//trim(var_name), lat1 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "lat2_"//trim(var_name), lat2 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "level1_"//trim(var_name), level1 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "level2_"//trim(var_name), level2 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_stat )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    dimlens(4)   = nperiod
!
!   ALLOCATE WORK ARRAYS.
!
    i = max( nlon, nlat, nlevel, nperiod )
    allocate( record(i), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   WRITE COORDINATE VARIABLES.
!
    do j = 1, 4
!
        j2 = dimlens(j)
!
        if ( use_coor_variable(j) ) then
!
            status = nf90_get_var( ncid, coor_id(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_stat, coor_id2(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            status = nf90_put_var( ncid_stat, coor_id2(j), arth(1_i4b,1_i4b,j2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( record )
!
    status = nf90_put_var( ncid_stat, periodid, arth(1_i4b,1_i4b,nperiod) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE STATISTIC VARIABLES.
!
    startlec(1:4) = (/ lon1, lat1, level1, 1_i4b /)
!
    do j = 1, 7
!
        data(:lon3,:lat3,:level3,:nperiod) = stat(:lon3,:lat3,:level3,j,:nperiod)
!
        if ( stat_type(j) == nf90_int .or. stat_type(j) == nf90_short .or. stat_type(j) == nf90_byte ) then 
           status = nf90_put_var( ncid_stat, stat_id(j), nint(data(:lon3,:lat3,:level3,:nperiod)),   &
                                  start=startlec )
        else
           status = nf90_put_var( ncid_stat, stat_id(j),      data(:lon3,:lat3,:level3,:nperiod) ,   &
                                  start=startlec )
        end if
!
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   WRITE NUMBER OF OBSERVATIONS.
!
    status = nf90_put_var( ncid_stat, nobsid, n_period )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   COMPUTE STANDARD ERRORS AND SIGNIFICANCE LEVELS FOR SKEWNESS AND KURTOSIS IF REQUIRED.
!
    if ( stderror ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( temp1(nperiod), temp2(nperiod),  stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        do j = 1, 2
!
            temp1(:nperiod) = real( n_period(:nperiod), stnd )
!
!           COMPUTE STANDARD ERRORS.
!
            if ( j==1 ) then
!
                temp2(:nperiod) = ( six*temp1(:nperiod)*(temp1(:nperiod)-one))  /   &
                                  ( (temp1(:nperiod)-two)*(temp1(:nperiod)+one)*(temp1(:nperiod)+three)  )
!
            else
!
                temp2(:nperiod) = ( c24*temp1(:nperiod)*(temp1(:nperiod)-one)**2)  /   &
             ( (temp1(:nperiod)-three)*(temp1(:nperiod)-two)*(temp1(:nperiod)+three)*(temp1(:nperiod)+five) )
!
            end if
!
            temp1(:nperiod) = sqrt( temp2(:nperiod) )
!
            if ( prob ) then
!
                j2 = j + 3
!
!               COMPUTE SIGNIFICANCE LEVELS OF SKEWNESS AND KURTOSIS.
!
                do i = 1, nperiod
!
                    where( stat(:lon3,:lat3,:level3,j2,i) == missing2 )
                        stat(:lon3,:lat3,:level3,1,i) = zero
                    elsewhere
                        stat(:lon3,:lat3,:level3,1,i) = abs( stat(:lon3,:lat3,:level3,j2,i) )*(one/temp1(i))
                    end where
!
                    do k = 1, level3
!
                        data(:lon3,:lat3,k,i)  = two*probn( stat(:lon3,:lat3,k,1,i), true )
!
                    end do
!
                    where( stat(:lon3,:lat3,:level3,j2,i) == missing2 )    &
                           data(:lon3,:lat3,:level3,i) = missing2
!
                end do
!
!               WRITE SIGNIFICANCE LEVELS OF SKEWNESS AND KURTOSIS.
!
                status = nf90_put_var( ncid_stat, prob_id(j), data(:lon3,:lat3,:level3,:nperiod) , start=startlec )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end if
!
!           WRITE STANDARD ERRORS OF SKEWNESS AND KURTOSIS.
!
            status = nf90_put_var( ncid_stat, se_id(j), temp1(:nperiod) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp1, temp2 )
!
    end if
!
!   CONSTRUCT THE MESH-MASK DATASET IF REQUIRED.
!
    if ( mask_name/=" " ) then
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( data, stat, n_period  )
!
!       ALLOCATE WORK ARRAYS.
!
        i = max( nlevel, nlat+1 )
        allocate( mask(nlon,nlat,nlevel), temp1(i), temp2(i), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       CONSTRUCT THE MASK.
!
        mask(:nlon,:nlat,:nlevel) = zero
!
        if ( allocated(maskl) ) then
!
            where( .not.maskl(:lon3,:lat3,:nlevel) )  mask(lon1:lon2,lat1:lat2,level1:level2) = one
!
!           DEALLOCATE LOGICAL MASK ARRAY.
!
            deallocate( maskl )
!
        else
!
            mask(lon1:lon2,lat1:lat2,level1:level2) = one
!
        end if
!
!       CREATE THE NEW NETCDF DATASET FOR THE MASK.
!
        status = nf90_create( trim(mask_name), cmode=nf90_clobber, ncid=ncid_mask )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_set_fill( ncid_mask, nf90_nofill, i )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       DEFINE THE DIMENSIONS FOR THE MESH-MASK NETCDF DATASET.
!
        do i = 1, 3
!
            status = nf90_def_dim( ncid_mask, name=trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
!       DEFINE DUMMY DIMENSIONS FOR THE THIRD SCALE FACTORS.
!
        dim_names(:2) = (/ "x_a", "y_a" /)
!
        do i = 1, 2
!
            status = nf90_def_dim( ncid_mask, name=trim(dim_names(i)), len=1, dimid=dimids2(i) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        dimids2(3) = dimids(3)
!
!       DEFINE THE VARIABLES FOR THE MESH-MASK NETCDF DATASET.
!
        status = nf90_def_var( ncid_mask, "nmask", nf90_real4, dimids=dimids(:3), varid=maskid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_mask, "e1n", nf90_real8, dimids=dimids(:2), varid=e1id )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_mask, "e2n", nf90_real8, dimids=dimids(:2), varid=e2id )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_mask, "e3n", nf90_real8, dimids=dimids2(:3), varid=e3id )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       ASSIGN ATTRIBUTES FOR THE MESH-MASK NETCDF DATASET.
!
        status = nf90_put_att( ncid_mask, maskid, "valid_min", zero )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "valid_max", one )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       ADD LONGITUDE / LATITUDE / LEVEL  BOUNDS ATTRIBUTES.
!
        status = nf90_put_att( ncid_mask, maskid, "lon1_Eastern_limit", lon1 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lon2_Western_limit", lon2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lat1_Southern_limit", lat1 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "lat2_Northern_limit", lat2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "level1_First_level", level1 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, maskid, "level2_Last_level", level2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, e1id, "units", "meters" )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, e1id, "missing_value", def_missing_scal )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, e2id, "units", "meters" )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, e2id, "missing_value", def_missing_scal )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( use_coor_variable(3) ) then
            status = nf90_copy_att(  ncid, coor_id(3), "units", ncid_mask, e3id )
        end if
!
        if ( status/=nf90_noerr .or. .not.use_coor_variable(3)) then
            status = nf90_put_att( ncid_mask, e3id, "units", "undefined" )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
        status = nf90_put_att( ncid_mask, e3id, "missing_value", def_missing_scal )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       SET GLOBAL ATTRIBUTES.
!
        status = nf90_put_att( ncid_mask, nf90_global, "title", trim(var_name)//" Mesh-mask" )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_mask, nf90_global, "history", trim(history) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       LEAVE DEFINE MODE FOR THE MESH-MASK NETCDF DATASET.
!
        status = nf90_enddef( ncid_mask )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       WRITE VARIABLES FOR THE MESH-MASK NETCDF DATASET.
!
        status = nf90_put_var( ncid_mask, maskid, mask )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       CONSTRUCT THE SCALE FACTORS.
!
!       START BY THE THIRD SCALE FACTOR.
!
        if ( use_coor_variable(3) ) then
!
            status = nf90_get_var( ncid, coor_id(3), temp1(:nlevel), count=(/ nlevel /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!           VERIFY THAT THE DEPTH OR HEIGHT VARIABLE IS STRICTLY MONOTONIC
!           AND THAT THERE IS AT LEAST TWO LEVELS.
!
            use_coor_variable(3) = ( all(temp1(:nlevel-1)>temp1(2:nlevel)) .or.               &
                                     all(temp1(:nlevel-1)<temp1(2:nlevel)) ) .and. nlevel/=1

        end if
!
        if ( use_coor_variable(3) ) then
!
            if ( method==1 ) then
!
!               COMPUTE THE THIRD SCALE FACTOR AS THE DIFFERENCES BETWEEN SUCESSIVE LEVELS.
!
                if ( temp1(1)>temp1(2) ) then
!
                    temp2(1:nlevel-1) = temp1(1:nlevel-1) - temp1(2:nlevel)
!
!                   IT IS ASSUMED HERE THAT THE LEVEL nlevel+1 IS THE VALUE OF z0 (ZERO, BY DEFAULT).
!
                    temp2(nlevel)     = temp1(nlevel) - z0
!
                else
!
                    temp2(2:nlevel) = temp1(2:nlevel) - temp1(1:nlevel-1)
!
!                   IT IS ASSUMED HERE THAT THE LEVEL 0 IS THE VALUE OF z0 (ZERO, BY DEFAULT).
!
                    temp2(1)        = temp1(1) - z0
!
                end if
!
            else
!
!               COMPUTE THE THIRD SCALE FACTOR AS IF EACH LEVEL IS IN THE MIDDLE OF THE CORRESPONDING LAYER.
!
                temp2(2:nlevel-1) = half*abs( temp1(3:nlevel) - temp1(1:nlevel-2)  )
!
                if ( temp1(1)>temp1(2) ) then
!
!                   FIRST ELEMENT OF SCALE FACTOR IS COMPUTED AS HALF OF THE STANDARD THICKNESS LAYER.
!
                    temp2(1)          = half*( temp1(1) - temp1(2) )
!
!                   IT IS ASSUMED HERE THAT THE LEVEL nlevel+1 IS THE VALUE OF z0 (ZERO, BY DEFAULT).
!
                    temp2(nlevel)     = half*( temp1(nlevel-1) -  z0 )
!
                else
!
!                   IT IS ASSUMED HERE THAT THE LEVEL 0 IS THE VALUE OF z0 (ZERO, BY DEFAULT).
!
                    temp2(1)          = half*( temp1(2) -  z0 )
!
!                   LAST ELEMENT OF SCALE FACTOR IS COMPUTED AS HALF OF THE STANDARD THICKNESS LAYER.
!
                    temp2(nlevel)     =  half*( temp1(nlevel) - temp1(nlevel-1) )
!
                end if
!
            end if

        else

            temp2(:nlevel) = def_missing_scal

        end if
!
        status = nf90_put_var( ncid_mask, e3id, temp2(:nlevel),                            &
                               start=(/1,1,1/), count=(/1,1,nlevel/) ) 
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       NOW DEFINE THE FIRST AND SECOND SCALE FACTORS.
!
        if ( use_coor_variable(2) ) then
!
!           GET THE COORDINATE VARIABLE (latitude).
!
            status = nf90_get_var( ncid, coor_id(2), temp1(:nlat) )     
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )

        else
!
!           TRY nav_lat VARIABLE, IF COORDINATE VARIABLE IS MISSING.
!
            status = nf90_inq_varid( ncid, "nav_lat", coor_id(2) )
!
            if ( status==nf90_noerr ) then
!
                use_coor_variable(2) = true
!
!               FIND THE DIMENSION LENGTHS FOR THE nav_lat VARIABLE.
!
                status = nf90_inquire_variable( ncid, coor_id(2), ndims=i,  dimids=dimids )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                if ( i/=2 ) call merror( name_proc//"Coordinate variable nav_lat is not twodimensional" )
!
                do i = 1, 2
                    status = nf90_inquire_dimension( ncid, dimids(i), len=dimlens(i) )
                    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
                end do
!
                if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )    &
                call merror( name_proc//"Shape disagreement with the coordinate variable nav_lat" )
!
                status = nf90_get_var( ncid, coor_id(2), temp1(:nlat), count=(/ 1, nlat /) )     
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )

            else
!
                use_coor_variable(2) = false

            end if

        end if
!
        if ( use_coor_variable(2) ) then
!
!           VERIFY THAT THE COORDINATE VARIABLE IS STRICTLY MONOTONIC.
!
            last                 = all(temp1(:nlat-1)<temp1(2:nlat))
            use_coor_variable(2) = all(temp1(:nlat-1)>temp1(2:nlat)) .or. last

        end if
!
        if ( use_coor_variable(2) ) then
!
            if ( use_yl ) then
!
!               CHECK THE SPECIFIED LATITUDE LIMITS AGAINST THE COORDINATE VARIABLE.
!
                first = merge( latl1>temp1(1) .or. latl2<temp1(nlat),    &
                               latl1<temp1(1) .or. latl2>temp1(nlat),    &
                               last )
                if ( first ) call merror( name_proc//"Wrong specification for the -yl= argument" )
!
                temp2(1)      = two*latl1
                temp2(nlat+1) = two*latl2
!
            else
!
!               IT IS ASSUMED HERE THAT THE NETCDF VARIABLE COVERS THE WHOLE GLOBE.
!
                temp2(1)      = merge( -c180, c180, temp1(1)<zero )
                temp2(nlat+1) = -temp2(1)
!
            end if
!
            temp2(2:nlat)  = ( temp1(:nlat-1) + temp1(2:nlat) )
            temp2(:nlat+1) = temp2(:nlat+1)*( pio2/c360 )
!
!           CONSTRUCT THE FIRST SCALE FACTOR.
!
            temp1(:nlat) = ( twopi*radius/real(nlon,stnd) )*abs( cos( temp2(2:nlat+1)+temp2(1:nlat) ) )
            mask(:nlon,:nlat,1) = spread( temp1(:nlat), dim=1, ncopies=nlon )
!
            status = nf90_put_var( ncid_mask, e1id, mask(:nlon,:nlat,1) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           CONSTRUCT THE SECOND SCALE FACTOR.
!
            temp1(:nlat) = ( two*radius )*abs( temp2(2:nlat+1)-temp2(1:nlat) )
            mask(:nlon,:nlat,1) = spread( temp1(:nlat), dim=1, ncopies=nlon )
!
        else
!
!           NO COORDINATE VARIABLES CONSTRUCT DUMMY SCALE FACTORS.
!
            mask(:nlon,:nlat,1) = def_missing_scal
!
            status = nf90_put_var( ncid_mask, e1id, mask(:nlon,:nlat,1) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
        status = nf90_put_var( ncid_mask, e2id, mask(:nlon,:nlat,1) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( mask, temp1, temp2  )
!
!       CLOSE THE NETCDF DATASET.
!
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( data, stat, n_period )

    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_stat )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM comp_stat_4d
! ===========================
!
end program comp_stat_4d
