program comp_stl_1d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_stl_1d:
!
! ===========
! comp_stl_1d
! ===========
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 13/09/2018
!
! Purpose
! =======
!
! Decompose a time series extracted from a unidimensional variable in
! a NetCDF dataset into seasonal and trend components by the Seasonal-Trend
! decomposition procedure based on Loess (STL). The STL procedure is a powerful
! statistical technique for describing a discrete time series [Cleveland_etal]_ [Terrayc]_ .
! In the STL procedure, the analyzed time series is decomposed into three terms:
!
!              **X**\ (*t*\ ) = **T**\ (*t*\ ) + **A**\ (*t*\ ) + **R**\ (*t*\ )
!
! where *t* refers to a time index, the **T** term is used to quantify the trend and low-frequency variations in the time series,
! the **A** term describes the harmonic component (e.g. diurnal or seasonal cycle)  and its modulation through time
! and, finally, the **R** term contains the residual component.
!
! All the terms are estimated through a sequence of applications of locally weighted regression or low-order polynomial (e.g. Loess)
! to data windows whose length is chosen by the user. The STL procedure is an iterative process, which may
! be interpreted as a frequency filter directly applicable to a non-stationary unidimensional time series including harmonic
! components  [Cleveland_etal]_ . Other important features of STL are the specification of
! the amounts of seasonal and trend smoothing according to the choice of the user, the ability to produce robust estimates
! of the trend and harmonic components that are not distorted by aberrant behavior in the data and the stationarity of
! the **R** time series. 
!
! This procedure returns the seasonal and trend components as estimated by the STL procedure and, optionally,
! the residuals or the sum of the residual and trend components in a NetCDF dataset, for the time series
! associated with a NetCDF variable.
!
! If the NetCDF variable is tridimensional or fourdimensional use :ref:`comp_stl_3d` or :ref:`comp_stl_4d`, respectively 
! instead of comp_stl_1d. It the time series have no seasonal (or diurnal) cycle,
! use :ref:`comp_trend_1d` instead of comp_stl_1d in order to estimate the trend component by the Loess procedure.
!
! The exact meaning and default values for most of the optional parameters of comp_stl_1d are exactly the same as in the original
! procedure described by [Cleveland_etal]_ and the user is referred to this publication for further
! details on the STL procedure.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_stl_1d \
!         -f=input_netcdf_file \
!         -v=input_netcdf_variable \
!         -p=periodicity \
!         -ns=seasonal_smoother_length     (ns) \
!         -t=time1,time2                             (optional) \
!         -a=type_of_analysis                        (optional : stl, residual,  \
!                                                                residual_trend) \
!         -o=output_netcdf_file                      (optional) \
!         -nt=trend_smoother_length        (nt)      (optional) \
!         -nl=low_pass_smoother_length     (nl)      (optional) \
!         -isdeg=seasonal_smoother_degree  (isdeg)   (optional : 0, 1 ) \
!         -itdeg=trend_smoother_degree     (itdeg)   (optional : 0, 1, 2 ) \
!         -ildeg=low_pass_smoother_degree  (ildeg)   (optional : 0, 1, 2 ) \
!         -nsjump=seasonal_skipping_value  (nsjump)  (optional) \
!         -ntjump=trend_skipping_value     (ntjump)  (optional) \
!         -nljump=low_pass_skipping_value  (nljump)  (optional) \
!         -maxit=max_robustness_iterations (maxit)   (optional) \
!         -ni=number_of_loops              (ni)      (optional) \
!         -sms=seasonal_smoothing_factor             (optional) \
!         -smt=trend_smoothing_factor                (optional) \
!         -robust                                    (optional) \
!         -double                                    (optional) \
!         -hdf5                                      (optional) \
!         -tlimited                                  (optional)
!
! By default
! ----------
!
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-a=*
!         the *type_of_analysis* is set to ``stl``. This means that the residual times series
!         are not computed and not stored in the *output_netcdf_file*
!        *-o=*
!         the *output_netcdf_file* is named  ``stl_``\ *netcdf_variable*\ ``.nc``
!        *-nt=*
!         the *trend_smoother_length* is set to the smallest odd integer greater than or equal to ``(1.5*periodicity) / (1-(1.5/ns))``
!        *-nl=*
!         the *low_pass_smoother_length* is set to the smallest odd integer greater than or equal to *periodicity*
!        *-isdeg=*
!         the *seasonal_smoother_degree* is set to ``1``
!        *-itdeg=*
!         the *trend_smoother_degree* is set to ``1``
!        *-ildeg=*
!         the *low_pass_smoother_degree* is set to the *trend_smoother_degree*
!        *-nsjump=*
!         the *seasonal_skipping_value* is set to ``ns/10``
!        *-ntjump=*
!         the *trend_skipping_value* is set to ``nt/10``
!        *-nljump=*
!         the *low_pass_skipping_value* is set to ``nl/10``
!        *-maxit=*
!         the *max_robustness_iterations* is set to ``15``
!        *-ni=*
!         the *number_of_loops* is set to ``1`` if -robust is specified and ``2`` otherwise
!        *-sms=*
!         no smoothing is applied to the seasonal component
!        *-smt=*
!         no smoothing is applied to the trend component
!        *-robust*
!         robustness iterations are not used, by default
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable from which the time series must
!      be decomposed and the **-f=**\ *input_netcdf_file* argument specifies that this NetCDF variable must
!      be extracted from the NetCDF file, *input_netcdf_file*.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used to decompose the time series.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) The **-p=**\ *periodicity* argument gives the periodicity of the input data.
!      For example, with monthly data **-p=**\ ``12`` should be specified.
!      The *periodicity* must be greater than ``2``. If your time series do not
!      have a periodic component, e.g. if *periodicity* is equal to ``1`` , use :ref:`comp_trend_1d` instead of comp_stl_1d.
!
!   #) The **-ns=** argument specifies the length of the seasonal smoother, ``ns``. The value
!      of ``ns`` should be an odd integer greater than or equal to ``3``; a value greater than ``6``
!      is recommended. As the value of the **-ns=** argument  increases  the  values  of  the
!      seasonal component at a given point in the seasonal cycle (e.g., January values of 
!      a monthly series with a yearly cycle) become smoother. However, the value of ``ns`` has no direct
!      effect on the smoothness of successive values of the seasonal component.
!
!   #) The **-nt=** argument specifies the length of the trend smoother, ``nt``. The value of ``nt``
!      should be an odd integer greater than or equal to ``3``; a value of ``nt`` between ``1.5``\ *\ *periodicity*
!      and ``2``\ *\ *periodicity* is  recommended. As ``nt`` increases the values of the trend component become
!      smoother.
!
!   #) The **-nl=** argument specifies the length of the low-pass smoother, ``nl``. The value
!      of ``nl`` should  be an odd integer greater than or equal to ``3``; the
!      smallest odd integer greater  than  or  equal  to  *periodicity*  is
!      recommended.
!
!   #) The **-isdeg=** argument specifies the degree of locally-fitted polynomial in seasonal
!      smoothing.  The value is ``0`` or ``1``.
!
!   #) The **-itdeg=** argument specifies the degree of locally-fitted polynomial in trend
!      smoothing.  The value is ``0``, ``1`` or ``2``.
!
!   #) The **-ildeg=** argument specifies the degree of locally-fitted polynomial in low-pass
!      smoothing.  The value is ``0``, ``1`` or ``2``.
!
!   #) The **-nsjump=** argument specifies the skipping  value  for  seasonal  smoothing.
!      The seasonal  smoother  skips  ahead  ``nsjump``  points and then
!      linearly interpolates in between.  The  value  of  ``nsjump``
!      should  be  a  positive  integer; if ``nsjump`` is set to ``1``, a seasonal
!      smooth is calculated  at  all points in the time series.   To  make  the
!      procedure  run  faster, a reasonable choice for ``nsjump`` is
!      ``10``\ % or ``20``\ % of ``ns``. The default value is ``ns/10``.
!
!   #) The **-ntjump=** argument specifies the skipping value for trend smoothing.
!      The default value is ``nt/10``.
!
!   #) The **-nljump=** argument specifies the skipping value for the low-pass filter.
!      The default value is ``nl/10``.
!
!   #) The **-ni=** argument specifies the number of loops for updating the seasonal
!      and trend components. The value of ``ni`` should be a positive integer.
!      If the data are well behaved without outliers, then robustness iterations
!      are not needed.  In this case do not use the **-robust** argument, and set ``ni`` between ``2`` and ``5``
!      depending on how much security you want that the seasonal-trend looping converges.
!      If  outliers are present then use the **-robust** argument and set ``ni`` to ``1`` or ``2``.
!
!   #) The **-a=** argument specifies if the residuals from the trend and seasonal components
!      are stored in the output NetCDF file. If:
!
!      * **-a=**\ ``stl`` means that the residuals are not computed
!
!      * **-a=**\ ``residual`` means that the residuals are computed and stored
!
!      * **-a=**\ ``residual_trend`` means that the residual and trend components are summed and the result is stored.
!
!      The default is **-a=**\ ``stl`` , e.g. the residuals are not stored. Note that in all cases, the
!      seasonal and trend components are computed and stored in the output NetCDF file.
!
!   #) If **-robust** is specified, robustness iterations are  carried out until
!      convergence of both seasonal and trend components or with a maximum of *max_robustness_iterations*
!      iterations as specified by the **-maxit=** argument. Convergence occurs if the maximum
!      changes in individual seasonal and  trend fits are less than ``1``\ % of the component's range
!      after the previous iteration.
!
!   #) **-sms=**\ *seasonal_smoothing_factor* means that the seasonal component extracted from
!      the netcdf_variable (e.g. the **-v=** argument) must be smoothed with
!      a moving average of approximately ``2``\ *\ *seasonal_smoothing_factor*\ +\ ``1`` terms.
!      The smoothing is applied before estimating the residuals (e.g. if **-a=** argument
!      is set to ``residual`` or ``residual_trend`` ).
!      *seasonal_smoothing_factor* must be a strictly positive integer (e.g. greater than ``0``).
!      Note that this additional smoothing is not implemented in the original STL procedure.
!
!   #) **-smt=**\ *trend_smoothing_factor* means that the trend component extracted from
!      the netcdf_variable (e.g. the **-v=** argument) must be smoothed with
!      a moving average of approximately ``2``\ *\ *trend_smoothing_factor*\ +\ ``1`` terms.
!      The smoothing is applied before estimating the residuals (e.g. if **-a=** argument
!      is set to ``residual`` or ``residual_trend`` ).
!      *trend_smoothing_factor* must be a strictly positive integer (e.g. greater than ``0``).
!      Note that this additional smoothing is not implemented in the original STL procedure.
!
!   #) The **-double** argument specifies that, the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the data has no missing values.
!
!   #) It the time series have no seasonal cycle, use :ref:`comp_trend_1d` instead of comp_stl_1d.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
!   #) For more details on the STL procedure and examples of use in the climate literature, see
!
!      * "A Seasonal-Trend Decomposition Procedure Based on Loess",
!        by R.B. Cleveland, W.S. Cleveland, J.E. McRae and I. Terpennings,
!        1990, Journal of Official Statistics, 6, 3-73. http://www.jos.nu/Articles/abstract.asp?article=613
!
!      * "Southern Hemisphere extra-tropical forcing: A new paradigm for El Nino-Southern Oscillation"
!        by P. Terray, 2011, Climate Dynamics, Vol. 36:2171-2199, doi: `10.1007/s00382-010-0825-z`_
!
! .. _`10.1007/s00382-010-0825-z` : http://dx.doi.org/10.1007/s00382-010-0825-z
!
! Outputs
! -------
!
!   comp_stl_1d creates an output NetCDF file that contains the trend and harmonic components extracted from the
!   time series associated with the input NetCDF variable. The output NetCDF dataset contains the
!   following NetCDF variables (in the description below, ntime is the selected number of time observations) :
!
!   #) *netcdf_variable*\ _trend\ ``(ntime)`` : the trend component for the time series 
!      associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _seasonal\ ``(ntime)`` : the harmonic component for
!      the time series associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _residual\ ``(ntime)`` : the residual component for
!      the time series associated with the input NetCDF variable.
!
!      This variable is stored only if the **-a**\ =\ ``residual`` argument has been specified when calling
!      comp_stl_1d.
!
!   #) *netcdf_variable*\ _residual_trend\ ``(ntime)`` : the sum of the residual and trend components
!      for the time series associated with the input NetCDF variable.
!
!      This variable is stored only if the **-a**\ =\ ``residual_trend`` argument has been specified when calling
!      comp_stl_1d.
!
! Examples
! --------
!
!   #)   For computing a STL decomposition from the unidimensional NetCDF variable called ``ts`` extracted from the
!        file ``ts_cnrm_cm3.picntrl_monthly_glob.nc``, which includes a global monthly time series, and store
!        the results in the NetCDF file ``ts_cnrm_cm3.picntrl_monthly_glob_stl.nc``, use the following
!        command :
!
!        ::
!
!          $ comp_stl_1d \
!            -f=ts_cnrm_cm3.picntrl_monthly_glob.nc \
!            -v=ts \
!            -p=12 \
!            -ns=35 \
!            -nt=240 \
!            -a=residual \
!            -robust \
!            -o=ts_cnrm_cm3.picntrl_monthly_glob_stl.nc
!
!-
! ================================================================================================
!
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,   only : nf90_max_var_dims, nf90_max_name, nf90_real4, nf90_real8, nf90_clobber, &
                         nf90_open, nf90_nowrite, nf90_noerr, nf90_strerror, nf90_get_var,       &
                         nf90_get_att, nf90_create, nf90_clobber, nf90_set_fill, nf90_nofill,    &
                         nf90_unlimited, nf90_def_dim, nf90_def_var, nf90_inq_attname,           &
                         nf90_copy_att, nf90_put_att, nf90_inquire_attribute, nf90_global,       &
                         nf90_enddef, nf90_put_var, nf90_close
!
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, one, ten, half,     & 
                         c1_5, c0_9, kint, tolower, case_change, string_to_val,      &
                         allocate_error, merror, arth, comp_stlez, comp_smooth
!
    use NCSTAT_Utilities, only : prtunit, nchar, get_listarg, get_variable_dimcoords
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_stl_1d : ',                       &
!
            cmethod(3)=(/"stl           ","residual      ", "residual_trend"/),     &
!
            cstltype(2)=(/"not_robust","robust    "/)
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_stl,                                        &
                     varid, residid, trendid, seasonid, coor_id2,                         &
                     i, ntime, time1, time2, time3, period,                               &
                     method, stltype, output_type, output_mode, maxiter, ns, nt, nl, ni,  &
                     isdeg, itdeg, ildeg, nsjump, ntjump, nljump,                         &
                     trend_smooth_factor, season_smooth_factor
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),     allocatable  :: y, trend, season
    real(stnd)                                 :: tmp
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: use_tlimited, robust, comp_resid, ok
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, stl_name, att_name
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Decompose a time series extracted from a unidimensional variable in ' 
        write (prtunit,*) ' a NetCDF dataset into seasonal and trend components. It returns the '
        write (prtunit,*) ' components and, optionally, the residuals in a NetCDF dataset. '
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_stl_1d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -p=periodicity'
        write (prtunit,*) '        -ns=seasonal_smoother_length'
        write (prtunit,*) '        -t=time1,time2                   (optional)'
        write (prtunit,*) '        -a=type_of_analysis              (optional : stl, residual, residual_trend)'
        write (prtunit,*) '        -o=output_netcdf_file            (optional)'
        write (prtunit,*) '        -nt=trend_smoother_length        (optional)'
        write (prtunit,*) '        -nl=low_pass_smoother_length     (optional)'
        write (prtunit,*) '        -isdeg=seasonal_smoother_degree  (optional : 0, 1)'
        write (prtunit,*) '        -itdeg=trend_smoother_degree     (optional : 0, 1, 2)'
        write (prtunit,*) '        -ildeg=low_pass_smoother_degree  (optional : 0, 1, 2)'
        write (prtunit,*) '        -nsjump=seasonal_skipping_value  (optional)'
        write (prtunit,*) '        -ntjump=trend_skipping_value     (optional)'
        write (prtunit,*) '        -nljump=low_pass_skipping_value  (optional)'
        write (prtunit,*) '        -maxit=max_robustness_iterations (optional)'
        write (prtunit,*) '        -ni=number_of_loops              (optional)'
        write (prtunit,*) '        -sms=seasonal_smoothing_factor   (optional)'
        write (prtunit,*) '        -smt=trend_smoothing_factor      (optional)'
        write (prtunit,*) '        -robust                          (optional)'
        write (prtunit,*) '        -double                          (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                            (optional)'
#endif
        write (prtunit,*) '        -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-a=stl'
        write (prtunit,*) '-o=stl_netcdf_variable.nc'
        write (prtunit,*) '-nt= smallest odd integer >= (1.5*periodicity) / (1-(1.5/ns))'
        write (prtunit,*) '-nl= smallest odd integer >=periodicity'
        write (prtunit,*) '-isdeg=1'
        write (prtunit,*) '-itdeg=1'
        write (prtunit,*) '-ildeg= ildeg is set to itdeg'
        write (prtunit,*) '-nsjump=ns/10'
        write (prtunit,*) '-ntjump=nt/10'
        write (prtunit,*) '-nljump=nl/10'
        write (prtunit,*) '-maxit=15'
        write (prtunit,*) '-ni= the number_of_loops is set to 1 if -robust is specified and 2 otherwise'
        write (prtunit,*) '-sms= by default, no smoothing is applied to the seasonal component'
        write (prtunit,*) '-smt= by default, no smoothing is applied to the trend component'
        write (prtunit,*) '-robust   : by default, no robustness iterations are to be used'
        write (prtunit,*) '-double   : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<4 .or. narg>23 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name   = " "
    var_name   = " "
    stl_name   = " "
!
    history = "comp_stl_1d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
!
    ns            = -1
    nt            = -1
    nl            = -1
    period        = -1
    isdeg         = 1
    itdeg         = 1
    ildeg         = -1
    nsjump        = -1
    ntjump        = -1
    nljump        = -1
    ni            = -1
    method        = 1
    stltype       = 1
    maxiter       = 15
!
    comp_resid    = false
    robust        = false
!
    season_smooth_factor = 0
    trend_smooth_factor  = 0
!
    time1     = 1
    time2     = huge( time2 )
!
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                stl_name = name(4:)
!
            case ('-p=')
!
!               GET THE PERIODICITY OF THE DATA.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode==kint ) then
                    read( name(4:), fmt ) period
                end if
!
                if ( kcode/=kint .or. period<=1 )      &
                call merror( name_proc//"The periodicity is an integer > 1 " )
!
            case ('-a=')
!
!               DETERMINE IF THE RESIDUALS ARE REQUIRED.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method     = 2
                        comp_resid = true
                    case ( cmethod(3) )
                        method     = 3
                        comp_resid = true
                    case default
                        call merror( name_proc//"Unknown analysis option" )
                end select
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-ns=')
!
!                       GET THE LENGTH OF THE SEASONAL SMOOTHER.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) ns
                        end if
                        if ( kcode/=kint .or. ns<=2  )                           &
                        call merror( name_proc//"The -ns= argument is an integer > 2" )
!
                    case ('-nt=')
!
!                       GET THE LENGTH OF THE TREND SMOOTHER.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) nt
                        end if
                        if ( kcode/=kint .or. nt<=2  )                           &
                        call merror( name_proc//"The -nt= argument is an integer > 2" )
!
                    case ('-nl=')
!
!                       GET THE LENGTH OF THE LOW PASS SMOOTHER.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) nl
                        end if
                        if ( kcode/=kint .or. nl<=2  )                           &
                        call merror( name_proc//"The -nl= argument is an integer > 2" )
!
                    case ('-ni=')
!
!                       GET THE NUMBER OF ITERATIONS FOR THE INNER LOOP OF STLEZ.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) ni
                        end if
                        if ( kcode/=kint .or. ni<=0  )                           &
                        call merror( name_proc//"The -ni= argument is a strictly positive integer" )
!
                    case default
!
                        call case_change( name(4:5), tolower )
!
                        select case ( name(1:5) )
!
                            case ('-sms=')
!
!                               DETERMINE THE SEASONAL SMOOTHING FACTOR.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                     read( name(6:), fmt ) season_smooth_factor
                                end if
                                if ( season_smooth_factor<=0 )        &
         call merror( name_proc//"The -sms= argument is a strictly positive integer value " )
!
                            case ('-smt=')
!
!                               DETERMINE THE TREND SMOOTHING FACTOR.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                     read( name(6:), fmt ) trend_smooth_factor
                                end if
                                if ( trend_smooth_factor<=0 )        &
         call merror( name_proc//"The -smt= argument is a strictly positive integer value " )
!
                            case default
!
                                call case_change( name(6:7), tolower )
!
                                select case ( name(1:7) )
!
                                    case ('-isdeg=')
!
!                                       GET THE DEGREE OF LOCALLY-FITTED POLYNOMIAL IN SEASONAL SMOOTHING.
!
                                        call string_to_val( name(8:), kcode, fmt )
                                        if ( kcode==kint ) then
                                            read( name(8:), fmt ) isdeg
                                        end if
                                        if ( kcode/=kint .or. (isdeg/=0 .and. isdeg/=1)  )                &
                                        call merror( name_proc//"The -isdeg= argument must be 0 or 1." )
!
                                    case ('-itdeg=')
!
!                                       GET THE DEGREE OF LOCALLY-FITTED POLYNOMIAL IN TREND SMOOTHING.
!
                                        call string_to_val( name(8:), kcode, fmt )
                                        if ( kcode==kint ) then
                                            read( name(8:), fmt ) itdeg
                                        end if
                                        if ( kcode/=kint .or. (itdeg<0 .or. itdeg>2)  )                &
                                        call merror( name_proc//"The -itdeg= argument must be 0, 1 or 2." )
!
                                    case ('-ildeg=')
!
!                                       GET THE DEGREE OF LOCALLY-FITTED POLYNOMIAL IN THE LOW PASS FILTER.
!
                                        call string_to_val( name(8:), kcode, fmt )
                                        if ( kcode==kint ) then
                                            read( name(8:), fmt ) ildeg
                                        end if
                                        if ( kcode/=kint .or. (ildeg<0 .or. ildeg>2)  )                &
                                        call merror( name_proc//"The -ildeg= argument must be 0, 1 or 2." )
!
                                    case ('-maxit=')
!
!                                       GET THE MAXIMUM NUMBER OF ROBUSTNESS ITERATIONS.
!
                                        call string_to_val( name(8:), kcode, fmt )
                                        if ( kcode==kint ) then
                                            read( name(8:), fmt ) maxiter
                                        end if
                                        if ( kcode/=kint .or. maxiter<=0  )                &
                                 call merror( name_proc//"The -maxit= argument is a strictly positive integer" )
!
                                    case default
!
                                        select case ( name(1:8) )
!
                                            case ('-nsjump=')
!
!                                               GET THE SEASONAL SKIPPING VALUE.
!
                                                call string_to_val( name(9:), kcode, fmt )
                                                if ( kcode==kint ) then
                                                    read( name(9:), fmt ) nsjump
                                                end if
                                                if ( kcode/=kint .or. nsjump<=0  )                &
                                 call merror( name_proc//"The -nsjump= argument is a stricly positive integer" )
!
                                            case ('-ntjump=')
!
!                                               GET THE TREND SKIPPING VALUE.
!
                                                call string_to_val( name(9:), kcode, fmt )
                                                if ( kcode==kint ) then
                                                    read( name(9:), fmt ) ntjump
                                                end if
                                                if ( kcode/=kint .or. ntjump<=0  )                &
                                 call merror( name_proc//"The -ntjump= argument is a stricly positive integer" )
!
                                            case ('-nljump=')
!
!                                               GET THE LOW PASS FILTER SKIPPING VALUE.
!
                                                call string_to_val( name(9:), kcode, fmt )
                                                if ( kcode==kint ) then
                                                    read( name(9:), fmt ) nljump
                                                end if
                                                if ( kcode/=kint .or. nljump<=0  )                &
                                 call merror( name_proc//"The -nljump= argument is a stricly positive integer" )
!
                                            case default
!
                                                call case_change( name(8:), tolower )
!
                                                select case ( name )
!
                                                    case ('-robust')
!
!                                                       ROBUST ITERATIONS ARE REQUIRED.
!
                                                        robust  = true
                                                        stltype = 2
!
                                                    case ('-double')
!
!                                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                                        output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                                                    case ('-hdf5')
!
!                                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                                        output_mode = nf90_hdf5
#endif
!
                                                    case ('-tlimited')
!
!                                                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                                                        use_tlimited = true
!
                                                    case default
!
!                                                       WRONG ARGUMENT.
!
                                                        call merror( name_proc//"Wrong argument" )
!
                                                end select
!
                                        end select
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS AND SET DEFAULT VALUES.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( period==-1_i4b )                                                &
    call merror( name_proc//"The periodicity argument is missing" )
!
    if ( ns==-1_i4b )                                                    &
    call merror( name_proc//"The -ns= argument is missing" )
    if ( mod(ns,2_i4b)==0_i4b) ns = ns + 1_i4b
!
    if ( nt==-1_i4b ) then
        nt = ( c1_5*real(period,stnd) )/( one - c1_5/real(ns,stnd) ) + half
        nt = max( 3_i4b, nt )
    end if
    if ( mod(nt,2_i4b)==0_i4b) nt = nt + 1_i4b
!
    if ( nl==-1_i4b ) then
        nl = period
    end if
    if ( mod(nl,2_i4b)==0_i4b) nl = nl + 1_i4b
!
    if ( ildeg==-1_i4b )   ildeg = itdeg
!
    if ( nsjump==-1_i4b ) then
        nsjump = max( 1_i4b, int(real(ns,stnd)/ten + c0_9,i4b) )
    end if
!
    if ( ntjump==-1_i4b ) then
        ntjump = max( 1_i4b, int(real(nt,stnd)/ten + c0_9,i4b) )
    end if
!
    if ( nljump==-1_i4b ) then
        nljump = max( 1_i4b, int(real(nl,stnd)/ten + c0_9,i4b) )
    end if
!
    if ( ni==-1_i4b ) then
        if ( robust ) then
            ni = 1_i4b
        else
            ni = 2_i4b
        end if
    end if
!
    if ( stl_name==" " ) stl_name="stl_"//trim(var_name)//".nc"
!
!   OPEN THE NETCDF DATASETS.
!
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                    &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names, &
                                  dimlens_out=dimlens, coor_id_out=coor_id,           &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type, &
                                  use_coor_variable_out=use_coor_variable             )
!
    if ( count(dimlens/=0)/=1 )           &
    call merror( name_proc//"Input NETCDF variable is not unidimensional" )
!
    ntime = dimlens(1)
!
!   CHECK THE TIME DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, ntime )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    if ( period>time3 )  call merror( name_proc//"-t=, -p= options mismatch" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( y(time3), trend(time3), season(time3), stat=status   )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   FIRST READ THE NETCDF VARIABLE.
!
    status = nf90_get_var( ncid, varid, y(:time3), start=(/ time1 /), count=(/ time3 /) )  
!
!   FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "missing_value", tmp )
    if ( status==nf90_noerr ) then
        if ( any( y(:time3)==tmp ) )      &
        call merror( name_proc//"Missing values are not allowed" )
    end if
!
!   FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "scale_factor", tmp )
    if ( status==nf90_noerr ) y(:time3) = y(:time3)*tmp
!
!   FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "add_offset", tmp )
    if ( status==nf90_noerr )  y(:time3) = y(:time3) + tmp
!
!   DECOMPOSES THE TIME SERIES INTO SEASONAL AND TREND COMPONENTS.
!
    call comp_stlez( y(:time3), period, ns, isdeg, itdeg, robust, season(:time3), trend(:time3),  &
                     ni=ni, nt=nt, nl=nl, ildeg=ildeg,                                            &
                     nsjump=nsjump, ntjump=ntjump, nljump=nljump,                                 &
                     maxiter=maxiter, ok=ok )
!
!   SMOOTH THE SEASONAL COMPONENT IF REQUIRED.
!
    if ( season_smooth_factor/=0 ) then
        call comp_smooth( season(:time3), season_smooth_factor )
    end if
!
!   SMOOTH THE TREND COMPONENT IF REQUIRED.
!
    if ( trend_smooth_factor/=0 ) then
        call comp_smooth( trend(:time3), trend_smooth_factor )
    end if
!
!   COMPUTE RESIDUALS IF REQUIRED.
!
    if ( method==2 ) then
         y(:time3) = y(:time3) - season(:time3) - trend(:time3)
    else if ( method==3 ) then
         y(:time3) = y(:time3) - season(:time3)
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE STL ANALYSIS.
!
    status = nf90_create( stl_name, cmode=output_mode, ncid=ncid_stl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_stl, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    if ( use_tlimited ) then
        dimlens(1)   = time3
    else
        dimlens(1)   = nf90_unlimited
    end if
!
    status = nf90_def_dim( ncid_stl, name=trim(dim_names(1)), len=dimlens(1), dimid=dimids(1) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    status = nf90_def_var( ncid_stl, trim(dim_names(1)), coor_type(1),                &
                           dimids=dimids(1), varid=coor_id2 )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    status = nf90_def_var( ncid_stl, trim(var_name)//"_trend", output_type, dimids=dimids(1), varid=trendid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_stl, trim(var_name)//"_seasonal", output_type, dimids=dimids(1), varid=seasonid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( comp_resid ) then
        status = nf90_def_var( ncid_stl, trim(var_name)//"_"//cmethod(method), output_type,    &
                               dimids=dimids(1), varid=residid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    if ( use_coor_variable(1) ) then
!
        do i = 1, coor_natts(1)
!
            status = nf90_inq_attname( ncid, coor_id(1), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE TIME VARIABLE.
!
            status = nf90_copy_att(  ncid, coor_id(1), att_name, ncid_stl, coor_id2 )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    else
!
        status = nf90_put_att( ncid_stl, coor_id2, "long_name", "time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_stl, coor_id2, "short_name", "time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_put_att( ncid_stl, trendid, "long_name", "trend of "//trim(var_name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, trendid, "short_name", "trend"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( trend_smooth_factor/=0_i4b ) then
        status = nf90_put_att( ncid_stl, trendid, "smoothing_factor", trend_smooth_factor )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_put_att( ncid_stl, seasonid, "long_name", "seasonal of "//trim(var_name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, seasonid, "short_name", "seasonal"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( season_smooth_factor/=0_i4b ) then
        status = nf90_put_att( ncid_stl, seasonid, "smoothing_factor", season_smooth_factor )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( comp_resid ) then
!
        status = nf90_put_att( ncid_stl, residid, "long_name", trim(cmethod(method))//" of "//trim(var_name)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_stl, residid, "short_name", trim(cmethod(method))  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_inquire_attribute( ncid, varid, "units" )
    if ( status==nf90_noerr ) then
!
        status = nf90_copy_att(  ncid, varid, "units", ncid_stl, trendid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_copy_att(  ncid, varid, "units", ncid_stl, seasonid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( comp_resid ) then
            status = nf90_copy_att(  ncid, varid, "units", ncid_stl, residid )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_stl, nf90_global, "title", "STL analysis of "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "periodicity", period )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "number_of_loops", ni )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "seasonal_smoother_length", ns )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "trend_smoother_length", nt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "low_pass_smoother_length", nl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "seasonal_smoother_degree", isdeg )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "trend_smoother_degree", itdeg )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "low_pass_smoother_degree", ildeg )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "seasonal_skipping_value", nsjump )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "trend_skipping_value", ntjump )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "low_pass_skipping_value", nljump )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( robust ) then
!
        if ( ok ) then
            fmt = "1 on 1"
        else
            fmt = "0 on 1"
        end if
!
        status = nf90_put_att( ncid_stl, nf90_global, "convergence_factor", fmt(1:6) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_stl, nf90_global, "max_robustness_iterations", maxiter )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_put_att( ncid_stl, nf90_global, "stltype", cstltype(stltype) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stl, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_stl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW  NETCDF DATASET.
!
!   WRITE TIME VARIABLE FOR THE NEW  NETCDF DATASET.
!
    if ( use_coor_variable(1) ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( record(time3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        status = nf90_get_var( ncid, coor_id(1), record, start=(/ time1 /) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_stl, coor_id2, record )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( record )
!
    else
!
        status = nf90_put_var( ncid_stl, coor_id2, arth(time1,1_i4b,time3 ) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   WRITE THE SEASONAL AND TREND COMPONENTS.
!
    status = nf90_put_var( ncid_stl, trendid, trend(:time3) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_var( ncid_stl, seasonid, season(:time3) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )

    if ( comp_resid ) then
!
!       WRITE THE RESIDUALS IF REQUIRED.
!
        status = nf90_put_var( ncid_stl, residid, y(:time3) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( y, trend, season )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_stl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_stl_1d
! ==========================
!
end program comp_stl_1d
