program pack_masked_data_3d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _pack_masked_data_3d:
!
! ===================
! pack_masked_data_3d
! ===================
!
! Authors
! =======
!
! Eric Maisonnave (CERFACS) and Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 05/01/2018
!
! Purpose
! =======
!
! Reduce the size of a tridimensional NetCDF variable extracted from a NetCDF dataset by
! storing only unmasked data associated with a constant land-sea mask in an output NetCDF
! dataset. The associated scale factors of the 2-D grid can also be stored in packed form in
! the output NetCDF dataset at the user option.
!
! The output NetCDF variable stored in packed form in the output NetCDF file can be unpacked
! with :ref:`unpack_masked_data_3d`.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ pack_masked_data_3d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -o=output_netcdf_file \
!         -x=lon1,lon2                      (optional) \
!         -y=lat1,lat2                      (optional) \
!         -t=time1,time2                    (optional) \
!         -g=grid_type                      (optional : n, t, u, v, w, f) \
!         -3d                               (optional) \
!         -scalfac                          (optional) \
!         -bigfile                          (optional) \
!         -hdf5                             (optional) \
!         -tlimited                         (optional)
!
! By default
! ----------
!
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input NetCDF variable is assumed to be regular or Gaussian
!        *-3d*
!         the packed data are defined as a twodimensional NetCDF variable. However,
!         if **-3d** is activated, the packed data is defined as an tridimensional NetCDF variable
!         but with one dummy dimension (e.g. with a length equal to ``1`` )
!        *-scalfac*
!         The scale factors NetCDF variables in the *input_mesh_mask_netcdf_file* are not packed
!         and copied to the output NetCDF file. If the **-scalfac** argument is specified,
!         the associated packed scale factors NetCDF variables are copied and a
!         dummy mask NetCDF variable is also written in the output NetCDF file
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable which
!      must be packed and the **-f=**\ *input_netcdf_file* argument
!      specifies that this NetCDF variable must be extracted from the NetCDF file
!      *input_netcdf_file*.
!
!   #) The argument **-m=**\ *input_mesh_mask_netcdf_file* specifies that the land-sea mask
!      to apply to the *netcdf_variable* must be read from the *input_mesh_mask_netcdf_file*.
!
!      This *input_mesh_mask_netcdf_file* may be created by :ref:`comp_clim_3d`
!      if the 2-D grid-mesh is regular or gaussian.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing, the geographical
!      domain used in the packing operation is determined from the attributes of the input mesh-mask
!      NetCDF variable named *grid_type*\ ``mask``\  (e.g. the lon1_Eastern_limit, lon2_Western_limit, lat1_Southern_limit
!      and lat2_Northern_limit attributes), which is read from the input NetCDF file *input_mesh_mask_netcdf_file*\ .
!      If these attributes are missing, the whole geographical domain associated with the *netcdf_variable*
!      is used to construct the packed NetCDF variable in the output dataset.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\  .
!      Negative values are not allowed for *lon1*.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices or generating an
!      appropriate mesh-mask before using pack_masked_data_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used to construct the packed NetCDF variable.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the NetCDF variable
!      is from an experiment with the ORCA or NEMO model.
!
!      If **-g=** is set to ``n``, it is assumed that the 2-D grid-mesh is regular or Gaussian.
!
!      The **-g=** argument is also used to determine the name of the NetCDF variable which
!      contains the 2-D mesh-mask in the *input_mesh_mask_netcdf_file* (e.g. this variable is named
!      *grid_type*\ ``mask``).
!
!   #) The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file* dataset) and the
!      land-sea mask (in the *input_mesh_mask_netcdf_file* dataset) must agree.
!
!   #) The **-3d** argument specifies that the packed data must be stored as
!      a tridimensional NetCDF variable with one dummy dimension in the output NetCDF file.
!      By default, the packed data are stored as a twodimensional NetCDF variable.
!
!   #) The **-scalfac** argument specifies that the scale factors from the *input_mesh_mask_netcdf_file*
!      must be packed and stored in the output NetCDF file if they exist. If the **-scalfac** argument
!      is specified, a dummy mask is also written in the output NetCDF file. This allows further processing
!      of the packed data by other NCSTAT procedures. Note that the *grid_type* of the packed mask and scale
!      factors in the output NetCDF file is set to ``n``.
!
!      By default, the packed scale factors are not stored.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
! Examples
! --------
!
!   #)   For packing a NetCDF variable ``sosstsst`` extracted from the NetCDF dataset
!        ``ST7_1m_0101_20012_grid_T_sosstsst.nc`` with the help of a land-sea mask extracted
!        from the NetCDF file ``mesh_ocean.nc`` and store the results in a NetCDF file named
!        ``packed_ST7_1m_0101_20012_grid_T_sosstsst.nc`` use the following command (note that
!        the variable ``sosstsst`` is from a NEMO simulation since **-g=**\ ``t``  is specified) :
!        ::
!
!          $ pack_masked_data_3d \
!            -f=ST7_1m_0101_20012_grid_T_sosstsst.nc \
!            -v=sosstsst \
!            -m=mesh_ocean.nc \
!            -o=packed_ST7_1m_0101_20012_grid_T_sosstsst.nc \
!            -g=t
!
!-
! ================================================================================================
!
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf, only   : nf90_global, nf90_noerr, nf90_clobber, nf90_max_name,              &
                         nf90_nowrite, nf90_max_var_dims, nf90_real4, nf90_real8,           &
                         nf90_sizehint_default, nf90_unlimited, nf90_strerror,              &
                         nf90_inquire_variable, nf90_open, nf90_get_att, nf90_get_var,      &
                         nf90_create, nf90_def_dim, nf90_def_var, nf90_inq_attname,         &
                         nf90_copy_att, nf90_put_att, nf90_enddef, nf90_put_var, nf90_close
!
    use Statpack, only : i4b, stnd, extd, lgl, zero, one, true, false,     &
                         allocate_error, merror, arth,                     &
                         tolower, case_change
!
    use NCSTAT_Utilities, only : prtunit, nchar, get_listarg, get_variable_dims, get_variable_dimcoords
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='pack_masked_data_3d : '
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_out, ncid_mask, varid, maskid, varoutid, maskoutid, & 
                     nlon, nlat, nt, natts, xtype, i, j, j1, j2, k,                               &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3, e1id, e2id,         &
                     e1outid, e2outid, nbdim, nvar, output_mode
    integer(i4b), dimension(3)                 :: coor_id2, startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
!
    real(stnd), allocatable, dimension(:,:) :: data
    real(stnd), allocatable, dimension(:)   :: pt_mat
    real(extd), allocatable, dimension(:)   :: record
!
    logical(lgl)                                            :: use_tlimited, use_x, use_y, use_3d, &
                                                               create_hdf5, copy_scalfac
    logical(lgl),              dimension(nf90_max_var_dims) :: use_coor_variable
    logical(lgl), allocatable, dimension(:,:)               :: mask
!
    character                                                  :: chr, grid_type
    character(len=nchar)                                       :: name, history
    character(len=nf90_max_name)                               :: fic_name, var_name, out_name,  &
                                                                  mask_name, att_name
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Pack a tridimensional NetCDF variable extracted from a NetCDF dataset'
        write (prtunit,*) ' and store the packed data in a NetCDF dataset.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' pack_masked_data_3d -f=input_netcdf_file'
        write (prtunit,*) '                     -v=netcdf_variable'
        write (prtunit,*) '                     -m=input_mesh_mask_netcdf_file'
        write (prtunit,*) '                     -o=output_netcdf_file'
        write (prtunit,*) '                     -x=lon1,lon2                    (optional)'
        write (prtunit,*) '                     -y=lat1,lat2                    (optional)'
        write (prtunit,*) '                     -t=time1,time2                  (optional)'
        write (prtunit,*) '                     -g=grid_type                    (optional : n, t, u, v, w, f)'
        write (prtunit,*) '                     -3d                             (optional)'
        write (prtunit,*) '                     -scalfac                        (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                     -bigfile                        (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                     -hdf5                           (optional)'
#endif
        write (prtunit,*) '                     -tlimited                       (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period associated with the netcdf_variable'
        write (prtunit,*) '-g=n'
        write (prtunit,*) '-3d       : by default, the output variable is a twodimensional NetCDF variable'
        write (prtunit,*) '-scalfac  : by default, the scale factors are not packed and stored'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<4 .or. narg>13 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name    = " "
    var_name    = " "
    mask_name   = " "
    out_name    = " "
!
    history = "pack_masked_data_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
!
    grid_type = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x        = false
    use_y        = false
    use_3d       = false
    copy_scalfac = false
    create_hdf5  = false
    use_tlimited = false
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-m=')
!
!               GET THE NAME OF THE MASK NETCDF DATASET.
!
                mask_name = name(4:)
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR THE PACKED DATA.
!
                out_name = name(4:)
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case default
!
                call case_change( name(4:), tolower )
!
                select case ( name(1:) )
!
                    case ('-3d')
!
!                       DEFINE THE OUTPUT VARIABLE AS A 3D NETCDF VARIABLE.
!
                        use_3d = true
!
                    case ('-scalfac')
!
!                       STORE PACKED SCALE FACTORS.
!
                        copy_scalfac = true
!
                    case ('-tlimited')
!
!                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                        use_tlimited = true
!
#ifdef _USE_NETCDF36
                    case ('-bigfile')
!
!                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                        output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                    case ('-hdf5')
!
!                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                        output_mode = nf90_hdf5
                        create_hdf5 = true
#endif
!
                    case default
!
!                       WRONG ARGUMENT.
!
                        call merror( name_proc//"Wrong argument" )
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " ) call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " ) call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( mask_name==" " ) call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( out_name==" " ) call merror( name_proc//"Output NETCDF file is missing" )
!
!   OPEN THE NETCDF DATASET.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   GET THE NUMBER OF ATTRIBUTS FOR THE NETCDF VARIABLE.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                     &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names,  &
                                  dimlens_out=dimlens, coor_id_out=coor_id,            &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type,  &
                                  use_coor_variable_out=use_coor_variable, natts=natts )
!
    if ( count(dimlens/=0)/=3 )           &
    call merror( name_proc//"Input NETCDF variable is not tridimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nt     = dimlens(3)
!
!   FIND THE EXTERNAL TYPE FOR THE NETCDF VARIABLE.
!
    status = nf90_inquire_variable( ncid, varid, xtype=xtype )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   OPEN THE NETCDF MESH-MASK DATASET.
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
    call merror( name_proc//"Shape disagreement with the mask" )
!
    if ( .not.use_x ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
        if ( status==nf90_noerr ) lon1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
        if ( status==nf90_noerr ) lon2 = i
!
    end if
!
    if ( .not.use_y ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
        if ( status==nf90_noerr ) lat1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
        if ( status==nf90_noerr ) lat2 = i
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
    if ( lon1<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
    lon2 = min( lon2, nlon )
    lon3 = lon2 - lon1 + 1
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0 ) call merror( name_proc//"Wrong time definition" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( data(lon3,lat3), mask(lon3,lat3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, data, start = (/ lon1, lat1 /), count = (/ lon3, lat3 /) )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED FROM THE MASK.
!
    mask(:lon3,:lat3) = data(:lon3,:lat3)==one
!
    nvar = count( mask )
!
    if ( copy_scalfac ) then
!
!       GET THE SCALE FACTORS IDs FROM THEIR NAMES.
!       VERIFY THE DIMENSION LENGTHS FOR THE SCALE FACTOR VARIABLES.
!
        call get_variable_dims( name_proc, "e1"//chr, ncid_mask,    &
                                e1id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the first scale factor" )
!
        call get_variable_dims( name_proc, "e2"//chr, ncid_mask,    &
                                e2id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the second scale factor" )
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE PACKED DATA.
!
    if ( create_hdf5 ) then
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out )
    else
        k = nf90_sizehint_default
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE PACKED NETCDF DATASET.
!
    dimlens(1)   = lon3
    dimlens(2)   = lat3
!
    if ( use_3d ) then
!
        nbdim = 5
!
        dimlens(4)   = 1
        dim_names(4) = "y_a"
!
    else
!
        nbdim = 4
!
    endif
!
    if ( use_tlimited ) then
        dimlens(nbdim)   = time3
    else
        dimlens(nbdim)   = nf90_unlimited
    end if
!
    if ( .not.use_3d ) then
        dim_names(nbdim) = dim_names(3)
    end if
!
    dim_names(3) = "x"
    dimlens(3)   = nvar
!
    do i = 1, nbdim
!
        status = nf90_def_dim( ncid_out, name=trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE PACKED NETCDF DATASET.
!
    status = nf90_def_var( ncid_out, trim(var_name), xtype, dimids=dimids(3_i4b:nbdim), varid=varoutid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
    do i = 1, 2
!
        status = nf90_def_var( ncid_out, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_out, trim(dim_names(nbdim)), coor_type(3), dimids=dimids(nbdim), varid=coor_id2(3) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
    if ( copy_scalfac ) then
!
        status = nf90_def_var( ncid_out, "nmask", nf90_real4, dimids=dimids(3), varid=maskoutid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
        status = nf90_def_var( ncid_out, "e1n", nf90_real8, dimids=dimids(3), varid=e1outid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
        status = nf90_def_var( ncid_out, "e2n", nf90_real8, dimids=dimids(3), varid=e2outid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
    end if
!
!   ASSIGN ATTRIBUTES FOR THE PACKED NETCDF DATASET.
!
!   COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_out, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
!   ASSIGN ATTRIBUTES FOR THE PACKED NETCDF VARIABLE.
!
    do i = 1, natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       COPY ATTRIBUTES FOR THE PACKED VARIABLE.
!
        status = nf90_copy_att(  ncid, varid, att_name, ncid_out, varoutid)
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    if ( copy_scalfac ) then
!
!       ADD ATTRIBUTES FOR THE PACKED MASK.
!
        status = nf90_put_att( ncid_out, maskoutid, "long_name", "Packed mask for "//trim(var_name) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, maskoutid, "short_name", "Packed "//chr//"mask" )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, maskoutid, "valid_min", zero )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, maskoutid, "valid_max", one )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       ADD ATTRIBUTES FOR THE PACKED SCALE FACTORS.
!
        status = nf90_put_att( ncid_out, e1outid, "long_name", "Packed first scale factor for "//trim(var_name) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, e1outid, "short_name", "Packed e1"//chr )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, e2outid, "long_name", "Packed second scale factor for "//trim(var_name) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, e2outid, "short_name", "Packed e2"//chr )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_out, nf90_global, "title", "Packed "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "lon1_"//trim(var_name), lon1 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "lon2_"//trim(var_name), lon2 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "lat1_"//trim(var_name), lat1 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "lat2_"//trim(var_name), lat2 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE PACKED NETCDF DATASET.
!
    status = nf90_enddef( ncid_out )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
!   WRITE VARIABLES FOR THE PACKED NETCDF DATASET.
!
!   ALLOCATE WORK ARRAYS TO READ THE DATA.
!
    i = max( nlon, nlat, time3 )
    allocate( record(i), pt_mat(nvar), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   WRITE COORDINATE VARIABLES.
!
    startlec(:3) = (/ lon1, lat1, time1 /)
    countlec(:3) = (/ lon3, lat3, time3 /)
!
    do j = 1, 3 
!
        j1 = startlec(j)
        j2 = countlec(j)
!
        if ( use_coor_variable(j) ) then
!
            status = nf90_get_var( ncid, coor_id(j), record(:j2), start=(/ j1 /), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_out, coor_id2(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            status = nf90_put_var( ncid_out, coor_id2(j), arth(j1,1_i4b,j2) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end do
!
!   WRITE PACKED DATA.
!
    countlec(3) = 1_i4b
!
    j = 0
!
    do i = time1, time2
!
        j = j + 1
!
        startlec(3) = i
!
!       FILL THE WORK ARRAY WITH THE NETCDF VARIABLE.
!
        status = nf90_get_var( ncid, varid, data, start=startlec(:3), count=countlec(:3)   )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       PACK THE DATA.
!
        pt_mat(:nvar) = pack( data(:lon3,:lat3), mask(:lon3,:lat3) )
!
        if ( use_3d ) then
            status = nf90_put_var( ncid_out, varoutid, pt_mat, start=(/ 1_i4b, 1_i4b, j /), &
                                   count=(/ nvar, 1_i4b, 1_i4b/) )
        else
            status = nf90_put_var( ncid_out, varoutid, pt_mat, start=(/ 1_i4b, j /),        &
                                   count=(/ nvar, 1_i4b/) )
        endif
!
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   GET AND PACK THE SCALE FACTORS AND THE MASK IF REQUIRED.
!
    if ( copy_scalfac ) then
!
!       PROCESS THE FIRST SCALE FACTOR.
!
        status = nf90_get_var( ncid_mask, e1id, data(:lon3,:lat3),             &
                               start=startlec(:2), count=countlec(:2) )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        pt_mat(:nvar) = pack( data(:lon3,:lat3), mask(:lon3,:lat3) )
!
        status = nf90_put_var( ncid_out, e1outid, pt_mat )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       PROCESS THE SECOND SCALE FACTOR.
!
        status = nf90_get_var( ncid_mask, e2id, data(:lon3,:lat3),             &     
                               start=startlec(:2), count=countlec(:2) )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        pt_mat(:nvar) = pack( data(:lon3,:lat3), mask(:lon3,:lat3) )
!
        status = nf90_put_var( ncid_out, e2outid, pt_mat )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       WRITE THE DUMMY MASK.
!
        pt_mat(:nvar) = one
!
        status = nf90_put_var( ncid_out, maskoutid, pt_mat )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( data, mask, pt_mat, record )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_out )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM pack_masked_data_3d
! ==================================
!
end program pack_masked_data_3d
