program unpack_masked_data_3d
!! #############################################################################
!!
!! Copyright 2018 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _unpack_masked_data_3d:
!
! =====================
! unpack_masked_data_3d
! =====================
!
! Authors
! =======
!
! Eric Maisonnave (CERFACS) and Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 05/01/2018
!
! Purpose
! =======
!
! Unpack a packed tridimensional NetCDF variable produced by :ref:`pack_masked_data_3d` or
! derived from an NCSTAT analysis of such packed tridimensional NetCDF variable and store
! the results in an output NetCDF dataset.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ unpack_masked_data_3d \
!         -f=input_netcdf_file \
!         -v=packed_netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -o=output_netcdf_file \
!         -t=time1,time2                    (optional) \
!         -g=grid_type                      (optional : n, t, u, v, w, f) \
!         -bigfile                          (optional) \
!         -hdf5                             (optional) \
!         -tlimited                         (optional)
!
! By default
! ----------
!
!        *-t=*
!         the whole time period associated with the *packed_netcdf_variable*
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the original (unpacked) NetCDF variable is assumed to be regular or Gaussian
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *packed_netcdf_variable* argument specifies the NetCDF variable which
!      must be unpacked and the **-f=**\ *input_netcdf_file* argument
!      specifies that this NetCDF variable must be extracted from the NetCDF file
!      *input_netcdf_file*.
!
!   #) The argument **-m=**\ *input_mesh_mask_netcdf_file* specifies that the land-sea mask
!      to use for unpacking the *packed_netcdf_variable* must be read from the *input_mesh_mask_netcdf_file*.
!
!      This *input_mesh_mask_netcdf_file* may be created by :ref:`comp_clim_3d`
!      if the 2-D grid-mesh is regular or gaussian and it must be the one used originally to pack
!      the data with :ref:`pack_masked_data_3d`.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *packed_netcdf_variable* is used to construct the output
!      NetCDF variable.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the original
!      NetCDF variable is from an experiment with the ORCA or NEMO model.
!
!      If **-g=** is set to ``n``, it is assumed that the 2-D grid-mesh is regular or Gaussian.
!
!      The **-g=** argument is also used to determine the name of the NetCDF variable which
!      contains the 2-D mesh-mask in the *input_mesh_mask_netcdf_file* (e.g. it is assumed that
!      this variable is named *grid_type*\ ``mask``).
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
! Examples
! --------
!
!   #)   For unpacking a NetCDF variable ``sosstsst`` extracted from the NetCDF dataset
!        ``packed_ST7_1m_0101_20012_grid_T_sosstsst.nc`` with the help of a land-sea mask extracted
!        from the NetCDF file ``mesh_ocean.nc`` and store the results in a NetCDF file named
!        ``ST7_1m_0101_20012_grid_T_sosstsst.nc`` use the following command (note that
!        the variable ``sosstsst`` is from a NEMO simulation since **-g=**\ ``t``  is specified) :
!        ::
!
!          $ unpack_masked_data_3d \
!            -f=ST7_1m_0101_20012_grid_T_sosstsst.nc \
!            -v=sosstsst \
!            -m=mesh_ocean.nc \
!            -o=ST7_1m_0101_20012_grid_T_sosstsst.nc \
!            -g=t
!
!-
! ================================================================================================
!
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf, only : nf90_max_name, nf90_max_var_dims, nf90_sizehint_default,          &
                       nf90_clobber, nf90_global, nf90_noerr, nf90_nowrite,              &
                       nf90_real4, nf90_real8, nf90_strerror, nf90_inquire_variable,     &
                       nf90_get_att, nf90_open, nf90_get_var, nf90_create, nf90_def_dim, &
                       nf90_def_var, nf90_inq_attname, nf90_copy_att, nf90_put_att,      &
                       nf90_enddef, nf90_put_var, nf90_close
!
    use Statpack, only : i4b, stnd, extd, lgl, zero, one, true, false,           &
                         allocate_error, merror, arth, tolower, case_change
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing, get_listarg,     &
                                 get_variable_dims, get_variable_dimcoords
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='pack_masked_data_3d : '
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_out, ncid_mask, varid, varoutid, maskid,       & 
                     nlon, nlat, nt, natts, xtype, i, j, j1, j2, k, nbdim, nvar, nvar_mask,  &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3,                &
                     output_mode
    integer(i4b), dimension(3)                 :: coor_id2, startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
!
    real(stnd)                              :: missing
    real(stnd), allocatable, dimension(:,:) :: data
    real(stnd), allocatable, dimension(:)   :: pt_mat
    real(extd), allocatable, dimension(:)   :: record
!
    logical(lgl)                               :: use_3d, use_tlimited, create_hdf5,  &
                                                  use_missing, use_fillvalue, use_def_missing
    logical(lgl),              dimension(nf90_max_var_dims) :: use_coor_variable
    logical(lgl), allocatable, dimension(:,:)  :: mask
!
    character                                  :: chr, grid_type
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, out_name, mask_name, att_name
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Unpack a packed tridimensional NetCDF variable extracted from a NetCDF dataset'
        write (prtunit,*) ' and store the unpacked data in a NetCDF dataset.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' pack_masked_data_3d -f=input_netcdf_file'
        write (prtunit,*) '                     -v=packed_netcdf_variable'
        write (prtunit,*) '                     -m=input_mesh_mask_netcdf_file  '
        write (prtunit,*) '                     -o=output_netcdf_file'
        write (prtunit,*) '                     -t=time1,time2                  (optional)'
        write (prtunit,*) '                     -g=grid_type                    (optional : n, t, u, v, w, f)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                     -bigfile                        (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                     -hdf5                           (optional)'
#endif
        write (prtunit,*) '                     -tlimited                       (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-t= the whole time period associated with the packed_netcdf_variable'
        write (prtunit,*) '-g=n'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<4 .or. narg>9 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name  = " "
    var_name  = " "
    mask_name = " "
    out_name  = " "
!
    history = "unpack_masked_data_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
!
    grid_type = "n"
!
    time1     = 1
    time2     = huge( time2 )
!
    create_hdf5  = false
    use_tlimited = false
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR THE REDUCED DATA.
!
                out_name = name(4:)
!
            case ('-m=')
!
!               GET THE NAME OF THE OPTIONAL MASK NETCDF DATASET.
!
                mask_name = name(4:)
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case default
!
                call case_change( name(4:), tolower )
!
                select case ( name(1:) )
!
                    case ('-tlimited')
!
!                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                        use_tlimited = true
!
#ifdef _USE_NETCDF36
                    case ('-bigfile')
!
!                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                        output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                    case ('-hdf5')
!
!                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                        output_mode = nf90_hdf5
                        create_hdf5 = true
#endif
!
                    case default
!
!                       WRONG ARGUMENT.
!
                        call merror( name_proc//"Wrong argument" )
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " ) call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " ) call merror( name_proc//"Input packed NETCDF variable is missing" )
!
    if ( mask_name==" " ) call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( out_name==" " ) call merror( name_proc//"Output NETCDF file is missing" )
!
!   OPEN THE NETCDF DATASET.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   GET THE NUMBER OF ATTRIBUTS FOR THE NETCDF VARIABLE.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                     &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names,  &
                                  dimlens_out=dimlens, coor_id_out=coor_id,            &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type,  &
                                  use_coor_variable_out=use_coor_variable, natts=natts )
!
    nbdim = count(dimlens/=0)
!
    if ( nbdim/=4 .and. nbdim/=5 )           &
    call merror( name_proc//"Input NETCDF variable is not packed" )
!
    lon3   = dimlens(1)
    lat3   = dimlens(2)
    nvar   = dimlens(3)
    nt     = dimlens(nbdim)
!
    use_3d = nbdim==5 .and. dimlens(4)==1
!
!   FIND THE EXTERNAL TYPE FOR THE NETCDF VARIABLE.
!
    status = nf90_inquire_variable( ncid, varid, xtype=xtype )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE MISSING INDICATOR VALUE.
!
    status = nf90_get_att( ncid, varid, "missing_value", missing )
!
    use_missing = status==nf90_noerr
    if (.not.use_missing ) then
!
!       TRY THE _FillValue ATTRIBUTE IF THE MISSING ATTRIBUTE IS ABSENT.
!
        status = nf90_get_att( ncid, varid, "_FillValue", missing )
        use_fillvalue = status==nf90_noerr
    else
        use_fillvalue = false
    end if
!
    use_def_missing = .not.use_missing .and. .not.use_fillvalue
!
    if ( use_def_missing ) then
        missing = def_missing
    end if
!
!   GET THE ORIGINAL BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    status = nf90_get_att( ncid, nf90_global, "lon1_"//trim(var_name), lon1 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_get_att( ncid, nf90_global, "lon2_"//trim(var_name), lon2 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_get_att( ncid, nf90_global, "lat1_"//trim(var_name), lat1 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_get_att( ncid, nf90_global, "lat2_"//trim(var_name), lat2 )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    nlon = lon2 - lon1 + 1
    nlat = lat2 - lat1 + 1
!
    if ( nlon/=lon3 .or. nlat/=lat3 )             &
    call merror( name_proc//"Input NETCDF variable is not packed" )
!
!   OPEN THE NETCDF MESH-MASK DATASET.
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    nlon = dimlens(1)
    nlat = dimlens(2)
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( (nlon*nlat<nvar) .or. (nlon<lon3) .or. (nlat<lat3) )       &
    call merror( name_proc//"Shape disagreement with the mask" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0 ) call merror( name_proc//"Wrong time definition" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( data(lon3,lat3), mask(lon3,lat3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, data, start = (/ lon1, lat1 /), count = (/ lon3, lat3 /) )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED FROM THE MASK.
!
    mask(:lon3,:lat3) = data(:lon3,:lat3)==one
!
    nvar_mask = count( mask )
!
    if ( nvar/=nvar_mask )             &
    call merror( name_proc//"Shape disagreement with the mask" )
!
!   CREATE THE NEW NETCDF DATASET FOR THE UNPACKED DATA.
!
    if ( create_hdf5 ) then
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out )
    else
        k = nf90_sizehint_default
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE UNPACKED NETCDF DATASET.
!
    dimlens(1)   = lon3
    dimlens(2)   = lat3
    dimlens(3)   = time3
!
    dim_names(3)         = dim_names(nbdim)
    coor_id(3)           = coor_id(nbdim)
    coor_type(3)         = coor_type(nbdim)
    use_coor_variable(3) = use_coor_variable(nbdim)
    coor_natts(3)        = coor_natts(nbdim)
!
    do i = 1, 3
!
        status = nf90_def_dim( ncid_out, name=trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE UNPACKED NETCDF DATASET.
!
    status = nf90_def_var( ncid_out, trim(var_name), xtype, dimids=dimids(:3), varid=varoutid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
    do i = 1, 3
!
        status = nf90_def_var( ncid_out, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   ASSIGN ATTRIBUTES FOR THE UNPACKED NETCDF DATASET.
!
!   COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_out, coor_id2(j) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
!   ASSIGN ATTRIBUTES FOR THE PACKED NETCDF VARIABLE.
!
    do i = 1, natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       COPY ATTRIBUTES FOR THE PACKED VARIABLE.
!
        status = nf90_copy_att(  ncid, varid, att_name, ncid_out, varoutid)
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    if ( use_def_missing ) then
!
        if (  xtype==nf90_real8 ) then
!
            status = nf90_put_att( ncid_out, varoutid, "missing_value", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_out, varoutid, "_FillValue", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        else if ( xtype==nf90_real4 ) then
!
            status = nf90_put_att( ncid_out, varoutid, "missing_value", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_out, varoutid, "_FillValue", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_out, nf90_global, "title", "Unpacked "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE UNPACKED NETCDF DATASET.
!
    status = nf90_enddef( ncid_out )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status )) )
!
!   WRITE VARIABLES FOR THE UNPACKED NETCDF DATASET.
!
!   ALLOCATE WORK ARRAYS TO READ THE DATA.
!
    i = max( lon3, lat3, time3 )
    allocate( record(i), pt_mat(nvar), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   WRITE COORDINATE VARIABLES.
!
    startlec(:3) = (/ 1_i4b, 1_i4b, time1 /)
    countlec(:3) = (/ lon3, lat3, time3 /)
!
    do j = 1, 3 
!
        j1 = startlec(j)
        j2 = countlec(j)
!
        if ( use_coor_variable(j) ) then
!
            status = nf90_get_var( ncid, coor_id(j), record(:j2), start=(/ j1 /), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_out, coor_id2(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            status = nf90_put_var( ncid_out, coor_id2(j), arth(j1,1_i4b,j2) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end do
!
!   WRITE UNPACKED DATA.
!
    countlec(3) = 1_i4b
!
    j = 0
!
    do i = time1, time2
!
        j = j + 1
!
!       FILL THE WORK ARRAY WITH THE NETCDF VARIABLE.
!
        if ( use_3d ) then
            status = nf90_get_var( ncid, varid, pt_mat, start=(/ 1_i4b, 1_i4b, i /), &
                                   count=(/ nvar, 1_i4b, 1_i4b/) )
        else
            status = nf90_get_var( ncid, varid, pt_mat, start=(/ 1_i4b, i /),        &
                                   count=(/ nvar, 1_i4b/) )
        endif
!
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       UNPACK THE DATA.
!
        data(:lon3,:lat3) = unpack( pt_mat(:nvar), mask(:lon3,:lat3), missing )
!
        startlec(3) = j
!
        status = nf90_put_var( ncid_out, varoutid, data, start=startlec(:3),     &
                               count=countlec(:3) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( data, mask, pt_mat, record )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_out )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM pack_masked_data_3d
! ==================================
!
end program unpack_masked_data_3d
